/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.antminer;

import java.util.Properties;

import net.sf.myra.datamining.AbstractObjectiveFunction;
import net.sf.myra.datamining.Classifier;
import net.sf.myra.datamining.ConvergenceTest;
import net.sf.myra.datamining.Evaluator;
import net.sf.myra.datamining.IterationTest;
import net.sf.myra.datamining.Main;
import net.sf.myra.datamining.MajorityClassAssignator;
import net.sf.myra.datamining.Pruner;
import net.sf.myra.datamining.Rule;
import net.sf.myra.datamining.data.Attribute;
import net.sf.myra.datamining.data.ContinuousAttribute;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.EntropyHeuristicInformation;
import net.sf.myra.datamining.data.Label;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.function.SensitivitySpecificityFunction;
import net.sf.myra.datamining.model.RuleList;
import net.sf.myra.framework.Colony;
import net.sf.myra.framework.CompoundStopCondition;
import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.Scheduler;

/**
 * This class represents the Ant-Miner classification algorithm. More
 * information is available in following paper:
 * 
 * <pre>
 * &#64;ARTICLE{Parpinelli02datamining,
 *    author = {Rafael S. Parpinelli and Heitor S. Lopes and Alex A. Freitas},
 *    title = {Data Mining with an Ant Colony Optimization Algorithm},
 *    journal = {IEEE Transactions on Evolutionary Computation},
 *    year = {2002},
 *    volume = {6},
 *    pages = {321--332}
 * }
 * </pre>
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class AntMiner
	extends Classifier
{
	/**
	 * Default options filename.
	 */
	private static final String OPTIONS_FILE = "myra-antminer.messages";

	/**
	 * Default constructor.
	 */
	public AntMiner()
	{
		super(OPTIONS_FILE);
	}

	@Override
	public Properties filter(Properties properties)
	{
		properties.remove(RULE_QUALITY_FUNCTION);
		properties.remove(Main.EXPORTER);
		return properties;
	}

	public RuleList run(Dataset dataset)
	{
		if (dataset.getMetadata().isHierarchical())
		{
			throw new IllegalArgumentException(
				"Hierarchical dataset not supported.");
		}

		for (Attribute attribute : dataset.getMetadata().getAttributes())
		{
			if (attribute instanceof ContinuousAttribute)
			{
				throw new IllegalArgumentException(
					"Continuous attributes are not supported.");
			}
		}

		RuleList discovered = new RuleList();
		Metadata metadata = dataset.getMetadata();

		int maxUncoveredCases =
			Configuration.getIntProperty(MAXIMUM_UNCOVERED_CASES,
				DEFAULT_MAXIMUM_UNCOVERED_CASES);

		Label majority = Evaluator.findMajorityClass(dataset);

		while (dataset.getSize() > maxUncoveredCases)
		{
			Rule rule = evolve(dataset);
			dataset.remove(Evaluator.findCoveredCases(rule, dataset));
			discovered.add(rule.clone());
		}

		if (!discovered.hasDefault())
		{
			if (!dataset.isEmpty())
			{
				majority = Evaluator.findMajorityClass(dataset);
			}

			Rule defaultRule = new Rule(metadata);
			defaultRule.setConsequent(majority);

			discovered.add(defaultRule);
		}

		return discovered;
	}

	/**
	 * Evolves a classification rule based on an ACO procedure.
	 * 
	 * @param dataset the data set to extract a rule.
	 * 
	 * @return a classification rule.
	 */
	private Rule evolve(Dataset dataset)
	{
		ProbabilisticRuleFactory factory =
			new ProbabilisticRuleFactory(dataset.getMetadata().toGraph(),
				dataset);

		Colony colony = new Colony();
		colony.setTrailFactory(factory);

		CompoundStopCondition condition = new CompoundStopCondition();
		condition.add(new ConvergenceTest());
		condition.add(new IterationTest());

		AbstractObjectiveFunction function =
			new SensitivitySpecificityFunction(dataset);

		Scheduler scheduler =
			new Scheduler(colony, condition, function,
				new DefaultPheromoneUpdater(), new EntropyHeuristicInformation(
					dataset));
		scheduler.setLocalSearch(new Pruner(function,
			new MajorityClassAssignator()));

		scheduler.run();

		// get the best ant trail constructed
		return (Rule) colony.getBest().getBestTrail();
	}
}