/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.antminer;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import net.sf.myra.datamining.Evaluator;
import net.sf.myra.datamining.Rule;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.data.Term;
import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.Graph;
import net.sf.myra.framework.NoAvailableVertexException;
import net.sf.myra.framework.RandomGenerator;
import net.sf.myra.framework.Trail;
import net.sf.myra.framework.TrailFactory;
import net.sf.myra.framework.Vertex;

/**
 * This class represents a probabilistic (roulette wheel) trail factory.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class ProbabilisticRuleFactory
	implements TrailFactory
{
	/**
	 * The property key under which the minimum rule cases value is stored.
	 */
	public static final String MINIMUM_RULE_CASES =
		"net.sf.myra.antminer.cases";

	/**
	 * Constant to indicate that there is no minimum number of covered cases.
	 */
	public static int NO_MINIMUM_CASES = 0;

	/**
	 * Minimum number of covered cases per rule.
	 */
	protected final int MINIMUM_CASES;

	/**
	 * The problem construction graph to create the trails.
	 */
	private Graph<?> graph;

	/**
	 * The dataset instance to validate the vertices selection.
	 */
	private Dataset dataset;

	/**
	 * Default constructor.
	 * 
	 * @param graph the problem construction graph.
	 * @param dataset the dataset reference.
	 */
	public ProbabilisticRuleFactory(Graph<?> graph, Dataset dataset)
	{
		this(graph, dataset, Configuration.getIntProperty(MINIMUM_RULE_CASES,
			NO_MINIMUM_CASES));
	}

	/**
	 * Default constructor.
	 * 
	 * @param graph the problem construction graph.
	 * @param dataset the dataset reference.
	 */
	protected ProbabilisticRuleFactory(Graph<?> graph, Dataset dataset,
			int mininumCases)
	{
		this.graph = graph;
		this.dataset = dataset;
		this.MINIMUM_CASES = mininumCases;
	}

	@SuppressWarnings("unchecked")
	public Trail createTrail()
	{
		Metadata metadata = dataset.getMetadata();
		Rule rule = new Rule(metadata);

		// visited vertices indexes
		Set<Integer> visited = new HashSet<Integer>();

		int available = dataset.getMetadata().getPredictor().size();
		List<Instance> covered = dataset.getInstances();
		Wheel wheel = new Wheel(((Graph<Term>) graph).vertices());

		try
		{
			while ((visited.size() < available) && !wheel.isEmpty()
				&& (covered.size() >= MINIMUM_CASES)
				&& !Evaluator.sameLabel(covered))
			{
				Rule clone = (Rule) rule.clone();
				double probability = RandomGenerator.getInstance().nextDouble();

				Slot slot = wheel.spin(probability);
				Vertex<Term> vertex = process(metadata, slot.vertex(), covered);

				clone.add(vertex);
				List<Instance> set = Evaluator.findCoveredCases(clone, covered);

				if (set.size() >= MINIMUM_CASES
					&& (set.size() != covered.size() || rule.isEmpty()))
				{
					rule.add(vertex);
					covered = set;

					int index = vertex.getInfo().getIndex();
					visited.add(index);
					wheel.remove(index);
				}
				else
				{
					wheel.remove(slot);
				}
			}
		}
		catch (NoAvailableVertexException e)
		{
			// end of the rule construction procedure
		}

		return rule;
	}

	/**
	 * Prepares the vertex to be added to a rule. The default implementation
	 * just returns the vertex.
	 * 
	 * @param vertex the vertex to be added to the rule.
	 * 
	 * @return the same vertex instance.
	 */
	protected Vertex<Term> process(Metadata metadata, Vertex<Term> vertex,
			List<Instance> covered)
	{
		return vertex;
	}

	public Graph<?> getGraph()
	{
		return graph;
	}

	/**
	 * Returns the dataset reference.
	 * 
	 * @return the dataset reference.
	 */
	protected Dataset getDataset()
	{
		return dataset;
	}

	/**
	 * This class represents the probabilistic selection roulette wheel.
	 */
	static class Wheel
	{
		/**
		 * The list of available slots (vertices).
		 */
		private ArrayList<Slot> slots;

		/**
		 * The total heuristic value.
		 */
		private double tHeuristic;

		/**
		 * Default constructor.
		 * 
		 * @param vertices the available vertices.
		 */
		Wheel(Vertex<Term>[] vertices)
		{
			slots = new ArrayList<Slot>(vertices.length);
			tHeuristic = 0.0;

			for (Vertex<Term> vertex : vertices)
			{
				slots.add(new Slot(vertex));
				tHeuristic += (vertex.getHeuristicValue().getValue());
			}

			normalize();
		}

		/**
		 * Selects a slot based on the probability specified.
		 * 
		 * @param probability the probability value.
		 * 
		 * @return the slot selected given the probability value.
		 * 
		 * @throws NoAvailableVertexException If a slot can not be selected.
		 */
		Slot spin(double probability)
			throws NoAvailableVertexException
		{
			double position = 0.0;

			for (int i = 0; i < slots.size(); i++)
			{
				position += slots.get(i).probability();

				if (probability <= position)
				{
					return slots.get(i);
				}
			}

			throw new NoAvailableVertexException(
				"No slot could be selected (wheel size " + slots.size() + ")");
		}

		/**
		 * Removes the specified slot.
		 * 
		 * @param slot the slot to be removed.
		 */
		void remove(Slot slot)
		{
			slots.remove(slot);
			tHeuristic -= (slot.vertex().getHeuristicValue().getValue());

			normalize();
		}

		/**
		 * Removes all slots that correspond to the specified attribute.
		 * 
		 * @param index the attribute index.
		 */
		void remove(int index)
		{
			for (Iterator<Slot> i = slots.iterator(); i.hasNext();)
			{
				Slot slot = i.next();

				if (slot.vertex().getInfo().getIndex() == index)
				{
					tHeuristic -= slot.vertex().getHeuristicValue().getValue();
					i.remove();
				}
			}
		}

		/**
		 * Tests if the wheel is empty (has no slots).
		 * 
		 * @return <code>true</code> is there are no slots in the wheel;
		 *         <code>false</code> otherwise.
		 */
		boolean isEmpty()
		{
			return slots.isEmpty();
		}

		/**
		 * Normalizes the probability values.
		 */
		private void normalize()
		{
			double tPheromone = 0.0;
			int size = slots.size();

			for (int i = 0; i < size; i++)
			{
				Slot slot = slots.get(i);
				slot.calculate(tHeuristic);
				tPheromone += slot.probability();
			}

			for (int i = 0; i < size; i++)
			{
				slots.get(i).normalize(tPheromone);
			}
		}
	}

	/**
	 * This class represents a slot of the probabilistic roulette wheel.
	 */
	static class Slot
	{
		/**
		 * The vertex at the slot.
		 */
		private Vertex<Term> vertex;

		/**
		 * The probability of selecting this slot.
		 */
		private double probability;

		/**
		 * Default constructor.
		 * 
		 * @param vertex the vertex at this slot.
		 */
		Slot(Vertex<Term> vertex)
		{
			this.vertex = vertex;
		}

		/**
		 * Returns the vertex at this slot.
		 * 
		 * @return the vertex at this slot.
		 */
		Vertex<Term> vertex()
		{
			return vertex;
		}

		/**
		 * Calculates the probability of selecting this slot based on the
		 * total heuristic value.
		 * 
		 * @param total the total heuristic value.
		 */
		void calculate(double total)
		{
			double heuristic = vertex.getHeuristicValue().getValue() / total;

			probability =
				vertex.getPheromone().getValue()
					* (Double.isNaN(heuristic) ? 0.0 : heuristic);
		}

		/**
		 * Normalizes the probability value.
		 * 
		 * @param total the total probability value.
		 */
		void normalize(double total)
		{
			probability /= total;
		}

		/**
		 * Returns the probability associated with this slot.
		 * 
		 * @return the probability associated with this slot.
		 */
		double probability()
		{
			return probability;
		}
	}
}