/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.antminer.baseline;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.Properties;

import net.sf.myra.datamining.Classifier;
import net.sf.myra.datamining.Gatherer;
import net.sf.myra.datamining.Main;
import net.sf.myra.datamining.Model;
import net.sf.myra.datamining.data.Attribute;
import net.sf.myra.datamining.data.ContinuousAttribute;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Label;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.io.ArffHelper;
import weka.classifiers.rules.PART;
import weka.core.Instances;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class Part
	extends Classifier
{
	/**
	 * Default constructor.
	 */
	public Part()
	{
		super(null);
	}

	@Override
	public Properties filter(Properties properties)
	{
		if (properties.getProperty(Main.MEASURE) != null)
		{
			Properties p = new Properties();
			p.setProperty(Main.MEASURE, properties.getProperty(Main.MEASURE));
			return p;
		}

		return new Properties();
	}

	@Override
	public Model run(Dataset dataset)
	{
		PartModel model = new PartModel();

		try
		{
			StringWriter output = new StringWriter();
			ArffHelper helper = new ArffHelper();
			helper.write(output, dataset);

			Instances training =
				new Instances(new StringReader(output.toString()));
			training.setClassIndex(training.numAttributes() - 1);

			PART part = new PART();
			part.buildClassifier(training);

			model.setClassifier(part);
			model.setDataset(training);
		}
		catch (Exception e)
		{
			throw new RuntimeException(e);
		}

		return model;
	}

	/**
	 * This class represents a J48 decision tree.
	 */
	private class PartModel
		implements Model
	{
		/**
		 * The (trained) J48 classifier.
		 */
		private PART classifier;

		/**
		 * The (WEKA) training dataset.
		 */
		private Instances dataset;

		/**
		 * Sets the classifier instance.
		 * 
		 * @param classifier the classifier to set
		 */
		public void setClassifier(PART classifier)
		{
			this.classifier = classifier;
		}

		/**
		 * Sets the dataset instance.
		 * 
		 * @param dataset the dataset to set
		 */
		public void setDataset(Instances dataset)
		{
			this.dataset = dataset;
		}

		public Gatherer getGatherer()
		{
			throw new UnsupportedOperationException();
		}

		/**
		 * Returns the predicted class label for the specified instance.
		 * 
		 * @param instance the instance to predict the class label.
		 *  
		 * @return the predicted class label for the specified instance.
		 */
		public Label classify(Instance instance)
		{
			weka.core.Instance wInstance =
				new weka.core.Instance(dataset.numAttributes());
			wInstance.setDataset(dataset);

			for (Attribute attribute : instance.getMetadata().getPredictor())
			{
				String name = attribute.getName().replaceAll("(\\s)+", "_");

				if (attribute instanceof ContinuousAttribute)
				{
					int index =
						instance.getMetadata().getIndex(attribute.getName());

					wInstance.setValue(dataset.attribute(name),
						instance.values()[index]);
				}
				else
				{
					String value = instance.getValue(attribute);

					if (Metadata.MISSING_VALUE.equals(value))
					{
						wInstance.setMissing(dataset.attribute(name));
					}
					else
					{
						wInstance.setValue(dataset.attribute(name), value);
					}
				}
			}

			try
			{
				return instance.getLabel().toLabel(
					wInstance.dataset().classAttribute()
						.value((int) classifier.classifyInstance(wInstance)));
			}
			catch (Exception e)
			{
				throw new RuntimeException(e);
			}
		}

		@Override
		public String toString()
		{
			String output = classifier.toString();
			BufferedReader in = new BufferedReader(new StringReader(output));
			String line = null;
			double length = 0.0;

			try
			{
				while ((line = in.readLine()) != null)
				{
					if (line.trim().endsWith("AND"))
					{
						length += 1;
					}
					else if (line.contains(")")
						&& line.substring(0, line.indexOf(":")).trim().length() > 0)
					{
						length += 1;
					}
				}
			}
			catch (IOException e)
			{
				length = 0.0;
			}

			return "PART (Rule List)\n" + "Model size: " + length
				+ "\nNumber of rules: " + classifier.measureNumRules()
				+ "\nAverage terms per rule: "
				+ (length / classifier.measureNumRules());
		}
	}
}