/*
 * $Id: AccuracyPruner.java 2329 2011-08-02 08:57:44Z febo $
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2010 Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.antree;

import java.util.Comparator;
import java.util.TreeSet;

import net.sf.myra.antree.representation.LeafNode;
import net.sf.myra.antree.representation.Node;
import net.sf.myra.antree.representation.Statistics;
import net.sf.myra.antree.representation.Tree;
import net.sf.myra.datamining.data.Dataset;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision: 2329 $ $Date:: 2011-08-02 09:57:44#$
 */
public class AccuracyPruner
	implements Pruner
{
	/**
	 * Prunes the tree based on the reduced-error pruning procedure of the
	 * C4.5 algorithm. The statistical test applied assume that the data set
	 * used in the pruning is the training data set.
	 * 
	 * @param tree the tree to be pruned.
	 * @param dataset the data set used during the pruning (this method assumes
	 *        that it is the training set).
	 *        
	 * @return the pruned tree.
	 */
	public Tree prune(Tree tree, Dataset dataset)
	{
		Statistics statistics = tree.gather(dataset);

		if (!tree.getRoot().isLeaf())
		{
			TreeSet<Node> candidates = new TreeSet<Node>(new Comparator<Node>()
			{
				public int compare(Node o1, Node o2)
				{
					int c = Double.compare(o1.getLevel(), o2.getLevel());

					if (c == 0 && !o1.equals(o2))
					{
						return o1.toString().compareTo(o2.toString());
					}

					return c;
				}
			});

			for (Node leaf : tree.leaves())
			{
				if (leaf.getParent() != null)
				{
					candidates.add(leaf.getParent());
				}
			}

			while (!candidates.isEmpty())
			{
				Node node = candidates.last();
				double total = statistics.getTotal(node);

				// (1) the (sub)tree accuracy
				double treeAccuracy =
					(total - ErrorBasedMetric.getInstance().rawError(node,
						statistics))
						/ total;

				// (2) the error if the (sub)tree is replaced by a leaf node
				double leafAccuracy =
					(total - statistics.getErrors(node, statistics
						.getMajorityLabel(node)))
						/ total;

				Tree clone = tree.clone();
				Node branchNode = replace(node, tree, statistics, clone);

				clone.adjust(dataset);
				Statistics cloneStats = clone.gather(dataset);

				// (3) the error if the (sub)tree is replaced by its most used
				// branch
				double branchAccuracy =
					(total - ErrorBasedMetric.getInstance().rawError(
						branchNode, cloneStats))
						/ total;

				if ((leafAccuracy >= treeAccuracy)
					&& (leafAccuracy >= branchAccuracy))
				{
					LeafNode leaf =
						new LeafNode(node.getLevel(), dataset.getMetadata(),
							statistics.getMajorityLabel(node));

					statistics.setDistribution(leaf, statistics
						.getDistribution(node));

					if (node.getParent() != null)
					{
						node.getParent().getBranch(node).setTo(leaf);
						candidates.add(leaf.getParent());
					}
					// we must be dealing with the root of the tree
					else
					{
						tree.setRoot(leaf);
						// if we replaced the tree for a leaf, we do not have
						// more leaf nodes to evaluate
						candidates.clear();
					}
				}
				else if (branchAccuracy >= treeAccuracy)
				{
					candidates.clear();

					for (Node leaf : clone.leaves())
					{
						if (leaf.getParent() != null)
						{
							candidates.add(leaf.getParent());
						}
					}

					tree = clone;
					statistics = cloneStats;
				}

				candidates.remove(node);
			}
		}

		return tree;
	}

	/**
	 * Replaces the specified node for its most used branch and returns the
	 * resultant tree.
	 * 
	 * @param node the node reference.
	 * @param tree the current tree.
	 * @param statistics the statistics of the tree.
	 * 
	 * @return a tree with the node replaced by its most used branch.
	 */
	private Node replace(Node node, Tree tree, Statistics statistics, Tree clone)
	{
		int branch = 0;

		for (int i = 1; i < node.getBranches().length; i++)
		{
			if (statistics.getTotal(node.getBranch(branch).getTo()) < statistics
				.getTotal(node.getBranch(i).getTo()))
			{
				branch = i;
			}
		}

		Node nodeClone = clone.get(tree.getIndex(node));
		Node branchNode = nodeClone.getBranch(branch).getTo();

		if (nodeClone.getParent() != null)
		{
			nodeClone.getParent().getBranch(nodeClone).setTo(branchNode);
			branchNode.setLevel(node.getLevel());
		}
		else
		{
			branchNode.setLevel(Node.INITIAL_LEVEL);
			clone.setRoot(branchNode);
		}

		return branchNode;
	}
}