/*
 * $Id: AntTreeMiner.java 2348 2011-09-14 16:05:59Z febo $
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2010 Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.antree;

import static net.sf.myra.datamining.ConvergenceTest.CONVERGENCE_TEST_SIZE;
import static net.sf.myra.datamining.IterationTest.MAXIMUM_ITERATIONS;
import static net.sf.myra.datamining.Main.EXPORTER;
import static net.sf.myra.datamining.Main.EXPORTER_FILE;

import java.util.Comparator;
import java.util.Properties;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadFactory;
import java.util.logging.Level;
import java.util.logging.Logger;

import net.sf.myra.antree.representation.Tree;
import net.sf.myra.datamining.Classifier;
import net.sf.myra.datamining.Evaluator;
import net.sf.myra.datamining.Measure;
import net.sf.myra.datamining.Model;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.measure.AccuracyMeasure;
import net.sf.myra.framework.Colony;
import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.ObjectFactory;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision: 2348 $ $Date:: 2011-09-14 17:05:59#$
 */
public class AntTreeMiner
	extends Classifier
{
	/**
	 * The property key under which the colony size is stored.
	 */
	public static final String COLONY_SIZE = "net.sf.myra.antree.colony";

	/**
	 * The property key under which the construction graph implementation class
	 * name is stored.
	 */
	public static final String GRAPH = "net.sf.myra.antree.graph";

	/**
	 * The property key under which the measure implementation class name
	 * is stored.
	 */
	public static final String MEASURE = "net.sf.myra.antree.measure";

	/**
	 * The property key under which the pruner implementation class name is
	 * stored.
	 */
	public static final String PRUNER = "net.sf.myra.antree.pruner";

	/**
	 * The property key under which the heuristic information implementation
	 * class name is stored.
	 */
	public static final String HEURISTIC = "net.sf.myra.antree.heuristic";

	/**
	 * The property key under which the number of worker threads is stored.
	 */
	public static final String WORKERS = "net.sf.myra.antree.workers";

	/**
	 * The property key under which the convergence test implementation class
	 * name is stored.
	 */
	public static final String CONVERGENCE_TEST =
		"net.sf.myra.antree.convergence";

	/**
	 * Default options filename.
	 */
	private static final String OPTIONS_FILE = "myra-antree.messages";

	/**
	 * Default colony size value.
	 */
	private static final int DEFAULT_COLONY_SIZE = 100;

	/**
	 * Local copy of the current dataset for each thread. This is necessary
	 * because instances can have different weights according to how the
	 * tree is structured.
	 */
	private static LocalDataset LOCAL_DATASET;

	/**
	 * The tree comparator instance.
	 */
	private static final Comparator<Tree> COMPARATOR = new TreeComparator();

	/**
	 * The logger instance.
	 */
	private Logger logger = Logger.getLogger(AntTreeMiner.class.getName());

	/**
	 * The tests to determine if the search has converged.
	 */
	private ConvergenceTest[] convergenceTests;

	/**
	 * Default constructor.
	 */
	public AntTreeMiner()
	{
		super(OPTIONS_FILE);

		convergenceTests = new ConvergenceTest[2];
		convergenceTests[0] = new IterationTest();
		convergenceTests[1] =
			(ConvergenceTest) ObjectFactory.create(Configuration.getProperty(
				CONVERGENCE_TEST, MaxMinStagnationTest.class.getName()));
	}

	@Override
	public Properties filter(Properties properties)
	{
		properties.remove(Colony.COLONY_SIZE);
		properties.remove(MAXIMUM_UNCOVERED_CASES);
		properties.remove(RULE_QUALITY_FUNCTION);
		properties.remove(CONVERGENCE_TEST_SIZE);
		properties.remove(MAXIMUM_ITERATIONS);
		properties.remove(EXPORTER);
		properties.remove(EXPORTER_FILE);

		return properties;
	}

	@Override
	public Model run(final Dataset dataset)
	{
		// this is necessary, so then we can adjust the threshold values the
		// same way that C4.5 does
		dataset.computeDomains();

		Heuristic heuristic =
			(Heuristic) ObjectFactory.create(Configuration.getProperty(
				HEURISTIC, GainRatioHeuristic.class.getName()),
				new Class[] { Dataset.class }, new Object[] { dataset });

		Graph graph =
			(Graph) ObjectFactory.create(Configuration.getProperty(GRAPH,
				MaxMinGraph.class.getName()), new Class[] { Metadata.class,
				Heuristic.class }, new Object[] { dataset.getMetadata(),
				heuristic });

		Stepper stepper =
			new Stepper(new Builder(graph, dataset.getMetadata(), Evaluator
				.findMajorityClass(dataset)), (Pruner) ObjectFactory
				.create(Configuration.getProperty(PRUNER,
					ErrorBasedPruner.class.getName())),
				(Measure) ObjectFactory.create(Configuration.getProperty(
					MEASURE, AccuracyMeasure.class.getName())));
		Ant.setStepper(stepper);

		Ant[] colony =
			new Ant[Configuration.getIntProperty(COLONY_SIZE,
				DEFAULT_COLONY_SIZE)];

		for (int i = 0; i < colony.length; i++)
		{
			colony[i] = new Ant();
		}

		int available =
			Configuration.getIntProperty(WORKERS, Runtime.getRuntime()
				.availableProcessors());

		ExecutorService executor =
			Executors.newFixedThreadPool(available, new ThreadFactory()
			{
				private int id = 0;

				public Thread newThread(Runnable r)
				{
					return new Thread(r, "Ant-Tree-Miner::worker" + (id++));
				}
			});
		Iteration iteration = null;
		int current = 0;

		LOCAL_DATASET = new LocalDataset(dataset);
		logger.info("run started");
		Tree currentBest = null;

		do
		{
			currentBest = null;
			createSolutions(executor, colony);

			// finds the best tree of the current iteration

			for (Ant ant : colony)
			{
				if (currentBest == null)
				{
					currentBest = ant.getTree();
				}
				else if (COMPARATOR.compare(ant.getTree(), currentBest) > 0)
				{
					currentBest = ant.getTree();
				}
			}

			updatePheromones(currentBest, graph);

			iteration = new Iteration(currentBest, current);
			current++;
		}
		while (!isFinished(graph, iteration));

		logger.info("run finished");

		executor.shutdown();

		if (logger.isLoggable(Level.INFO))
		{
			logger.info("Accuracy: "
				+ new AccuracyMeasure().evaluate(dataset, currentBest));
			logger.info("Best tree size: " + currentBest.countLeaves());
			logger.info("Number of nodes: " + currentBest.size());
		}

		return currentBest;
	}

	private void createSolutions(ExecutorService executor, Ant[] colony)
	{
		CountDownLatch latch = new CountDownLatch(colony.length);

		for (Ant ant : colony)
		{
			executor.execute(new Task(ant, latch));
		}

		try
		{
			// waits for the creation of the trees
			latch.await();
		}
		catch (InterruptedException e)
		{
			throw new RuntimeException(e);
		}
	}

	private void updatePheromones(Tree tree, Graph graph)
	{
		graph.update(tree);
		graph.evaporate();
	}

	private boolean isFinished(Graph graph, Iteration iteration)
	{
		for (ConvergenceTest test : convergenceTests)
		{
			if (test.hasConverged(graph, iteration))
			{
				logger.info(test.getClass().getName()
					+ " has converged at iteration " + iteration.getNumber());
				return true;
			}
		}

		return false;
	}

	/**
	 * A task representing an ant.
	 */
	private static class Task
		implements Runnable
	{
		/**
		 * The ant reference of this task.
		 */
		private Ant ant;

		/**
		 * The synchronization count down aid.
		 */
		private CountDownLatch latch;

		/**
		 * Default constructor.
		 * 
		 *  @param id the worker id.
		 */
		public Task(Ant ant, CountDownLatch latch)
		{
			this.ant = ant;
			this.latch = latch;
		}

		public void run()
		{
			Dataset local = LOCAL_DATASET.get();
			local.reset();
			ant.walk(local);
			latch.countDown();
		}
	}

	private static class LocalDataset
		extends ThreadLocal<Dataset>
	{
		private final Dataset dataset;

		public LocalDataset(Dataset dataset)
		{
			this.dataset = dataset;
		}

		@Override
		public Dataset initialValue()
		{
			return dataset.clone();
		}
	}
}