/*
 * $Id: Builder.java 2341 2011-08-29 11:43:18Z febo $
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2010 Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.antree;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;

import net.sf.myra.antree.Graph.Vertex;
import net.sf.myra.antree.representation.Branch;
import net.sf.myra.antree.representation.InternalNode;
import net.sf.myra.antree.representation.LeafNode;
import net.sf.myra.antree.representation.Node;
import net.sf.myra.antree.representation.NodeFactory;
import net.sf.myra.antree.representation.Tree;
import net.sf.myra.datamining.Evaluator;
import net.sf.myra.datamining.data.Attribute;
import net.sf.myra.datamining.data.ContinuousAttribute;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Label;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.RandomGenerator;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision: 2341 $ $Date:: 2011-08-29 12:43:18#$
 */
public class Builder
{
	/**
	 * The property key under which the minimum number of instances for each
	 * branch is stored.
	 */
	public static final String MINIMUM_INSTANCES =
		"net.sf.myra.antree.builder.minimum";

	/**
	 * The property key under which the alpha parameter is stored.
	 */
	public static final String ALPHA = "net.sf.myra.antree.builder.alpha";

	/**
	 * The property key under which the beta parameter is stored.
	 */
	public static final String BETA = "net.sf.myra.antree.builder.beta";

	/**
	 * The construction graph.
	 */
	private final Graph graph;

	/**
	 * The current metadata.
	 */
	private final Metadata metadata;

	/**
	 * The majority class label of the data set.
	 */
	private final Label majority;

	/**
	 * The minimum number of instances required for each branch.
	 */
	private final int minimum =
		Configuration.getIntProperty(MINIMUM_INSTANCES, 2);

	/**
	 * The exponent for the pheromone value.
	 */
	private final double alpha = Configuration.getDoubleProperty(ALPHA, 1.0);

	/**
	 * The exponent for the heuristic value.
	 */
	private final double beta = Configuration.getDoubleProperty(BETA, 1.0);

	/**
	 * Default constructor.
	 * 
	 * @param graph the constructions graph.
	 * @param metadata the dataset's metadata.
	 * @param majority the majority label.
	 */
	public Builder(Graph graph, Metadata metadata, Label majority)
	{
		this.graph = graph;
		this.metadata = metadata;
		this.majority = majority;
	}

	/**
	 * Returns a decision tree.
	 * 
	 * @return a decision tree. 
	 */
	public Tree build(Dataset dataset)
	{
		Tree tree =
			new Tree(follow(null, new ArrayList<Attribute>(1), dataset
				.getInstances()));
		return tree;
	}

	/**
	 * Recursively follows the specified branch, adding nodes until a leaf
	 * node is reached.
	 *  
	 * @param branch the branch to follow.
	 * @param used the list of used attributes.
	 * @param instances the list of instances.
	 */
	private Node follow(Branch branch, Collection<Attribute> used,
			List<Instance> instances)
	{
		// if all instances have the same class label we just create a
		// leaf node
		//if (Evaluator.sameLabel(instances) || instances.size() < (2 * minimum))
		if (Evaluator.sameLabel(instances)
			|| (branch != null && Util.subset(instances, branch.getTerm(),
				metadata).size() < (2 * minimum)))
		{
			if (branch == null)
			{
				return new LeafNode(Node.INITIAL_LEVEL, metadata, Evaluator
					.findMajorityClass(instances, metadata, majority));
			}

			Node node =
				new LeafNode(branch.getFrom().getLevel() + 1, metadata,
					Evaluator.findMajorityClass(instances, metadata, majority));
			branch.setTo(node);

			return node;
		}

		Attribute attribute = null;
		int level = 0;

		if (branch == null)
		{
			attribute = select(graph.getVertices(null));
		}
		else
		{
			attribute = select(graph.getVertices(branch, used));
			level = branch.getFrom().getLevel() + 1;
		}

		Node node = null;

		// do we have an attribute?
		if (attribute == null)
		{
			node =
				new LeafNode(level, metadata, Evaluator.findMajorityClass(
					instances, metadata, majority));
		}
		else
		{
			node = NodeFactory.create(attribute, metadata, instances, level);

			// do the generated subsets have a reasonable number of instances?
			if (!verify(((InternalNode) node).getBranches(), instances,
				metadata))
			{
				//node =
				//	new LeafNode(level, metadata, Evaluator.findMajorityClass(
				//		instances, metadata, majority));

				// the selected attributes did not produce valid subsets, so
				// we will try to select another one
				used.add(attribute);
				return follow(branch, used, instances);
			}
			// yes, so let's create the partitions
			else
			{
				InternalNode internal = (InternalNode) node;

				for (Branch child : internal.getBranches())
				{
					Collection<Attribute> expanded =
						new HashSet<Attribute>(used);

					// continuous attributes can be selected more than one
					// time per path to allow multiple discretisation points
					if (!(attribute instanceof ContinuousAttribute))
					{
						expanded.add(attribute);
					}

					List<Instance> filtered =
						Util.split(instances, child.getTerm(), metadata);

					// if the list of instances is empty, we add a leaf node
					// with the majority class label of the parent node
					if (filtered.isEmpty())
					{
						LeafNode leaf =
							new LeafNode(level + 1, metadata, Evaluator
								.findMajorityClass(instances, metadata,
									majority));
						child.setTo(leaf);
					}
					// if the number of instances is smaller than 2 times the
					// minimum size, we automatically add a leaf node
					else if (filtered.size() < (2 * minimum))
					{
						LeafNode leaf =
							new LeafNode(level + 1, metadata,
								Evaluator.findMajorityClass(filtered, metadata,
									majority));
						child.setTo(leaf);
					}
					// recursively partition the branch
					else
					{
						follow(child, expanded, filtered);
					}
				}
			}
		}

		// adds the created node to the branch

		if (branch != null)
		{
			branch.setTo(node);
		}

		return node;
	}

	/**
	 * Selects the next attribute to visit based on the vertices' pheromone and
	 * heuristic information values.
	 *   
	 * @param vertices the list of candidate vertices (representing the
	 *        attributes to be selected).
	 * 
	 * @return the next attribute to visit. 
	 */
	private Attribute select(Collection<Vertex> vertices)
	{
		Attribute[] attributes = new Attribute[vertices.size()];
		double[] probabilities = new double[attributes.length];

		double total = 0;
		int index = 0;

		// computes the probability of each vertex

		for (Vertex vertex : vertices)
		{
			attributes[index] = vertex.getAttribute();
			probabilities[index] =
				Math.pow(vertex.getPheromone(), alpha)
					* Math.pow(vertex.getHeuristic(), beta);
			total += probabilities[index];

			index++;
		}

		if (total > 0.0)
		{
			// normalises the probability values

			for (int i = 0; i < probabilities.length; i++)
			{
				probabilities[i] /= total;
			}

			double probability = RandomGenerator.getInstance().nextDouble();
			double position = 0.0;

			for (int i = 0; i < probabilities.length; i++)
			{
				position += probabilities[i];

				if (probability <= position)
				{
					return attributes[i];
				}
			}
		}

		// there are two possibilities to reach this: either we already used
		// all the attributes or the remaining attributes have 0.0 heuristic
		// information or pheromone
		return null;
	}

	/**
	 * Verifies if the subsets, represented by the specified branches, have
	 * a reasonable number of instances.
	 * 
	 * @param branches the branches representing the subsets.
	 * @param instances the list of instances.
	 * @param index the index of attribute to be used as a filter.
	 * 
	 * @return <code>true</code> if 2 or more of the subsets have a reasonable
	 *         number of instances; <code>false</code> otherwise.
	 */
	private boolean verify(Branch[] branches, List<Instance> instances,
			Metadata metadata)
	{
		int violated = 0;

		for (Branch branch : branches)
		{
			if (Util.subset(instances, branch.getTerm(), metadata).size() < minimum)
			{
				violated++;
			}
		}

		// at least 2 of the subsets must be grater than or equal to
		// the minumum value and the number of instances should be at least
		// equal to the number of branches
		return (branches.length - violated) >= 2;
	}
}