/*
 * $Id: EntropyHelper.java 2349 2011-09-14 16:07:02Z febo $
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2010 Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.antree;

import static net.sf.myra.antree.Builder.MINIMUM_INSTANCES;

import java.util.Iterator;
import java.util.List;

import net.sf.myra.datamining.data.Attribute;
import net.sf.myra.datamining.data.ContinuousAttribute;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Label;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.data.NominalAttribute;
import net.sf.myra.datamining.data.Operator;
import net.sf.myra.datamining.data.Term;
import net.sf.myra.framework.Configuration;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision: 2349 $ $Date:: 2011-09-14 17:07:02#$
 */
public class EntropyHelper
{
	/**
	 * Attributes' frequencies per class label
	 */
	private Frequency[] frequencies;

	/**
	 * The maximum entropy value.
	 */
	private double maximum;

	/**
	 * The number of instances of the data set.
	 */
	private double length;

	/**
	 * Default constructor.
	 * 
	 * @param dataset the data set instance.
	 */
	public EntropyHelper(Dataset dataset)
	{
		this(dataset.getMetadata(), dataset.getInstances());
	}

	public EntropyHelper(Metadata metadata, List<Instance> instances)
	{
		int size = metadata.getTarget().getValues().size();
		frequencies = new Frequency[metadata.getPredictor().size()];

		Iterator<Attribute> predictor = metadata.getPredictor().iterator();

		for (int i = 0; i < frequencies.length; i++)
		{
			Attribute attribute = predictor.next();
			frequencies[i] =
				new Frequency(metadata, instances, attribute, metadata
					.getIndex(attribute.getName()), size);
		}

		for (Instance instance : instances)
		{
			length += instance.getWeight();
			Label label = instance.getLabel();

			for (int i = 0; i < size; i++)
			{
				for (int j = 0; j < frequencies.length; j++)
				{
					Attribute attribute = frequencies[j].attribute;
					double value =
						instance.values()[metadata
							.getIndex(attribute.getName())];

					if (attribute.isMissing(value))
					{
						frequencies[j].missing += instance.getWeight();
					}
					else
					{
						for (int k = 0; k < frequencies[j].terms.length; k++)
						{
							Term term = frequencies[j].terms[k];

							if (term.satisfies(value))
							{
								if (label.active(i))
								{
									frequencies[j].distribution[k][i] +=
										instance.getWeight();
								}

								frequencies[j].counter[k] +=
									instance.getWeight();
							}
						}

						if (label.active(i))
						{
							frequencies[j].overall[i] += instance.getWeight();
						}
					}
				}
			}
		}

		// fix the counter values

		for (int j = 0; j < frequencies.length; j++)
		{
			for (int k = 0; k < frequencies[j].terms.length; k++)
			{
				frequencies[j].counter[k] /= size;
			}

			frequencies[j].missing /= size;
			frequencies[j].correction /= length;
		}

		//length = dataset.getSize();
		maximum = Math.log(size) / Math.log(2.0);
	}

	/**
	 * Returns the maximum (worse) possible entropy value.
	 *  
	 * @return the maximum (worse) possible entropy value.
	 */
	public double maximum()
	{
		return maximum;
	}

	/**
	 * Returns the gain of the specified attribute.
	 * 
	 * @param attribute the attribute to determine the gain value.
	 * 
	 * @return the gain of the specified attribute.
	 */
	public double gain(Attribute attribute)
	{
		for (Frequency frequency : frequencies)
		{
			if (frequency.attribute.equals(attribute))
			{
				double factor = 1.0;

				if (frequency.missing > 0)
				{
					factor = frequency.size() / (double) length;
				}

				return (factor * gain(entropy(frequency.overall, frequency
					.size()), frequency, frequency.size()))
					- frequency.correction;

			}
		}

		throw new IllegalArgumentException("Attribute not found: " + attribute);
	}

	/**
	 * Returns the gain ratio of the specified attribute.
	 * 
	 * @param attribute the attribute to determine the gain ratio value.
	 * 
	 * @return the gain ratio of the specified attribute.
	 */
	public double gainRatio(Attribute attribute)
	{
		for (Frequency frequency : frequencies)
		{
			if (frequency.attribute.equals(attribute))
			{
				double gain = gain(attribute);
				
				if (Double.isNaN(gain) || Double.isInfinite(gain))
				{
					gain = 0.0;
				}

				double split = splitInformation(frequency.counter, length);

				if (frequency.missing > 0)
				{
					double p = frequency.missing / (double) length;
					split += (-p * (Math.log(p) / Math.log(2.0)));
				}

				return (split == 0.0) ? 0.0 : gain / split;
			}
		}

		throw new IllegalArgumentException("Attribute not found: " + attribute);
	}

	public boolean verify(Attribute attribute)
	{
		int minimum = Configuration.getIntProperty(MINIMUM_INSTANCES, 2);

		for (Frequency frequency : frequencies)
		{
			if (frequency.attribute.equals(attribute))
			{
				int valid = 0;

				for (int i = 0; i < frequency.counter.length; i++)
				{
					if (frequency.counter[i] >= minimum)
					{
						valid++;
					}
				}

				return valid >= 2;
			}
		}

		throw new IllegalArgumentException("Attribute not found: " + attribute);
	}

	//--------------------------------------------------------- Utility methods

	/**
	 * Returns the entropy of the specified distribution of values.
	 * 
	 * @param distribution the distribution of values.
	 * @param length the total number of values.
	 * 
	 * @return the entropy of the specified distribution of values.
	 */
	private double entropy(double[] distribution, double length)
	{
		double entropy = 0;

		for (int i = 0; i < distribution.length; i++)
		{
			if (distribution[i] > 0)
			{
				double p = distribution[i] / (double) length;
				entropy -= (p * (Math.log(p) / Math.log(2.0)));
			}
		}

		return entropy;
	}

	/**
	 * Returns the information gain of the specified attribute's frequency
	 * relative to the entropy value.
	 * 
	 * @param entropy the entropy value.
	 * @param frequency the attribute's frequency.
	 * @param length the total number of values.
	 * 
	 * @return the information gain of the specified attribute's frequency
	 * relative to the entropy value.
	 */
	private double gain(double entropy, Frequency frequency, double length)
	{
		for (int i = 0; i < frequency.terms.length; i++)
		{
			if (frequency.counter[i] > 0)
			{
				entropy -=
					(frequency.counter[i] / (double) length)
						* entropy(frequency.distribution[i],
							frequency.counter[i]);
			}
		}

		return entropy;
	}

	/**
	 * Returns the split information of the specified distribution.
	 * 
	 * @param counter the distribution of values.
	 * @param length the total number of values.
	 * 
	 * @return the split information of the specified distribution.
	 */
	private double splitInformation(double[] counter, double length)
	{
		double information = 0;

		for (int i = 0; i < counter.length; i++)
		{
			if (counter[i] > 0)
			{
				double p = counter[i] / (double) length;
				information += (p * (Math.log(p) / Math.log(2.0)));
			}
		}

		return (-information);
	}

	/**
	 * This class represents the frequency information of an attribute and its
	 * values.
	 */
	private static class Frequency
	{
		/**
		 * The set of terms.
		 */
		Term[] terms;

		/**
		 * The class frequency distribution per term.
		 */
		double[][] distribution;

		/**
		 * The number of instances per term.
		 */
		double[] counter;

		/**
		 * Overall distribution for this attribute.
		 */
		double[] overall;

		/**
		 * The number of instance with missing values.
		 */
		double missing;

		/**
		 * The attribute represented.
		 */
		Attribute attribute;

		/**
		 * The gain correction value (this is based on J48 algorithm). 
		 */
		double correction = 0.0;

		/**
		 * Default constructor.
		 * 
		 * @param dataset the data set instance.
		 * @param attribute the attribute to be represented.
		 * @param index the index of the attribute.
		 * @param length the number of class labels.
		 */
		Frequency(Metadata metadata, List<Instance> instances,
				Attribute attribute, int index, int length)
		{
			this.attribute = attribute;

			if (attribute instanceof NominalAttribute)
			{
				NominalAttribute nominal = (NominalAttribute) attribute;
				terms = new Term[nominal.getValues().size()];

				for (int i = 0; i < terms.length; i++)
				{
					terms[i] = new Term(attribute.getName(), index);
					terms[i].setOperator(Operator.EQUAL_TO);

					String value = nominal.getValue(i);
					terms[i].setDescription(value);
					terms[i].setValues(nominal.getIndex(value));
				}
			}
			else if (attribute instanceof ContinuousAttribute)
			{
				ContinuousAttribute continuous =
					(ContinuousAttribute) attribute;
				terms = ThresholdHelper.create(metadata, continuous, instances);

				correction =
					(Math.log(ThresholdHelper.count(metadata, continuous,
						instances)) / Math.log(2.0));
			}

			distribution = new double[terms.length][length];
			counter = new double[terms.length];
			overall = new double[length];
			missing = 0;
		}

		double size()
		{
			double total = 0;

			for (int i = 0; i < counter.length; i++)
			{
				total += counter[i];
			}

			return total;
		}
	}
}