/*
 * $Id$ 
 */

package net.sf.myra.antree;

import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;
import java.util.TreeSet;

import net.sf.myra.antree.representation.LeafNode;
import net.sf.myra.antree.representation.Node;
import net.sf.myra.antree.representation.Tree;
import net.sf.myra.datamining.Evaluator;
import net.sf.myra.datamining.data.Attribute;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.data.Term;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                    #$
 */
public class EntropyPruner
	implements Pruner
{
	private ErrorBasedPruner pruner;

	public EntropyPruner()
	{
		pruner = new ErrorBasedPruner();
	}

	public Tree prune(Tree tree, Dataset dataset)
	{
		Metadata metadata = dataset.getMetadata();
		TreeSet<Node> candidates = new TreeSet<Node>(new Comparator<Node>()
		{
			public int compare(Node o1, Node o2)
			{
				int c = Double.compare(o1.getLevel(), o2.getLevel());

				if (c == 0 && !o1.equals(o2))
				{
					return o1.toString().compareTo(o2.toString());
				}

				return c;
			}
		});

		for (Node leaf : tree.leaves())
		{
			if (leaf.getParent() != null)
			{
				candidates.add(leaf.getParent());
			}
		}

		while (!candidates.isEmpty())
		{
			Node node = candidates.last();
			Term[] path = collapse(node);
			List<Instance> filtered = dataset.getInstances();

			for (Term term : path)
			{
				filtered = Util.split(filtered, term, metadata);
			}

			EntropyHelper helper = new EntropyHelper(metadata, filtered);
			double gain = 0.0;
			int count = 0;

			// accumulates the gain value of the available attributes

			for (Attribute attribute : metadata.getPredictor())
			{
				if (helper.verify(attribute))
				{
					double v = helper.gain(attribute);

					if (v > 0.0 && !Double.isInfinite(v) && !Double.isNaN(v))
					{
						gain += helper.gain(attribute);
						count++;
					}
				}
			}

			if (count > 0)
			{
				gain = (gain / (double) count) - 1E-3;
			}

			// if the gain of the node's attribute is smaller than the average
			// gain at this level, we replace the node for a leaf node

			if (gain <= 0.0 || helper.gain(node.getAttribute()) < gain)
			{
				LeafNode leaf =
					new LeafNode(node.getLevel(), metadata, Evaluator
						.findMajorityClass(filtered, metadata));

				if (node.getParent() != null)
				{
					node.getParent().getBranch(node).setTo(leaf);
					candidates.add(leaf.getParent());
				}
				// we must be dealing with the root of the tree
				else
				{
					tree.setRoot(leaf);
					// if we replaced the tree for a leaf, we do not have
					// more leaf nodes to evaluate
					candidates.clear();
				}
			}

			candidates.remove(node);
		}

		return pruner.prune(tree, dataset);
	}

	private Term[] collapse(Node node)
	{
		LinkedList<Term> path = new LinkedList<Term>();

		while (node.getParent() != null)
		{
			Node parent = node.getParent();
			path.addFirst(parent.getBranch(node).getTerm());
			node = parent;
		}

		return path.toArray(new Term[path.size()]);
	}
}