/*
 * $Id: ReducedErrorMetric.java 2329 2011-08-02 08:57:44Z febo $
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2010 Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.antree;

import net.sf.myra.antree.representation.Branch;
import net.sf.myra.antree.representation.InternalNode;
import net.sf.myra.antree.representation.Node;
import net.sf.myra.antree.representation.Statistics;
import net.sf.myra.framework.Configuration;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision: 2329 $ $Date:: 2011-08-02 09:57:44#$
 */
public class ErrorBasedMetric
{
	/**
	 * The property key under which the reduced-error pruner confidence level
	 * is stored.
	 */
	public static final String CONFIDENCE_LEVEL =
		"net.sf.myra.antree.error.confidence.level";

	/**
	 * The confidence level constant.
	 */
	private static final double DEFAULT_CONFIDENCE_LEVEL = 0.25;

	/**
	 * The values used to compute the coefficient value.
	 */
	private static final double values[] =
		{ 0, 0.001, 0.005, 0.01, 0.05, 0.10, 0.20, 0.40, 1.00 };

	/**
	 * The deviation used to compute the coefficient value.
	 */
	private static final double deviation[] =
		{ 4.0, 3.09, 2.58, 2.33, 1.65, 1.28, 0.84, 0.25, 0.00 };

	/**
	 * The singleton instance.
	 */
	private static final ErrorBasedMetric singleton =
		new ErrorBasedMetric();

	/**
	 * The confidence level value.
	 */
	private double confidence;

	/**
	 * The coefficient value.
	 */
	private double coefficient;

	/**
	 * Default constructor.
	 */
	public ErrorBasedMetric()
	{
		confidence =
			Configuration.getDoubleProperty(CONFIDENCE_LEVEL,
				DEFAULT_CONFIDENCE_LEVEL);

		// computes the coefficient value based on C4.5

		int i = 0;

		while (confidence > values[i])
		{
			i++;
		}

		coefficient =
			deviation[i - 1] + (deviation[i] - deviation[i - 1])
				* (confidence - values[i - 1]) / (values[i] - values[i - 1]);
		coefficient = coefficient * coefficient;
	}

	/**
	 * Returns the predicted error of the subtree represented by the
	 * specified node.
	 * 
	 * @param node the node reference.
	 * @param statistics the statistics of the node.
	 * 
	 * @return the predicted error of the subtree represented by the
	 *         specified node.
	 */
	public double error(Node node, Statistics statistics)
	{
		double error = 0;

		if (node.isLeaf())
		{
			error = statistics.getErrors(node, node.getPrediction());
			return error + ucf(error, statistics.getTotal(node));
		}
		else
		{
			for (Branch branch : ((InternalNode) node).getBranches())
			{
				error += error(branch.getTo(), statistics);
			}

			return error;
		}
	}
	
	/**
	 * Returns the (raw) error of the subtree represented by the
	 * specified node.
	 * 
	 * @param node the node reference.
	 * @param statistics the statistics of the node.
	 * 
	 * @return the (raw) error of the subtree represented by the
	 *         specified node.
	 */
	public double rawError(Node node, Statistics statistics)
	{
		double error = 0;

		if (node.isLeaf())
		{
			return statistics.getErrors(node, node.getPrediction());
		}
		else
		{
			for (Branch branch : ((InternalNode) node).getBranches())
			{
				error += rawError(branch.getTo(), statistics);
			}

			return error;
		}
	}

	/**
	 * Returns the estimate error rate value based on the upper confidence
	 * level. This implementation is based on C4.5 algorithm.
	 * 
	 * @param error the number of misclassified instances.
	 * @param total the total number of instances.
	 * 
	 * @return the estimate error rate value.
	 */
	public double ucf(double error, double total)
	{
		if (error < 1E-6)
		{
			return total * (1 - Math.exp(Math.log(confidence) / total));
		}
		else if (error < 0.9999)
		{
			double v = total * (1 - Math.exp(Math.log(confidence) / total));
			return v + error * (ucf(1.0, total) - v);
		}
		else if (error + 0.5 >= total)
		{
			return 0.67 * (total - error);
		}
		else
		{
			double pr =
				(error + 0.5 + coefficient / 2 + Math
					.sqrt(coefficient
						* ((error + 0.5) * (1 - (error + 0.5) / total) + coefficient / 4)))
					/ (total + coefficient);
			return (total * pr - error);
		}
	}

	/**
	 * Returns the singleton instance.
	 * 
	 * @return the singleton instance.
	 */
	public static ErrorBasedMetric getInstance()
	{
		return singleton;
	}
}