/*
 * $Id: Graph.java 2344 2011-08-29 11:46:17Z febo $
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2010 Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.antree;

import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;

import net.sf.myra.antree.representation.Branch;
import net.sf.myra.antree.representation.InternalNode;
import net.sf.myra.antree.representation.Node;
import net.sf.myra.antree.representation.Tree;
import net.sf.myra.datamining.data.Attribute;
import net.sf.myra.datamining.data.Metadata;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision: 2344 $ $Date:: 2011-08-29 12:46:17#$
 */
public class Graph
{
	/**
	 * The (dynamic) pheromone matrix.
	 */
	protected double[][] matrix;

	/**
	 * The cache for <code>Entry</code> objects [encoding,entry].
	 */
	protected Map<Integer, Entry> cache;

	/**
	 * The attribute array representing the vertices.
	 */
	protected Attribute[] vertices;

	/**
	 * The mapping of attribute indexes.
	 */
	protected Map<String, Integer> mapping;

	/**
	 * Template pheromone array. This array is used for new entries on the
	 * pheromone matrix.
	 */
	protected double[] template;

	/**
	 * The heuristic information.
	 */
	protected Heuristic heuristic;

	/**
	 * Creates a new <code>Graph</code> instance.
	 * 
	 * @param metadata the current dataset's metadata.
	 * @param heuristic the heuristic information.
	 */
	public Graph(Metadata metadata, Heuristic heuristic)
	{
		this.heuristic = heuristic;

		Collection<Attribute> predictor = metadata.getPredictor();
		vertices = new Attribute[predictor.size()];
		int current = 0;

		for (Attribute attribute : predictor)
		{
			vertices[current] = attribute;
			current++;
		}

		template = new double[vertices.length];

		for (int i = 0; i < template.length; i++)
		{
			template[i] = 1 / (double) template.length;
		}

		mapping = new HashMap<String, Integer>(vertices.length, 1.0f);

		for (int i = 0; i < vertices.length; i++)
		{
			mapping.put(vertices[i].getName(), i);
		}

		// although the matrix is an array, it is resized to fit more entries
		// when necessary
		matrix = new double[metadata.getAttributes().size()][template.length];

		for (int i = 0; i < matrix.length; i++)
		{
			matrix[i] = template.clone();
		}

		cache = new HashMap<Integer, Entry>(matrix.length, 1.0f);
		cache.put(0, Entry.ROOT);
	}

	/**
	 * Returns the collection of vertices for the specified branch. If the
	 * branch is <code>null</code>, it returns the collection of vertices at
	 * the start of the graph.
	 * 
	 * @param branch the current branch being followed.
	 * 
	 * @return the collection of vertices for the specified branch.
	 */
	@SuppressWarnings("unchecked")
	public Collection<Vertex> getVertices(Branch branch)
	{
		return getVertices(branch, Collections.EMPTY_SET);
	}

	/**
	 * Returns the collection of vertices for the specified branch. If the
	 * branch is <code>null</code>, it returns the collection of vertices at
	 * the start of the graph.
	 * 
	 * @param branch the current branch being followed.
	 * @param used the list of used attributes, which cannot be reused.
	 * 
	 * @return the collection of vertices for the specified branch.
	 */
	public Collection<Vertex> getVertices(Branch branch,
			Collection<Attribute> used)
	{
		Entry entry =
			(branch == null) ? Entry.ROOT : cache.get(branch.getCode());
		double[] pheromone = null;

		if (entry != null)
		{
			entry.increment();
			pheromone = matrix[entry.getIndex()];
		}
		else
		{
			pheromone = template.clone();
		}

		ArrayList<Vertex> current = new ArrayList<Vertex>(pheromone.length);

		for (int i = 0; i < pheromone.length; i++)
		{
			if (!used.contains(vertices[i]))
			{
				Vertex v =
					new Vertex(vertices[i], pheromone[i], heuristic.valueFor(
						branch, vertices[i]));
				current.add(v);
			}
		}

		return current;
	}

	/**
	 * Updates the pheromone values of the branches used in the specified
	 * tree. <b>Note:</b> This method is not thread-safe.
	 * 
	 * @param tree the tree used to update the pheromone values.
	 */
	public void update(Tree tree)
	{
		Node root = tree.getRoot();

		// TODO what to do when the root is a leaf node?
		if (!root.isLeaf())
		{
			int index = mapping.get(root.getAttribute().getName());
			double value = matrix[Entry.ROOT.getIndex()][index];

			matrix[Entry.ROOT.getIndex()][index] =
				value + (value * tree.getQuality().doubleValue());

			LinkedList<Branch> branches = new LinkedList<Branch>();
			branches.addAll(Arrays.asList(((InternalNode) root).getBranches()));

			while (!branches.isEmpty())
			{
				Branch branch = branches.removeFirst();

				if (!branch.getTo().isLeaf())
				{
					Entry entry = getEntry(branch);

					index =
						mapping.get(branch.getTo().getAttribute().getName());
					value = matrix[entry.getIndex()][index];

					matrix[entry.getIndex()][index] =
						value + (value * tree.getQuality().doubleValue());

					branches
						.addAll(Arrays.asList(branch.getTo().getBranches()));
				}
			}
		}
	}

	/**
	 * Evaporates the pheromone values of each branch by normalising the values.
	 * Therefore, the values that have not been updated are decreased.
	 */
	public void evaporate()
	{
		for (int i = 0; i < matrix.length; i++)
		{
			double total = 0.0;

			for (int j = 0; j < matrix[i].length; j++)
			{
				total += matrix[i][j];
			}

			for (int j = 0; j < matrix[i].length; j++)
			{
				matrix[i][j] /= total;
			}
		}
	}

	/**
	 * Returns a correspondent <code>Entry</code> object for the specified
	 * branch. <b>Note:</b> This implementation is not thread-safe.
	 * 
	 * @param branch the branch to return the entry.
	 * 
	 * @return a correspondent <code>Entry</code> object for the specified
	 *         branch.
	 */
	protected Entry getEntry(Branch branch)
	{
		int code = branch.getCode();
		Entry entry = cache.get(code);

		if (entry == null)
		{
			entry = new Entry(cache.size());
			cache.put(code, entry);

			if (entry.getIndex() >= matrix.length)
			{
				double[][] clone =
					new double[matrix.length + template.length][];

				for (int i = 0; i < matrix.length; i++)
				{
					clone[i] = matrix[i].clone();
				}

				for (int i = matrix.length; i < clone.length; i++)
				{
					clone[i] = template.clone();
				}

				matrix = clone;
			}
		}

		return entry;
	}

	public String stats()
	{
		DecimalFormat formatter = new DecimalFormat("#0.00");

		long max = Long.MIN_VALUE;
		long min = Long.MAX_VALUE;
		// 0 - < 2 hits
		// 1 - < 5
		// 2 - < 10
		// 3 - < 15
		// 4 - hits >= 20
		double[] average = new double[5];

		for (Entry entry : cache.values())
		{
			if (entry != cache.get(0))
			{
				long hits = entry.hits();

				if (hits > max)
				{
					max = hits;
				}

				if (hits < min)
				{
					min = hits;
				}

				int index = 0;

				if (hits < 2)
				{
					index = 0;
				}
				else if (hits < 5)
				{
					index = 1;
				}
				else if (hits < 10)
				{
					index = 2;
				}
				else if (hits < 15)
				{
					index = 3;
				}
				else if (hits < 20)
				{
					index = 4;
				}

				average[index]++;
			}
		}

		for (int i = 0; i < average.length; i++)
		{
			average[i] /= cache.size();
		}

		StringBuffer buffer = new StringBuffer();
		buffer.append("Pheromone matrix entries: ");
		buffer.append(cache.size());
		buffer.append("\nAverage hits per entry: ");

		buffer.append(formatter.format(average[0]));
		buffer.append(" | ");
		buffer.append(formatter.format(average[1]));
		buffer.append(" | ");
		buffer.append(formatter.format(average[2]));
		buffer.append(" | ");
		buffer.append(formatter.format(average[3]));
		buffer.append(" | ");
		buffer.append(formatter.format(average[4]));

		buffer.append("\nMax entry hits: ");
		buffer.append(max);
		buffer.append("\nMin entry hits: ");
		buffer.append(min);

		return buffer.toString();
	}

	/**
	 * This class is used as the key of an entry in the pheromone matrix.
	 */
	public static class Entry
	{
		/**
		 * The <code>Id</code> instance representing the start of the graph.
		 */
		public static final Entry ROOT = new Entry(0);

		/**
		 * The index of this entry.
		 */
		private int index;

		/**
		 * The number of times that this id was used for statistics purpose.
		 */
		private long hits;

		/**
		 * Creates a new <code>Entry</code> instance.
		 * 
		 * @param index the index of this entry in the pheromone matrix.
		 */
		Entry(int index)
		{
			this.index = index;
		}

		/**
		 * Returns the index of this entry in the pheromone matrix. 
		 * 
		 * @return the index of this entry in the pheromone matrix.
		 */
		public int getIndex()
		{
			return index;
		}

		/**
		 * Returns the number of times that this id was used.
		 * 
		 * @return the number of times that this id was used.
		 */
		public long hits()
		{
			return hits;
		}

		/**
		 * Increments the number of hits by one.
		 */
		public void increment()
		{
			hits++;
		}
	}

	/**
	 * Struct-like class that represents a vertex of the construction graph.
	 */
	public static class Vertex
	{
		/**
		 * The attribute represented by this vertex.
		 */
		private Attribute attribute;

		/**
		 * The pheromone associted with the vertex.
		 */
		private double pheromone;

		/**
		 * The heuristic information associated with the vertex.
		 */
		private double heuristic;

		/**
		 * Default constructor.
		 * 
		 * @param attribute the attribute represented by the vertex.
		 * @param pheromone the pheromone to be associated with the vertex.
		 * @param heuristic the heuristic information to be associated with
		 *        the vertex.
		 */
		private Vertex(Attribute attribute, double pheromone, double heuristic)
		{
			this.attribute = attribute;
			this.pheromone = pheromone;
			this.heuristic = heuristic;
		}

		/**
		 * Returns the attribute represented by this vertex.
		 * 
		 * @return the attribute represented by this vertex.
		 */
		public Attribute getAttribute()
		{
			return attribute;
		}

		/**
		 * Returns the pheromone associated with this vertex.
		 * 
		 * @return the pheromone associated with this vertex.
		 */
		public double getPheromone()
		{
			return pheromone;
		}

		/**
		 * Returns the heuristic information associated with this vertex.
		 * 
		 * @return the heuristic information associated with this vertex.
		 */
		public double getHeuristic()
		{
			return heuristic;
		}

		@Override
		public String toString()
		{
			return attribute.getName();
		}
	}
}