/*
 * $Id$ 
 */

package net.sf.myra.antree;

import static net.sf.myra.antree.Builder.MINIMUM_INSTANCES;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import net.sf.myra.antree.representation.Branch;
import net.sf.myra.antree.representation.InternalNode;
import net.sf.myra.antree.representation.LeafNode;
import net.sf.myra.antree.representation.Node;
import net.sf.myra.antree.representation.NodeFactory;
import net.sf.myra.antree.representation.Statistics;
import net.sf.myra.antree.representation.Tree;
import net.sf.myra.datamining.Evaluator;
import net.sf.myra.datamining.data.Attribute;
import net.sf.myra.datamining.data.ContinuousAttribute;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Label;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.io.Helper;
import net.sf.myra.datamining.measure.AccuracyMeasure;
import net.sf.myra.framework.Configuration;

/**
 * This class is reponsible for creating the initial tree (using a similar
 * approach as the C4.5 algorithm) that will be used to deposit the initial
 * pheromone.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                    #$
 */
public class Initialiser
{
	/**
	 * The minimum number of instances required for each branch.
	 */
	private final int minimum =
		Configuration.getIntProperty(MINIMUM_INSTANCES, 2);

	/**
	 * The current data set.
	 */
	private Dataset dataset;

	/**
	 * The current metadata.
	 */
	private Metadata metadata;

	/**
	 * The majority class label of the data set.
	 */
	private Label majority;

	/**
	 * The heuristic information gathered during the tree induction process.
	 */
	private C45Heuristic heuristic;

	/**
	 * Default constructor.
	 * 
	 * @param dataset the data set to be covered.
	 */
	public Initialiser(Dataset dataset)
	{
		this(dataset, null);
	}

	/**
	 * Default constructor.
	 * 
	 * @param dataset the data set to be covered.
	 * @param heuristic indicates whether we are collecting heuristic information
	 *                  or not.
	 */
	public Initialiser(Dataset dataset, C45Heuristic heuristic)
	{
		this.dataset = dataset;
		this.heuristic = heuristic;
		this.metadata = dataset.getMetadata();
		this.majority = Evaluator.findMajorityClass(dataset);

		// this is necessary, so then we can adjust the threshold values the
		// same way that C4.5 does
		dataset.computeDomains();
	}

	/**
	 * Returns a decision tree.
	 * 
	 * @return a decision tree. 
	 */
	@SuppressWarnings("unchecked")
	public Tree build()
	{
		Tree tree =
			new Tree(
				follow(null, Collections.EMPTY_SET, dataset.getInstances()));

		ErrorBasedPruner pruner = new ErrorBasedPruner();
		return pruner.prune(tree, dataset);
	}

	/**
	 * Recursively follows the specified branch, adding nodes until a leaf
	 * node is reached.
	 *  
	 * @param branch the branch to follow.
	 * @param used the list of used attributes.
	 * @param instances the list of instances.
	 */
	private Node follow(Branch branch, Collection<Attribute> used,
			List<Instance> instances)
	{
		Dataset current = new Dataset(metadata);
		current.setInstances(instances);

		// if all instances have the same class label we just create a
		// leaf node
		if (Evaluator.sameLabel(instances))
		{
			if (branch == null)
			{
				return new LeafNode(Node.INITIAL_LEVEL, metadata, Evaluator
					.findMajorityClass(instances, metadata, majority));
			}

			Node node =
				new LeafNode(branch.getFrom().getLevel() + 1, metadata,
					Evaluator.findMajorityClass(instances, metadata, majority));
			branch.setTo(node);

			return node;
		}

		Attribute attribute = select(current, used, branch);
		int level = (branch == null) ? 0 : branch.getFrom().getLevel() + 1;

		Node node = null;

		// do we have an attribute?
		if (attribute == null)
		{
			node =
				new LeafNode(level, metadata, Evaluator.findMajorityClass(
					instances, metadata, majority));
		}
		else
		{
			node =
				NodeFactory.create(attribute, dataset.getMetadata(), instances,
					level);

			// do the generated subsets have a reasonable number of instances?
			if (!verify(((InternalNode) node).getBranches(), instances,
				metadata))
			{
				node =
					new LeafNode(level, metadata, Evaluator.findMajorityClass(
						instances, metadata, majority));
			}
			// yes, so let's create the partitions
			else
			{
				InternalNode internal = (InternalNode) node;

				for (Branch child : internal.getBranches())
				{
					Collection<Attribute> expanded =
						new HashSet<Attribute>(used);

					// continuous attributes can be selected more than one
					// time per path to allow multiple discretisation points
					if (!(attribute instanceof ContinuousAttribute))
					{
						expanded.add(attribute);
					}

					List<Instance> filtered =
						Util.split(instances, child.getTerm(), metadata);

					// if the list of instances is empty, we add a leaf node
					// with the majority class label of the parent node
					if (filtered.isEmpty())
					{
						LeafNode leaf =
							new LeafNode(level + 1, dataset.getMetadata(),
								Evaluator.findMajorityClass(instances,
									metadata, majority));
						child.setTo(leaf);
					}
					// if the number of instances is smaller than 2 times the
					// minimum size, we automatically add a leaf node
					else if (filtered.size() < (2 * minimum))
					{
						LeafNode leaf =
							new LeafNode(level + 1, metadata,
								Evaluator.findMajorityClass(filtered, metadata,
									majority));
						child.setTo(leaf);
					}
					// recursively partition the branch
					else
					{
						follow(child, expanded, filtered);
					}
				}
			}
		}

		// adds the created node to the branch

		if (branch != null)
		{
			branch.setTo(node);
		}

		return node;
	}

	private Attribute select(Dataset current, Collection<Attribute> used,
			Branch branch)
	{
		EntropyHelper helper = new EntropyHelper(current);
		double gain = 0.0;
		int count = 0;

		// accumulates the gain value of the available attributes
		Set<String> valid = new HashSet<String>();

		for (Attribute attribute : current.getMetadata().getPredictor())
		{
			if (!used.contains(attribute) && helper.verify(attribute))
			{
				double v = helper.gain(attribute);

				if (v > 0.0 && !Double.isInfinite(v) && !Double.isNaN(v))
				{
					valid.add(attribute.getName());
					gain += helper.gain(attribute);
					count++;
				}
			}
		}

		gain = (gain / (double) count) - 1E-3;

		// we now selects the attribute with the highest gain ratio and gain
		// greater or equal to the average gain

		Attribute selected = null;
		double maxGainRatio = 0.0;

		double[] information =
			new double[current.getMetadata().getPredictor().size()];
		Arrays.fill(information, 0.0);
		int index = 0;

		for (Attribute attribute : current.getMetadata().getPredictor())
		{
			if (!used.contains(attribute)
				&& valid.contains(attribute.getName()))
			{
				double gainRatio = helper.gainRatio(attribute);

				if (helper.gain(attribute) >= gain)
				{
					information[index] += gainRatio;

					if (gainRatio > (maxGainRatio + 1e-6))
					{
						selected = attribute;
						maxGainRatio = gainRatio;
					}
				}
			}

			index++;
		}

		if (heuristic != null)
		{
			heuristic.set(branch, information);
		}

		return selected;
	}

	/**
	 * Verifies if the subsets, represented by the specified branches, have
	 * a reasonable number of instances.
	 * 
	 * @param branches the branches representing the subsets.
	 * @param instances the list of instances.
	 * @param index the index of attribute to be used as a filter.
	 * 
	 * @return <code>true</code> if 2 or more of the subsets have a reasonable
	 *         number of instances; <code>false</code> otherwise.
	 */
	private boolean verify(Branch[] branches, List<Instance> instances,
			Metadata metadata)
	{
		int violated = 0;

		for (Branch branch : branches)
		{
			if (Util.filter(instances, branch.getTerm(), metadata).size() < minimum)
			{
				violated++;
			}
		}

		// at least 2 of the subsets must be grater than the minumum value
		return (branches.length - violated) >= 2;
	}

	public static void main(String[] args)
		throws Exception
	{
		Dataset dataset = Helper.open(args[0]);
		Initialiser init = new Initialiser(dataset);

		Tree tree = init.build();
		Statistics statistics = tree.gather(dataset);
		tree.setStatistics(statistics);

		System.out.println(tree.toString(statistics));

		AccuracyMeasure measure1 = new AccuracyMeasure();
		System.out.println(measure1.evaluate(dataset, tree));
		
		ErrorBasedMeasure measure2 = new ErrorBasedMeasure();
		System.out.println(measure2.evaluate(dataset, tree));
	}
}