/*
 * $Id: LeafNodePruner.java 2323 2011-01-25 13:14:50Z febo $
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2010 Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.antree;

import static net.sf.myra.antree.AntTreeMiner.MEASURE;

import java.util.Comparator;
import java.util.TreeSet;

import net.sf.myra.antree.representation.LeafNode;
import net.sf.myra.antree.representation.Node;
import net.sf.myra.antree.representation.Statistics;
import net.sf.myra.antree.representation.Tree;
import net.sf.myra.datamining.Measure;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.measure.AccuracyMeasure;
import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.ObjectFactory;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision: 2323 $ $Date:: 2011-01-25 13:14:50#$
 */
public class LeafNodePruner
	implements Pruner
{
	private static final TreeComparator COMPARATOR = new TreeComparator();

	/**
	 * The measure used to evaluate the trees.
	 */
	protected Measure measure;

	/**
	 * Default constructor.
	 */
	public LeafNodePruner()
	{
		this((Measure) ObjectFactory.create(Configuration.getProperty(
			MEASURE, AccuracyMeasure.class.getName())));
	}

	/**
	 * Creates a new <code>LeafNodePruner</code> instance.
	 * 
	 * @param measure the measure to be used during pruning.
	 */
	public LeafNodePruner(Measure measure)
	{
		this.measure = measure;
	}

	public Tree prune(Tree tree, Dataset dataset)
	{
		tree.setStatistics(tree.gather(dataset));
		tree.setQuality(measure.evaluate(dataset, tree));

		if (!tree.getRoot().isLeaf())
		{
			TreeSet<Node> candidates = candidates(tree);

			while (!candidates.isEmpty())
			{
				Node node = candidates.first();
				Statistics statistics = tree.gather(dataset);

				// (1) replaces the subtree by a leaf node
				Tree leafClone = tree.clone();
				Node parent = leafClone.get(tree.getIndex(node));

				LeafNode leaf =
					new LeafNode(parent.getLevel(), dataset.getMetadata(),
						statistics.getMajorityLabel(node));

				if (parent.getParent() != null)
				{
					parent.getParent().getBranch(parent).setTo(leaf);
				}
				else
				{
					leafClone.setRoot(leaf);
				}

				leafClone.adjust(dataset);
				leafClone.setStatistics(leafClone.gather(dataset));
				leafClone.setQuality(measure.evaluate(dataset, leafClone));

				// (2) replaces the subtree by its most used branch
				Tree branchClone = tree.clone();
				replace(node, tree, statistics, branchClone);

				branchClone.adjust(dataset);
				branchClone.setStatistics(branchClone.gather(dataset));
				branchClone.setQuality(measure.evaluate(dataset, branchClone));

				if (COMPARATOR.compare(leafClone, tree) >= 0
					&& COMPARATOR.compare(leafClone, branchClone) >= 0)
				{
					tree = leafClone;
					candidates = candidates(tree);
				}
				else if (COMPARATOR.compare(branchClone, tree) >= 0)
				{
					tree = branchClone;
					candidates = candidates(tree);
				}

				candidates.remove(node);
			}
		}

		return tree;
	}

	/**
	 * Replaces the specified node for its most used branch and returns the
	 * resultant tree.
	 * 
	 * @param node the node reference.
	 * @param tree the current tree.
	 * @param statistics the statistics of the tree.
	 * 
	 * @return a tree with the node replaced by its most used branch.
	 */
	private Node replace(Node node, Tree tree, Statistics statistics, Tree clone)
	{
		int branch = 0;

		for (int i = 1; i < node.getBranches().length; i++)
		{
			if (statistics.getTotal(node.getBranch(branch).getTo()) < statistics
				.getTotal(node.getBranch(i).getTo()))
			{
				branch = i;
			}
		}

		Node nodeClone = clone.get(tree.getIndex(node));
		Node branchNode = nodeClone.getBranch(branch).getTo();

		if (nodeClone.getParent() != null)
		{
			nodeClone.getParent().getBranch(nodeClone).setTo(branchNode);
			branchNode.setLevel(node.getLevel());
		}
		else
		{
			branchNode.setLevel(Node.INITIAL_LEVEL);
			clone.setRoot(branchNode);
		}

		return branchNode;
	}

	/**
	 * Returns the list of candidate nodes for pruning.
	 *  
	 * @param tree the tree undergoing pruning.
	 * 
	 * @return the list of candidate nodes for pruning.
	 */
	private TreeSet<Node> candidates(Tree tree)
	{
		TreeSet<Node> candidates = new TreeSet<Node>(new Comparator<Node>()
		{
			public int compare(Node o1, Node o2)
			{
				int c = Double.compare(o1.getLevel(), o2.getLevel());

				if (c == 0 && !o1.equals(o2))
				{
					return o1.toString().compareTo(o2.toString());
				}

				return c;
			}
		});

		for (Node leaf : tree.leaves())
		{
			if (leaf.getParent() != null)
			{
				candidates.add(leaf.getParent());
			}
		}

		return candidates;
	}
}