/*
 * $Id: MaxMinGraph.java 2339 2011-08-15 15:08:03Z febo $
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2010 Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.antree;

import java.util.Arrays;
import java.util.HashSet;
import java.util.LinkedList;

import net.sf.myra.antree.representation.Branch;
import net.sf.myra.antree.representation.InternalNode;
import net.sf.myra.antree.representation.Node;
import net.sf.myra.antree.representation.Tree;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.framework.Configuration;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision: 2339 $ $Date:: 2011-08-15 16:08:03#$
 */
public class MaxMinGraph
	extends Graph
{
	/**
	 * The property key under which the evaporation factor is stored.
	 */
	public static final String EVAPORATION_FACTOR =
		"net.sf.myra.antree.maxmin.evaporation";

	/**
	 * The property key under which the p_best probability is stored.
	 */
	public static final String P_BEST = "net.sf.myra.antree.maxmin.pbest";

	/**
	 * The initial pheromone value (this is an arbitrary 'big' value).
	 */
	public static final double INITIAL = 10.0;
	
	/**
	 * The fraction of the tree quality used to update the pheromones.
	 */
	private static final double FRACTION = 10.0;

	/**
	 * The pheromone evaporation factor (default 0.90).
	 */
	private double factor;

	/**
	 * The p_best probability (default 0.05) to estimate the tMin value.
	 */
	private double pBest;

	/**
	 * The maximum pheromone value.
	 */
	private double tMax;

	/**
	 * The minimum pheromone value.
	 */
	private double tMin;

	/**
	 * The quality of the current best solution.
	 */
	private double currentBest = 0.0;

	/**
	 * Creates a new <code>MaxMinGraph</code> instance.
	 * 
	 * @param metadata the current dataset's metadata.
	 * @param heuristic the heuristic information.
	 */
	public MaxMinGraph(Metadata metadata, Heuristic heuristic)
	{
		super(metadata, heuristic);

		// at the begining of the search, tMax = tMin, so only the heuristic
		// will be used
		tMax = INITIAL;
		tMin = INITIAL;

		factor = Configuration.getDoubleProperty(EVAPORATION_FACTOR, 0.90);
		pBest = Configuration.getDoubleProperty(P_BEST, 0.05);

		// the template for new entries on the pheromone matrix is also set to
		// INITIAL, so the choice will depend only on the heuristic information
		for (int i = 0; i < template.length; i++)
		{
			template[i] = INITIAL;
		}

		for (int i = 0; i < matrix.length; i++)
		{
			matrix[i] = template.clone();
		}
	}

	@Override
	public void update(Tree tree)
	{
		// if we have found a new best solution, updates the values of tMax
		// and tMin
		if (tree.getQuality().doubleValue() > currentBest)
		{
			currentBest = tree.getQuality().doubleValue();

			int n = vertices.length;
			// the average should represent the average number of options an
			// ant has at each choice point
			double average = (n / 2.0) * tree.size();
			double pDec = Math.pow(pBest, 1 / (double) n);

			tMax = (1 / (1 - factor)) * (currentBest / FRACTION);
			tMin = (tMax * (1 - pDec)) / ((average - 1) * pDec);

			if (tMin > tMax)
			{
				tMin = tMax;
			}
		}

		HashSet<Integer> updated = new HashSet<Integer>();
		double delta = tree.getQuality().doubleValue() / FRACTION;

		// updates the branches of the iteration best solution

		Node root = tree.getRoot();

		if (!root.isLeaf())
		{
			updated.add(Entry.ROOT.getIndex());

			int index = mapping.get(root.getAttribute().getName());
			update(matrix[Entry.ROOT.getIndex()], index, delta);

			LinkedList<Branch> branches = new LinkedList<Branch>();
			branches.addAll(Arrays.asList(((InternalNode) root).getBranches()));

			while (!branches.isEmpty())
			{
				Branch branch = branches.removeFirst();

				if (!branch.getTo().isLeaf())
				{
					Entry entry = getEntry(branch);
					updated.add(entry.getIndex());

					index =
						mapping.get(branch.getTo().getAttribute().getName());
					update(matrix[entry.getIndex()], index, delta);

					branches
						.addAll(Arrays.asList(branch.getTo().getBranches()));
				}
			}
		}

		// evaporates the pheromone of the unused branches

		for (int i = 0; i < matrix.length; i++)
		{
			if (!updated.contains(i))
			{
				update(matrix[i], -1, 0.0);
			}
		}
	}

	@Override
	public void evaporate()
	{
		// the evaporation is implicit in the pheromone update
	}

	/**
	 * Returns the maximum pheromone value allowed.
	 * 
	 * @return the maximum pheromone value allowed.
	 */
	public double getTMax()
	{
		return tMax;
	}

	/**
	 * Returns the minimum pheromone value allowed.
	 * 
	 * @return the minimum pheromone value allowed.
	 */
	public double getTMin()
	{
		return tMin;
	}

	/**
	 * Returns the evaporation factor.
	 * 
	 * @return the evaporation factor.
	 */
	public double getEvaporationFactor()
	{
		return factor;
	}

	/**
	 * Updates the specified pheromone array.
	 * 
	 * @param slots the pheromone array.
	 * @param index the index of the array which was present in the current best
	 *        solution; <code>-1</code> in case that none of the components were
	 *        present in the current best solution.
	 * @param delta the increment factor.
	 */
	private void update(double[] slots, int index, double delta)
	{
		for (int i = 0; i < slots.length; i++)
		{
			slots[i] = (factor * slots[i]) + (i == index ? delta : 0.0);

			if (slots[i] > Util.precision(tMax))
			{
				slots[i] = tMax;
			}
			else if (slots[i] < Util.precision(tMin))
			{
				slots[i] = tMin;
			}
		}
	}
}