/*
 * $Id: Util.java 2336 2011-08-09 16:39:15Z febo $
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2010 Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.antree;

import java.util.LinkedList;
import java.util.List;

import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.data.Term;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision: 2336 $ $Date:: 2011-08-09 17:39:15#$
 */
public class Util
{
	/**
	 * Filters the specified list of instances according to the test represented
	 * by the term.
	 * 
	 * @param instances the list of instances to be filtered.
	 * @param term the filter criterion.
	 * 
	 * @return a list of instances that satisfies the term test.
	 */
	public static List<Instance> filter(List<Instance> instances, Term term,
			Metadata metadata)
	{
		LinkedList<Instance> filtered = new LinkedList<Instance>();

		for (Instance instance : instances)
		{
			double value = instance.values()[term.getIndex()];

			if (term.satisfies(value)
				|| metadata.isMissing(term.getIndex(), value))
			{
				filtered.add(instance);
			}
		}

		return filtered;
	}
	
	/**
	 * Returns the subset of instances according to the test represented
	 * by the term, not includung instances with missing values.
	 * 
	 * @param instances the list of instances to be filtered.
	 * @param term the filter criterion.
	 * 
	 * @return the subset of instances according to the test represented
	 *         by the term, not includung instances with missing values.
	 */
	public static List<Instance> subset(List<Instance> instances, Term term,
			Metadata metadata)
	{
		LinkedList<Instance> filtered = new LinkedList<Instance>();

		for (Instance instance : instances)
		{
			double value = instance.values()[term.getIndex()];

			if (term.satisfies(value))
			{
				filtered.add(instance);
			}
		}

		return filtered;
	}

	public static List<Instance> split(List<Instance> instances, Term term,
			Metadata metadata)
	{
		Dataset filtered = new Dataset(metadata);
		double weight = 0.0;
		double total = 0.0;

		for (Instance instance : instances)
		{
			double value = instance.values()[term.getIndex()];

			if (term.satisfies(value)
				|| (metadata.isMissing(term.getIndex(), value)))
			{
				Instance i = filtered.newInstance();
				i.set(instance.getValues());
				i.setWeight(instance.getWeight());
				
				if (!metadata.isMissing(term.getIndex(), value))
				{
					weight += instance.getWeight();
					total += instance.getWeight();
				}
			}
			else
			{
				total += instance.getWeight();
			}
		}

		// fix the weight for instances with missing values

		for (Instance instance : filtered.getInstances())
		{
			double value = instance.values()[term.getIndex()];

			if (metadata.isMissing(term.getIndex(), value))
			{
				instance.setWeight(instance.getWeight() * (weight / total));
			}
		}

		return filtered.getInstances();
	}

	/**
	 * Returns the double value with up to four decimal digits.
	 * 
	 * @param value the value to be converted.
	 * 
	 * @return the double value with up to four decimal digits.
	 */
	public static final double precision(double value)
	{
		int v = (int) (value * 100);
		return v / 100.0;
	}
}
