/*
 * $Id: Branch.java 2329 2011-08-02 08:57:44Z febo $
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2010 Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.antree.representation;

import java.util.List;

import net.sf.myra.antree.ThresholdHelper;
import net.sf.myra.datamining.data.Attribute;
import net.sf.myra.datamining.data.ContinuousAttribute;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.data.NominalAttribute;
import net.sf.myra.datamining.data.Operator;
import net.sf.myra.datamining.data.Term;
import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.ObjectFactory;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision: 2329 $ $Date:: 2011-08-02 09:57:44#$
 */
public class Branch
{
	/**
	 * The property key under which the branch encoder class name is stored.
	 */
	public static final String ENCODER_CLASS =
		"net.sf.myra.antree.branch.encoder";

	/**
	 * The branch encoder used to generate the code for the pheromone matrix.
	 */
	private static final BranchEncoder ENCODER =
		(BranchEncoder) ObjectFactory.create(Configuration.getProperty(
			ENCODER_CLASS, LevelBranchEncoder.class.getName()));

	/**
	 * Constant to represent an undefined code.
	 */
	private static final int UNDEFINED = 0;

	/**
	 * The node from where this branch originates.
	 */
	private Node from;

	/**
	 * The node from where this branch originates.
	 */
	private Node to;

	/**
	 * The term represented by this branch.
	 */
	private Term term;

	/**
	 * The hash code of the branch.
	 */
	private int code;

	/**
	 * Creates a new branch.
	 * 
	 * @param term the term represented by the branch.
	 */
	public Branch(Term term)
	{
		this.term = term;
		this.code = UNDEFINED;
	}

	/**
	 * Returns the node where this branch originates.
	 * 
	 * @return the node where this branch originates.
	 */
	public Node getFrom()
	{
		return from;
	}

	/**
	 * Sets the node where this branch originates. When the origin node is
	 * changes, the code of the branch is set to <code>UNDEFINED</code>.
	 * <strong>Note:</strong> This method is not thread safe.
	 * 
	 * @param from the node to set.
	 * 
	 * @see #getCode()
	 */
	public void setFrom(Node from)
	{
		if (from.isLeaf())
		{
			throw new IllegalArgumentException(
				"Leaf nodes cannot have branches");
		}

		this.from = from;
		code = UNDEFINED;
	}

	/**
	 * Returns the node that this branch leads to.
	 * 
	 * @return the node that this branch leads to.
	 */
	public Node getTo()
	{
		return to;
	}

	/**
	 * Sets the node that this branch leads to.
	 * 
	 * @param to the node to set.
	 */
	public void setTo(Node to)
	{
		this.to = to;
		to.setParent(getFrom());
	}

	/**
	 * Returns the term represented by this branch.
	 * 
	 * @return the term represented by this branch.
	 */
	public Term getTerm()
	{
		return term;
	}

	/**
	 * Returns the code of this branch. This should be used as the pheromone
	 * entry for this branch. If the code of the branch is
	 * <code>UNDEFINED</code>, it will be computed. <strong>Note:</strong>
	 * This method is not thread safe.
	 * 
	 * @return the code of this branch.
	 * 
	 * @see #setFrom(Node)
	 */
	public int getCode()
	{
		if (code == UNDEFINED)
		{
			code = ENCODER.encode(this);
		}

		return code;
	}

	/**
	 * Returns a branch array for the specified attribute, one branch for each
	 * attribute value.
	 * 
	 * @param metadata the current data set metadata.
	 * @param attribute the attribute to get the values from.
	 * @param instances the list of instances.
	 *  
	 * @return a branch array for the specified attribute, one branch for each
	 *         attribute value.
	 */
	public static Branch[] create(Metadata metadata, Attribute attribute,
			List<Instance> instances)
	{
		if (attribute instanceof NominalAttribute)
		{
			return create(metadata, (NominalAttribute) attribute);
		}
		else if (attribute instanceof ContinuousAttribute)
		{
			return create(metadata, (ContinuousAttribute) attribute, instances);
		}

		throw new IllegalArgumentException("Invalid attribute type: "
			+ attribute.getClass());
	}

	/**
	 * Returns a branch array for the specified attribute, one branch for each
	 * attribute value.
	 * 
	 * @param metadata the metadata object.
	 * @param attribute the nominal attribute to get the values from.
	 *  
	 * @return a branch array for the specified attribute, one branch for each
	 *         nominal value.
	 */
	public static Branch[] create(Metadata metadata, NominalAttribute attribute)
	{
		Branch[] branches = new Branch[attribute.getValues().size()];
		int index = metadata.getIndex(attribute.getName());

		for (int i = 0; i < branches.length; i++)
		{
			Term term = new Term(attribute.getName(), index);
			term.setOperator(Operator.EQUAL_TO);

			String value = attribute.getValue(i);
			term.setDescription(value);
			term.setValues(attribute.getIndex(value));

			branches[i] = new Branch(term);
		}

		return branches;
	}

	/**
	 * Returns a branch array for the specified continuous attribute, one branch
	 * for each discrete interval.
	 * 
	 * @param metadata the metadata object.
	 * @param attribute the continuous attribute to get the values from.
	 * @param instances the instances used for creating the discrete intervals.
	 *  
	 * @return a branch array for the specified continuous attribute, one branch
	 *         for each discrete interval.
	 */
	public static Branch[] create(Metadata metadata,
			ContinuousAttribute attribute, List<Instance> instances)
	{
		Term[] terms = ThresholdHelper.create(metadata, attribute, instances);
		Branch[] branches = new Branch[terms.length];

		for (int i = 0; i < branches.length; i++)
		{
			branches[i] = new Branch(terms[i]);
		}

		return branches;
	}

	@Override
	public String toString()
	{
		StringBuffer buffer = new StringBuffer();

		buffer.append(from == null ? "[START]" : from.getAttribute().getName());
		buffer.append(" -> ");
		buffer.append(to.getAttribute().getName());
		buffer.append(" (");
		buffer.append(term);
		buffer.append(")");

		return buffer.toString();
	}
}