/*
 * $Id: Node.java 2320 2011-01-25 13:06:10Z febo $
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2010 Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.antree.representation;

import net.sf.myra.datamining.data.Attribute;
import net.sf.myra.datamining.data.Label;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision: 2320 $ $Date:: 2011-01-25 13:06:10#$
 */
public abstract class Node
	implements Cloneable
{
	/**
	 * The first level of the tree (the level of the root node).
	 */
	public static final int INITIAL_LEVEL = 0;

	/**
	 * An empty array representing the branches of a leaf node.
	 */
	private static final Branch[] LEAF_BRANCHES = new Branch[0];

	/**
	 * The level in the tree where this node is placed (<code>0</code>
	 * represents the first level of the tree).
	 */
	private int level;

	/**
	 * The attribute that this node represents.
	 */
	private Attribute attribute;

	/**
	 * The parent node.
	 */
	private Node parent;

	/**
	 * The branches originating from this node.
	 */
	private Branch[] branches;

	/**
	 * Creates a new node.
	 * 
	 * @param level the level of the node.
	 * @param attribute the attribute that this node represents. 
	 */
	public Node(int level, Attribute attribute)
	{
		this(level, attribute, LEAF_BRANCHES);
	}

	/**
	 * Creates a new node.
	 * 
	 * @param level the level of the node.
	 * @param attribute the attribute represented by the node.
	 * @param branches the branches originating from this node.
	 */
	public Node(int level, Attribute attribute, Branch[] branches)
	{
		this.level = level;
		this.attribute = attribute;
		this.branches = branches;

		for (Branch branch : branches)
		{
			branch.setFrom(this);
		}
	}

	/**
	 * Returns the level of this node.
	 * 
	 * @return the level of this node.
	 */
	public int getLevel()
	{
		return level;
	}

	/**
	 * Sets the level of this node.
	 * 
	 * @param level the level to set.
	 */
	public void setLevel(int level)
	{
		this.level = level;
	}

	/**
	 * Returns the attribute that this node represents.
	 * 
	 * @return the attribute that this node represents.
	 */
	public Attribute getAttribute()
	{
		return attribute;
	}

	/**
	 * Returns the parent node of this node.
	 * 
	 * @return the parent node of this node.
	 */
	public Node getParent()
	{
		return parent;
	}

	/**
	 * Sets the parent node.
	 * 
	 * @param parent the node to set.
	 */
	void setParent(Node parent)
	{
		if (parent != null && parent.isLeaf())
		{
			throw new IllegalArgumentException("Leaf node cannot be parent.");
		}

		this.parent = parent;
	}

	/**
	 * Returns the branch given the specified index.
	 * 
	 * @param index the index of the branch.
	 * 
	 * @return the branch given the specified index.
	 */
	public Branch getBranch(int index)
	{
		return branches[index];
	}

	/**
	 * Returns the branch that leads to the specified node.
	 * 
	 * @param node
	 * 
	 * @return the branch that leads to the specified node.
	 */
	public Branch getBranch(Node node)
	{
		for (Branch branch : branches)
		{
			if (branch.getTo().equals(node))
			{
				return branch;
			}
		}

		throw new IllegalArgumentException("Branch for node not found: "
			+ node.label());
	}

	/**
	 * Returns the branch that corresponds to the specified value.
	 * 
	 * @param value the value to test.
	 * 
	 * @return the branch that corresponds to the specified value.
	 */
	public Branch getBranch(double value)
	{
		for (Branch branch : branches)
		{
			if (branch.getTerm().satisfies(value))
			{
				return branch;
			}
		}

		throw new IllegalArgumentException("Branch for value not found: "
			+ value);
	}

	/**
	 * Returns the branches originating from this node.
	 * 
	 * @return the branches originating from this node.
	 */
	public Branch[] getBranches()
	{
		return branches.clone();
	}

	/**
	 * Indicates if this node is a leaf node or not. If the node is a leaf
	 * node, it is safe to call the {{@link #getPrediction()} method.
	 * 
	 * @return <code>true</code> if this node is a leaf node; <code>false</code>
	 *         otherwise.
	 *         
	 * @see #getPrediction()
	 */
	public final boolean isLeaf()
	{
		return (branches.length == 0);
	}

	/**
	 * Returns a label representation of this node.
	 * 
	 * @return a label representation of this node.
	 */
	public abstract String label();

	/**
	 * Returns the label predicted by the node.
	 * 
	 * @return the label predicted by the node.
	 */
	public abstract Label getPrediction();

	/**
	 * Sets the prediction of the node.
	 * 
	 * @param prediction the prediction to set.
	 */
	public abstract void setPrediction(Label prediction);

	@Override
	public final boolean equals(Object o)
	{
		return super.equals(o);
	}

	@Override
	public final int hashCode()
	{
		return super.hashCode();
	}

	@Override
	public Node clone()
	{
		try
		{
			Node clone = (Node) super.clone();
			clone.branches = branches.clone();
			clone.parent = null;

			for (int i = 0; i < branches.length; i++)
			{
				clone.branches[i] = new Branch(branches[i].getTerm().clone());
				clone.branches[i].setFrom(clone);
				clone.branches[i].setTo(branches[i].getTo().clone());
			}

			return clone;
		}
		catch (CloneNotSupportedException e)
		{
			throw new InternalError(e.getMessage());
		}
	}

	@Override
	public final String toString()
	{
		return super.toString();
	}

	public boolean validate()
	{
		boolean valid = true;

		if (parent != null)
		{
			valid =
				(parent.getBranch(this) != null)
					&& ((parent.getLevel() + 1) == level);

		}

		for (Branch branch : branches)
		{
			valid =
				(branch.getFrom() == this)
					&& (branch.getTo().getLevel() == (level + 1))
					&& branch.getTo().validate();

			if (!valid)
			{
				break;
			}
		}

		return valid;
	}
}