/*
 * $Id: Statistics.java 2331 2011-08-06 21:23:05Z febo $
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2010 Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.antree.representation;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Label;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.framework.RandomGenerator;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision: 2331 $ $Date:: 2011-08-06 22:23:05#$
 */
public class Statistics
{
	/**
	 * The mapping of each node to its class distribution.
	 */
	private Map<Node, double[]> distribution;

	/**
	 * The number of target values.
	 */
	private int size;

	/**
	 * The majority class label of the data set.
	 */
	private Label majority;

	/**
	 * Default constructor.
	 * 
	 * @param metadata the current data set metadata.
	 */
	public Statistics(Metadata metadata)
	{
		size = metadata.getTarget().getValues().size();
		distribution = new HashMap<Node, double[]>();
	}

	/**
	 * Increments the class distribution of the specified node.
	 * 
	 * @param node the node reference.
	 * @param instance the instance reaching the node.
	 * @param weight the weight of the instance.
	 */
	void increment(Node node, Instance instance)
	{
		increment(node, instance, 1.0);
	}

	/**
	 * Increments the class distribution of the specified node.
	 * 
	 * @param node the node reference.
	 * @param instance the instance reaching the node.
	 * @param weight the weight of the instance.
	 */
	void increment(Node node, Instance instance, double weight)
	{
		double[] current = distribution.get(node);

		if (current == null)
		{
			current = new double[size];
			distribution.put(node, current);
		}

		for (int i = 0; i < current.length; i++)
		{
			if (instance.labels()[i])
			{
				current[i] += weight;
			}
		}
	}

	/**
	 * Returns the class distribution of the specified node.
	 * 
	 * @param node the node reference.
	 * 
	 * @return the class distribution of the specified node.
	 */
	public double[] getDistribution(Node node)
	{
		return safe(node);
	}

	/**
	 * Sets the class distribution of the specified node.
	 * 
	 * @param node the node reference.
	 * @param distribution the distribution to set.
	 */
	public void setDistribution(Node node, double[] distribution)
	{
		this.distribution.put(node, distribution);
	}

	/**
	 * Returns the majority class label.
	 * 
	 * @return the majority class label.
	 */
	public Label getMajority()
	{
		return majority;
	}

	/**
	 * Sets the majority class label.
	 * 
	 * @param majority the majority class label to set.
	 */
	public void setMajority(Label majority)
	{
		this.majority = majority;
	}

	/**
	 * Returns the total number of instances reaching the specified node.
	 * 
	 * @param node the node reference.
	 * 
	 * @return the total number of instances reaching the specified node.
	 */
	public double getTotal(Node node)
	{
		double[] current = safe(node);
		double total = 0;

		for (int i = 0; i < current.length; i++)
		{
			total += current[i];
		}

		return total;
	}

	/**
	 * Returns the majority class label for the specified node. <b>Note:</b>
	 * this method should only be used for flat single-label problems.
	 * 
	 * @param node the node reference.
	 * 
	 * @return the majority class label for the specified node.
	 */
	public Label getMajorityLabel(Node node)
	{
		double[] current = safe(node);
		int index = 0;
		ArrayList<Label> candidates = new ArrayList<Label>();
		candidates.add(majority.toLabel(majority.get(index)));

		for (int i = 1; i < current.length; i++)
		{
			if (current[i] > current[index])
			{
				index = i;

				candidates.clear();
				candidates.add(majority.toLabel(majority.get(index)));
			}
			else if (current[i] == current[index])
			{
				candidates.add(majority.toLabel(majority.get(i)));
			}
		}

		if (candidates.size() > 1)
		{
			if (candidates.contains(majority))
			{
				return majority;
			}
			else
			{
				return candidates.get(RandomGenerator.getInstance().nextInt(
					candidates.size()));
			}
		}

		return candidates.get(0);
	}

	/**
	 * Returns the error rate at the specified node given a prediction.
	 * 
	 * @param node the node reference.
	 * @param prediction the prediction at the node.
	 * 
	 * @return the error rate at the specified node given a prediction.
	 */
	public double getErrors(Node node, Label prediction)
	{
		double[] current = safe(node);
		double errors = 0;

		for (int i = 0; i < current.length; i++)
		{
			if (!prediction.active(i))
			{
				errors += current[i];
			}
		}

		return errors;
	}

	/**
	 * Returns the distribution for the specified node or an empty distribution
	 * in case one is not found.
	 * 
	 * @param node the node reference.
	 * 
	 * @return the distribution for the specified node or an empty distribution
	 *         in case one is not found.
	 */
	private double[] safe(Node node)
	{
		double[] current = distribution.get(node);
		return (current == null) ? new double[size] : current.clone();
	}
}