/*
 * $Id: Tree.java 2332 2011-08-06 21:30:28Z febo $
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2010 Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.antree.representation;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import net.sf.myra.antree.Util;
import net.sf.myra.datamining.Evaluator;
import net.sf.myra.datamining.Gatherer;
import net.sf.myra.datamining.Model;
import net.sf.myra.datamining.Rule;
import net.sf.myra.datamining.Unit;
import net.sf.myra.datamining.UnitAveragingMode;
import net.sf.myra.datamining.data.Attribute;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Label;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.data.Term;
import net.sf.myra.datamining.model.RuleSet;
import net.sf.myra.datamining.model.RuleSet.ConflictResolutionStrategy;
import net.sf.myra.framework.Vertex;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision: 2332 $ $Date:: 2011-08-06 22:30:28#$
 */
public class Tree
	implements Model, Cloneable
{
	/**
	 * Number formatter used to output the statistics .
	 */
	private static NumberFormat FORMATTER = new DecimalFormat("##0.0");

	/**
	 * The root node of the tree.
	 */
	private Node root;

	/**
	 * The quality of the decision tree.
	 */
	private Unit<?> quality;

	/**
	 * Node statistics used during classification to deal with missing values.
	 */
	private Statistics statistics;

	/**
	 * Default constructor.
	 */
	public Tree()
	{
		this(null);
	}

	/**
	 * Creates a new tree.
	 * 
	 * @param root the root node of the tree.
	 */
	public Tree(Node root)
	{
		this.root = root;
	}

	/**
	 * Sets the root node of the tree.
	 * 
	 * @param root the root node to set.
	 */
	public void setRoot(Node root)
	{
		this.root = root;
		root.setParent(null);
	}

	/**
	 * Returns the root of the tree.
	 * 
	 * @return the root of the tree.
	 */
	public Node getRoot()
	{
		return root;
	}

	/**
	 * Sets the nodes statistics to be used to deal with missing values.
	 * 
	 * @param statistics the statistics to set.
	 */
	public void setStatistics(Statistics statistics)
	{
		this.statistics = statistics;
	}

	/**
	 * Returns a node of the tree specified by its index.
	 * 
	 * @param index the node index.
	 * 
	 * @return a node of the tree specified by its index.
	 */
	public Node get(int index)
	{
		LinkedList<Node> nodes = new LinkedList<Node>();
		nodes.add(root);
		int current = 0;

		while (!nodes.isEmpty())
		{
			Node node = nodes.removeFirst();

			if (index == current)
			{
				return node;
			}

			current++;
			Branch[] branches = node.getBranches();

			for (int i = branches.length - 1; i >= 0; i--)
			{
				nodes.addFirst(branches[i].getTo());
			}
		}

		throw new IllegalArgumentException("Node at index not found: " + index);
	}

	/**
	 * Returns the index of the specified node.
	 * 
	 * @param node the node to look for the index.
	 * 
	 * @return the index of the specified node.
	 */
	public int getIndex(Node node)
	{
		LinkedList<Node> nodes = new LinkedList<Node>();
		nodes.add(root);
		int current = 0;

		while (!nodes.isEmpty())
		{
			Node n = nodes.removeFirst();

			if (n.equals(node))
			{
				return current;
			}

			current++;
			Branch[] branches = n.getBranches();

			for (int i = branches.length - 1; i >= 0; i--)
			{
				nodes.addFirst(branches[i].getTo());
			}
		}

		throw new IllegalArgumentException("Node not found: " + node);
	}

	/**
	 * Returns the branch leading to the specified node.
	 * 
	 * @param node the node at the end of the branch.
	 * 
	 * @return the branch leading to the specified node.
	 */
	public Branch getBranch(Node node)
	{
		LinkedList<Node> nodes = new LinkedList<Node>();
		nodes.add(node);

		while (!nodes.isEmpty())
		{
			Node current = nodes.removeFirst();

			for (Branch branch : current.getBranches())
			{
				if (branch.getTo().equals(node))
				{
					return branch;
				}

				nodes.addLast(branch.getTo());
			}
		}

		throw new IllegalArgumentException("Branch to node not found: " + node);
	}

	/**
	 * Returns the indexes of nodes.
	 * 
	 * @param type the type of the nodes.
	 * 
	 * @return the indexes of nodes.
	 */
	public Collection<Integer> indexes(Class<? extends Node> type)
	{
		LinkedList<Node> nodes = new LinkedList<Node>();
		nodes.add(root);

		LinkedList<Integer> indexes = new LinkedList<Integer>();
		int current = 0;

		while (!nodes.isEmpty())
		{
			Node node = nodes.removeFirst();

			if (type.isInstance(node))
			{
				indexes.add(current);
			}

			current++;
			Branch[] branches = node.getBranches();

			for (int i = branches.length - 1; i >= 0; i--)
			{
				nodes.addFirst(branches[i].getTo());
			}
		}

		return indexes;
	}

	/**
	 * Adjust the leaf node predictions based on the specified data set. This
	 * method should be used if the structure of the tree is changed (e.g.
	 * during the pruning of the tree).
	 * 
	 * @param dataset the current data set.
	 */
	public void adjust(Dataset dataset)
	{
		Label majority = Evaluator.findMajorityClass(dataset);

		if (root.isLeaf())
		{
			root.setPrediction(majority);
		}
		else
		{
			for (LeafNode leaf : leaves())
			{
				Metadata metadata = dataset.getMetadata();
				List<Instance> instances = dataset.getInstances();
				Node parent = leaf;

				while (parent.getParent() != null)
				{
					instances =
						Util.filter(instances, parent.getParent().getBranch(
							parent).getTerm(), metadata);
					parent = parent.getParent();
				}

				if (!instances.isEmpty())
				{
					leaf.setPrediction(Evaluator.findMajorityClass(instances,
						metadata, majority));
				}
			}
		}
	}

	/**
	 * Returns the quality of the tree.
	 * 
	 * @return the quality of the tree.
	 */
	public Unit<?> getQuality()
	{
		return quality;
	}

	/**
	 * Sets the quality of the tree.
	 * 
	 * @param quality the quality to set.
	 */
	public void setQuality(Unit<?> quality)
	{
		this.quality = quality;
	}

	/**
	 * Returns the number of nodes of the tree.
	 * 
	 * @return the number of nodes of the tree.
	 */
	public int size()
	{
		LinkedList<Node> nodes = new LinkedList<Node>();
		nodes.add(root);
		int count = 0;

		while (!nodes.isEmpty())
		{
			Node node = nodes.removeFirst();
			count++;

			for (Branch branch : node.getBranches())
			{
				nodes.addLast(branch.getTo());
			}
		}

		return count;
	}

	/**
	 * Returns the number of leaf nodes of the tree.
	 * 
	 * @return the number of leaf nodes of the tree.
	 */
	public int countInternalNodes()
	{
		LinkedList<Node> nodes = new LinkedList<Node>();
		nodes.add(root);
		int count = 0;

		while (!nodes.isEmpty())
		{
			Node node = nodes.removeFirst();

			if (!node.isLeaf())
			{
				count++;
			}

			for (Branch branch : node.getBranches())
			{
				nodes.addLast(branch.getTo());
			}
		}

		return count;
	}

	/**
	 * Returns the number of leaf nodes of the tree.
	 * 
	 * @return the number of leaf nodes of the tree.
	 */
	public int countLeaves()
	{
		LinkedList<Node> nodes = new LinkedList<Node>();
		nodes.add(root);
		int count = 0;

		while (!nodes.isEmpty())
		{
			Node node = nodes.removeFirst();

			if (node.isLeaf())
			{
				count++;
			}
			else
			{
				for (Branch branch : node.getBranches())
				{
					nodes.addLast(branch.getTo());
				}
			}
		}

		return count;
	}

	/**
	 * Returns the number of leaf nodes of the tree.
	 * 
	 * @return the number of leaf nodes of the tree.
	 */
	public Collection<LeafNode> leaves()
	{
		LinkedList<Node> nodes = new LinkedList<Node>();
		nodes.add(root);

		LinkedList<LeafNode> leaves = new LinkedList<LeafNode>();

		while (!nodes.isEmpty())
		{
			Node node = nodes.removeFirst();

			if (node.isLeaf())
			{
				leaves.add((LeafNode) node);
			}
			else
			{
				for (Branch branch : node.getBranches())
				{
					nodes.addLast(branch.getTo());
				}
			}
		}

		return leaves;
	}

	/**
	 * Returns the label predicted for the specified instance.
	 * 
	 * @param instance the instance to classify.
	 * 
	 * @return the label predicted for the specified instance.
	 */
	public Label classify(Instance instance)
	{
		/*
				Node node = root;

				while (!node.isLeaf())
				{
					int index =
						instance.getMetadata().getIndex(node.getAttribute().getName());
					boolean found = false;

					for (Branch branch : node.getBranches())
					{
						if (branch.getTerm().satisfies(instance.values()[index]))
						{
							node = branch.getTo();
							found = true;
							break;
						}
					}

					if (!found)
					{
						throw new IllegalStateException(
							"Invalid tree structure: branch not found.");
					}
				}

				return ((LeafNode) node).getPrediction();
		*/
		List<LeafNode> candidates = null;

		if (root.isLeaf())
		{
			candidates = new LinkedList<LeafNode>();
			candidates.add((LeafNode) root);
		}
		else
		{
			candidates = prediction(root, instance);
		}

		if (candidates.size() == 1)
		{
			return candidates.get(0).getPrediction();
		}
		else
		{
			double total = 0;

			for (LeafNode candidate : candidates)
			{
				total += statistics.getTotal(candidate);
			}

			Label template = instance.getMetadata().getLabel();
			double[] probabilities = new double[template.length()];

			for (LeafNode candidate : candidates)
			{
				Label prediction = candidate.getPrediction();

				double n = statistics.getTotal(candidate);
				double e = statistics.getErrors(candidate, prediction);

				for (int i = 0; i < probabilities.length; i++)
				{
					if (prediction.active(i))
					{
						probabilities[i] += ((n / total) * ((n - e) / n));
					}
					else
					{
						probabilities[i] += ((n / total) * (e / n));
					}
				}
			}

			int current = 0;

			for (int i = 1; i < probabilities.length; i++)
			{
				if (probabilities[i] > probabilities[current])
				{
					current = i;
				}
			}

			return template.toLabel(template.get(current));
		}
	}

	/**
	 * Computes the nodes statistics.
	 * 
	 * @param dataset the current dataset.
	 * 
	 * @return the statistics of the nodes based on the specified dataset.
	 */
	public Statistics gather(Dataset dataset)
	{
		/*		
				Statistics statistics = new Statistics(dataset.getMetadata());

				for (Instance instance : dataset)
				{
					Node node = root;
					statistics.increment(node, instance);

					while (!node.isLeaf())
					{
						int index =
							instance.getMetadata().getIndex(
								node.getAttribute().getName());
						boolean found = false;

						for (Branch branch : node.getBranches())
						{
							if (branch.getTerm().satisfies(instance.values()[index]))
							{
								node = branch.getTo();
								statistics.increment(node, instance);

								found = true;
								break;
							}
						}

						if (!found)
						{
							throw new IllegalStateException(
								"Invalid tree structure: branch not found.");
						}
					}
				}

				return statistics;
		*/
		Statistics statistics = new Statistics(dataset.getMetadata());
		List<WeightedInstance> instances = new LinkedList<WeightedInstance>();

		for (Instance instance : dataset)
		{
			WeightedInstance weighted = new WeightedInstance();
			weighted.weight = 1.0;
			weighted.instance = instance;

			instances.add(weighted);
		}

		gather(root, instances, statistics);
		statistics.setMajority(Evaluator.findMajorityClass(dataset));

		return statistics;
	}

	private void gather(Node node, List<WeightedInstance> instances,
			Statistics statistics)
	{
		for (WeightedInstance weighted : instances)
		{
			statistics.increment(node, weighted.instance, weighted.weight);
		}

		if (!node.isLeaf())
		{
			ArrayList<List<WeightedInstance>> subsets =
				new ArrayList<List<WeightedInstance>>();

			for (Branch branch : node.getBranches())
			{
				List<WeightedInstance> subset =
					new LinkedList<WeightedInstance>();
				Iterator<WeightedInstance> iter = instances.iterator();
				Term term = branch.getTerm();

				while (iter.hasNext())
				{
					WeightedInstance w = iter.next();

					if (term.satisfies(w.instance.values()[term.getIndex()]))
					{
						subset.add(w);
						iter.remove();
					}
				}

				subsets.add(subset);
			}

			// do we have instances with missing values?
			if (!instances.isEmpty())
			{
				double[] percentage = new double[subsets.size()];
				int total = 0;

				for (int i = 0; i < percentage.length; i++)
				{
					percentage[i] = subsets.get(i).size();
					total += percentage[i];
				}

				for (int i = 0; i < percentage.length; i++)
				{
					percentage[i] /= total;
				}

				for (int i = 0; i < percentage.length; i++)
				{
					List<WeightedInstance> subset = subsets.get(i);

					for (WeightedInstance w : instances)
					{
						WeightedInstance copy = new WeightedInstance();
						copy.instance = w.instance;
						copy.weight = w.weight * percentage[i];

						subset.add(copy);
					}
				}
			}

			int index = 0;

			for (Branch branch : node.getBranches())
			{
				gather(branch.getTo(), subsets.get(index), statistics);
				index++;
			}
		}
	}

	@Override
	public String toString()
	{
		return toString(null);
	}

	public String toString(Statistics statistics)
	{
		StringWriter output = new StringWriter();
		PrintWriter writer = new PrintWriter(output);

		print(writer, root, "", statistics);

		return output.toString();
	}

	/**
	 * Returns the tree represented as a set of rules.
	 * 
	 * @param metadata the data set metadata.
	 * 
	 * @return the tree represented as a set of rules.
	 * 
	 * @see Rule
	 */
	public RuleSet toRuleSet(Metadata metadata)
	{
		RuleSet ruleSet = new RuleSet(ConflictResolutionStrategy.UNSUPPORTED);

		Rule rule = new Rule(metadata);
		expand(ruleSet, rule, root);

		return ruleSet;
	}

	/**
	 * Expands the specified rule by following the branches originated from the
	 * current node.
	 * 
	 * @param ruleSet the set of rules being built.
	 * @param rule the rule to be expanded.
	 * @param node the current node.
	 */
	private void expand(RuleSet ruleSet, Rule rule, Node node)
	{
		if (node.isLeaf())
		{
			rule.setConsequent(node.getPrediction());
			ruleSet.add(rule);
		}
		else
		{
			for (Branch branch : node.getBranches())
			{
				Rule r = rule.clone();

				Vertex<Term> vertex =
					new Vertex<Term>(branch.getTerm().toString());
				vertex.setInfo(branch.getTerm());

				r.add(vertex);

				expand(ruleSet, r, branch.getTo());
			}
		}
	}

	/**
	 * Recursively builds the string representation of the specified node.
	 * 
	 * @param writer the output writer.
	 * @param node the current node.
	 * @param last indicates if the node is the last of the branch.
	 */
	private void print(PrintWriter writer, Node node, String indent,
			Statistics statistics)
	{
		if (!node.isLeaf())
		{
			writer.println(node.label());

			Branch[] branches = node.getBranches();

			for (int i = 0; i < branches.length; i++)
			{
				writer.println(indent + "|");
				writer.print(indent + "+- " + branches[i].getTerm() + ": ");

				boolean last = (i + 1) == branches.length;
				String next = indent + (last ? " " : "|") + "    ";

				print(writer, branches[i].getTo(), next, statistics);
			}
		}
		else
		{
			writer.print(node.label());

			if (statistics == null)
			{
				writer.println();
			}
			else
			{
				// to output the predicted error
				// ReducedErrorMetric.getInstance().error(node, statistics)

				writer.println(" ("
					+ FORMATTER.format(statistics.getTotal(node))
					+ "/"
					+ FORMATTER.format(statistics.getErrors(node, node
						.getPrediction())) + ")");
			}
		}
	}

	private List<LeafNode> prediction(Node node, Instance instance)
	{
		LinkedList<LeafNode> candidates = new LinkedList<LeafNode>();

		while (!node.isLeaf())
		{
			Attribute attribute = node.getAttribute();
			int index = instance.getMetadata().getIndex(attribute.getName());
			boolean found = false;

			// the only case where a test cannot be found is when the instance
			// has missing values, so we push the instances down all the
			// branches if there is a statistics object to assist a
			// probabilistic prediction

			if (attribute.isMissing(instance.values()[index])
				&& (statistics != null))
			{
				for (Branch branch : node.getBranches())
				{
					Node candidate = branch.getTo();

					if (candidate.isLeaf())
					{
						candidates.add((LeafNode) candidate);
					}
					else
					{
						candidates.addAll(prediction(branch.getTo(), instance));
					}
				}

				break;
			}
			else
			{
				for (Branch branch : node.getBranches())
				{
					if (branch.getTerm().satisfies(instance.values()[index]))
					{
						node = branch.getTo();
						found = true;
						break;
					}
				}

				if (!found)
				{
					throw new IllegalStateException(
						"Invalid tree structure: branch not found.");
				}
				else if (node.isLeaf())
				{
					candidates.add((LeafNode) node);
				}
			}
		}

		return candidates;
	}

	@Override
	public Tree clone()
	{
		try
		{
			Tree clone = (Tree) super.clone();
			clone.root = root.clone();
			clone.statistics = null;
			return clone;
		}
		catch (CloneNotSupportedException e)
		{
			throw new InternalError(e.getMessage());
		}
	}

	public Gatherer getGatherer()
	{
		return new DefaultGatherer();
	}

	/**
	 * Struct-like class to associate a weight to each instance.
	 */
	private static class WeightedInstance
	{
		/**
		 * The weight of the instance.
		 */
		double weight;

		/**
		 * The instance reference.
		 */
		Instance instance;
	}

	public static class DefaultGatherer
		implements Gatherer
	{
		/**
		 * The total number of leaf nodes.
		 */
		private ArrayList<Integer> leaves = new ArrayList<Integer>();

		/**
		 * The number of nodes.
		 */
		private ArrayList<Integer> size = new ArrayList<Integer>();

		public void add(Model model)
		{
			if (model instanceof Tree)
			{
				Tree tree = (Tree) model;
				leaves.add(tree.countLeaves());
				size.add(tree.size());
			}
			else
			{
				throw new IllegalArgumentException("Invalid model class: "
					+ model.getClass().getName());
			}
		}

		@Override
		public String toString()
		{
			StringBuffer buffer = new StringBuffer();

			buffer.append("Model size: ");
			buffer.append(UnitAveragingMode.stdev(size) + "\n");

			buffer.append("Number of leaf nodes: ");
			buffer.append(UnitAveragingMode.stdev(leaves) + "\n");

			return buffer.toString();
		}
	}
}