/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.cantminer;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import net.sf.myra.datamining.Evaluator;
import net.sf.myra.datamining.Rule;
import net.sf.myra.datamining.data.Attribute;
import net.sf.myra.datamining.data.ContinuousAttribute;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.data.Term;
import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.Edge;
import net.sf.myra.framework.Graph;
import net.sf.myra.framework.NoAvailableVertexException;
import net.sf.myra.framework.ObjectFactory;
import net.sf.myra.framework.RandomGenerator;
import net.sf.myra.framework.Trail;
import net.sf.myra.framework.TrailFactory;
import net.sf.myra.framework.Vertex;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class EdgeTrailFactory
	implements TrailFactory
{
	/**
	 * The property key under which the minimum rule cases value is stored.
	 */
	public static final String MINIMUM_RULE_CASES =
		"net.sf.myra.cantminer.cases";

	/**
	 * The property key under which the rule implementation class is stored.
	 */
	public static final String RULE = "net.sf.myra.cantminer.rule";

	/**
	 * Constant to indicate that there is no minimum number of covered cases.
	 */
	public static final int NO_MINIMUM_CASES = 0;

	/**
	 * Minimum number of covered cases per rule.
	 */
	private final int MINIMUM_CASES =
		Configuration.getIntProperty(MINIMUM_RULE_CASES, NO_MINIMUM_CASES);

	/**
	 * The problem construction graph to create the trails.
	 */
	protected Graph<Term> graph;

	/**
	 * The dataset instance to validate the vertices selection.
	 */
	protected Dataset dataset;

	/**
	 * The threshold selection helper.
	 */
	protected ThresholdHelper helper;

	/**
	 * Default constructor.
	 * 
	 * @param graph the problem construction graph.
	 * @param dataset the dataset reference.
	 */
	public EdgeTrailFactory(Graph<Term> graph, Dataset dataset)
	{
		this.graph = graph;
		this.dataset = dataset;
		this.helper = new ThresholdHelper(dataset.getMetadata());
	}

	@SuppressWarnings("unchecked")
	public Trail createTrail()
	{
		Metadata metadata = dataset.getMetadata();

		Rule rule =
			(Rule) ObjectFactory.create(Configuration.getProperty(RULE,
				Rule.class.getName()), new Class<?>[] { Metadata.class },
				new Object[] { metadata });

		List<Instance> covered = dataset.getInstances();
		Set<Vertex<Term>> incompatible = new HashSet<Vertex<Term>>();

		try
		{
			while (covered.size() >= MINIMUM_CASES)
			{
				Vertex<Term> last = (Vertex<Term>) rule.getLast();

				if (last == null)
				{
					last = graph.get(cAntMiner2.VIRTUAL_VERTEX);
				}

				Wheel wheel =
					new Wheel(rule, graph.getEdges(last), incompatible);

				while (!wheel.isEmpty())
				{
					Rule clone = (Rule) rule.clone();
					double probability =
						RandomGenerator.getInstance().nextDouble();

					Slot slot = wheel.spin(probability);

					Vertex<Term> vertex =
						process(metadata, slot.vertex(), covered, clone
							.isEmpty());
					clone.add(vertex);

					List<Instance> set =
						Evaluator.findCoveredCases(clone, covered);
					int size = set.size();

					// if the candidate rule cover at least MINIMUM_CASES and
					// either the size of examples covered have changed or the
					// current rule is empty
					if ((size >= MINIMUM_CASES)
						&& (size != covered.size() || rule.isEmpty()))
					{
						rule.add(vertex);
						covered = set;

						break;
					}
					else
					{
						Attribute attribute =
							metadata.get(vertex.getInfo().getAttribute());

						if (!(attribute instanceof ContinuousAttribute))
						{
							incompatible.add(vertex);
						}
					}

					wheel.remove(slot);
				}

				if (wheel.isEmpty())
				{
					// there are no more attributes to be added to the rule
					break;
				}
			}
		}
		catch (NoAvailableVertexException e)
		{
			// end of the rule construction procedure
		}

		return rule;
	}

	/**
	 * Prepares the vertex to be added to a rule. The default implementation
	 * just returns the vertex.
	 * 
	 * @param vertex the vertex to be added to the rule.
	 * 
	 * @return the same vertex instance.
	 */
	protected Vertex<Term> process(Metadata metadata, Vertex<Term> vertex,
			List<Instance> covered, boolean first)
	{
		Attribute attribute = metadata.get(vertex.getInfo().getIndex());

		if (attribute instanceof ContinuousAttribute)
		{
			return helper.create((ContinuousAttribute) attribute, vertex,
				covered, first);
		}

		return vertex;
	}

	public Graph<?> getGraph()
	{
		return graph;
	}

	/**
	 * Returns the dataset reference.
	 * 
	 * @return the dataset reference.
	 */
	protected Dataset getDataset()
	{
		return dataset;
	}

	/**
	 * This class represents the probabilistic selection roulette wheel.
	 */
	static class Wheel
	{
		/**
		 * The list of available slots (vertices).
		 */
		private ArrayList<Slot> slots;

		/**
		 * Default constructor.
		 * 
		 * @param vertices the available vertices.
		 */
		@SuppressWarnings("unchecked")
		Wheel(Rule rule, Collection<Edge<Term>> edges,
				Set<Vertex<Term>> incompatible)
		{
			slots = new ArrayList<Slot>(edges.size());

			Set<String> used = new HashSet<String>();

			for (Vertex<?> vertex : rule.getVertices())
			{
				Term term = ((Vertex<Term>) vertex).getInfo();
				used.add(term.getAttribute());
			}

			for (Edge<Term> edge : edges)
			{
				Vertex<Term> to = edge.getTo();

				if (!(incompatible.contains(to) || used.contains(to.getInfo()
					.getAttribute())))
				{
					Slot slot = new Slot(edge);
					slots.add(slot);
				}
			}

			normalize();
		}

		/**
		 * Selects a slot based on the probability specified.
		 * 
		 * @param probability the probability value.
		 * 
		 * @return the slot selected given the probability value.
		 * 
		 * @throws NoAvailableVertexException If a slot can not be selected.
		 */
		Slot spin(double probability)
			throws NoAvailableVertexException
		{
			double position = 0.0;

			for (int i = 0; i < slots.size(); i++)
			{
				position += slots.get(i).probability();

				if (probability <= position)
				{
					return slots.get(i);
				}
			}

			throw new NoAvailableVertexException(
				"No slot could be selected (wheel size " + slots.size() + ")");
		}

		/**
		 * Removes the specified slot.
		 * 
		 * @param slot the slot to be removed.
		 */
		void remove(Slot slot)
		{
			slots.remove(slot);
			normalize();
		}

		/**
		 * Tests if the wheel is empty (has no slots) or all slots have zero
		 * probability.
		 * 
		 * @return <code>true</code> is there are no available slots in the
		 *         wheel; <code>false</code> otherwise.
		 */
		boolean isEmpty()
		{
			boolean empty = slots.isEmpty();

			if (!empty)
			{
				for (Slot slot : slots)
				{
					if (slot.probability() > 0)
					{
						return empty;
					}
				}

				// all slots have 0 (zero) probability
				return !empty;
			}

			return empty;
		}

		/**
		 * Normalizes the probability values.
		 */
		private void normalize()
		{
			double total = 0.0;
			int size = slots.size();

			for (int i = 0; i < size; i++)
			{
				Slot slot = slots.get(i);
				slot.calculate();
				total += slot.probability();
			}

			for (int i = 0; i < size; i++)
			{
				slots.get(i).normalize(total);
			}
		}
	}

	/**
	 * This class represents a slot of the probabilistic roulette wheel.
	 */
	static class Slot
	{
		/**
		 * The vertex at the slot.
		 */
		private Edge<Term> edge;

		/**
		 * The probability of selecting this slot.
		 */
		private double probability;

		/**
		 * Default constructor.
		 * 
		 * @param vertex the vertex at this slot.
		 */
		Slot(Edge<Term> edge)
		{
			this.edge = edge;
		}

		/**
		 * Returns the vertex at this slot.
		 * 
		 * @return the vertex at this slot.
		 */
		Vertex<Term> vertex()
		{
			return edge.getTo();
		}

		/**
		 * Calculates the probability of selecting this slot. Note that in this
		 * implementation, the heuristic value is stored on the end-side vertex
		 * and the pheromone value is stored on the edge.
		 * 
		 * @param total the total heuristic value.
		 */
		void calculate()
		{
			probability =
				edge.getPheromone().getValue()
					* edge.getTo().getHeuristicValue().getValue();
		}

		/**
		 * Normalizes the probability value.
		 * 
		 * @param total the total probability value.
		 */
		void normalize(double total)
		{
			probability /= total;
		}

		/**
		 * Returns the probability associated with this slot.
		 * 
		 * @return the probability associated with this slot.
		 */
		double probability()
		{
			return probability;
		}
	}
}