
package net.sf.myra.cantminer;

import static net.sf.myra.datamining.Main.TR_PREFIX;
import static net.sf.myra.datamining.Main.TS_PREFIX;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.StringReader;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import net.sf.myra.datamining.Rule;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.function.AccuracyFunction;
import net.sf.myra.datamining.io.Helper;
import net.sf.myra.datamining.measure.AccuracyMeasure;
import net.sf.myra.datamining.model.RuleList;
import net.sf.myra.datamining.util.RuleParser;

public class ListPostProcessing
{
	private final static String START_RULE_LIST = "=== Discovered Model ===";

	private final static String START_RULE = "IF";

	private static List<RuleList> parseLists(String text, Metadata metadata)
	{
		ArrayList<RuleList> lists = new ArrayList<RuleList>();

		RuleParser parser = new RuleParser();
		BufferedReader input = new BufferedReader(new StringReader(text));
		String line = null;

		try
		{
			ArrayList<String> current = null;

			while ((line = input.readLine()) != null)
			{
				if (line.indexOf(START_RULE_LIST) != -1)
				{
					current = new ArrayList<String>();
				}
				else if (line.indexOf(START_RULE) != -1)
				{
					current.add(line);
				}
				// do we completed the previous list of rules?
				else if (current != null && !current.isEmpty())
				{
					RuleList list = new RuleList();

					for (String r : current)
					{
						list.add(parser.parse(r).convert(metadata));
					}

					lists.add(list);
					current = null;
				}
			}
		}
		catch (IOException e)
		{
			throw new RuntimeException(e);
		}

		return lists;
	}

	private static RuleList process(RuleList list, Dataset dataset)
	{
		AccuracyFunction function = new AccuracyFunction(dataset);
		ArrayList<Rule> rules = new ArrayList<Rule>();
		Rule defaultRule = null;

		for (Rule rule : list.getRules())
		{
			if (!rule.isEmpty())
			{
				rule.setCost(function.evaluate(rule));
				rules.add(rule);
			}
			else
			{
				defaultRule = rule;
			}
		}

		// (1) we sort rules based on their accuracy

		Collections.sort(rules, new Comparator<Rule>()
			{
				public int compare(Rule o1, Rule o2)
				{
					return Double.compare(o2.getCost().getValue(), o1.getCost()
						.getValue());
				}
			});
		RuleList modified = new RuleList();

		for (Rule rule : rules)
		{
			modified.add(rule);
		}

		modified.add(defaultRule);

		AccuracyMeasure measure = new AccuracyMeasure();
		double quality = measure.evaluate(dataset, modified).getValue();

		// (2) we try to prune rules and rule terms while the quality of
		// the candidate list does not decrease

		while (!rules.isEmpty())
		{
			Rule removed = rules.remove(rules.size() - 1);
			RuleList candidate = new RuleList();

			for (Rule rule : rules)
			{
				candidate.add(rule);
			}

			candidate.add(defaultRule);
			double q = measure.evaluate(dataset, candidate).getValue();

			// is it better without the rule?
			if (q >= quality)
			{
				modified = candidate;
				quality = q;
			}
			// so we try to prune the removed rule
			else
			{
				Rule clone = removed.clone();

				do
				{
					clone.remove(clone.getLast());
					candidate = new RuleList();

					for (Rule rule : rules)
					{
						candidate.add(rule);
					}

					candidate.add(clone);
					candidate.add(defaultRule);
					q = measure.evaluate(dataset, candidate).getValue();
				} while (q >= quality && !clone.isEmpty());

				if (q >= quality)
				{
					modified = candidate;
					break;
				}
			}
		}

		return modified;
	}

	public static void main(String[] args)
		throws Exception
	{
		final File output = new File(args[0]);
		final File directory = new File(args[1]);
		final String pattern = args[2];

		AccuracyMeasure measure = new AccuracyMeasure();
		ArrayList<Double> summary = new ArrayList<Double>();

		for (String file : output.list(new FilenameFilter()
			{
				public boolean accept(File dir, String name)
				{
					return name.contains(pattern) && name.endsWith(".out");
				}
			}))
		{
			StringBuffer buffer = new StringBuffer();
			BufferedReader reader =
				new BufferedReader(new FileReader(new File(output, file)));
			String line = null;

			while ((line = reader.readLine()) != null)
			{
				buffer.append(line);
				buffer.append("\n");
			}

			reader.close();

			List<RuleList> lists = null;
			ArrayList<Double> values = new ArrayList<Double>();

			for (int i = 0;; i++)
			{
				String fold = (i < 10) ? "0" + i : Integer.toString(i);
				File tr =
					new File(directory, TR_PREFIX + fold + "_" + pattern
						+ ".arff");
				File ts =
					new File(directory, TS_PREFIX + fold + "_" + pattern
						+ ".arff");

				if (tr.exists() && ts.exists())
				{
					Dataset trainingSet = Helper.open(tr);
					Dataset testSet = Helper.open(ts);

					// do we need to read the lists?
					if (lists == null)
					{
						lists =
							parseLists(buffer.toString(),
								trainingSet.getMetadata());
					}

					RuleList model = process(lists.get(i), trainingSet);
					values
						.add(measure.evaluate(testSet, model).getValue() * 100);
				}
				else
				{
					break;
				}
			}

			String current = stdev((Double[]) values.toArray(new Double[0]));
			summary.add(new Double(current.substring(0, current.indexOf(' '))));
		}

		System.out.println(stdev((Double[]) summary.toArray(new Double[0])));
		System.out.println("(" + summary.size() + ")");
	}

	public static String stdev(Double[] numbers)
	{
		if (numbers.length == 0)
		{
			return "0.0 +/- 0.0";
		}

		double mean = 0;

		for (int i = 0; i < numbers.length; i++)
		{
			mean += numbers[i];
		}

		mean /= numbers.length;
		double std = 0;

		for (int i = 0; i < numbers.length; i++)
		{
			std += Math.pow(numbers[i] - mean, 2);
		}

		std /= (numbers.length - 1);
		std /= numbers.length;
		std = Math.sqrt(std);

		if (Double.isNaN(std) || Double.isInfinite(std))
		{
			std = 0.0;
		}

		NumberFormat formatter = new DecimalFormat("#,##0.000");
		return formatter.format(mean) + " +/- " + formatter.format(std);
	}
}