/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.cantminer;

import java.util.List;
import net.sf.myra.datamining.data.ContinuousAttribute;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.IntervalBuilder;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.data.Term;
import net.sf.myra.datamining.data.IntervalBuilder.Interval;
import net.sf.myra.framework.Vertex;

/**
 * Utility class to calculate threshold values for continuous attributes.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class ThresholdHelper
{
	/**
	 * The interval builder helper instance.
	 */
	private IntervalBuilder builder;

	/**
	 * Default constructor.
	 * 
	 * @param metadata the current metadata instance.
	 */
	public ThresholdHelper(Metadata metadata)
	{
		builder = IntervalBuilder.getInstance(metadata);
	}

	/**
	 * Returns a copy of the specified vertex with a continuous attribute
	 * threshold information.
	 * 
	 * @param attribute the continuous attribute name instance.
	 * @param vertex the vertex representing the continuous attribute.
	 * @param instances the instances used in the threshold calculation.
	 * 
	 * @return a copy of the specified vertex with a continuous attribute
	 *         threshold information.
	 */
	public Vertex<Term> create(ContinuousAttribute attribute,
			Vertex<Term> vertex, List<Instance> instances)
	{
		return create(attribute, vertex, instances, null);
	}

	/**
	 * Returns a copy of the specified vertex with a continuous attribute
	 * threshold information.
	 * 
	 * @param attribute the continuous attribute name instance.
	 * @param vertex the vertex representing the continuous attribute.
	 * @param instances the instances used in the threshold calculation.
	 * @param value the target class value.
	 * 
	 * @return a copy of the specified vertex with a continuous attribute
	 *         threshold information.
	 */
	public Vertex<Term> create(ContinuousAttribute attribute,
			Vertex<Term> vertex, List<Instance> instances, String value)
	{
		Vertex<Term> clone = (Vertex<Term>) vertex.clone();
		Term info = clone.getInfo();

		// creates a continuous attribute interval

		Interval interval = null;

		if (value == null)
		{
			interval = builder.createSingle(instances, attribute);
		}
		else
		{
			interval = builder.createSingle(instances, attribute, value);
		}

		info.setValues(interval.getValues());
		info.setOperator(interval.getOperator());

		return clone;
	}

	/**
	 * Returns a copy of the specified vertex with a continuous attribute
	 * threshold information.
	 * 
	 * @param attribute the continuous attribute name instance.
	 * @param vertex the vertex representing the continuous attribute.
	 * @param instances the instances used in the threshold calculation.
	 * @param first indicates if the specified vertex is the first of the trail
	 * 
	 * @return a copy of the specified vertex with a continuous attribute
	 *         threshold information.
	 */
	public Vertex<Term> create(ContinuousAttribute attribute,
			Vertex<Term> vertex, List<Instance> instances, boolean first)
	{
		if (first && (attribute.getInitial() != null))
		{
			Interval interval = attribute.getInitial();
			Vertex<Term> clone = (Vertex<Term>) vertex.clone();

			Term info = clone.getInfo();
			info.setValues(interval.getValues().clone());
			info.setOperator(interval.getOperator());

			return clone;
		}
		else
		{
			return create(attribute, vertex, instances, null);
		}
	}
}