/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.cantminer;

import java.util.ArrayList;
import java.util.List;
import net.sf.myra.datamining.AbstractObjectiveFunction;
import net.sf.myra.datamining.ClassAssignator;
import net.sf.myra.datamining.Evaluator;
import net.sf.myra.datamining.Pruner;
import net.sf.myra.datamining.Rule;
import net.sf.myra.datamining.data.Attribute;
import net.sf.myra.datamining.data.ContinuousAttribute;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.data.Term;
import net.sf.myra.framework.Cost;
import net.sf.myra.framework.Trail;
import net.sf.myra.framework.Vertex;

/**
 * This class represents a threshold-aware pruner, which adds one term at a
 * time until the addition of a term does not improve the rule quality. If
 * the term being added represents a continuous attribute, a threshold value
 * is chosen based on the partial rule.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class ThresholdPruner
	extends Pruner
{
	public ThresholdPruner(AbstractObjectiveFunction function,
			ClassAssignator assignator)
	{
		super(function, assignator);
	}

	@SuppressWarnings("unchecked")
	@Override
	public Trail explore(Trail trail)
	{
		ClassAssignator assignator = getClassAssignator();
		AbstractObjectiveFunction function = getFunction();
		Metadata metadata = function.getDataset().getMetadata();
		ThresholdHelper helper = new ThresholdHelper(metadata);

		Rule rule = new Rule(metadata);
		Rule candidate = null;
		Cost cost = null;

		ArrayList<Vertex<?>> vertices =
			new ArrayList<Vertex<?>>(trail.getVertices());
		List<Instance> covered = function.getDataset().getInstances();

		while (!vertices.isEmpty() && !rule.equals(candidate))
		{
			candidate = rule;
			Vertex<?> vertex = null;

			for (Vertex<?> v : vertices)
			{
				Rule r = candidate.clone();

				Attribute attribute =
					metadata.get(((Vertex<Term>) v).getInfo().getIndex());

				if (attribute instanceof ContinuousAttribute)
				{
					ContinuousAttribute c = (ContinuousAttribute) attribute;
					v = helper.create(c, (Vertex<Term>) v, covered);
				}

				r.add(v);
				assignator.assign(r, function.getDataset());
				Cost c = function.evaluate(r);

				// we are favoring shorter rules, so if the candidate has the
				// same cost, we keep the original
				if (cost == null || c.compareTo(cost) == 1)
				{
					rule = r;
					cost = c;
					vertex = v;
				}
			}

			if (vertex != null)
			{
				vertices.remove(vertex);
				covered = Evaluator.findCoveredCases(rule, covered);
			}
		}

		return rule;
	}
}