/*
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2012 Fernando Esteban Barril Otero 
 */

package net.sf.myra.cantminer.alpha;

import static net.sf.myra.cantminer.alpha.Alpha.START_VERTEX;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

import net.sf.myra.cantminer.alpha.step.ContinuousStep;
import net.sf.myra.cantminer.alpha.step.NominalStep;
import net.sf.myra.cantminer.alpha.step.Step;
import net.sf.myra.datamining.Evaluator;
import net.sf.myra.datamining.Rule;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.data.Term;
import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.Edge;
import net.sf.myra.framework.Graph;
import net.sf.myra.framework.RandomGenerator;
import net.sf.myra.framework.Vertex;

/**
 * @author Fernando Esteban Barril Otero
 */
public class RuleFactory
{
	/**
	 * The property key under which the minimum rule cases value is stored.
	 */
	public static final String MINIMUM_RULE_CASES =
		"net.sf.myra.cantminer.alpha.rule.minimum";

	/**
	 * Minimum number of covered cases per rule.
	 */
	private final int MINIMUM_CASES = Configuration.getIntProperty(
		MINIMUM_RULE_CASES, 2);

	/**
	 * The problem construction graph to create the trails.
	 */
	private Graph<Step> graph;

	/**
	 * The dataset instance to validate the vertices selection.
	 */
	private Dataset dataset;

	/**
	 * Cache for continuous intervals. Only intervals for the first rule of
	 * a list are stored in the cache.
	 */
	private RuleThresholdCache cache;

	/**
	 * Default constructor.
	 * 
	 * @param graph the problem construction graph.
	 * @param dataset the dataset reference.
	 */
	public RuleFactory(Graph<Step> graph, Dataset dataset,
		RuleThresholdCache cache)
	{
		this.graph = graph;
		this.dataset = dataset;
		this.cache = cache;
	}

	public Rule create(final int level)
	{
		Metadata metadata = dataset.getMetadata();
		Rule rule = new Rule(metadata);
		Set<String> used = new HashSet<String>();

		List<Instance> covered = dataset.getInstances();
		Vertex<Step>[] vertices = graph.vertices();
		Vertex<Step> current = graph.get(START_VERTEX);
		used.add(START_VERTEX);

		// the partial representation of the rule
		String partial = rule.toString();

		while (covered.size() >= MINIMUM_CASES && !Evaluator.sameLabel(covered))
		{
			Set<String> incompatible = new HashSet<String>(used);

			while (incompatible.size() < vertices.length)
			{
				double total = 0.0;

				for (int i = 0; i < vertices.length; i++)
				{
					if (!incompatible.contains(vertices[i].getLabel()))
					{
						Edge<Step> edge = graph.getEdge(current, vertices[i]);
						total +=
							(edge.getPheromone(level).getValue() * vertices[i]
								.getHeuristicValue().getValue());
					}
				}

				// in case that we have vertices left but none of them is
				// suitable, we force the rule creation procedure to stop
				if (total == 0)
				{
					for (int i = 0; i < vertices.length; i++)
					{
						incompatible.add(vertices[i].getLabel());
					}

					break;
				}

				int selected = -1;
				double slot = RandomGenerator.getInstance().nextDouble();
				double probability = 0.0;

				for (int i = 0; i < vertices.length; i++)
				{
					if (!incompatible.contains(vertices[i].getLabel()))
					{
						Edge<Step> edge = graph.getEdge(current, vertices[i]);
						probability +=
							(edge.getPheromone(level).getValue() * vertices[i]
								.getHeuristicValue().getValue()) / total;

						if (slot <= probability)
						{
							selected = i;
							break;
						}
					}
				}

				if (selected == -1)
				{
					throw new RuntimeException(
						"Could not select a vertex for rule: " + rule);
				}

				Term term =
					select(vertices[selected], covered, level, rule.getSize(),
						partial);
				//Term term =
				//	vertices[selected].getInfo().select(covered, level,
				//		MINIMUM_CASES);

				if (term != null)
				{
					Rule clone = (Rule) rule.clone();

					Vertex<Term> vertex =
						new Vertex<Term>(vertices[selected].getLabel());
					vertex.setInfo(term);
					clone.add(vertex);

					List<Instance> set =
						Evaluator.findCoveredCases(clone, covered);
					int size = set.size();

					// if the size of covered examples have changed or the
					// current rule is empty, we add the term to the rule
					if (size != covered.size() || rule.isEmpty())
					{
						used.add(vertices[selected].getLabel());
						rule.add(vertex);

						covered = set;
						current = vertices[selected];

						break;
					}
					else
					{
						incompatible.add(vertices[selected].getLabel());
					}
				}
				// in case that we could not create a term for the selected
				// vertex
				else
				{
					incompatible.add(vertices[selected].getLabel());
				}
			}

			if (incompatible.size() == vertices.length)
			{
				// if we could not add a term during an iteration, we stop
				// the rule construction process
				break;
			}
		}

		return rule;
	}

	/**
	 * Returns the term represeting the selected vertex.
	 * 
	 * @param vertex the vertex to be selected.
	 * @param instances the current set of instances.
	 * @param level the level of the rule being created.
	 * @param position the position (index) of the vertex to be selected.
	 * 
	 * @return the term representing the selected vertex.
	 */
	private Term select(Vertex<Step> vertex, List<Instance> instances,
		int level, int position, String partial)
	{
		Term term = null;

		if (vertex.getInfo() instanceof NominalStep)
		{
			term = vertex.getInfo().select(instances, level, MINIMUM_CASES);
		}
		else if (vertex.getInfo() instanceof ContinuousStep)
		{
			if (level == 0)
			{
				term = cache.get(partial + "," + vertex.getLabel());

				if (term == null)
				{
					term =
						vertex.getInfo()
							.select(instances, level, MINIMUM_CASES);

					if (term != null)
					{
						cache.add(partial + "," + vertex.getLabel(), term);
					}
				}
			}
			else
			{
				term = vertex.getInfo().select(instances, level, MINIMUM_CASES);
			}
		}

		return term;
	}

	public Graph<Step> getGraph()
	{
		return graph;
	}
}