/*
 * $Id$ 
 */

package net.sf.myra.cantminer.alpha.step;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;

import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.NominalAttribute;
import net.sf.myra.datamining.data.Operator;
import net.sf.myra.datamining.data.Term;
import net.sf.myra.framework.RandomGenerator;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                    #$
 */
public class NominalStep
	extends AbstractStep
{
	private static final long serialVersionUID = 6203891233092276598L;

	/**
	 * The attribute index.
	 */
	private int index;

	/**
	 * The attribute represented by this step.
	 */
	private NominalAttribute attribute;

	/**
	 * The pheromone array for the values of the attribute. This array is
	 * only used by steps representing nominal attributes.
	 */
	private List<double[]> pheromone;

	/**
	 * The initial pheromone array.
	 */
	private double[] initial;

	/**
	 * Creates a new <code>NominalAttribute</code> instance.
	 * 
	 * @param attribute the attribute represented by this step.
	 */
	public NominalStep(NominalAttribute attribute, int index)
	{
		this.attribute = attribute;
		this.index = index;
		pheromone = new ArrayList<double[]>(1);
	}

	public Term select(List<Instance> instances, int level, int minimum)
	{
		HashSet<Integer> invalid = new HashSet<Integer>();
		Term term = null;

		// if the pheromone array is null, use the initial
		double[] choices =
			(level < pheromone.size()) ? pheromone.get(level) : initial;

		while ((term == null) && (invalid.size() < choices.length))
		{
			// selected an index value based on the pheromone array
			double total = 0.0;

			for (int i = 0; i < choices.length; i++)
			{
				if (!invalid.contains(i))
				{
					total += choices[i];
				}
			}

			double probability = RandomGenerator.getInstance().nextDouble();
			double position = 0.0;
			int value = -1;

			for (int i = 0; i < choices.length; i++)
			{
				if (!invalid.contains(i))
				{
					position += (choices[i] / total);

					if (probability <= position)
					{
						value = i;
						break;
					}
				}
			}

			if (value == -1)
			{
				// something is wrong, tell the world
				throw new RuntimeException("Could not select a value: "
					+ attribute.getName());
			}

			// let's try to create a term for the selected value

			term = new Term(attribute.getName(), index);
			term.setValues(value);
			term.setDescription(attribute.getValue(value));
			term.setOperator(Operator.EQUAL_TO);

			if (!accept(term, instances, minimum))
			{
				invalid.add(value);
				term = null;
			}
		}

		return term;
	}

	public NominalAttribute getAttribute()
	{
		return attribute;
	}

	public void setPheromone(double[] values, int level)
	{
		if (initial == null)
		{
			initial = values.clone();
		}

		if (level < pheromone.size())
		{
			pheromone.set(level, values.clone());
		}
		else
		{
			while (level >= pheromone.size())
			{
				pheromone.add(initial.clone());
			}

			pheromone.set(level, values.clone());
		}
	}

	public double[] getPheromone(int level)
	{
		if (level < pheromone.size())
		{
			return pheromone.get(level).clone();
		}
		else if (initial == null)
		{
			return new double[attribute.getValues().size()];
		}
		else
		{
			return initial.clone();
		}
	}

	public int levels()
	{
		return pheromone.size();
	}

	@Override
	public NominalStep clone()
	{
		NominalStep clone = (NominalStep) super.clone();
		clone.pheromone = new ArrayList<double[]>(pheromone.size());

		for (double[] values : pheromone)
		{
			clone.pheromone.add((double[]) values.clone());
		}

		return clone;
	}
}