/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.cantminer;

import java.util.List;
import java.util.Properties;

import net.sf.myra.antminer.DefaultPheromoneUpdater;
import net.sf.myra.antminer.ProbabilisticRuleFactory;
import net.sf.myra.datamining.AbstractObjectiveFunction;
import net.sf.myra.datamining.ClassAssignator;
import net.sf.myra.datamining.Classifier;
import net.sf.myra.datamining.ConvergenceTest;
import net.sf.myra.datamining.Evaluator;
import net.sf.myra.datamining.IterationTest;
import net.sf.myra.datamining.Main;
import net.sf.myra.datamining.MajorityClassAssignator;
import net.sf.myra.datamining.Pruner;
import net.sf.myra.datamining.Rule;
import net.sf.myra.datamining.data.Attribute;
import net.sf.myra.datamining.data.ContinuousAttribute;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.EntropyHeuristicInformation;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Label;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.data.Term;
import net.sf.myra.datamining.function.SensitivitySpecificityFunction;
import net.sf.myra.datamining.model.AbstractRuleModel;
import net.sf.myra.datamining.model.RuleList;
import net.sf.myra.framework.Colony;
import net.sf.myra.framework.CompoundStopCondition;
import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.Graph;
import net.sf.myra.framework.ObjectFactory;
import net.sf.myra.framework.Scheduler;
import net.sf.myra.framework.Vertex;

/**
 * This class represents the cAnt-Miner classification algorithm. More
 * information is available in following paper:
 * 
 * <pre>
 * &#64;ARTICLE{Otero08datamining,
 *    author = {F.E.B. Otero and A.A. Freitas and C.G. Johnson},
 *    title = {c{A}nt-{M}iner: an ant colony classification algorithm to cope with continuous attributes},
 *    booktitle = {Ant Colony Optimization and Swarm Intelligence (Proc. ANTS 2008), LNCS 5217},
 *    editor = {M. Dorigo et al.},
 *    publisher = {Springer},
 *    year = {2008},
 *    pages = {48-59}
 * }
 * </pre>
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class cAntMiner
	extends Classifier
{
	/**
	 * Default options filename.
	 */
	public static final String OPTIONS_FILE = "myra-cantminer.messages";

	/**
	 * Default constructor.
	 */
	public cAntMiner()
	{
		super(OPTIONS_FILE);
	}

	@Override
	public Properties filter(Properties properties)
	{
		properties.remove(Main.EXPORTER);
		return properties;
	}

	public AbstractRuleModel run(Dataset dataset)
	{
		if (dataset.getMetadata().isHierarchical())
		{
			throw new IllegalArgumentException(
				"Hierarchical dataset not supported.");
		}

		AbstractRuleModel discovered = new RuleList();

		int maxUncoveredCases =
			Configuration.getIntProperty(MAXIMUM_UNCOVERED_CASES,
				DEFAULT_MAXIMUM_UNCOVERED_CASES);

		Metadata metadata = dataset.getMetadata();
		Label majority = Evaluator.findMajorityClass(dataset);

		while (dataset.getSize() > maxUncoveredCases)
		{
			Rule rule = evolve(dataset);
			dataset.remove(Evaluator.findCoveredCases(rule, dataset));
			discovered.add(rule.clone());
		}

		if (!discovered.hasDefault())
		{
			if (!dataset.isEmpty())
			{
				majority = Evaluator.findMajorityClass(dataset);
			}

			Rule defaultRule = new Rule(metadata);
			defaultRule.setConsequent(majority);

			discovered.add(defaultRule);
		}

		return discovered;
	}

	/**
	 * Ant Colony Optimization (ACO) procedure to evolve a rule.
	 * 
	 * @param dataset
	 * @return
	 */
	private Rule evolve(Dataset dataset)
	{
		TrailFactory factory =
			new TrailFactory(dataset.getMetadata().toGraph(), dataset);

		Colony colony = new Colony();
		colony.setTrailFactory(factory);

		CompoundStopCondition condition = new CompoundStopCondition();
		condition.add(new ConvergenceTest());
		condition.add(new IterationTest());

		ClassAssignator assignator = new MajorityClassAssignator();

		AbstractObjectiveFunction function =
			(AbstractObjectiveFunction) ObjectFactory.create(Configuration
				.getProperty(RULE_QUALITY_FUNCTION,
					SensitivitySpecificityFunction.class.getName()),
				new Class<?>[] { Dataset.class }, new Object[] { dataset });

		Scheduler scheduler =
			new Scheduler(colony, condition, function,
				new DefaultPheromoneUpdater(), new EntropyHeuristicInformation(
					dataset));
		scheduler.setLocalSearch(new Pruner(function, assignator));

		scheduler.run();

		// get the best ant trail constructed
		return (Rule) colony.getBest().getBestTrail();
	}

	private static class TrailFactory
		extends ProbabilisticRuleFactory
	{
		/**
		 * The property key under which the minimum rule cases value is stored.
		 */
		public static final String MINIMUM_RULE_CASES =
			"net.sf.myra.cantminer.cases";

		/**
		 * The threshold selection helper.
		 */
		private ThresholdHelper helper;

		/**
		 * Defautl constructor.
		 * 
		 * @param graph the problem construction graph.
		 * @param dataset the dataset reference.
		 */
		public TrailFactory(Graph<?> graph, Dataset dataset)
		{
			super(graph, dataset, Configuration.getIntProperty(
				MINIMUM_RULE_CASES, NO_MINIMUM_CASES));
			this.helper = new ThresholdHelper(dataset.getMetadata());
		}

		@Override
		protected Vertex<Term> process(Metadata metadata, Vertex<Term> vertex,
				List<Instance> covered)
		{
			Attribute attribute = metadata.get(vertex.getInfo().getIndex());

			if (attribute instanceof ContinuousAttribute)
			{
				return helper.create((ContinuousAttribute) attribute, vertex,
					covered);
			}

			return vertex;
		}
	}
}