/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.cantminer;

import java.util.Properties;

import net.sf.myra.datamining.AbstractObjectiveFunction;
import net.sf.myra.datamining.BacktrackPruner;
import net.sf.myra.datamining.ClassAssignator;
import net.sf.myra.datamining.Classifier;
import net.sf.myra.datamining.ConvergenceTest;
import net.sf.myra.datamining.Evaluator;
import net.sf.myra.datamining.IterationTest;
import net.sf.myra.datamining.Main;
import net.sf.myra.datamining.MajorityClassAssignator;
import net.sf.myra.datamining.Rule;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.EntropyHeuristicInformation;
import net.sf.myra.datamining.data.Label;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.data.Term;
import net.sf.myra.datamining.function.SensitivitySpecificityFunction;
import net.sf.myra.datamining.model.AbstractRuleModel;
import net.sf.myra.datamining.model.RuleList;
import net.sf.myra.framework.Colony;
import net.sf.myra.framework.CompoundStopCondition;
import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.Graph;
import net.sf.myra.framework.ObjectFactory;
import net.sf.myra.framework.Scheduler;
import net.sf.myra.framework.TrailFactory;
import net.sf.myra.framework.Vertex;

/**
 * This class represents the cAnt-Miner2 classification algorithm.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class cAntMiner2
	extends Classifier
{
	/**
	 * Virtual starting vertex label.
	 */
	public static final String VIRTUAL_VERTEX = "[VIRTUAL]";

	/**
	 * Default constructor.
	 */
	public cAntMiner2()
	{
		super(cAntMiner.OPTIONS_FILE);
	}

	@Override
	public Properties filter(Properties properties)
	{
		properties.remove(Main.EXPORTER);
		return properties;
	}

	public AbstractRuleModel run(Dataset dataset)
	{
		if (dataset.getMetadata().isHierarchical())
		{
			throw new IllegalArgumentException(
				"Hierarchical dataset not supported.");
		}

		AbstractRuleModel discovered = new RuleList();

		int maxUncoveredCases =
			Configuration.getIntProperty(MAXIMUM_UNCOVERED_CASES,
				DEFAULT_MAXIMUM_UNCOVERED_CASES);

		Metadata metadata = dataset.getMetadata();
		Label majority = Evaluator.findMajorityClass(dataset);

		while (dataset.getSize() > maxUncoveredCases)
		{
			Rule rule = evolve(dataset);
			dataset.remove(Evaluator.findCoveredCases(rule, dataset));
			discovered.add(rule.clone());
		}

		if (!discovered.hasDefault())
		{
			if (!dataset.isEmpty())
			{
				majority = Evaluator.findMajorityClass(dataset);
			}

			Rule defaultRule = new Rule(metadata);
			defaultRule.setConsequent(majority);

			discovered.add(defaultRule);
		}

		return discovered;
	}

	/**
	 * Ant Colony Optimization (ACO) procedure to evolve a rule.
	 * 
	 * @param dataset
	 * @return
	 */
	private Rule evolve(Dataset dataset)
	{
		Graph<Term> virtual = new Graph<Term>();
		virtual.add(new Vertex<Term>(VIRTUAL_VERTEX));

		// creates the virtual starting vertex
		virtual.merge(dataset.getMetadata().toGraph());

		TrailFactory factory = new EdgeTrailFactory(virtual, dataset);

		Colony colony = new Colony();
		colony.setTrailFactory(factory);

		CompoundStopCondition condition = new CompoundStopCondition();
		condition.add(new ConvergenceTest());
		condition.add(new IterationTest());

		ClassAssignator assignator = new MajorityClassAssignator();

		AbstractObjectiveFunction function =
			(AbstractObjectiveFunction) ObjectFactory.create(Configuration
				.getProperty(RULE_QUALITY_FUNCTION,
					SensitivitySpecificityFunction.class.getName()),
				new Class<?>[] { Dataset.class }, new Object[] { dataset });

		Scheduler scheduler =
			new Scheduler(colony, condition, function,
				new EdgePheromoneUpdater(), new EntropyHeuristicInformation(
					dataset));
		scheduler.setLocalSearch(new BacktrackPruner(function, assignator));

		scheduler.run();

		// get the best ant trail constructed
		return (Rule) colony.getBest().getBestTrail();
	}
}