/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2011  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.cantminer.pittsburgh;

import java.util.List;

import net.sf.myra.datamining.Rule;
import net.sf.myra.datamining.model.RuleList;

/**
 * A candidate rule list reprents a rule list created during the search for the
 * best rule list.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class CandidateRuleList
	extends RuleList
	implements Comparable<CandidateRuleList>
{
	private static final long serialVersionUID = -3080400314456188558L;

	/**
	 * The quality of the candidate rule list. The higher the value the better
	 * the quality.
	 */
	private double quality;

	/**
	 * The iteration in which this candidate rule list was created. The
	 * default value is <code>-1</code>.
	 */
	private int iteration = -1;

	/**
	 * Returns the iteration in which this candidate rule list was created. If
	 * the value has not been set, returns <code>-1</code>.
	 * 
	 * @return the iteration in which this candidate rule list was created.
	 */
	public int getIteration()
	{
		return iteration;
	}

	/**
	 * Sets the iteration in which this candidate rule list was created.
	 * 
	 * @param iteration the value to set.
	 */
	public void setIteration(int iteration)
	{
		this.iteration = iteration;
	}

	/**
	 * Returns the quality of this rule list.
	 * 
	 * @return the quality of this rule list.
	 */
	public double getQuality()
	{
		return quality;
	}

	/**
	 * Sets the quality of this rule list.
	 * 
	 * @param quality the quality to set.
	 */
	public void setQuality(double quality)
	{
		this.quality = quality;
	}

	public int compareTo(CandidateRuleList o)
	{
		// (1) compare the quality

		int c = Double.compare(quality, o.quality);

		if (c == 0)
		{
			// (2) compare the number of rules

			c = Double.compare(o.size(), size());

			if (c == 0)
			{
				int total1 = 0;

				for (Rule rule : getRules())
				{
					total1 += rule.getSize();
				}

				int total2 = 0;

				for (Rule rule : o.getRules())
				{
					total2 += rule.getSize();
				}

				// (3) compare the total number of terms

				return Double.compare(total2, total1);
			}
		}

		return c;
	}

	@Override
	public boolean equals(Object o)
	{
		return equals((CandidateRuleList) o);
	}

	private boolean equals(CandidateRuleList other)
	{
		if (compareTo(other) == 0)
		{
			List<Rule> set1 = getRules();
			List<Rule> set2 = other.getRules();

			for (int i = 0; i < set1.size(); i++)
			{
				if (!set1.get(i).equals(set2.get(i)))
				{
					return false;
				}
			}

			return true;
		}

		return false;
	}
}