/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2011  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.cantminer.pittsburgh;

import static net.sf.myra.datamining.IterationTest.DEFAULT_MAXIMUM_ITERATIONS;
import static net.sf.myra.datamining.IterationTest.MAXIMUM_ITERATIONS;

import java.util.List;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import net.sf.myra.datamining.AbstractObjectiveFunction;
import net.sf.myra.datamining.BacktrackPruner;
import net.sf.myra.datamining.ClassAssignator;
import net.sf.myra.datamining.Classifier;
import net.sf.myra.datamining.ConvergenceTest;
import net.sf.myra.datamining.Evaluator;
import net.sf.myra.datamining.Main;
import net.sf.myra.datamining.MajorityClassAssignator;
import net.sf.myra.datamining.Measure;
import net.sf.myra.datamining.Model;
import net.sf.myra.datamining.Rule;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.EntropyContent;
import net.sf.myra.datamining.data.EntropyHeuristicInformation;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Label;
import net.sf.myra.datamining.data.Term;
import net.sf.myra.datamining.function.SensitivitySpecificityFunction;
import net.sf.myra.datamining.measure.AccuracyMeasure;
import net.sf.myra.framework.Colony;
import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.Graph;
import net.sf.myra.framework.LocalSearch;
import net.sf.myra.framework.ObjectFactory;
import net.sf.myra.framework.Vertex;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class cAntMinerPB
	extends Classifier
{
	/**
	 * Virtual starting vertex label.
	 */
	public static final String START_VERTEX = "[START]";

	/**
	 * The default configuration file name.
	 */
	public static final String OPTIONS_FILE = "myra-cantminer-pb.messages";

	/**
	 * The property key name under which the percentage of uncovered cases is
	 * stored.
	 */
	public static final String UNCOVERED_CASES_PCT =
		"net.sf.myra.cantminer.pittsburgh.uncovered";

	/**
	 * The property key name under which the measure class name to be used to
	 * evaluate a list of rules is stored.
	 */
	public static final String LIST_MEASURE =
		"net.sf.myra.cantminer.pittsburgh.list.measure";

	/**
	 * The property key name under which the pruner class name is stored.
	 */
	public static final String PRUNER =
		"net.sf.myra.cantminer.pittsburgh.pruner";

	/**
	 * Default uncovered percentage.
	 */
	public static final double DEFAULT_PERCENTAGE = 1.0;

	/**
	 * The logger instance.
	 */
	private Logger logger = Logger.getLogger(cAntMinerPB.class.getName());

	/**
	 * Default constructor.
	 */
	public cAntMinerPB()
	{
		super(OPTIONS_FILE);
	}

	@Override
	public Properties filter(Properties properties)
	{
		properties.remove(Main.EXPORTER);
		properties.remove(ConvergenceTest.CONVERGENCE_TEST_SIZE);
		properties.remove(Classifier.MAXIMUM_UNCOVERED_CASES);
		return properties;
	}

	@Override
	public Model run(Dataset dataset)
	{
		if (dataset.getMetadata().isHierarchical())
		{
			throw new IllegalArgumentException(
				"Hierarchical dataset not supported.");
		}

		final int colony = Configuration.getIntProperty(Colony.COLONY_SIZE);

		final int maxUncoveredCases =
			(int) ((Configuration.getDoubleProperty(UNCOVERED_CASES_PCT,
				DEFAULT_PERCENTAGE) / 100) * dataset.getSize() + 0.5);

		// (1) creates the construction graph

		Graph<Term> graph = new Graph<Term>();
		graph.add(new Vertex<Term>(START_VERTEX));
		graph.merge(dataset.getMetadata().toGraph());

		Measure measure =
			(Measure) ObjectFactory.create(Configuration.getProperty(
				LIST_MEASURE, AccuracyMeasure.class.getName()));

		PheromoneUpdater updater = new PheromoneUpdater();
		updater.initialize(graph);

		Label majority = Evaluator.findMajorityClass(dataset);
		CandidateRuleList discovered = null;

		// (2) while the algorithm has not converged

		logger.info("Run started.");

		final int maximum =
			Configuration.getIntProperty(MAXIMUM_ITERATIONS,
				DEFAULT_MAXIMUM_ITERATIONS);
		int iteration = 0;

		// the entropy information of the complete data set (we cache this
		// to improve the running time of the algorithm)
		EntropyContent initial = new EntropyContent(dataset);
		// cache for continuous intervals
		RuleThresholdCache cache = new RuleThresholdCache();

		while (iteration < maximum)
		{
			CandidateRuleList current = null;

			for (int i = 0; i < colony; i++)
			{
				Dataset clone = dataset.clone();
				RuleFactory factory = new RuleFactory(graph, clone, cache);
				CandidateRuleList candidate = new CandidateRuleList();

				while (clone.getSize() > maxUncoveredCases)
				{
					EntropyHeuristicInformation heuristic =
						new EntropyHeuristicInformation(clone);

					if (candidate.getRules().isEmpty())
					{
						// if is the first rule, we use the cached entropy
						// content
						heuristic.initialize(graph, initial);
					}
					else
					{
						heuristic.initialize(graph);
					}

					AbstractObjectiveFunction function =
						(AbstractObjectiveFunction) ObjectFactory
							.create(
								Configuration.getProperty(
									RULE_QUALITY_FUNCTION,
									SensitivitySpecificityFunction.class
										.getName()),
								new Class<?>[] { Dataset.class },
								new Object[] { clone });

					LocalSearch pruner =
						(LocalSearch) ObjectFactory.create(Configuration
							.getProperty(PRUNER,
								BacktrackPruner.class.getName()),
							new Class<?>[] { AbstractObjectiveFunction.class,
								ClassAssignator.class }, new Object[] {
								function, new MajorityClassAssignator() });

					Rule rule =
						(Rule) pruner.explore(factory.create(candidate.size()));
					clone.remove(Evaluator.findCoveredCases(rule, clone));
					candidate.add(rule.clone());
				}

				if (!candidate.hasDefault())
				{
					if (!clone.isEmpty())
					{
						majority = Evaluator.findMajorityClass(clone);
					}

					Rule defaultRule = new Rule(clone.getMetadata());
					defaultRule.setConsequent(majority);
					candidate.add(defaultRule);
				}

				candidate.setQuality(measure.evaluate(dataset, candidate)
					.doubleValue());

				if ((current == null) || (candidate.compareTo(current) == 1))
				{
					current = candidate;
				}
			}

			updater.update(current, graph);

			if ((discovered == null) || (current.compareTo(discovered) == 1))
			{
				discovered = current;
			}

			if (updater.hasConverged(graph, current))
			{
				logger.info("Search has converged at iteration " + iteration);
				break;
			}
			// this is experimental and assumes that the highest quality is
			// 1.0 (which is usually the case)
			else if (discovered.getQuality() == 1.0
				&& current.getQuality() == 1.0)
			{
				logger
					.info("Search has found the best list on the training data at iteration "
						+ iteration);
				break;
			}

			iteration++;
		}

		if (iteration == maximum)
		{
			logger.info("Maximum number of iterations reached.");
		}

		logger.info("Rule list quality during training: "
			+ discovered.getQuality());

		if (logger.isLoggable(Level.FINE))
		{
			StringBuffer coverage = new StringBuffer();
			int index = 1;

			for (Rule rule : discovered.getRules())
			{
				List<Instance> covered =
					Evaluator.findCoveredCases(rule, dataset);

				coverage.append("R");
				coverage.append(index);
				coverage.append("=");
				coverage.append(covered.size());

				if (!rule.isEmpty())
				{
					coverage.append(" | ");
				}

				dataset.remove(covered);
				index++;
			}

			logger.fine("Coverage per rule: " + coverage.toString());
		}

		return discovered;
	}
}