/*
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2012 Fernando Esteban Barril Otero and Matthew Medland
 */

package net.sf.myra.cantminer.pittsburgh.dynamic;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import net.sf.myra.datamining.AbstractObjectiveFunction;
import net.sf.myra.datamining.data.Term;
import net.sf.myra.datamining.function.ConfidenceCoverageFunction;
import net.sf.myra.datamining.function.CostMeasureFunction;
import net.sf.myra.datamining.function.ErrorBasedFunction;
import net.sf.myra.datamining.function.FmeasureFunction;
import net.sf.myra.datamining.function.JaccardFunction;
import net.sf.myra.datamining.function.KlosgenFunction;
import net.sf.myra.datamining.function.LaplaceFunction;
import net.sf.myra.datamining.function.MEstimateFunction;
import net.sf.myra.datamining.function.RelativeCostMeasureFunction;
import net.sf.myra.datamining.function.SensitivitySpecificityFunction;
import net.sf.myra.framework.RandomGenerator;

/**
 * @author Fernando Esteban Barril Otero
 * @author Matthew Medland
 */
public class FunctionTerm
	extends Term
{
	private static final long serialVersionUID = -8971120544330445841L;

	private List<double[]> pheromone;

	/**
	 * The initial pheromone array.
	 */
	private double[] initial;

	private AbstractObjectiveFunction[] functions;

	private HashMap<String, Integer> indexes;

	public FunctionTerm()
	{
		super(null, -1);

		ArrayList<AbstractObjectiveFunction> available =
			new ArrayList<AbstractObjectiveFunction>();

		// non-parametric functions

		//available.add(new AccuracyFunction(null));
		available.add(new LaplaceFunction(null));
		available.add(new ConfidenceCoverageFunction(null));
		available.add(new JaccardFunction(null));
		available.add(new ErrorBasedFunction(null));
		available.add(new SensitivitySpecificityFunction(null));

		// for the parametric functions we use the optimum values from
		// Janssen and Furnkranze (2010)

		available.add(new CostMeasureFunction(null, 0.437));
		available.add(new RelativeCostMeasureFunction(null, 0.342));
		available.add(new KlosgenFunction(null, 0.4324));
		available.add(new MEstimateFunction(null, 22.466));
		available.add(new FmeasureFunction(null, 0.5));

		functions =
			available.toArray(new AbstractObjectiveFunction[available.size()]);
		pheromone = new ArrayList<double[]>(1);

		indexes = new HashMap<String, Integer>(functions.length, 1.0f);
		int index = 0;

		for (AbstractObjectiveFunction function : functions)
		{
			indexes.put(function.getClass().getSimpleName(), index);
			index++;
		}
	}

	public AbstractObjectiveFunction select(int level)
	{
		// if the pheromone array is null, use the initial
		double[] choices =
			(level < pheromone.size()) ? pheromone.get(level) : initial;

		// selected an index value based on the pheromone array
		double total = 0.0;

		for (int i = 0; i < choices.length; i++)
		{
			total += choices[i];
		}

		double probability = RandomGenerator.getInstance().nextDouble();
		double position = 0.0;
		int value = -1;

		for (int i = 0; i < choices.length; i++)
		{
			position += (choices[i] / total);

			if (probability <= position)
			{
				value = i;
				break;
			}
		}

		if (value == -1)
		{
			// something is wrong, tell the world
			throw new RuntimeException("Could not select a function for level "
				+ level);
		}

		return functions[value];
	}

	public void setPheromone(double[] values, int level)
	{
		if (initial == null)
		{
			initial = values.clone();
		}

		if (level < pheromone.size())
		{
			pheromone.set(level, values.clone());
		}
		else
		{
			while (level >= pheromone.size())
			{
				pheromone.add(initial.clone());
			}

			pheromone.set(level, values.clone());
		}
	}

	public double[] getPheromone(int level)
	{
		if (level < pheromone.size())
		{
			return pheromone.get(level).clone();
		}
		else if (initial == null)
		{
			return new double[functions.length];
		}
		else
		{
			return initial.clone();
		}
	}

	public int levels()
	{
		return pheromone.size();
	}

	public int getIndex(AbstractObjectiveFunction function)
	{
		return indexes.get(function.getClass().getSimpleName());
	}
}