/*
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2011  Fernando Esteban Barril Otero and Matthew Medland
 */

package net.sf.myra.cantminer.pittsburgh.dynamic;

import static net.sf.myra.cantminer.pittsburgh.PheromoneUpdater.EVAPORATION_FACTOR;
import static net.sf.myra.cantminer.pittsburgh.PheromoneUpdater.INITIAL;
import static net.sf.myra.cantminer.pittsburgh.PheromoneUpdater.P_BEST;
import static net.sf.myra.cantminer.pittsburgh.cAntMinerPB.START_VERTEX;
import net.sf.myra.cantminer.pittsburgh.CandidateRuleList;
import net.sf.myra.datamining.Rule;
import net.sf.myra.datamining.data.Term;
import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.Edge;
import net.sf.myra.framework.Graph;
import net.sf.myra.framework.Pheromone;
import net.sf.myra.framework.Vertex;

/**
 * @author Fernando Esteban Barril Otero
 * @author Matthew Medland
 */
public class FunctionalPheromoneUpdater
{
	/**
	 * The pheromone evaporation factor (default 0.90);
	 */
	private double factor;

	/**
	 * The p_best probability (default 0.05) to estimate the tMin value.
	 */
	private double pBest;

	/**
	 * The maximum pheromone value.
	 */
	private double tMax;

	/**
	 * The minimum pheromone value.
	 */
	private double tMin;

	/**
	 * The quality of the current best solution.
	 */
	private double current = 0.0;

	public void initialize(Graph<Term> graph)
	{
		graph.apply(new Graph.EdgeInitializationTransform(INITIAL));

		// at the begining of the search, tMax = tMin, therefore only the
		// heuristic information will be used
		tMax = INITIAL;
		tMin = INITIAL;

		factor = Configuration.getDoubleProperty(EVAPORATION_FACTOR, 0.90);
		pBest = Configuration.getDoubleProperty(P_BEST, 0.05);

		FunctionTerm term = (FunctionTerm) graph.get(START_VERTEX).getInfo();
		double[] values = term.getPheromone(0);

		for (int i = 0; i < values.length; i++)
		{
			values[i] = INITIAL;
		}

		term.setPheromone(values, 0);
	}

	public void update(CandidateRuleList candidate, Graph<Term> graph)
	{
		// (1) updates tMin and tMan values if we have found a better candidate 

		if (candidate.getQuality() > current)
		{
			current = candidate.getQuality();
			double n = graph.getVertices().size();

			// the average should represent the average number of options an
			// ant has at each choice point
			double average = (n / 2.0) * candidate.size();
			double pDec = Math.pow(pBest, 1 / (double) n);

			tMax = (1 / (1 - factor)) * (current / 5.0);
			tMin = (tMax * (1 - pDec)) / ((average - 1) * pDec);

			if (tMin > tMax)
			{
				tMin = tMax;
			}
		}

		for (int i = 0; i < candidate.size(); i++)
		{
			FunctionTerm term =
				(FunctionTerm) graph.get(START_VERTEX).getInfo();
			term.setPheromone(term.getPheromone(i), i);

			for (Vertex<Term> vertex : graph.vertices())
			{
				for (Edge<Term> edge : graph.getEdges(vertex))
				{
					edge.setPheromone(edge.getPheromone(i), i);
				}
			}
		}

		// (2) evaporates the pheromone (in all levels) of all edges

		graph.apply(new Graph.EdgeEvaporateTransform(factor));

		FunctionTerm term = (FunctionTerm) graph.get(START_VERTEX).getInfo();

		for (int i = 0; i < term.levels(); i++)
		{
			double[] pheromone = term.getPheromone(i);

			for (int j = 0; j < pheromone.length; j++)
			{
				pheromone[j] = pheromone[j] * factor;

				if (pheromone[j] > tMax)
				{
					pheromone[j] = tMax;
				}
				else if (pheromone[j] < tMin)
				{
					pheromone[j] = tMin;
				}
			}

			term.setPheromone(pheromone, i);
		}

		// (3) updates the pheromone of the used edges at each level (a rule
		//     in the candidate rule list represents a new level)

		double delta = candidate.getQuality() / 5.0;
		int level = 0;

		for (Rule rule : candidate.getRules())
		{
			if (!rule.isEmpty())
			{
				Vertex<Term> current = graph.get(START_VERTEX);

				term = (FunctionTerm) current.getInfo();
				double[] pheromone = term.getPheromone(level);
				int index =
					term.getIndex(((FunctionalRule) rule).getFunction());
				pheromone[index] = pheromone[index] + delta;

				if (pheromone[index] > tMax)
				{
					pheromone[index] = tMax;
				}

				term.setPheromone(pheromone, level);

				for (Vertex<?> vertex : rule.getVertices())
				{
					Vertex<Term> to = graph.get(vertex.getLabel());
					Edge<Term> edge = graph.getEdge(current, to);

					Pheromone p = edge.getPheromone(level);
					p.setValue(p.getValue() + delta);

					current = to;
				}
			}

			level++;
		}

		// (4) guarantees that pehromone values are within [tMin,tMax] range

		graph.apply(new Graph.EdgeRangeTransform(tMin, tMax));
	}

	@SuppressWarnings("unchecked")
	public boolean hasConverged(Graph<Term> graph, CandidateRuleList best)
	{
		int level = 0;

		for (Rule rule : best.getRules())
		{
			if (!rule.isEmpty())
			{
				Vertex<Term> current = graph.get(START_VERTEX);

				for (Vertex<?> vertex : rule.getVertices())
				{
					int maxCount = 0;
					int minCount = 0;

					for (Edge<Term> edge : graph.getEdges(current))
					{
						double value = edge.getPheromone(level).getValue();

						if (precision(value) == precision(tMin))
						{
							minCount++;
						}
						else if (precision(value) == precision(tMax))
						{
							maxCount++;
						}
					}

					int size = graph.getEdges(current).size();

					if ((minCount != (size - 1)) || (maxCount != 1))
					{
						return false;
					}

					current = (Vertex<Term>) vertex;
				}
			}

			level++;
		}

		return true;
	}

	/**
	 * Returns the double value with up to four decimal digits.
	 * 
	 * @param value the value to be converted.
	 * 
	 * @return the double value with up to four decimal digits.
	 */
	private final double precision(double value)
	{
		int v = (int) (value * 100);
		return v / 100.0;
	}
}