/*
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2012 Fernando Esteban Barril Otero
 */

package net.sf.myra.cantminer.pittsburgh.unordered;

import static net.sf.myra.datamining.IterationTest.DEFAULT_MAXIMUM_ITERATIONS;
import static net.sf.myra.datamining.IterationTest.MAXIMUM_ITERATIONS;

import java.util.Collection;
import java.util.List;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import net.sf.myra.cantminer.pittsburgh.RuleThresholdCache;
import net.sf.myra.cantminer.pittsburgh.cAntMinerPB;
import net.sf.myra.cantminer.pittsburgh.dynamic.FunctionTerm;
import net.sf.myra.cantminer.pittsburgh.dynamic.FunctionalPheromoneUpdater;
import net.sf.myra.cantminer.pittsburgh.dynamic.FunctionalRule;
import net.sf.myra.cantminer.preselect.ClassHeuristicInformation;
import net.sf.myra.datamining.AbstractObjectiveFunction;
import net.sf.myra.datamining.BacktrackPruner;
import net.sf.myra.datamining.ClassAssignator;
import net.sf.myra.datamining.Classifier;
import net.sf.myra.datamining.ConvergenceTest;
import net.sf.myra.datamining.Evaluator;
import net.sf.myra.datamining.Main;
import net.sf.myra.datamining.Measure;
import net.sf.myra.datamining.Model;
import net.sf.myra.datamining.NoClassAssignator;
import net.sf.myra.datamining.Rule;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Label;
import net.sf.myra.datamining.data.Term;
import net.sf.myra.datamining.function.LaplaceFunction;
import net.sf.myra.datamining.measure.AccuracyMeasure;
import net.sf.myra.framework.Colony;
import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.Graph;
import net.sf.myra.framework.LocalSearch;
import net.sf.myra.framework.ObjectFactory;
import net.sf.myra.framework.Vertex;

/**
 * This class implements the Unordered <i>c</i>Ant-Miner<sub>PB</sub>
 * algorithm.
 * 
 * @author Fernando Esteban Barril Otero
 */
public class UcAntMinerPB
	extends Classifier
{
	/**
	 * The logger instance.
	 */
	private Logger logger = Logger.getLogger(UcAntMinerPB.class.getName());
	
	/**
	 * The default configuration file name.
	 */
	public static final String OPTIONS_FILE = "myra-ucantminer-pb.messages";

	/**
	 * Default constructor.
	 */
	public UcAntMinerPB()
	{
		super(OPTIONS_FILE);
	}

	@Override
	public Properties filter(Properties properties)
	{
		properties.remove(Main.EXPORTER);
		properties.remove(ConvergenceTest.CONVERGENCE_TEST_SIZE);
		properties.remove(Classifier.MAXIMUM_UNCOVERED_CASES);
		return properties;
	}

	@Override
	public Model run(Dataset dataset)
	{
		if (dataset.getMetadata().isHierarchical())
		{
			throw new IllegalArgumentException(
				"Hierarchical dataset not supported.");
		}

		final int colony = Configuration.getIntProperty(Colony.COLONY_SIZE);

		// (1) creates the construction graph

		Graph<Term> graph = new Graph<Term>();

		Vertex<Term> start = new Vertex<Term>(cAntMinerPB.START_VERTEX);
		start.setInfo(new FunctionTerm());

		graph.add(start);
		graph.merge(dataset.getMetadata().toGraph());

		Measure measure =
			(Measure) ObjectFactory.create(Configuration.getProperty(
				cAntMinerPB.LIST_MEASURE, AccuracyMeasure.class.getName()));

		FunctionalPheromoneUpdater updater = new FunctionalPheromoneUpdater();
		updater.initialize(graph);

		UnorderedRuleList discovered = null;
		Collection<String> labels =
			dataset.getMetadata().getTarget().getValues();

		// (2) while the algorithm has not converged

		logger.info("Run started.");

		final int maximum =
			Configuration.getIntProperty(MAXIMUM_ITERATIONS,
				DEFAULT_MAXIMUM_ITERATIONS);
		int iteration = 0;

		Label majority = Evaluator.findMajorityClass(dataset);
		// cache for continuous intervals
		RuleThresholdCache cache = new RuleThresholdCache();

		while (iteration < maximum)
		{
			UnorderedRuleList current = null;

			for (int i = 0; i < colony; i++)
			{
				UnorderedRuleList candidate = new UnorderedRuleList();

				for (String label : labels)
				{
					Dataset clone = dataset.clone();
					FixedClassRuleFactory factory =
						new FixedClassRuleFactory(graph, clone, cache);

					int maxUncoveredCases =
						(int) ((Configuration.getDoubleProperty(
							cAntMinerPB.UNCOVERED_CASES_PCT,
							cAntMinerPB.DEFAULT_PERCENTAGE) / 100)
							* dataset.getSize(label) + 0.5);
					Label predicted =
						clone.getMetadata().getLabel().toLabel(label);

					while (clone.getSize(label) > maxUncoveredCases)
					{
						ClassHeuristicInformation heuristic =
							new ClassHeuristicInformation(clone);
						heuristic.initialize(graph);

						FunctionalRule rule =
							new FunctionalRule(clone.getMetadata());
						rule.setConsequent(predicted);

						for (Vertex<?> vertex : factory.create(
							candidate.size(), predicted).getVertices())
						{
							rule.add(vertex);
						}

						// selects an objective function for the rule

						rule.setFunction(((FunctionTerm) start.getInfo())
							.select(candidate.size()));
						AbstractObjectiveFunction function = rule.getFunction();
						function.setDataset(clone);

						if (!rule.isEmpty())
						{
							LocalSearch pruner =
								(LocalSearch) ObjectFactory.create(
									Configuration.getProperty(
										cAntMinerPB.PRUNER,
										BacktrackPruner.class.getName()),
									new Class<?>[] {
										AbstractObjectiveFunction.class,
										ClassAssignator.class }, new Object[] {
										function, new NoClassAssignator() });

							rule = (FunctionalRule) pruner.explore(rule);
						}

						clone.remove(Evaluator.findCorrectCases(rule, clone));
						candidate.add(rule.clone());
					}
				}

				// calculates the confidence of each rule

				AbstractObjectiveFunction function =
					(AbstractObjectiveFunction) ObjectFactory.create(
						Configuration.getProperty(RULE_QUALITY_FUNCTION,
							LaplaceFunction.class.getName()),
						new Class<?>[] { Dataset.class },
						new Object[] { dataset });

				for (Rule rule : candidate.getRules())
				{
					rule.setConfidence(Evaluator.getConfidenceFactor(rule,
						dataset));
					rule.setCost(function.evaluate(rule));

				}

				// evaluates the unordered list

				candidate.setMajority(majority);
				candidate.setQuality(measure.evaluate(dataset, candidate)
					.doubleValue());

				if ((current == null) || (candidate.compareTo(current) == 1))
				{
					current = candidate;
				}
			}

			updater.update(current, graph);

			if ((discovered == null) || (current.compareTo(discovered) == 1))
			{
				discovered = current;
			}

			if (updater.hasConverged(graph, current))
			{
				logger.info("Search has converged at iteration " + iteration);
				break;
			}
			// this is experimental and assumes that the highest quality is
			// 1.0 (which is usually the case)
			else if (discovered.getQuality() == 1.0
				&& current.getQuality() == 1.0)
			{
				logger
					.info("Search has found the best list on the training data at iteration "
						+ iteration);
				break;
			}

			iteration++;
		}

		if (iteration == maximum)
		{
			logger.info("Maximum number of iterations reached.");
		}

		logger.info("Rule list quality during training: "
			+ discovered.getQuality());

		if (logger.isLoggable(Level.FINE))
		{
			StringBuffer coverage = new StringBuffer();
			int index = 1;

			for (Rule rule : discovered.getRules())
			{
				List<Instance> covered =
					Evaluator.findCoveredCases(rule, dataset);

				coverage.append("R");
				coverage.append(index);
				coverage.append("=");
				coverage.append(covered.size());

				if (!rule.isEmpty())
				{
					coverage.append(" | ");
				}

				dataset.remove(covered);
				index++;
			}

			logger.fine("Coverage per rule: " + coverage.toString());
		}

		if (logger.isLoggable(Level.FINE))
		{
			StringBuffer rules = new StringBuffer();

			for (Rule rule : discovered.getRules())
			{
				rules.append("\n");
				rules.append(rule.toString());
				rules.append(" USING ");
				rules.append(((FunctionalRule) rule).getFunctionName());
			}

			logger.fine("Discovered list of rules (with functions):"
				+ rules.toString());
		}

		return discovered;
	}
}