/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.cantminer.preselect;

import static net.sf.myra.cantminer.EdgeTrailFactory.MINIMUM_RULE_CASES;
import static net.sf.myra.cantminer.EdgeTrailFactory.NO_MINIMUM_CASES;
import static net.sf.myra.cantminer.cAntMiner2.VIRTUAL_VERTEX;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

import net.sf.myra.cantminer.ThresholdHelper;
import net.sf.myra.datamining.Evaluator;
import net.sf.myra.datamining.Rule;
import net.sf.myra.datamining.data.Attribute;
import net.sf.myra.datamining.data.ContinuousAttribute;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.data.Term;
import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.Edge;
import net.sf.myra.framework.Graph;
import net.sf.myra.framework.RandomGenerator;
import net.sf.myra.framework.Trail;
import net.sf.myra.framework.TrailFactory;
import net.sf.myra.framework.Vertex;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class ClassEdgeTrailFactory
	implements TrailFactory
{
	/**
	 * Minimum number of covered cases per rule.
	 */
	private final int MINIMUM_CASES = Configuration.getIntProperty(
		MINIMUM_RULE_CASES, NO_MINIMUM_CASES);

	/**
	 * The problem construction graph to create the trails.
	 */
	protected Graph<Term> graph;

	/**
	 * The dataset instance to validate the vertices selection.
	 */
	protected Dataset dataset;

	/**
	 * The threshold selection helper.
	 */
	protected ThresholdHelper helper;

	/**
	 * Default constructor.
	 * 
	 * @param graph the problem construction graph.
	 * @param dataset the dataset reference.
	 */
	public ClassEdgeTrailFactory(Graph<Term> graph, Dataset dataset)
	{
		this.graph = graph;
		this.dataset = dataset;
		this.helper = new ThresholdHelper(dataset.getMetadata());
	}

	@SuppressWarnings("unchecked")
	public Trail createTrail()
	{
		Metadata metadata = dataset.getMetadata();
		Rule rule = new Rule(metadata);

		List<Instance> covered = dataset.getInstances();
		Set<String> used = new HashSet<String>();

		int level = 0;
		String consequent = null;

		Vertex<Term> current = graph.get(VIRTUAL_VERTEX);
		used.add(VIRTUAL_VERTEX);

		while (covered.size() >= MINIMUM_CASES)
		{
			Set<String> incompatible = new HashSet<String>();
			Vertex<Term>[] vertices =
				graph.getNeighbours(current).toArray(new Vertex[0]);

			for (int i = 0; i < vertices.length; i++)
			{
				if (used.contains(vertices[i].getLabel()))
				{
					incompatible.add(vertices[i].getLabel());
				}
			}

			while (incompatible.size() < vertices.length)
			{
				double total = 0.0;

				for (int i = 0; i < vertices.length; i++)
				{
					if (!incompatible.contains(vertices[i].getLabel()))
					{
						Edge<Term> edge = graph.getEdge(current, vertices[i]);
						total +=
							(edge.getPheromone().getValue() * vertices[i]
								.getHeuristicValue(level).getValue());
					}
				}

				// in case that we have vertices left but none of them is
				// suitable, we force the rule creation procedure to stop
				if (total == 0)
				{
					for (int i = 0; i < vertices.length; i++)
					{
						incompatible.add(vertices[i].getLabel());
					}

					break;
				}

				int selected = -1;
				double slot = RandomGenerator.getInstance().nextDouble();
				double probability = 0.0;

				for (int i = 0; i < vertices.length; i++)
				{
					if (!incompatible.contains(vertices[i].getLabel()))
					{
						Edge<Term> edge = graph.getEdge(current, vertices[i]);
						probability +=
							(edge.getPheromone().getValue() * vertices[i]
								.getHeuristicValue(level).getValue()) / total;

						if (slot <= probability)
						{
							selected = i;
							break;
						}
					}
				}

				if (selected == -1)
				{
					throw new RuntimeException(
						"Could not select a vertex for rule: " + rule);
				}

				if (vertices[selected].getInfo() == null)
				{
					// we have a class vertex
					consequent = vertices[selected].getLabel();
					rule.setConsequent(metadata.getLabel().toLabel(consequent));
					level = metadata.getTarget().getIndex(consequent);

					current = vertices[selected];
					break;
				}
				else
				{
					Vertex<Term> vertex =
						process(metadata, vertices[selected], covered,
							consequent);

					Rule clone = (Rule) rule.clone();
					clone.add(vertex);

					List<Instance> set =
						Evaluator.findCoveredCases(clone, covered);
					int size = set.size();

					// if the candidate rule cover at least MINIMUM_CASES and
					// either the size of examples covered have changed or the
					// current rule is empty
					if ((size >= MINIMUM_CASES)
						&& (size != covered.size() || rule.isEmpty()))
					{
						used.add(vertex.getLabel());
						rule.add(vertex);

						covered = set;
						current = vertex;

						break;
					}
					else
					{
						incompatible.add(vertex.getLabel());
					}
				}
			}

			if (incompatible.size() == vertices.length)
			{
				// if we could not add a term during an iteration, we stop
				// the rule construction process
				break;
			}
		}

		return rule;
	}

	/**
	 * Prepares the vertex to be added to a rule. The default implementation
	 * just returns the vertex.
	 * 
	 * @param vertex the vertex to be added to the rule.
	 * 
	 * @return the same vertex instance.
	 */
	protected Vertex<Term> process(Metadata metadata, Vertex<Term> vertex,
		List<Instance> covered, String consequent)
	{
		Attribute attribute = metadata.get(vertex.getInfo().getIndex());

		if (attribute instanceof ContinuousAttribute)
		{
			return helper.create((ContinuousAttribute) attribute, vertex,
				covered, consequent);
		}

		return vertex;
	}

	public Graph<?> getGraph()
	{
		return graph;
	}

	/**
	 * Returns the dataset reference.
	 * 
	 * @return the dataset reference.
	 */
	protected Dataset getDataset()
	{
		return dataset;
	}
}