/*
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2012 Fernando Esteban Barril Otero
 */

package net.sf.myra.cantminer.preselect;

import static net.sf.myra.cantminer.cAntMiner2.VIRTUAL_VERTEX;

import java.util.Properties;

import net.sf.myra.cantminer.cAntMiner;
import net.sf.myra.datamining.AbstractObjectiveFunction;
import net.sf.myra.datamining.BacktrackPruner;
import net.sf.myra.datamining.ClassAssignator;
import net.sf.myra.datamining.Classifier;
import net.sf.myra.datamining.ConvergenceTest;
import net.sf.myra.datamining.Evaluator;
import net.sf.myra.datamining.IterationTest;
import net.sf.myra.datamining.Main;
import net.sf.myra.datamining.NoClassAssignator;
import net.sf.myra.datamining.Rule;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Label;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.data.NominalAttribute;
import net.sf.myra.datamining.data.Term;
import net.sf.myra.datamining.function.SensitivitySpecificityFunction;
import net.sf.myra.datamining.model.AbstractRuleModel;
import net.sf.myra.datamining.model.RuleList;
import net.sf.myra.framework.Colony;
import net.sf.myra.framework.CompoundStopCondition;
import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.Graph;
import net.sf.myra.framework.ObjectFactory;
import net.sf.myra.framework.Scheduler;
import net.sf.myra.framework.TrailFactory;
import net.sf.myra.framework.Vertex;

/**
 * This class represents the cAnt-MinerS classification algorithm.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class cAntMinerS
	extends Classifier
{
	/**
	 * Default constructor.
	 */
	public cAntMinerS()
	{
		super(cAntMiner.OPTIONS_FILE);
	}

	@Override
	public Properties filter(Properties properties)
	{
		properties.remove(Main.EXPORTER);
		return properties;
	}

	public AbstractRuleModel run(Dataset dataset)
	{
		if (dataset.getMetadata().isHierarchical())
		{
			throw new IllegalArgumentException(
				"Hierarchical dataset not supported.");
		}

		AbstractRuleModel discovered = new RuleList();

		Metadata metadata = dataset.getMetadata();
		Label majority = Evaluator.findMajorityClass(dataset);

		while (!Evaluator.sameLabel(dataset.getInstances()))
		{
			Rule rule = evolve(dataset);
			
			if (!rule.isEmpty())
			{
				dataset.remove(Evaluator.findCoveredCases(rule, dataset));
				discovered.add(rule.clone());
			}
			else
			{
				// a default rule has been created, so it is time to stop
				break;
			}
		}

		if (discovered.hasDefault())
		{
			throw new IllegalArgumentException("Default rule has been created.");
		}
		else if (!dataset.isEmpty())
		{
			majority = Evaluator.findMajorityClass(dataset);
		}

		Rule defaultRule = new Rule(metadata);
		defaultRule.setConsequent(majority);
		discovered.add(defaultRule);

		return discovered;
	}

	/**
	 * Ant Colony Optimization (ACO) procedure to evolve a rule.
	 * 
	 * @param dataset
	 * @return
	 */
	private Rule evolve(Dataset dataset)
	{
		Graph<Term> graph = new Graph<Term>();
		NominalAttribute target = dataset.getMetadata().getTarget();

		// adds the class values (apart from the majority) to the graph
		for (String value : target.getValues())
		{
			graph.add(new Vertex<Term>(value));
		}

		// adds the attribute-value pair to the graph
		graph.merge(dataset.getMetadata().toGraph());

		// adds the virtual start vertex
		Vertex<Term> start = new Vertex<Term>(VIRTUAL_VERTEX);
		graph.add(start);

		for (String value : target.getValues())
		{
			graph.connect(start, graph.get(value));
		}

		TrailFactory factory = new ClassEdgeTrailFactory(graph, dataset);

		Colony colony = new Colony();
		colony.setTrailFactory(factory);

		CompoundStopCondition condition = new CompoundStopCondition();
		condition.add(new ConvergenceTest());
		condition.add(new IterationTest());

		ClassAssignator assignator = new NoClassAssignator();

		AbstractObjectiveFunction function =
			(AbstractObjectiveFunction) ObjectFactory.create(Configuration
				.getProperty(RULE_QUALITY_FUNCTION,
					SensitivitySpecificityFunction.class.getName()),
				new Class<?>[] { Dataset.class }, new Object[] { dataset });

		Scheduler scheduler =
			new Scheduler(colony, condition, function,
				new ClassEdgePheromoneUpdater(), new ClassHeuristicInformation(
					dataset));
		scheduler.setLocalSearch(new BacktrackPruner(function, assignator));

		scheduler.run();

		// get the best ant trail constructed
		return (Rule) colony.getBest().getBestTrail();
	}
}