/*
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2011 Fernando Esteban Barril Otero
 */

package net.sf.myra.cantminer.sequential;

import java.util.Collection;

import net.sf.myra.cantminer.EdgePheromoneUpdater;
import net.sf.myra.cantminer.cAntMiner2;
import net.sf.myra.cantminer.pittsburgh.PheromoneUpdater;
import net.sf.myra.datamining.data.Term;
import net.sf.myra.framework.Colony;
import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.Edge;
import net.sf.myra.framework.Graph;
import net.sf.myra.framework.Trail;
import net.sf.myra.framework.Vertex;

/**
 * @author Fernando Esteban Barril Otero
 */
public class MaxMinEdgePheromoneUpdater
	extends EdgePheromoneUpdater
{
	/**
	 * The property key under which the evaporation factor is stored.
	 */
	public static final String EVAPORATION_FACTOR =
		PheromoneUpdater.EVAPORATION_FACTOR;

	/**
	 * The property key under which the p_best probability is stored.
	 */
	public static final String P_BEST = PheromoneUpdater.P_BEST;

	/**
	 * The initial pheromone value (this is an arbitrary 'big' value).
	 */
	public static final double INITIAL = 10.0;

	/**
	 * The pheromone evaporation factor (default 0.90);
	 */
	private double factor;

	/**
	 * The p_best probability (default 0.05) to estimate the tMin value.
	 */
	private double pBest;

	/**
	 * The maximum pheromone value.
	 */
	private double tMax;

	/**
	 * The minimum pheromone value.
	 */
	private double tMin;

	/**
	 * The quality of the current best solution.
	 */
	private double currentBest = 0.0;

	public void evaporate(Colony colony)
	{
		// empty implementation, the update implements the evaporation
	}

	public void initialize(Colony colony)
	{
		colony.getTrailFactory().getGraph()
			.apply(new Graph.EdgeInitializationTransform(INITIAL));

		// at the begining of the search, tMax = tMin, therefore only the
		// heuristic information will be used
		tMax = INITIAL;
		tMin = INITIAL;

		factor = Configuration.getDoubleProperty(EVAPORATION_FACTOR, 0.90);
		pBest = Configuration.getDoubleProperty(P_BEST, 0.05);
	}

	@SuppressWarnings("unchecked")
	public void update(Trail trail, Colony colony)
	{
		Graph<Term> graph = (Graph<Term>) colony.getTrailFactory().getGraph();

		// if we have found a new best solution, updates the values of tMax
		// and tMin
		if (trail.getCost().getValue() > currentBest)
		{
			currentBest = trail.getCost().getValue();
			double n = graph.getVertices().size();

			// the average should represent the average number of options an
			// ant has at each choice point
			double average = (n / 2.0);
			double pDec = Math.pow(pBest, 1 / (double) n);

			tMax = (1 / (1 - factor)) * (currentBest / 5.0);
			tMin = (tMax * (1 - pDec)) / ((average - 1) * pDec);

			if (tMin > tMax)
			{
				tMin = tMax;
			}
		}

		// (1) evaporates the pheromone
		colony.getTrailFactory().getGraph()
			.apply(new Graph.EdgeEvaporateTransform(factor));

		double delta = trail.getCost().getValue();
		Collection<Vertex<?>> used = trail.getVertices();
		Vertex<Term> from = graph.get(cAntMiner2.VIRTUAL_VERTEX);
		int level = 0;

		// (2) updates the pheromone value of the used terms (vertices)
		for (Vertex<?> to : used)
		{
			Edge<?> edge = graph.getEdge(from, (Vertex<Term>) to);
			double value = edge.getPheromone(level).getValue() + delta;
			edge.getPheromone(level).setValue(value);
			from = (Vertex<Term>) to;

			level++;
		}

		// (3) makes sure that pheromone values are int the range [tMin, tMax]
		colony.getTrailFactory().getGraph()
			.apply(new Graph.EdgeRangeTransform(tMin, tMax));
	}

	/**
	 * Returns the maximum pheromone value allowed.
	 * 
	 * @return the maximum pheromone value allowed.
	 */
	public double getTMax()
	{
		return tMax;
	}

	/**
	 * Returns the minimum pheromone value allowed.
	 * 
	 * @return the minimum pheromone value allowed.
	 */
	public double getTMin()
	{
		return tMin;
	}
}