/*
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2011 Fernando Esteban Barril Otero
 */

package net.sf.myra.cantminer.sequential;

import static net.sf.myra.cantminer.pittsburgh.cAntMinerPB.START_VERTEX;

import java.util.Collection;
import java.util.logging.Logger;

import net.sf.myra.datamining.Rule;
import net.sf.myra.datamining.data.Term;
import net.sf.myra.framework.Edge;
import net.sf.myra.framework.Environment;
import net.sf.myra.framework.Graph;
import net.sf.myra.framework.PheromoneUpdater;
import net.sf.myra.framework.StopCondition;
import net.sf.myra.framework.Vertex;

/**
 * @author Fernando Esteban Barril Otero
 */
public class MaxMinEdgeStagnationTest
	implements StopCondition
{
	/**
	 * The logger instance.
	 */
	private Logger logger = Logger.getLogger(MaxMinEdgeStagnationTest.class
		.getName());

	@SuppressWarnings("unchecked")
	public boolean isReached(Environment environment)
	{
		PheromoneUpdater updater =
			environment.getScheduler().getPheromoneUpdater();

		if (updater instanceof MaxMinEdgePheromoneUpdater)
		{
			double tMin = ((MaxMinEdgePheromoneUpdater) updater).getTMin();
			double tMax = ((MaxMinEdgePheromoneUpdater) updater).getTMax();

			Graph<Term> graph =
				(Graph<Term>) environment.getColony().getTrailFactory()
					.getGraph();
			Rule rule = (Rule) environment.getColony().getCurrent().getTrail();
			int level = 0;

			if (rule != null)
			{
				Vertex<Term> current = graph.get(START_VERTEX);

				for (Vertex<?> vertex : rule.getVertices())
				{
					if (!hasConverged(graph.getEdges(current), level, tMin,
						tMax))
					{
						return false;
					}

					level++;
					current = (Vertex<Term>) vertex;
				}

				logger.info("Search has converged at iteration "
					+ environment.getIteration());

				return true;
			}
			else
			{
				return false;
			}
		}

		throw new IllegalArgumentException(
			"Not a max-min pheromone updater instance: "
				+ updater.getClass().getName());
	}

	private boolean hasConverged(Collection<Edge<Term>> edges, int level,
		double tMin, double tMax)
	{
		int minCount = 0;
		int maxCount = 0;

		for (Edge<Term> edge : edges)
		{
			double value = precision(edge.getPheromone(level).getValue());

			if (value == precision(tMin))
			{
				minCount++;
			}

			if (value == precision(tMax))
			{
				maxCount++;
			}
		}

		return (minCount == (edges.size() - 1)) && (maxCount == 1);
	}

	/**
	 * Returns the double value with up to four decimal digits.
	 * 
	 * @param value the value to be converted.
	 * 
	 * @return the double value with up to four decimal digits.
	 */
	private final double precision(double value)
	{
		int v = (int) (value * 100);
		return v / 100.0;
	}
}