/*
 * $Id: MaxMinPheromoneUpdater.java 2339 2011-01-25 11:35:30Z febo2 $ 
 */

package net.sf.myra.cantminer.sequential;

import java.util.Collection;

import net.sf.myra.antminer.DefaultPheromoneUpdater;
import net.sf.myra.datamining.data.Term;
import net.sf.myra.framework.Colony;
import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.Graph;
import net.sf.myra.framework.Pheromone;
import net.sf.myra.framework.Trail;
import net.sf.myra.framework.Vertex;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision: 2339 $ $Date:: 2011-01-25 11:35:30#$
 */
public class MaxMinPheromoneUpdater
	extends DefaultPheromoneUpdater
{
	/**
	 * The property key under which the evaporation factor is stored.
	 */
	public static final String EVAPORATION_FACTOR =
		"net.sf.myra.cantminer.maxmin.evaporation";

	/**
	 * The property key under which the p_best probability is stored.
	 */
	public static final String P_BEST = "net.sf.myra.cantminer.maxmin.pbest";

	/**
	 * The initial pheromone value (this is an arbitrary 'big' value).
	 */
	public static final double INITIAL = 10.0;

	/**
	 * The pheromone evaporation factor (default 0.95);
	 */
	protected double factor;

	/**
	 * The p_best probability (default 0.05) to estimate the tMin value.
	 */
	protected double pBest;

	/**
	 * The maximum pheromone value.
	 */
	protected double tMax;

	/**
	 * The minimum pheromone value.
	 */
	protected double tMin;

	/**
	 * The quality of the current best solution.
	 */
	protected double currentBest = 0.0;

	@SuppressWarnings("unchecked")
	public void initialize(Colony colony)
	{
		Graph<Term> graph = (Graph<Term>) colony.getTrailFactory().getGraph();
		Pheromone pheromone = new Pheromone(INITIAL);

		for (Vertex<Term> vertex : graph.vertices())
		{
			vertex.setPheromone(pheromone);
		}

		// at the begining of the search, tMax = tMin, so only the heuristic
		// will be used
		tMax = INITIAL;
		tMin = INITIAL;

		factor = Configuration.getDoubleProperty(EVAPORATION_FACTOR, 0.90);
		pBest = Configuration.getDoubleProperty(P_BEST, 0.05);
	}

	public void evaporate(Colony colony)
	{
		// empty implementation, the update implements the evaporation
	}

	@SuppressWarnings("unchecked")
	public void update(Trail trail, Colony colony)
	{
		Graph<Term> graph = (Graph<Term>) colony.getTrailFactory().getGraph();

		// if we have found a new best solution, updates the values of tMax
		// and tMin
		if (trail.getCost().getValue() > currentBest)
		{
			currentBest = trail.getCost().getValue();
			double n = graph.getVertices().size() - 1;

			// the average should represent the average number of options an
			// ant has at each choice point
			double average = n / 2.0;
			double pDec = Math.pow(pBest, 1 / (double) n);

			//tMax = (1 / (1 - factor)) * currentBest;
			tMax = (1 / (1 - factor)) * (currentBest / 10);
			tMin = (tMax * (1 - pDec)) / ((average - 1) * pDec);

			if (tMin > tMax)
			{
				tMin = tMax;
			}
		}

		int levels = trail.getSize();

		// (1) ensures that all vertices have the same number of levels

		for (Vertex<?> vertex : graph.getVertices())
		{
			for (int i = 0; i < levels; i++)
			{
				vertex.setPheromone(vertex.getPheromone(i), i);
			}
		}

		// (2) evaporates the pheromone

		for (Vertex<Term> vertex : graph.vertices())
		{
			levels = vertex.levels();

			for (int i = 0; i < levels; i++)
			{
				double value = vertex.getPheromone(i).getValue();
				vertex.getPheromone(i).setValue(value * factor);
			}
		}

		// (3) updates the pheromone of the used vertices

		double delta = trail.getCost().getValue();
		Collection<Vertex<?>> used = trail.getVertices();
		int index = 0;

		for (Vertex<?> vertex : used)
		{
			Pheromone pheromone =
				graph.get(vertex.getLabel()).getPheromone(index);
			pheromone.setValue(pheromone.getValue() + delta);
			index++;
		}

		// (4) ensures that the pehromones are in the range [tMin, tMax]

		for (Vertex<Term> vertex : graph.vertices())
		{
			levels = vertex.levels();

			for (int i = 0; i < levels; i++)
			{
				Pheromone pheromone = vertex.getPheromone(i);
				double value = pheromone.getValue();

				if (value > tMax)
				{
					value = tMax;
				}
				else if (value < tMin)
				{
					value = tMin;
				}

				pheromone.setValue(value);
			}
		}
	}

	/**
	 * Returns the maximum pheromone value allowed.
	 * 
	 * @return the maximum pheromone value allowed.
	 */
	public double getTMax()
	{
		return tMax;
	}

	/**
	 * Returns the minimum pheromone value allowed.
	 * 
	 * @return the minimum pheromone value allowed.
	 */
	public double getTMin()
	{
		return tMin;
	}
}