/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining;

import java.util.Properties;
import net.sf.myra.datamining.data.Dataset;

/**
 * This class defines a skeletal implementation of a classifier algorithm. A
 * classifier instance is responsible for creating a <code>Model</code>
 * instance during training, which in turn can be used to classify unseen
 * instances.
 * 
 * @see Model
 *  
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public abstract class Classifier
{
	/**
	 * The property key under which the maximum uncovered value is stored.
	 */
	public static final String MAXIMUM_UNCOVERED_CASES =
		"net.sf.myra.datamining.uncovered";

	/**
	 * The default maximum number of uncovered cases.
	 */
	public static final int DEFAULT_MAXIMUM_UNCOVERED_CASES = 10;

	/**
	 * The property key under which the rule quality function value is stored.
	 */
	public static final String RULE_QUALITY_FUNCTION =
		"net.sf.myra.datamining.function";

	/**
	 * Classifier options filename.
	 */
	private String optionsFile;

	/**
	 * Default constructor.
	 * 
	 * @param optionsFile the classifier options filename.
	 */
	public Classifier(String optionsFile)
	{
		this.optionsFile = optionsFile;
	}

	/**
	 * Returns the configuration property options.
	 * 
	 * @return the configuration property options.
	 */
	public Properties getOptions()
	{
		try
		{
			Properties messages = new Properties();

			if (optionsFile != null)
			{
				messages.load(getClass().getResourceAsStream(optionsFile));
			}

			return messages;
		}
		catch (Exception e)
		{
			throw new IllegalArgumentException("Options file not found: "
				+ optionsFile);
		}
	}

	/**
	 * Filters the specified properties. The classifier implementation should
	 * remove all properties from the framework that are not used.
	 * 
	 * @param properties the framework (default) properties.
	 * 
	 * @return the filtered properties.
	 */
	public Properties filter(Properties properties)
	{
		return properties;
	}

	/**
	 * Trains the classifier using the specified dataset.
	 * 
	 * @param dataset the training dataset.
	 * 
	 * @return the classification model.
	 * 
	 * @see Model
	 */
	public abstract Model run(Dataset dataset);
}