/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining;

/**
 * This class represents a confusion matrix implementation.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public final class ConfusionMatrix
	implements Cloneable
{
	/**
	 * The true positive's index.
	 */
	public static final int TP = 0;

	/**
	 * The false positive's index.
	 */
	public static final int FP = 1;

	/**
	 * The true negative's index.
	 */
	public static final int TN = 2;

	/**
	 * The false negative's index.
	 */
	public static final int FN = 3;

	/**
	 * The mapping between a matrix cell and a counter.
	 */
	private double[] matrix;

	/**
	 * Default constructor.
	 */
	public ConfusionMatrix()
	{
		this(0, 0, 0, 0);
	}

	/**
	 * Creates a new <code>ConfusionMatrix</code> instance.
	 * 
	 * @param tp true positive initial value.
	 * @param fp false positive initial value.
	 * @param fn false negative initial value.
	 * @param tn true negative initial value.
	 */
	public ConfusionMatrix(double tp, double fp, double fn, double tn)
	{
		matrix = new double[4];

		matrix[TP] = tp;
		matrix[FP] = fp;
		matrix[FN] = fn;
		matrix[TN] = tn;
	}

	/**
	 * Adds the values from the specified confusion matrix.
	 * 
	 * @param other the confusion matrix the add.
	 */
	public void add(ConfusionMatrix other)
	{
		matrix[TP] += other.matrix[TP];
		matrix[FP] += other.matrix[FP];
		matrix[FN] += other.matrix[FN];
		matrix[TN] += other.matrix[TN];
	}
	
	/**
	 * Adds the value to the specified cell.
	 * 
	 * @param field the cell of the matrix.
	 * @param value the value to add.
	 */
	public void add(int field, double value)
	{
		matrix[field] = matrix[field] + value;
	}

	/**
	 * Increments the specified cell in the matrix.
	 * 
	 * @param field the cell to increment.
	 */
	public void increment(int field)
	{
		matrix[field]++;
	}

	/**
	 * Returns the value of the specified cell.
	 * 
	 * @param field the cell.
	 * 
	 * @return the value of the specified cell.
	 */
	public double get(int field)
	{
		return matrix[field];
	}

	/**
	 * Sets the value of the specified cell.
	 * 
	 * @param field the field to set.
	 * @param value the value to set.
	 */
	public void set(int field, double value)
	{
		matrix[field] = value;
	}

	/**
	 * Returns the sensitivity value.
	 * 
	 * @return the sensitivity value.
	 */
	public double getSensitivity()
	{
		return filter(matrix[TP] / (double) (matrix[TP] + matrix[FN]));
	}

	/**
	 * Returns the specificity value.
	 * 
	 * @return the specificity value.
	 */
	public double getSpecificity()
	{
		return filter(matrix[TN] / (double) (matrix[FP] + matrix[TN]));
	}

	/**
	 * Returns the precision value.
	 * 
	 * @return the precision value.
	 */
	public double getPrecision()
	{
		return filter(matrix[TP] / (double) (matrix[TP] + matrix[FP]));
	}

	/**
	 * Returns the recall value.
	 * 
	 * @return the recall value.
	 */
	public double getRecall()
	{
		return getSensitivity();
	}

	/**
	 * Returns the accuracy value.
	 * 
	 * @return the accuracy value.
	 */
	public double getAccuracy()
	{
		return filter((matrix[TP] + matrix[TN])
			/ (double) (matrix[TP] + matrix[FP] + matrix[TN] + matrix[FN]));
	}

	/**
	 * Filters <code>NaN</code> (not-a-number) values.
	 * 
	 * @param value the value to be filtered.
	 * 
	 * @return <code>0.0</code> if a <code>NaN</code> is found;
	 *         <code>value</code> otherwise.
	 */
	private double filter(double value)
	{
		return Double.isNaN(value) ? 0.0 : value;
	}

	@Override
	public String toString()
	{
		StringBuffer output = new StringBuffer();

		output.append("[TP=");
		output.append(matrix[TP]);
		output.append(",FP=");
		output.append(matrix[FP]);
		output.append(",FN=");
		output.append(matrix[FN]);
		output.append(",TN=");
		output.append(matrix[TN]);
		output.append("]");

		return output.toString();
	}

	@Override
	public ConfusionMatrix clone()
	{
		try
		{
			ConfusionMatrix clone = (ConfusionMatrix) super.clone();
			clone.matrix = matrix.clone();
			return clone;
		}
		catch (CloneNotSupportedException e)
		{
			throw new UnknownError(getClass() + ":clone()");
		}
	}
}