/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining;

import java.util.logging.Logger;

import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.Environment;
import net.sf.myra.framework.StopCondition;
import net.sf.myra.framework.Trail;

/**
 * This class represents the convergence stop condition. It tests if the
 * colony of ants have converged to a single solution over a specified
 * number of iterations.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class ConvergenceTest
	implements StopCondition
{
	/**
	 * The property key under which the convergence test size value is stored.
	 */
	public static final String CONVERGENCE_TEST_SIZE =
		"net.sf.myra.datamining.convergence";

	/**
	 * The default convergence test size.
	 */
	public static final int DEFAULT_CONVERGENCE_TEST_SIZE = 10;

	/**
	 * The logger instance.
	 */
	private Logger logger = Logger.getLogger(ConvergenceTest.class.getName());

	/**
	 * The number of rules used to test the convergence of the ants.
	 */
	private int size;

	/**
	 * The current test number.
	 */
	private int testNumber;

	/**
	 * The best trail (rule) found in the previous iteration.
	 */
	private Trail previous;

	/**
	 * Creates a new <code>ConvergenceTest</code> instance.
	 */
	public ConvergenceTest()
	{
		this(Configuration.getIntProperty(CONVERGENCE_TEST_SIZE,
			DEFAULT_CONVERGENCE_TEST_SIZE));
	}

	/**
	 * Creates a new <code>ConvergenceTest</code> instance.
	 * 
	 * @param size the number of rules used to test the convergence of the ants.
	 */
	public ConvergenceTest(int size)
	{
		this.size = size;
		testNumber = 0;
		previous = null;
	}

	public boolean isReached(Environment environment)
	{
		Trail current = environment.getColony().getCurrent().getTrail();

		if (previous == null)
		{
			previous = current;
			testNumber++;
		}
		else if (previous.equals(current))
		{
			testNumber++;
		}
		else
		{
			previous = current;
			testNumber = 1;
		}

		boolean reached = (testNumber >= size);

		if (reached)
		{
			logger.info("Search has converged at iteration "
				+ environment.getIteration());
		}

		return reached;
	}
}