/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import net.sf.myra.framework.Cost;

/**
 * This class represents the confidence factor measure.
 * 
 * @see ConfidenceEvaluator
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class DefaultConfidenceFactor
	implements ConfidenceFactor
{
	private static final long serialVersionUID = 1655387385272070599L;

	private HashMap<String, Integer> mapping;

	public DefaultConfidenceFactor(Map<String, Integer> mapping)
	{
		this.mapping = new HashMap<String, Integer>(mapping);
	}

	/**
	 * Returns the rule's covering examples mapping.
	 * 
	 * @return the rule's covering examples mapping.
	 */
	public Map<String, Integer> getMapping()
	{
		return Collections.unmodifiableMap(mapping);
	}
	
	/**
	 * Returns the number of covered examples for the specified class label.
	 * 
	 * @param label the class label.
	 * 
	 * @return the number of covered examples for the specified class label.
	 */
	public int get(String label)
	{
		return mapping.containsKey(label) ? mapping.get(label) : 0;
	}

	/**
	 * Returns the confidence factor value.
	 * 
	 * @return the confidence factor value.
	 */
	public double getValue()
	{
		int total = 0;
		int majority = 0;

		for (String value : mapping.keySet())
		{
			int current = mapping.get(value);
			total += current;

			if (current > majority)
			{
				majority = mapping.get(value);
			}
		}

		return majority / (double) total;
	}

	public int compareTo(Cost o)
	{
		return compareTo((DefaultConfidenceFactor) o);
	}

	public int compareTo(DefaultConfidenceFactor o)
	{
		return Double.compare(getValue(), o.getValue());
	}

	@Override
	public ConfidenceFactor clone()
	{
		try
		{
			DefaultConfidenceFactor clone =
				(DefaultConfidenceFactor) super.clone();
			clone.mapping = new HashMap<String, Integer>(mapping);
			return clone;
		}
		catch (CloneNotSupportedException e)
		{
			throw new UnknownError(getClass() + ": clone()");
		}
	}
}