/*
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2011 Fernando Esteban Barril Otero
 */

package net.sf.myra.datamining;

import static net.sf.myra.datamining.ConfusionMatrix.FP;
import static net.sf.myra.datamining.ConfusionMatrix.TP;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.framework.Trail;

/**
 * This class implements an error based pruning procedure based on the metric
 * used by C4.5. The procedure consists of removing the last term of the rule
 * while either the quality of the rule does not decrease or the reduced error
 * is (sufficient) smaller than the original rule.
 * 
 * @author Fernando Esteban Barril Otero
 * 
 * @see ErrorBasedMetric
 */
public class ErrorBasedPruner
	extends AbstractPruner
{
	/**
	 * Creates a new <code>ReducedErrorPruner</code> instance.
	 * 
	 * @param function
	 * @param assignator
	 */
	public ErrorBasedPruner(AbstractObjectiveFunction function,
		ClassAssignator assignator)
	{
		super(function, assignator);
	}

	public Trail explore(Trail trail)
	{
		Rule rule = (Rule) trail;

		AbstractObjectiveFunction function = getFunction();
		Dataset dataset = function.getDataset();

		ClassAssignator assignator = getClassAssignator();
		assignator.assign(rule, dataset);

		ConfusionMatrix matrix = Evaluator.getConfusionMatrix(rule, dataset);
		double error =
			ErrorBasedMetric.estimate(matrix.get(FP),
				matrix.get(TP) + matrix.get(FP));

		while (rule.getSize() > 1)
		{
			Rule improved = rule.clone();
			improved.remove(rule.getLast());
			assignator.assign(improved, function.getDataset());

			matrix = Evaluator.getConfusionMatrix(improved, dataset);
			double improvedError =
				ErrorBasedMetric.estimate(matrix.get(FP), matrix.get(TP)
					+ matrix.get(FP));

			// if the rule quality has improved (or has the same quality) or
			// the (estimated) reduced error metric has improved
			if (improvedError <= error)
			{
				rule = improved;
				error = improvedError;
			}
			// otherwise we stop
			else
			{
				break;
			}
		}

		return rule;
	}
}