/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining;

import static net.sf.myra.datamining.ConfusionMatrix.FN;
import static net.sf.myra.datamining.ConfusionMatrix.FP;
import static net.sf.myra.datamining.ConfusionMatrix.TN;
import static net.sf.myra.datamining.ConfusionMatrix.TP;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Label;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.data.Node;
import net.sf.myra.datamining.data.NominalAttribute;
import net.sf.myra.datamining.data.Term;
import net.sf.myra.framework.RandomGenerator;

/**
 * Utility class to evaluate rules discovered by ants.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public final class Evaluator
{
	/**
	 * Private constructor.
	 */
	private Evaluator()
	{
	}

	/**
	 * Returns the covered instances by the specified rule.
	 * 
	 * @param rule the trail to be used as a filter.
	 * @param dataset the dataset to get the instances.
	 * 
	 * @return the covered instances by the specified rule.
	 */
	public static List<Instance> findCoveredCases(Rule rule, Dataset dataset)
	{
		return findCoveredCases(rule, dataset.getInstances());
	}

	/**
	 * Returns the covered instances by the specified rule.
	 * 
	 * @param rule the trail to be used as a filter.
	 * @param instances the instances list to check.
	 * 
	 * @return the covered instances by the specified rule.
	 */
	public static List<Instance> findCoveredCases(Rule rule,
		List<Instance> instances)
	{
		List<Instance> covered = new LinkedList<Instance>();

		for (Instance instance : instances)
		{
			if (rule.isCovered(instance))
			{
				covered.add(instance);
			}
		}

		return covered;
	}

	/**
	 * Returns the not covered instances by the specified rule.
	 * 
	 * @param rule the trail to be used as a filter.
	 * @param dataset the dataset to get the instances.
	 * 
	 * @return the not covered instances by the specified rule.
	 */
	public static List<Instance> findNotCoveredCases(Rule rule, Dataset dataset)
	{
		return findNotCoveredCases(rule, dataset.getInstances());
	}

	/**
	 * Returns the not covered instances by the specified rule.
	 * 
	 * @param rule the trail to be used as a filter.
	 * @param instances the instances list to check.
	 * 
	 * @return the not covered instances by the specified rule.
	 */
	public static List<Instance> findNotCoveredCases(Rule rule,
		List<Instance> instances)
	{
		List<Instance> notCovered = new LinkedList<Instance>();

		for (Instance instance : instances)
		{
			if (!rule.isCovered(instance))
			{
				notCovered.add(instance);
			}
		}

		return notCovered;
	}

	/**
	 * Returns a set of indexes of the covered instances by the specified rule.
	 * 
	 * @param rule the trail to be used as a filter.
	 * @param dataset the dataset to get the instances.
	 * 
	 * @return a set of indexes of the covered instances by the specified rule.
	 */
	public static Set<Integer> findIndexedCoveredCases(Rule rule,
		Dataset dataset)
	{
		return findIndexedCoveredCases(rule, dataset.getInstances());
	}

	/**
	 * Returns a set of indexes of the covered instances by the specified rule.
	 * 
	 * @param rule the trail to be used as a filter.
	 * @param instances the instances list to check.
	 * 
	 * @return a set of indexes of the covered instances by the specified rule.
	 */
	public static Set<Integer> findIndexedCoveredCases(Rule rule,
		List<Instance> instances)
	{
		Set<Integer> covered = new HashSet<Integer>();

		for (Instance instance : instances)
		{
			if (rule.isCovered(instance))
			{
				covered.add(instance.getIndex());
			}
		}

		return covered;
	}

	/**
	 * Returns the positive covered (the instance class value contains the
	 * specified rule consequent) instances by the specified rule. For flat and
	 * single-label classification problems, this method returns the same set
	 * 
	 * @see #findCorrectCases(Rule, Dataset)
	 */
	public static List<Instance> findPositiveCases(Rule rule, Dataset dataset)
	{
		List<Instance> instances = new LinkedList<Instance>();
		Label consequent = rule.getConsequent();

		for (Instance instance : findCoveredCases(rule, dataset))
		{
			if (instance.getLabel().contains(consequent))
			{
				instances.add(instance);
			}
		}

		return instances;
	}

	/**
	 * Returns the correct covered (the instance class value equals to the
	 * specified rule consequent) instances by the specified rule. For flat and
	 * single-label classification problems, this method returns the same set
	 * of instances as the {@link #findPositiveCases(Rule, Dataset)} method.
	 * 
	 * @param rule the rule to be used as a filter.
	 * @param dataset the dataset to get the instances.
	 * 
	 * @return the correct covered instances by the specified rule.
	 * 
	 * @see #findPositiveCases(Rule, Dataset)
	 */
	public static List<Instance> findCorrectCases(Rule rule, Dataset dataset)
	{
		List<Instance> instances = new LinkedList<Instance>();
		Label consequent = rule.getConsequent();

		for (Instance instance : findCoveredCases(rule, dataset))
		{
			if (instance.getLabel().equals(consequent))
			{
				instances.add(instance);
			}
		}

		return instances;
	}

	/**
	 * Returns the confusion matrix of the specified rule when used to classify
	 * the dataset instances. For hierarchical problems, the confusion matrix
	 * is computed across all class labels, so each field of the matrix 
	 * epresents the accumulated value. 
	 * 
	 * @param rule the rule to be used in the classification.
	 * @param dataset the dataset to get the instances.
	 * 
	 * @return the confusion matrix of the specified rule.
	 */
	public static ConfusionMatrix getConfusionMatrix(Rule rule, Dataset dataset)
	{
		return getConfusionMatrix(rule, dataset.getInstances());
	}

	/**
	 * Returns the confusion matrix of the specified rule when used to classify
	 * the specified instances. For hierarchical problems, the confusion matrix
	 * is computed across all class labels, so each field of the matrix 
	 * epresents the accumulated value.
	 * 
	 * @param rule the rule to be used in the classification.
	 * @param instances the instances to classify.
	 * 
	 * @return the confusion matrix of the specified rule.
	 */
	public static ConfusionMatrix getConfusionMatrix(Rule rule,
		List<Instance> instances)
	{
		ConfusionMatrix matrix = new ConfusionMatrix();

		if (instances.isEmpty())
		{
			return matrix;
		}

		Label consequent = rule.getConsequent();
		Metadata metadata = instances.iterator().next().getMetadata();

		if (metadata.isHierarchical())
		{
			String root = metadata.getClassHierarchy().getRoot().getLabel();

			for (Instance instance : instances)
			{
				Label tLabel = instance.getLabel();

				if (rule.isCovered(instance))
				{
					for (String value : consequent.names())
					{
						if (!root.equals(value))
						{
							int index = consequent.getIndex(value);

							if (tLabel.active(index)
								&& consequent.active(index))
							{
								matrix.increment(TP);
							}
							else if (!tLabel.active(index)
								&& consequent.active(index))
							{
								matrix.increment(FP);
							}
							else if (tLabel.active(index)
								&& !consequent.active(index))
							{
								matrix.increment(FN);
							}
						}
					}
				}
				else
				{
					for (String value : consequent.getValues())
					{
						if (!root.equals(value))
						{
							int index = consequent.getIndex(value);

							if (tLabel.active(index)
								&& consequent.active(index))
							{
								matrix.increment(FN);
							}
							else if (!tLabel.active(index)
								&& consequent.active(index))
							{
								matrix.increment(TN);
							}
						}
					}
				}
			}
		}
		// in case of flat problems
		else
		{
			for (Instance instance : instances)
			{
				boolean isCovered = rule.isCovered(instance);

				if (instance.getLabel().equals(consequent))
				{
					matrix.increment(isCovered ? TP : FN);
				}
				else
				{
					matrix.increment(isCovered ? FP : TN);
				}
			}
		}

		return matrix;
	}

	/**
	 * Returns the confusion matrix of the specified rule when used to classify
	 * the specified instances. For hierarchical problems, the confusion matrix
	 * is computed across all class labels, so each field of the matrix 
	 * epresents the accumulated value.
	 * 
	 * @param rule the rule to be used in the classification.
	 * @param instances the instances to classify.
	 * 
	 * @return the confusion matrix of the specified rule.
	 */
	public static ConfusionMatrix getConfusionMatrix(Rule rule,
		List<Instance> instances, UsefulnessHelper helper)
	{
		ConfusionMatrix matrix = new ConfusionMatrix();

		if (instances.isEmpty())
		{
			return matrix;
		}

		Label consequent = rule.getConsequent();
		Metadata metadata = instances.iterator().next().getMetadata();
		String root = metadata.getClassHierarchy().getRoot().getLabel();

		if (metadata.isHierarchical())
		{
			throw new IllegalArgumentException("Not a hierarchical problem.");
		}
		else if (helper == null)
		{
			throw new IllegalArgumentException("UsefulnessHelper is null.");
		}

		for (Instance instance : instances)
		{
			Label tLabel = instance.getLabel();

			if (rule.isCovered(instance))
			{
				for (String value : consequent.names())
				{
					if (!root.equals(value))
					{
						int index = consequent.getIndex(value);

						if (tLabel.active(index) && consequent.active(index))
						{
							matrix.add(TP, helper.get(value));
						}
						else if (!tLabel.active(index)
							&& consequent.active(index))
						{
							matrix.add(FP, helper.get(value));
						}
						else if (tLabel.active(index)
							&& !consequent.active(index))
						{
							matrix.add(FN, helper.get(value));
						}
					}
				}
			}
			else
			{
				for (String value : consequent.getValues())
				{
					if (!root.equals(value))
					{
						int index = consequent.getIndex(value);

						if (tLabel.active(index) && consequent.active(index))
						{
							matrix.add(FN, helper.get(value));
						}
						else if (!tLabel.active(index)
							&& consequent.active(index))
						{
							matrix.add(TN, helper.get(value));
						}
					}
				}
			}
		}

		return matrix;
	}

	/**
	 * Returns the confusion matrix of the rule when used to classify the
	 * dataset instances, predicting the specified class label. This method
	 * is useful for calculating the confusion matrix for an individual
	 * class label in hierarchical problems.
	 * 
	 * @param rule the rule to be used in the classification.
	 * @param dataset the dataset to get the instances from.
	 * @param label the class label to be used as the prediction. The label
	 *        should be present in the consequent of the rule.
	 * 
	 * @return the confusion matrix of the specified rule.
	 */
	public static ConfusionMatrix getConfusionMatrix(Rule rule,
		Dataset dataset, String label)
	{
		return getConfusionMatrix(rule, dataset.getInstances(), label);
	}

	/**
	 * Returns the confusion matrix of the rule when used to classify the
	 * specified instances, predicting a class label. This method
	 * is useful for calculating the confusion matrix for an individual
	 * class label in hierarchical problems.
	 * 
	 * @param rule the rule to be used in the classification.
	 * @param instances the instances to classify.
	 * @param label the class label to be used as the prediction. The label
	 *        should be present in the consequent of the rule.
	 * 
	 * @return the confusion matrix of the specified rule.
	 */
	public static ConfusionMatrix getConfusionMatrix(Rule rule,
		List<Instance> instances, String label)
	{
		Label consequent = rule.getConsequent();

		if (!consequent.active(consequent.getIndex(label)))
		{
			throw new IllegalArgumentException(
				"Label is not active in the consequent of the rule: " + label);
		}

		ConfusionMatrix matrix = new ConfusionMatrix();

		for (Instance instance : instances)
		{
			boolean isCovered = rule.isCovered(instance);

			if (instance.belongs(label))
			{
				matrix.increment(isCovered ? TP : FN);
			}
			else
			{
				matrix.increment(isCovered ? FP : TN);
			}
		}

		return matrix;
	}

	/**
	 * Returns the confidence factor of the specified rule when used to classify
	 * the dataset instances.
	 * 
	 * @param rule the rule to be used in the classification.
	 * @param dataset the dataset to get the instances.
	 * 
	 * @return the confusion matrix of the specified rule.
	 */
	public static DefaultConfidenceFactor getConfidenceFactor(Rule rule,
		Dataset dataset)
	{
		NominalAttribute target = dataset.getMetadata().getTarget();
		HashMap<String, Integer> mapping = new HashMap<String, Integer>();

		for (Instance instance : dataset)
		{
			if (rule.isCovered(instance))
			{
				String value = instance.getValue(target);
				Integer count = mapping.get(value);

				if (count == null)
				{
					count = 0;
				}

				mapping.put(value, count + 1);
			}
		}

		return new DefaultConfidenceFactor(mapping);
		/*
		Set<Integer> covered = findIndexedCoveredCases(rule, dataset);
		Metadata metadata = dataset.getMetadata();
		NominalAttribute target = metadata.getTarget();

		HashMap<String, Integer> mapping = new HashMap<String, Integer>();

		for (Instance instance : dataset)
		{
			if (instance.getLabel().getValues().size() > 1)
			{
				throw new UnsupportedOperationException(
					"Multi-label instace found.");
			}

			if (covered.contains(instance.getIndex()))
			{
				String value = instance.getValue(target);
				Integer count = mapping.get(value);

				if (count == null)
				{
					count = 0;
				}

				mapping.put(value, count + 1);
			}
		}

		return new DefaultConfidenceFactor(mapping);
		*/
	}

	/**
	 * Returns the confusion matrix of the model for the specified class label.
	 * 
	 * @param model the model to be used in the classification.
	 * @param dataset the dataset to get the instances.
	 * @param label the desired class label.
	 * 
	 * @return the confusion matrix for the specified class label.
	 */
	public static ConfusionMatrix getConfusionMatrix(Model model,
		Dataset dataset, String label)
	{
		ConfusionMatrix matrix = new ConfusionMatrix();
		int index = dataset.getMetadata().getTarget().getIndex(label);

		for (Instance instance : dataset)
		{
			Label predicted = model.classify(instance);

			if (instance.labels()[index])
			{
				matrix.increment(predicted.active(index) ? TP : FN);
			}
			else
			{
				matrix.increment(predicted.active(index) ? FP : TN);
			}
		}

		return matrix;
	}

	/**
	 * Returns the confusion matrix of the specified rule set when used to
	 * classify the dataset instances. This method assumes that the majority
	 * class represents the negative examples and that there are only two
	 * different class labels.
	 * 
	 * @param model the model to be used in the classification.
	 * @param dataset the dataset to get the instances.
	 * 
	 * @return the confusion matrix of the classification.
	 */
	public static ConfusionMatrix getConfusionMatrix(Model model,
		Dataset dataset)
	{
		ConfusionMatrix matrix = new ConfusionMatrix();

		// the majority class represent the "negative" examples, since it is
		// easier to be predicted
		Label majority = Evaluator.findMajorityClass(dataset);
		NominalAttribute target = dataset.getMetadata().getTarget();

		// checks if we are dealing with a binary classification problem
		if (target.getValues().size() > 2)
		{
			throw new IllegalArgumentException("Not a binary problem.");
		}

		for (Instance instance : dataset)
		{
			if (instance.getLabel().getValues().size() > 1)
			{
				throw new UnsupportedOperationException(
					"Multi-label instace found: " + instance.getLabel());
			}

			Label p = model.classify(instance);

			// are we predicting the "negative" class?
			if (p.equals(majority))
			{
				matrix.increment(instance.getLabel().equals(p) ? TN : FN);
			}
			else
			{
				matrix.increment(instance.getLabel().equals(p) ? TP : FP);
			}
		}

		return matrix;
	}

	/**
	 * Returns the majority class value among the specified instances.
	 * 
	 * @param instances the list of instances.
	 * @param index the class attribute index.
	 * 
	 * @return the majority class value among the specified instances.
	 */
	public static Label findMajorityClassValue(List<Instance> instances,
		int index)
	{
		return findMajorityClassValue(instances, index, null);
	}

	/**
	 * Returns the majority class value among the specified instances.
	 * 
	 * @param instances the list of instances.
	 * @param index the class attribute index.
	 * @param label the label returned in case of a tie. Usually this label
	 *        will represent the overall majority of the data set.
	 * 
	 * @return the majority class value among the specified instances.
	 */
	public static Label findMajorityClassValue(List<Instance> instances,
		int index, Label label)
	{
		Map<String, Integer> distribution =
			new HashMap<String, Integer>(instances.size(), 1.0f);

		for (Instance instance : instances)
		{
			for (String value : instance.getLabel().getValues())
			{
				Integer count = distribution.get(value);
				distribution.put(value, (count == null ? 1 : count + 1));
			}
		}

		Metadata metadata = instances.iterator().next().getMetadata();

		if (metadata.isHierarchical())
		{
			// does not take into account the root label, since all instances
			// belong to it
			distribution.remove(metadata.getClassHierarchy().getRoot()
				.getLabel());
		}

		ArrayList<Label> candidates = new ArrayList<Label>();
		String majority = null;
		int max = -1;

		for (String value : distribution.keySet())
		{
			int size = distribution.get(value);

			if (size > max)
			{
				majority = value;
				max = size;

				candidates.clear();
				candidates.add(metadata.expand(metadata.getLabel().toLabel(
					value)));
			}
			else if (size == max)
			{
				if (metadata.isHierarchical())
				{
					Node node1 = metadata.getClassHierarchy().find(majority);
					Node node2 = metadata.getClassHierarchy().find(value);

					// if the majority node is an ancestor of the current node,
					// we keep the current node (the most specific node)
					if (node2.getAncestors().contains(node1))
					{
						majority = value;

						candidates.clear();
						candidates.add(metadata.expand(metadata.getLabel()
							.toLabel(value)));
					}
					// in case they are not related
					else
					{
						candidates.add(metadata.expand(metadata.getLabel()
							.toLabel(value)));
					}
				}
				else
				{
					candidates.add(metadata.expand(metadata.getLabel().toLabel(
						value)));
				}
			}
		}

		if (candidates.size() > 1)
		{
			if (candidates.contains(label))
			{
				return label;
			}
			else
			{
				return candidates.get(RandomGenerator.getInstance().nextInt(
					candidates.size()));
			}
		}

		return candidates.get(0);
	}

	/**
	 * Returns the majority class value among the specified instances.
	 * 
	 * @param instances the list of instances.
	 * @param metadata the dataset metadata.
	 * 
	 * @return the majority class value among the specified instances.
	 */
	public static Label findMajorityClass(List<Instance> instances,
		Metadata metadata)
	{
		String name = metadata.getTarget().getName();
		return findMajorityClassValue(instances, metadata.getIndex(name));
	}

	/**
	 * Returns the majority class value among the specified instances.
	 * 
	 * @param instances the list of instances.
	 * @param metadata the dataset metadata.
	 * @param label the label returned in case of a tie. Usually this label
	 *        will represent the overall majority of the data set.
	 * 
	 * @return the majority class value among the specified instances.
	 */
	public static Label findMajorityClass(List<Instance> instances,
		Metadata metadata, Label label)
	{
		String name = metadata.getTarget().getName();
		return findMajorityClassValue(instances, metadata.getIndex(name), label);
	}

	/**
	 * Returns the majority class value among the specified dataset.
	 * 
	 * @param dataset the dataset reference.
	 * 
	 * @return the majority class value among the specified dataset.
	 */
	public static Label findMajorityClass(Dataset dataset)
	{
		return findMajorityClass(dataset.getInstances(), dataset.getMetadata());
	}

	/**
	 * Filters the specified list of instances according to the test represented
	 * by the term.
	 * 
	 * @param instances the list of instances to be filtered.
	 * @param term the filter criterion.
	 * 
	 * @return a list of instances that satisfies the term test.
	 */
	public static List<Instance> filter(List<Instance> instances, Term term)
	{
		LinkedList<Instance> filtered = new LinkedList<Instance>();

		for (Instance instance : instances)
		{
			if (term.satisfies(instance.values()[term.getIndex()]))
			{
				filtered.add(instance);
			}
		}

		return filtered;
	}

	/**
	 * Verifies if the the list of instances have the same class label.
	 * 
	 * @param instances the list of instances to check.
	 * 
	 * @return <code>true</code> of all instances have the same label;
	 *         <code>false</code> otherwise.
	 */
	public static boolean sameLabel(List<Instance> instances)
	{
		Label label = null;

		for (Instance instance : instances)
		{
			if (label == null)
			{
				label = instance.getLabel();
			}
			else if (!label.equals(instance.getLabel()))
			{
				return false;
			}
		}

		return true;
	}

	/**
	 * Returns a label instance with all present class labels from from the
	 * specified instances active.
	 * 
	 * @param instances the list of instances to check.
	 * 
	 * @return a label instance with all present class labels from from the
	 * specified instances active.
	 */
	public static Label getPresentLabel(List<Instance> instances)
	{
		if (instances.isEmpty())
		{
			throw new IllegalArgumentException("Empty instance list.");
		}

		Metadata metadata = instances.iterator().next().getMetadata();
		Label prototype = metadata.getLabel();
		boolean[] present = new boolean[prototype.length()];
		Arrays.fill(present, false);

		for (Instance instance : instances)
		{
			boolean[] flags = instance.labels();

			for (int i = 0; i < flags.length; i++)
			{
				if (flags[i])
				{
					present[i] = true;
				}
			}
		}

		return prototype.toLabel(present);
	}
}