/*
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2011 Fernando Esteban Barril Otero
 */

package net.sf.myra.datamining;

import java.util.Collection;

import net.sf.myra.datamining.data.ClassHierarchy;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Label;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.framework.Cost;
import net.sf.myra.framework.Trail;

/**
 * @author Fernando Esteban Barril Otero
 */
public class HierarchicalPruner
	extends AbstractPruner
{
	public HierarchicalPruner(AbstractObjectiveFunction function,
		ClassAssignator assignator)
	{
		super(function, assignator);
	}

	public Trail explore(Trail trail)
	{
		Rule rule = (Rule) trail;
		Dataset dataset = getFunction().getDataset();
		Metadata metadata = dataset.getMetadata();
		ClassHierarchy hierarchy = metadata.getClassHierarchy();

		// class label present in the covered examples
		Label present =
			Evaluator
				.getPresentLabel(Evaluator.findCoveredCases(rule, dataset));
		rule.setConsequent(present);

		Cost cost = getFunction().evaluate(rule);
		rule.setCost(cost);

		while (true)
		{
			// removes irrelevant terms from the antecedent
			Rule improved = rule.clone();
			improved = pruneConsequent(pruneAntecedent(improved), hierarchy);

			if (improved.equals(rule))
			{
				// we could not improve the rule
				break;
			}
			else
			{
				rule = improved;
			}
		}

		return rule;
	}

	private Rule pruneAntecedent(Rule rule)
	{
		Cost cost = rule.getCost();

		while (rule.getSize() > 1)
		{
			Rule improved = rule.clone();
			improved.remove(rule.getLast());

			Cost c = getFunction().evaluate(improved);

			// if the rule quality has improved or has the same quality
			if (c.compareTo(cost) > -1)
			{
				improved.setCost(c);
				rule = improved;
				cost = c;
			}
			// otherwise we stop
			else
			{
				break;
			}
		}

		return rule;
	}

	/**
	 * Removes class labels from the consequent of the rule until the quality
	 * of the rule decreases.
	 * 
	 * @param rule
	 * @param consequent
	 * @param hierarchy
	 * 
	 * @return
	 */
	private Rule pruneConsequent(Rule rule, ClassHierarchy hierarchy)
	{
		Label consequent = rule.getConsequent();
		Collection<String> values = consequent.getValues();

		boolean removed = false;
		Cost cost = rule.getCost();

		do
		{
			removed = false;
			Collection<String> leaves = hierarchy.leaves(values);

			for (String label : leaves)
			{
				values.remove(label);
				rule.setConsequent(consequent.toLabel(values));

				Cost c = getFunction().evaluate(rule);

				if (c.compareTo(cost) < 0)
				{
					// the quality decreased, so we do not remove the label
					values.add(label);
				}
				else
				{
					removed = true;
					cost = c;
				}
			}
		} while (removed);

		rule.setConsequent(consequent.toLabel(values));
		rule.setCost(cost);

		return rule;
	}
}