/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Map.Entry;
import java.util.logging.Logger;

import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.io.Exporter;
import net.sf.myra.datamining.io.Helper;
import net.sf.myra.datamining.io.HierarchicalExporter;
import net.sf.myra.datamining.util.FlatCrossValidation;
import net.sf.myra.datamining.util.HierarchicalCrossValidation;
import net.sf.myra.datamining.util.FlatCrossValidation.Mode;
import net.sf.myra.datamining.util.FlatCrossValidation.Partition;
import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.ObjectFactory;
import net.sf.myra.framework.RandomGenerator;

/**
 * Default executable class to classification algorithms.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class Main
{
	/**
	 * The property key under which the cross validation folds value is stored.
	 */
	public static final String CROSS_VALIDATION_FOLDS =
		"net.sf.myra.datamining.folds";

	/**
	 * The property key under which the bulk value is stored.
	 */
	public static final String BULK_SIZE = "net.sf.myra.datamining.bulk";

	/**
	 * The property key under which the classifier class name is stored.
	 */
	public static final String CLASSIFIER = "net.sf.myra.datamining.classifier";

	/**
	 * The property key under which the measure implementation class name
	 * is stored.
	 */
	public static final String MEASURE = "net.sf.myra.datamining.measure";

	/**
	 * The property key under which the exporter implementation class name
	 * is stored.
	 */
	public static final String EXPORTER = "net.sf.myra.datamining.exporter";

	/**
	 * The property key under which the exporter output file name is stored.
	 */
	public static final String EXPORTER_FILE =
		"net.sf.myra.datamining.exporter.file";

	/**
	 * Default training file name prefix.
	 */
	public static final String TR_PREFIX = "TR";

	/**
	 * Default test file name prefix.
	 */
	public static final String TS_PREFIX = "TS";

	/**
	 * Default test file name prefix.
	 */
	public static final String FOLD_PREFIX = "FOLD";

	/**
	 * Configuration messages file.
	 */
	private static final String MESSAGE_FILE = "myra-datamining.messages";

	/**
	 * The logger instance.
	 */
	private Logger logger = Logger.getLogger(Main.class.getName());

	/**
	 * Returns the classifier instance to be used.
	 * 
	 * @return the classifier instance to be used.
	 */
	public Classifier getClassifier()
	{
		return (Classifier) ObjectFactory.create(Configuration
			.getProperty(CLASSIFIER));
	}

	/**
	 * Returns the measure instance to be used.
	 * 
	 * @return the measure instance to be used.
	 */
	public Measure getMeasure()
	{
		return (Measure) ObjectFactory.create(Configuration
			.getProperty(MEASURE));
	}

	/**
	 * Returns the classifier options.
	 * 
	 * @return the classifier options.
	 */
	public Properties getOptions(String file)
	{
		try
		{
			Properties options = new Properties();
			options.load(getClass().getResourceAsStream(file));
			return options;
		}
		catch (IOException e)
		{
			throw new RuntimeException(e);
		}
	}

	/**
	 * Program entry point.
	 * 
	 * @param args an array of command line arguments.
	 */
	public static void main(String[] args)
		throws IOException
	{
		new Main().run(args);
	}

	//------------------------------------------------------------- Run Methods

	/**
	 * Executes a classification algorithm.
	 * 
	 * @param args an array of command line arguments.
	 */
	public void run(String[] args)
		throws IOException
	{
		if (args.length == 0)
		{
			StringBuffer usage = new StringBuffer();

			usage.append("usage: java [options] ");
			usage.append(getClass().getName());
			usage.append(" ([-test <test file>]");
			usage.append(" <training file> | -batch <pattern> <directory>)");
			usage.append("\n\n");

			usage.append("General options:");
			usage.append("\n\n");

			try
			{
				Classifier c = getClassifier();
				append(usage, c.filter(Configuration.getOptions()));
				append(usage, c.filter(getOptions(MESSAGE_FILE)));

				if (!c.getOptions().isEmpty())
				{
					usage.append("Classifier options:");
					usage.append("\n\n");

					append(usage, c.getOptions());
				}
			}
			catch (IllegalArgumentException e)
			{
				// the classifier is not defined, prints the default properties
				append(usage, Configuration.getOptions());
				append(usage, getOptions(MESSAGE_FILE));
			}

			System.out.print(usage.toString());
			System.exit(1);
		}

		Map<String, String> arguments = new LinkedHashMap<String, String>();
		boolean value = false;

		// parses the command line arguments

		for (int i = 0; i < args.length; i++)
		{
			if (args[i].startsWith("-"))
			{
				if ("-batch".equals(args[i]) || "-test".equals(args[i]))
				{
					arguments.put(args[i], null);
					value = true;
				}
				else
				{
					throw new IllegalArgumentException("Invalid option: "
						+ args[i]);
				}
			}
			else if (value)
			{
				arguments.put(args[i - 1], args[i]);
				value = false;
			}
			else
			{
				arguments.put(args[i], null);
			}
		}

		long startTime = System.nanoTime();

		// batch run
		if (arguments.containsKey("-batch"))
		{
			String pattern = null;
			String directory = null;

			for (Entry<String, String> entry : arguments.entrySet())
			{
				if ("-batch".equals(entry.getKey()))
				{
					pattern = entry.getValue();
				}
				else if (pattern != null)
				{
					directory = entry.getKey();
					break;
				}
			}

			if (pattern == null || directory == null)
			{
				throw new IllegalArgumentException("Missing batch arguments: "
					+ "pattern={" + (pattern == null ? "" : pattern)
					+ "},directory={" + (directory == null ? "" : directory)
					+ "}");
			}

			runBatch(pattern, directory);
		}
		// single training/test run
		else
		{
			if (arguments.containsKey("-test"))
			{
				String trFile = null;
				String tsFile = null;

				for (Entry<String, String> entry : arguments.entrySet())
				{
					if ("-test".equals(entry.getKey()))
					{
						tsFile = entry.getValue();
					}
					else if (tsFile != null)
					{
						trFile = entry.getKey();
						break;
					}
				}

				Dataset trainingSet = Helper.open(trFile);
				Dataset testSet = Helper.open(tsFile);

				runTest(trainingSet, testSet);
			}
			else
			{
				String ds = arguments.entrySet().iterator().next().getKey();
				Dataset dataset = Helper.open(ds);

				// cross-validation run
				if (Configuration.isPresent(CROSS_VALIDATION_FOLDS))
				{
					runCrossValidation(dataset);
				}
				// single training run
				else
				{
					print(dataset);
					runTrain(dataset);
				}
			}
		}

		// elapsed time in seconds
		long elapsedTime = (System.nanoTime() - startTime) / 1000000000;
		print("\nElapsed time: " + elapsedTime + " second(s)\n");
	}

	protected void runTest(Dataset trainingSet, Dataset testSet)
	{
		if (Configuration.isPresent(BULK_SIZE))
		{
			int size = Configuration.getIntProperty(BULK_SIZE);
			Measure measure = getMeasure();

			print(trainingSet);
			print("Test mode: " + size + "-bulk(s)");
			print("\n");

			Unit<?>[] trResult = new Unit<?>[size];
			Unit<?>[] tsResult = new Unit<?>[size];

			Gatherer gatherer = null;
			UnitAveragingMode<?> mode = null;

			for (int i = 0; i < size; i++)
			{
				Model model = getClassifier().run(trainingSet.clone());
				trResult[i] = measure.evaluate(trainingSet, model);
				tsResult[i] = measure.evaluate(testSet, model);

				logger.info("Model evaluation: TR=" + trResult[i].doubleValue()
					+ ", TS=" + tsResult[i].doubleValue());

				if (gatherer == null)
				{
					gatherer = model.getGatherer();
					mode =
						UnitAveragingMode
							.get(trResult[i].getValue().getClass());
				}

				gatherer.add(model);
				print(".");
			}

			print("\n\n");
			print("Measure (" + measure.getClass().getSimpleName() + "): "
				+ mode.average(tsResult) + " [" + mode.average(trResult)
				+ "]\n\n");

			print(gatherer.toString());
			print("\n");

			for (int i = 0; i < tsResult.length; i++)
			{
				print((i < 9 ? "0" : "") + (i + 1) + ": ");
				print(tsResult[i].toString());
				print(" [" + trResult[i].toString() + "]");
				print("\n");
			}

			print("\n");
		}
		else
		{
			print(trainingSet);
			print("\n");

			Model discovered = getClassifier().run(trainingSet.clone());
			print("=== Discovered Model ===\n\n");
			print(discovered.toString());

			Measure measure = getMeasure();
			print("\nMeasure (" + measure.getClass().getSimpleName() + "): "
				+ measure.evaluate(testSet, discovered).toString() + " ["
				+ measure.evaluate(trainingSet, discovered) + "]\n");

			// export individual predictions for test instances (if the options
			// has been enabled)

			if (Configuration.isPresent(EXPORTER_FILE))
			{
				Exporter exporter = null;

				try
				{
					exporter =
						(Exporter) ObjectFactory.create(Configuration
							.getProperty(EXPORTER, (trainingSet.getMetadata()
								.isHierarchical() ? HierarchicalExporter.class
								.getName() : Exporter.class.getName())));

					exporter
						.initialize(new File(Configuration
							.getProperty(EXPORTER_FILE)), trainingSet
							.getMetadata());

					exporter.write(testSet, discovered);
				}
				catch (IOException e)
				{
					e.printStackTrace();
				}
				finally
				{
					if (exporter != null)
					{
						exporter.close();
					}
				}
			}
		}
	}

	protected void runTrain(Dataset dataset)
	{
		print("\n");
		Model discovered = getClassifier().run(dataset);
		print("=== Discovered Model ===\n\n");
		print(discovered.toString());
	}

	protected void runCrossValidation(Dataset dataset)
	{
		int folds = Configuration.getIntProperty(CROSS_VALIDATION_FOLDS);

		Partition[] partitions = null;

		if (dataset.getMetadata().isHierarchical())
		{
			partitions = HierarchicalCrossValidation.partition(dataset, folds);
		}
		else
		{
			partitions =
				FlatCrossValidation.split(dataset, folds, Mode.STRATIFIED);
		}

		Metadata metadata = dataset.getMetadata();
		String name = metadata.getName();

		for (int i = 0; i < folds; i++)
		{
			// prepares the training data

			List<Instance> instances = new ArrayList<Instance>();

			for (int j = 0; j < folds; j++)
			{
				if (i != j)
				{
					instances.addAll(partitions[j].getInstances());
				}
			}

			metadata.setName(name + (i < 9 ? "-0" : "-") + (i + 1));

			Dataset trainingSet = new Dataset(metadata);
			trainingSet.setInstances(instances);

			Dataset testSet = new Dataset(metadata);
			testSet.setInstances(partitions[i].getInstances());

			runTest(trainingSet, testSet);
			print("\n");
		}
	}

	protected void runBatch(String pattern, String directory)
		throws IOException
	{
		for (int i = 0;; i++)
		{
			String fold = (i < 10) ? "0" + i : Integer.toString(i);
			File tr = new File(directory, TR_PREFIX + fold + "_" + pattern);
			File ts = new File(directory, TS_PREFIX + fold + "_" + pattern);

			if (tr.exists() && ts.exists())
			{
				Dataset trainingSet = Helper.open(tr);
				Dataset testSet = Helper.open(ts);

				runTest(trainingSet, testSet);
				print("\n");
			}
			else
			{
				break;
			}
		}
	}

	//---------------------------------------------------------- Output Methods

	/**
	 * Prints the specified string to the default output stream.
	 * 
	 * @param s the string to be printed.
	 */
	protected void print(String s)
	{
		System.out.print(s);
		System.out.flush();
	}

	/**
	 * Prints the dataset information.
	 * 
	 * @param dataset thje current dataset.
	 */
	protected void print(Dataset dataset)
	{
		StringBuffer output = new StringBuffer();

		output.append("=== Run Information ===\n\n");
		output.append("Random seed: ");
		output.append(RandomGenerator.getInstance().getSeed());
		output.append("\nRelation: ");
		output.append(dataset.getMetadata().getName());
		output.append("\nInstances: ");
		output.append(dataset.getSize());
		output.append("\nAttributes: ");
		output.append(dataset.getMetadata().getPredictor().size());
		output.append("\nDuplicates: ");
		output.append(dataset.hasDuplicates() ? "yes" : "no");
		output.append("\nMissing values: ");
		output.append(dataset.hasMissing() ? "yes" : "no");
		output.append("\nClasses: ");
		output.append(dataset.getMetadata().getTarget().getValues().size());
		output.append("\nHierarchical: ");
		output.append(dataset.getMetadata().isHierarchical() ? "yes" : "no");
		output.append("\n");

		System.out.print(output);
	}

	/**
	 * Appends the specified options to the output buffer.
	 * 
	 * @param output the output buffer.
	 * @param options the options mapping.
	 */
	protected void append(StringBuffer output, Properties options)
	{
		for (Object key : options.keySet())
		{
			output.append("-D");
			output.append(key);
			output.append("=<value>\n");
			output.append("\t");
			output.append(options.get(key));
			output.append("\n\n");
		}
	}
}