/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining;

import net.sf.myra.framework.Cost;
import net.sf.myra.framework.Trail;
import net.sf.myra.framework.Vertex;

/**
 * This class represents a prune procedure where one term is removed at a
 * time, until no terms can be removed to improve the rule quality.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class Pruner
	extends AbstractPruner
{
	/**
	 * Creates a new <code>Pruner</code> instance.
	 * 
	 * @param function the <code>ObjectiveFunction</code> instance.
	 */
	public Pruner(AbstractObjectiveFunction function, ClassAssignator assignator)
	{
		super(function, assignator);
	}

	public Trail explore(Trail trail)
	{
		Rule rule = (Rule) trail;
		ClassAssignator assignator = getClassAssignator();
		AbstractObjectiveFunction function = getFunction();

		assignator.assign(rule, function.getDataset());
		Cost quality = function.evaluate(rule);
		Rule candidate = null;

		// removes one term at a time until the removal of one term does
		// not improve the rule quality or the rule has only one term left
		while (rule.getSize() > 1 && !rule.equals(candidate))
		{
			// at the begining of the removal loop, the candidate is
			// equal to the best rule found so far
			candidate = rule;

			for (Vertex<?> v : candidate.getVertices())
			{
				Rule r = candidate.clone();
				r.remove(v);

				assignator.assign(r, function.getDataset());
				Cost c = function.evaluate(r);

				// we are favoring shorter rules, so if the candidate has the
				// same cost, we keep the candidate (occam's razor rule)
				if (c.compareTo(quality) > -1)
				{
					rule = r;
					quality = c;
				}
			}
		}

		return rule;
	}
}