/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining;

import java.util.ArrayList;
import net.sf.myra.framework.Cost;
import net.sf.myra.framework.Trail;
import net.sf.myra.framework.Vertex;

/**
 * This class represents a prune procedure where one term is added at a
 * time, until no terms can be added to improve the rule quality.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class ReversePruner
	extends AbstractPruner
{
	/**
	 * Creates a new <code>ReversePruner</code> instance.
	 * 
	 * @param function the <code>ObjectiveFunction</code> instance.
	 */
	public ReversePruner(AbstractObjectiveFunction function,
			ClassAssignator assignator)
	{
		super(function, assignator);
	}

	public Trail explore(Trail trail)
	{
		ClassAssignator assignator = getClassAssignator();
		AbstractObjectiveFunction function = getFunction();
		Rule original = (Rule) trail;

		Rule rule =
			new Rule(original.getAntecedent().length, original.getConsequent());
		Rule candidate = null;
		Cost cost = null;

		ArrayList<Vertex<?>> vertices =
			new ArrayList<Vertex<?>>(trail.getVertices());

		// adds one term at a time until the addition of one term does
		// not improve the rule quality or there is no more terms to add
		while (!vertices.isEmpty() && !rule.equals(candidate))
		{
			candidate = rule;
			Vertex<?> vertex = null;

			for (Vertex<?> v : vertices)
			{
				Rule r = candidate.clone();
				r.add(v);

				assignator.assign(r, function.getDataset());
				Cost c = function.evaluate(r);

				// we are favoring shorter rules, so if the candidate has the
				// same cost, we keep the original (occam's razor rule)
				if (cost == null || c.compareTo(cost) == 1)
				{
					rule = r;
					cost = c;
					vertex = v;
				}
			}

			if (vertex != null)
			{
				vertices.remove(vertex);
			}
		}

		return rule;
	}
}