/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining;

import java.util.Arrays;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Label;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.data.Term;
import net.sf.myra.framework.DefaultTrail;
import net.sf.myra.framework.Vertex;

/**
 * This class represents a classification rule, which is built from ant tours
 * on the problem construction graph.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date: 2008-02-08 15:54:20#$
 */
public class Rule
	extends DefaultTrail
{
	private static final long serialVersionUID = 77455399685800795L;

	/**
	 * The rule consequent array.
	 */
	private Label consequent;

	/**
	 * The rule antecedent array.
	 */
	private Term[] antecedent;

	/**
	 * The rule confidence factor.
	 */
	private ConfidenceFactor confidence;

	/**
	 * Default constructor.
	 */
	public Rule(Metadata metadata)
	{
		this(metadata.getPredictor().size(), metadata.getLabel());
	}

	/**
	 * Creates a new <code>Rule</code> instance.
	 * 
	 * @param size the maximum size of the rule antecedent.
	 * @param prototype the label prototype instance.
	 */
	public Rule(int size, Label prototype)
	{
		antecedent = new Term[size];
		Arrays.fill(antecedent, null);

		consequent = prototype.toLabel();
	}

	/**
	 * Returns the rule consequent.
	 * 
	 * @return the rule consequent.
	 */
	public Label getConsequent()
	{
		return consequent;
	}

	/**
	 * Returns the rule antecedent.
	 * 
	 * @return the rule antecedent.
	 */
	public Term[] getAntecedent()
	{
		return antecedent.clone();
	}

	/**
	 * Sets the rule consequent.
	 * 
	 * @param consequent the rule consequent to set.
	 */
	public void setConsequent(Label consequent)
	{
		this.consequent = consequent;
	}

	/**
	 * @return the confidence
	 */
	public ConfidenceFactor getConfidence()
	{
		return confidence;
	}

	/**
	 * @param confidence the confidence to set
	 */
	public void setConfidence(ConfidenceFactor confidence)
	{
		this.confidence = confidence;
	}

	@Override
	@SuppressWarnings("unchecked")
	public void add(Vertex<?> vertex)
	{
		super.add(vertex);

		Term term = ((Vertex<Term>) vertex).getInfo();
		antecedent[term.getIndex()] = term;
	}

	@Override
	@SuppressWarnings("unchecked")
	public void remove(Vertex<?> vertex)
	{
		super.remove(vertex);

		Term term = ((Vertex<Term>) vertex).getInfo();
		antecedent[term.getIndex()] = null;
	}

	/**
	 * Checks if this rule covers the specified instance.
	 * 
	 * @param instance the instance to check.
	 * 
	 * @return <code>true</code> if this rule covers the specified instances;
	 *         <code>false</code> otherwise.
	 */
	public boolean isCovered(Instance instance)
	{
		boolean valid = true;

		for (int i = 0; (i < antecedent.length) && valid; i++)
		{
			if (antecedent[i] != null)
			{
				valid = antecedent[i].satisfies(instance.values()[i]);
			}
		}

		return valid;
	}

	@Override
	@SuppressWarnings("unchecked")
	public String toString()
	{
		StringBuffer buffer = new StringBuffer();
		buffer.append("IF ");

		if (getVertices().isEmpty())
		{
			buffer.append("<empty>");
		}
		else
		{
			int index = 0;

			for (Vertex<?> vertex : getVertices())
			{
				Vertex<Term> v = (Vertex<Term>) vertex;

				if (index != 0)
				{
					buffer.append(" AND ");
				}

				buffer.append(v.getInfo().toString());
				index++;
			}
		}

		buffer.append(" THEN ");
		buffer.append((consequent.cardinality() == 0) ? "<undefined>"
			: consequent);

		if (confidence != null && confidence instanceof DefaultConfidenceFactor)
		{
			DefaultConfidenceFactor factor = (DefaultConfidenceFactor)confidence;
			buffer.append(" [");
			String[] values = consequent.names();
			
			for (int i = 0; i < values.length; i++)
			{
				if (i > 0)
				{
					buffer.append(',');
				}
				
				buffer.append(factor.get(values[i]));
			}
			
			buffer.append("]");
		}
		
		return buffer.toString();
	}

	@Override
	public Rule clone()
	{
		Rule rule = (Rule) super.clone();
		rule.antecedent = antecedent.clone();

		if (confidence != null)
		{
			rule.confidence = confidence.clone();
		}

		return rule;
	}

	@Override
	public boolean equals(Object o)
	{
		return equals((Rule) o);
	}

	/**
	 * Compares this rule against the specified rule for equality.
	 * 
	 * @param o the rule to be compared to.
	 * 
	 * @return <code>true</code> if both rules are equal; <code>false</code>
	 *         otherwise.
	 */
	private boolean equals(Rule o)
	{
		if (this == o)
		{
			return true;
		}
		else if (o == null)
		{
			return false;
		}

		// sanity check: makes sure the consequent is not null
		if ((consequent == null) && (o.consequent != null))
		{
			return false;
		}

		return ((consequent == o.consequent) || consequent.equals(o.consequent))
			&& Arrays.equals(antecedent, o.antecedent);
	}
}