/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining;

import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                    #$
 */
public abstract class UnitAveragingMode<V extends Comparable<V>>
{
	private static Map<String, UnitAveragingMode<?>> mapping =
		new HashMap<String, UnitAveragingMode<?>>();

	static
	{
		UnitAveragingMode.register(Double.class, new DoubleUnitAveragingMode());
	}

	/**
	 * Registers the specified averaging mode.
	 * 
	 * @param c the target unit class name.
	 * @param mode the averaging mode instance.
	 */
	public static void register(Class<?> c, UnitAveragingMode<?> mode)
	{
		mapping.put(c.toString(), mode);
	}

	public static UnitAveragingMode<?> get(Class<?> c)
	{
		return mapping.get(c.toString());
	}

	public static String stdev(Collection<? extends Number> numbers)
	{
		double[] values = new double[numbers.size()];
		int index = 0;

		for (Number number : numbers)
		{
			values[index] = number.doubleValue();
			index++;
		}

		return stdev(values);
	}

	/**
	 * Caculates the mean and standard deviation. This method calculates the
	 * traditional (not cross-validation) standard deviation measure.
	 * 
	 * @param numbers the distribution.
	 */
	public static String stdev(double[] numbers)
	{
		double mean = 0;

		for (int i = 0; i < numbers.length; i++)
		{
			mean += numbers[i];
		}

		mean /= numbers.length;
		double std = 0;

		if (numbers.length > 1)
		{
			for (int i = 0; i < numbers.length; i++)
			{
				std += Math.pow(numbers[i] - mean, 2);
			}

			std /= (numbers.length - 1);
			std = Math.sqrt(std);
		}

		NumberFormat formatter = new DecimalFormat("#,##0.000000");
		return formatter.format(mean) + " +/- " + formatter.format(std);
	}

	/**
	 * Returns a <code>String</code> which corresponds to the average values
	 * of the specified units.
	 * 
	 * @param values the array of unit values.
	 * 
	 * @return a <code>String</code> which corresponds to the average values
	 * of the specified units.
	 */
	@SuppressWarnings("unchecked")
	public String average(Unit<?>[] values)
	{
		return (getAverage((Unit<V>[]) values));
	}

	@SuppressWarnings("unchecked")
	public int compare(Unit<?> u1, Unit<?> u2)
	{
		return compareTo((Unit<V>) u1, (Unit<V>) u2);
	}

	/**
	 * Returns a <code>String</code> which corresponds to the average values
	 * of the specified units.
	 * 
	 * @param values the array of unit values.
	 * 
	 * @return a <code>String</code> which corresponds to the average values
	 * of the specified units.
	 */
	protected abstract String getAverage(Unit<V>[] values);

	/**
	 * Compares the specified unit instances.
	 * 
	 * @param u1 the first unit. 
	 * @param u2 the second unit.
	 * 
	 * @return a negative integer, zero, or a positive integer as
	 *         <code>u1</code> is less than, equal to, or greater than the
	 *         specified <code>u2</code>.
	 */
	protected int compareTo(Unit<V> u1, Unit<V> u2)
	{
		return u1.compareTo(u2);
	}

	/**
	 * Default averaging mode for <code>Double</code> values.
	 */
	public static class DoubleUnitAveragingMode
		extends UnitAveragingMode<Double>
	{
		@Override
		public String getAverage(Unit<Double>[] values)
		{
			double[] v = new double[values.length];

			for (int i = 0; i < v.length; i++)
			{
				v[i] = values[i].getValue();
			}

			return UnitAveragingMode.stdev(v);
		}
	}
}