/*
 * $Id: AbstractDistanceBasedBuilder.java 2304 2010-08-06 22:58:15Z febo2 $
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2009  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining.data;

import java.util.Arrays;
import java.util.HashSet;
import java.util.List;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision: 2304 $ $Date:: 2010-08-06 23:58:15#$
 */
public abstract class AbstractDistanceBasedBuilder
	extends IntervalBuilder
{
	/**
	 * The class-wise weights. The order of weights is the same as the order
	 * of labels
	 */
	private double[] weight;

	/**
	 * Default constructor.
	 * 
	 * @param metadata the dataset metadata.
	 */
	public AbstractDistanceBasedBuilder(Metadata metadata)
	{
		super(metadata);

		ClassHierarchy hierarchy = metadata.getClassHierarchy();

		Label label = metadata.getLabel();
		weight = new double[label.length()];

		for (int i = 0; i < weight.length; i++)
		{
			weight[i] = hierarchy.find(label.names()[i]).getWeight();
		}
	}

	/**
	 * Prepared the data for the discretisation procedure.
	 * 
	 * @param instances the instances used to determine the candidate values.
	 * @param attribute the continuous attribute.
	 * 
	 * @return an array of the candidate values for discretisation.
	 */
	private Candidate[] prepare(List<Instance> instances,
			ContinuousAttribute attribute)
	{
		if (instances.isEmpty())
		{
			throw new IllegalArgumentException("Empty instance list");
		}

		Metadata metadata = instances.iterator().next().getMetadata();
		int klass = metadata.getIndex(metadata.getTarget().getName());
		int attributeIndex = metadata.getIndex(attribute.getName());

		Candidate[] candidates = new Candidate[instances.size()];
		int index = 0;

		for (Instance instance : instances)
		{
			double[] indexes = instance.values();

			if (!Double.isNaN(indexes[attributeIndex]))
			{
				Candidate point = new Candidate();
				point.flags = instance.getLabel().flags();
				point.label = indexes[klass];
				point.value = indexes[attributeIndex];

				candidates[index++] = point;
			}
		}

		if (index < candidates.length)
		{
			Candidate[] resized = new Candidate[index];
			System.arraycopy(candidates, 0, resized, 0, index);
			candidates = resized;
		}

		if (candidates.length > 0)
		{
			Arrays.sort(candidates);
		}

		return candidates;
	}

	@Override
	public Interval[] create(List<Instance> instances,
			ContinuousAttribute attribute)
	{
		Candidate[] candidates = prepare(instances, attribute);

		if (candidates.length > 0)
		{
			CutPoint[] points = create(candidates, false);

			if (points.length > 0)
			{
				Interval[] intervals = new Interval[points.length + 1];

				// calculates the average value for the threshold (only if it is
				// not the first value of the array)

				double value = points[0].value;

				if (points[0].threshold > 0)
				{
					value =
						(candidates[points[0].threshold - 1].value + value) / 2;
				}

				intervals[0] =
					new Interval(Double.NaN, Operator.LESS_THAN, value);

				for (int i = 1; i < points.length; i++)
				{
					double lower = points[i - 1].value;
					double upper = points[i].value;

					// calculates the average value for the threshold (only if it is
					// not the first value of the array)
					if (points[i - 1].threshold > 0)
					{
						lower =
							(candidates[points[i - 1].threshold - 1].value + lower) / 2;
					}

					// the upper bound should always be greater than the first value,
					// so it is safe to get the previous value
					upper =
						(candidates[points[i].threshold - 1].value + upper) / 2;

					intervals[i] =
						new Interval(Double.NaN, Operator.IN_RANGE, lower,
							upper);
				}

				// calculates the average value for the threshold (only if it is
				// not the first value of the array)

				value = points[points.length - 1].value;

				if (points[points.length - 1].threshold > 0)
				{
					value =
						(candidates[points[points.length - 1].threshold - 1].value + value) / 2;
				}

				intervals[intervals.length - 1] =
					new Interval(Double.NaN, Operator.GREATER_THAN_EQUAL_TO,
						value);

				return intervals;
			}
		}

		return new Interval[0];
	}

	@Override
	public Interval createSingle(List<Instance> instances,
			ContinuousAttribute attribute)
	{
		Candidate[] candidates = prepare(instances, attribute);

		if (candidates.length > 0)
		{
			Arrays.sort(candidates);
			CutPoint[] points = create(candidates, true);

			if (points.length == 1)
			{
				if (!Double.isInfinite(points[0].gain))
				{
					double value = points[0].value;

					// calculates the average value for the threshold (only if
					// it is not the first value of the array)
					if (points[0].threshold > 0)
					{
						value =
							(candidates[points[0].threshold - 1].value + value) / 2;
					}

					return new Interval(
						Math.min(points[0].impurity_s1, points[0].impurity_s2),
						(points[0].impurity_s1 > points[0].impurity_s2) ? Operator.GREATER_THAN_EQUAL_TO
							: Operator.LESS_THAN, value);
				}
			}
			else if (points.length == 2)
			{
				double lower = points[0].value;
				double upper = points[1].value;

				// calculates the average value for the threshold (only if it is
				// not the first value of the array)
				if (points[0].threshold > 0)
				{
					lower =
						(candidates[points[0].threshold - 1].value + lower) / 2;
				}

				// the upper bound should always be greater than the first value,
				// so it is safe to get the previous value
				upper = (candidates[points[1].threshold - 1].value + upper) / 2;

				return new Interval(points[0].impurity_s2, Operator.IN_RANGE,
					lower, upper);
			}
		}

		// candidates list may be empty or it was not possible to create
		// a continuous attribute (maybe all instances have the same value
		// or class label)
		return new Interval(Double.POSITIVE_INFINITY, Operator.LESS_THAN,
			Double.NaN);
	}

	@Override
	public Interval createSingle(List<Instance> instances,
			ContinuousAttribute attribute, String c)
	{
		throw new UnsupportedOperationException();
	}
	
	@Override
	public int count(List<Instance> instances, ContinuousAttribute attribute)
	{
		return prepare(instances, attribute).length;
	}

	/**
	 * Returns candidate cut point(s) given the data distribution.
	 * 
	 * @param candidates the data distribution.
	 * @param filter indicated if the cut points should be filtered. When
	 *        filtering cut points, it is guarenteed that this method will
	 *        return 1 or 2 values and when 2 values is returned, the interval
	 *        to be selected should be the one between the two values.
	 * 
	 * @return candidate cut point(s) given the data distribution.
	 */
	protected abstract CutPoint[] create(Candidate[] candidates, boolean filter);

	/**
	 * Returns the number of different class labels present in the specified
	 * distribution.
	 * 
	 * @param candidates the candidate points distributions.
	 * @param index the initial index.
	 * @param length the number of candidate points to consider.
	 * 
	 * @return the number of different class labels present in the specified
	 *         distribution.
	 */
	protected int count(Candidate[] candidates, int index, int length)
	{
		HashSet<Double> labels = new HashSet<Double>();

		for (int i = index; i < length; i++)
		{
			labels.add(candidates[i].label);
		}

		return labels.size();
	}

	/**
	 * Returns the variance of the specified class values distribution.
	 * 
	 * @param candidates the class values distribution.
	 * @param index the start index of the distribution.
	 * @param length the number of candidates to be includes in the evaluation.
	 * 
	 * @return the variance of the specified class values distribution.
	 */
	public double variance(Candidate[] candidates, int index, int length)
	{
		double[] mean = new double[candidates[index].flags.length];
		int size = length - index;

		for (int i = index; i < length; i++)
		{
			for (int j = 0; j < mean.length; j++)
			{
				if (candidates[i].flags[j])
				{
					mean[j]++;
				}
			}
		}

		for (int i = 0; i < mean.length; i++)
		{
			mean[i] /= size;
		}

		return variance(candidates, index, length, mean);
	}

	/**
	 * Returns the variance of the specified class values distribution.
	 * 
	 * @param candidates the class values distribution.
	 * @param index the start index of the distribution.
	 * @param length the number of candidates to be includes in the evaluation.
	 * 
	 * @return the variance of the specified class values distribution.
	 */
	public double variance(Candidate[] candidates, int index, int length,
			double[] mean)
	{
		double total = 0.0;

		for (int i = index; i < length; i++)
		{
			double distance = 0.0;

			for (int j = 0; j < mean.length; j++)
			{
				double value = (candidates[i].flags[j] ? 1 : 0) - mean[j];
				distance += weight[j] * (value * value);
			}

			total += distance;
		}

		return total / (length - index);
	}

	/**
	 * Struct-like class to represent the label information of an instance.
	 */
	protected static class Candidate
		implements Comparable<Candidate>
	{
		protected double value;

		protected double label;

		protected boolean[] flags;

		public int compareTo(Candidate o)
		{
			return Double.compare(value, o.value);
		}
	}
}