/*
 * $Id: IntervalBuilder.java 1981 2009-02-20 11:46:44Z febo2 $
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2009  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining.data;

import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision: 1981 $ $Date:: 2009-02-20 11:46:44#$
 */
public abstract class AbstractEntropyBasedBuilder
	extends IntervalBuilder
{
	/**
	 * Default constructor.
	 * 
	 * @param metadata the dataset metadata.
	 */
	public AbstractEntropyBasedBuilder(Metadata metadata)
	{
		super(metadata);
	}

	private double[][] prepare(List<Instance> instances,
		ContinuousAttribute attribute, Map<Double, Integer> mapping)
	{
		if (instances.isEmpty())
		{
			throw new IllegalArgumentException("Empty instance list");
		}

		Metadata metadata = instances.iterator().next().getMetadata();
		int klass = metadata.getIndex(metadata.getTarget().getName());
		int attributeIndex = metadata.getIndex(attribute.getName());

		double[][] matrix = new double[instances.size()][3];
		int index = 0;

		for (Instance instance : instances)
		{
			double[] indexes = instance.values();

			if (!Double.isNaN(indexes[attributeIndex]))
			{
				Integer value = mapping.get(indexes[klass]);

				if (value == null)
				{
					value = mapping.size();
					mapping.put(indexes[klass], value);
				}

				matrix[index][0] = indexes[attributeIndex];
				matrix[index][1] = value;
				matrix[index][2] = instance.getWeight();

				index++;
			}
			else
			{
				double[][] m = new double[matrix.length - 1][3];
				System.arraycopy(matrix, 0, m, 0, index);
				matrix = m;
			}
		}

		// sort the matrix based on the attribute values
		sort(matrix);

		return matrix;
	}

	@Override
	public Interval[] create(List<Instance> instances,
		ContinuousAttribute attribute)
	{
		// mapping of class label values
		Map<Double, Integer> mapping = new HashMap<Double, Integer>();

		double[][] matrix = prepare(instances, attribute, mapping);
		CutPoint[] points = create(matrix, mapping.size(), false);

		if (points.length > 0)
		{
			Interval[] intervals = new Interval[points.length + 1];

			// calculates the average value for the threshold (only if it is
			// not the first value of the array)

			double value = points[0].value;

			if (points[0].threshold > 0)
			{
				value = average(matrix[points[0].threshold - 1][0], value);
			}

			intervals[0] =
				new Interval(Double.NaN, Operator.LESS_THAN_EQUAL_TO, value);

			for (int i = 1; i < points.length; i++)
			{
				double lower = points[i - 1].value;
				double upper = points[i].value;

				// calculates the average value for the threshold (only if it is
				// not the first value of the array)
				if (points[i - 1].threshold > 0)
				{
					lower =
						average(matrix[points[i - 1].threshold - 1][0], lower);
				}

				// the upper bound should always be greater than the first value,
				// so it is safe to get the previous value
				upper = average(matrix[points[i].threshold - 1][0], upper);

				intervals[i] =
					new Interval(Double.NaN, Operator.IN_RANGE, lower, upper);
			}

			// calculates the average value for the threshold (only if it is
			// not the first value of the array)

			value = points[points.length - 1].value;

			if (points[points.length - 1].threshold > 0)
			{
				value =
					average(matrix[points[points.length - 1].threshold - 1][0],
						value);
			}

			intervals[intervals.length - 1] =
				new Interval(Double.NaN, Operator.GREATER_THAN, value);

			return intervals;
		}

		return new Interval[0];
	}

	/**
	 * Returns an interval for the specified continuous attribute tailored
	 * for the specified instance.
	 * 
	 * @param instances the list of instances.
	 * @param attribute the continuous attribute.
	 * 
	 * @return an interval for the specified continuous attribute tailored
	 * for the specified instance.
	 */
	@Override
	public Interval createSingle(List<Instance> instances,
		ContinuousAttribute attribute)
	{
		// mapping of class label values
		Map<Double, Integer> mapping = new HashMap<Double, Integer>();

		double[][] matrix = prepare(instances, attribute, mapping);
		CutPoint[] points = create(matrix, mapping.size());

		if (points.length == 1)
		{
			double value = points[0].value;

			// calculates the average value for the threshold (only if it is
			// not the first value of the array)
			if (points[0].threshold > 0)
			{
				value = average(matrix[points[0].threshold - 1][0], value);
			}

			return new Interval(
				Math.min(points[0].impurity_s1, points[0].impurity_s2),
				(points[0].impurity_s1 > points[0].impurity_s2) ? Operator.GREATER_THAN
					: Operator.LESS_THAN_EQUAL_TO, value);
		}
		else if (points.length == 2)
		{
			double lower = points[0].value;
			double upper = points[1].value;

			// calculates the average value for the threshold (only if it is
			// not the first value of the array)
			if (points[0].threshold > 0)
			{
				lower = average(matrix[points[0].threshold - 1][0], lower);
			}

			// the upper bound should always be greater than the first value,
			// so it is safe to get the previous value
			upper = average(matrix[points[1].threshold - 1][0], upper);

			return new Interval(points[0].impurity_s2, Operator.IN_RANGE,
				lower, upper);
		}

		// candidates list may be empty or it was not possible to create
		// a continuous attribute (maybe all instances have the same value
		// or class label)
		return new Interval(Double.POSITIVE_INFINITY, Operator.LESS_THAN,
			Double.NaN);
	}

	/**
	 * Returns a discrete interval for the specified continuous attribute
	 * tailored for the specified instances and class value. The instances
	 * are arranged into a binary distribution (i.e. instances that belong to
	 * the specified class and instances that do not belong to the specified
	 * class).
	 * 
	 * @param instances the list of instances.
	 * @param attribute the continuous attribute.
	 * @param label the class value.
	 * 
	 * @return a discrete interval for the specified continuous attribute
	 *         tailored for the specified instances and class value.
	 */
	@Override
	public Interval createSingle(List<Instance> instances,
		ContinuousAttribute attribute, String label)
	{
		if (instances.isEmpty())
		{
			throw new IllegalArgumentException("Empty instance list");
		}

		Metadata metadata = instances.iterator().next().getMetadata();
		int cIndex = metadata.getLabel().getIndex(label);
		int aIndex = metadata.getIndex(attribute.getName());

		double[][] matrix = new double[instances.size()][3];
		int index = 0;

		for (Instance instance : instances)
		{
			double[] indexes = instance.values();

			if (!Double.isNaN(indexes[aIndex]))
			{
				matrix[index][0] = indexes[aIndex];
				matrix[index][1] = (instance.labels()[cIndex] ? 1 : 0);
				matrix[index][2] = instance.getWeight();

				index++;
			}
			else
			{
				double[][] m = new double[matrix.length - 1][3];
				System.arraycopy(matrix, 0, m, 0, index);
				matrix = m;
			}
		}

		// sort the matrix based on the attribute values
		sort(matrix);

		CutPoint[] points = create(matrix, 2, false);

		if (points.length == 1 && Double.isNaN(points[0].value))
		{
			return new Interval(Double.POSITIVE_INFINITY, Operator.LESS_THAN,
				Double.NaN);
		}

		// calculates the 'positive' class distribution of each interval

		int[] frequencies = new int[points.length + 1];
		double[] proportion = new double[points.length + 1];
		int[] size = new int[points.length + 1];
		int start = 0;
		index = 0;

		for (CutPoint point : points)
		{
			for (int i = start; i < point.threshold; i++)
			{
				// if this value corresponds to the 'postive' class
				if (matrix[i][1] == 1)
				{
					frequencies[index]++;
				}
			}

			size[index] = point.threshold - start;
			proportion[index] = frequencies[index] / (double) size[index];
			start = point.threshold;
			index++;
		}

		// calculates the 'positive' class distribution of the last interval

		for (int i = start; i < matrix.length; i++)
		{
			// if this value corresponds to the 'postive' class
			if (matrix[i][1] == 1)
			{
				frequencies[index]++;
			}
		}

		size[index] = matrix.length - start;
		proportion[index] = frequencies[index] / (double) size[index];

		// find the best (higher frequency value) interval index

		int biggest = Integer.MIN_VALUE;
		int shorter = Integer.MAX_VALUE;
		double purity = Double.MIN_VALUE;
		index = -1;

		for (int i = 0; i < frequencies.length; i++)
		{
			// if the frequency is higher
			if (frequencies[i] > biggest)
			{
				biggest = frequencies[i];
				shorter = size[i];
				purity = proportion[i];
				index = i;
			}
			// otherwise we check the proportion and the size
			else if (frequencies[i] == biggest)
			{
				if (proportion[i] > purity
					|| (proportion[i] == purity && size[i] < shorter))
				{
					biggest = frequencies[i];
					shorter = size[i];
					purity = proportion[i];
					index = i;
				}
			}
		}

		if (points.length > 0)
		{
			if (index == 0)
			{
				double value = points[0].value;
	
				// calculates the average value for the threshold (only if it is
				// not the first value of the array)
				if (points[0].threshold > 0)
				{
					value = average(matrix[points[0].threshold - 1][0], value);
				}
	
				return new Interval(points[0].impurity_s1,
					Operator.LESS_THAN_EQUAL_TO, value);
			}
			else if (index == (frequencies.length - 1))
			{
				double value = points[index - 1].value;
	
				// calculates the average value for the threshold (only if it is
				// not the first value of the array)
				if (points[index - 1].threshold > 0)
				{
					value =
						average(matrix[points[index - 1].threshold - 1][0], value);
				}
	
				return new Interval(points[index - 1].impurity_s2,
					Operator.GREATER_THAN, value);
			}
			else if (index > -1)
			{
				CutPoint p1 = points[index - 1];
				CutPoint p2 = points[index];
	
				double lower = p1.value;
				double upper = p2.value;
	
				// calculates the average value for the threshold (only if it is
				// not the first value of the array)
				if (p1.threshold > 0)
				{
					lower = average(matrix[p1.threshold - 1][0], lower);
				}
	
				// the upper bound should always be greater than the first value,
				// so it is safe to get the previous value
				upper = average(matrix[p2.threshold - 1][0], upper);
	
				return new Interval(entropy(matrix, p1.threshold, p2.threshold, 2),
					Operator.IN_RANGE, lower, upper);
			}
		}

		// candidates list may be empty or it was not possible to create
		// a continuous attribute (maybe all instances have the same value
		// or class label)
		return new Interval(Double.POSITIVE_INFINITY, Operator.LESS_THAN,
			Double.NaN);
	}

	@Override
	public int count(List<Instance> instances, ContinuousAttribute attribute)
	{
		Map<Double, Integer> mapping = new HashMap<Double, Integer>();
		double[][] matrix = prepare(instances, attribute, mapping);

		// the minimum number of instances of each partition (similar to
		// J48 algorithm)
		double minimum =
			0.1
				* (instances.size())
				/ ((double) instances.iterator().next().getMetadata()
					.getTarget().getValues().size());

		if (minimum <= minimumLimit)
		{
			minimum = minimumLimit;
		}
		else if (minimum > 25.0)
		{
			minimum = 25.0;
		}

		int count = 0;

		for (int i = 1; i < matrix.length; i++)
		{
			if (matrix[i - 1][0] + 1e-5 < matrix[i][0])
			{
				if (weightedLength(matrix, 0, i - 1) >= minimum
					&& weightedLength(matrix, i, matrix.length - 1) >= minimum)
				{
					count++;
				}
			}
		}

		return count;
	}

	/**
	 * Sorts (based on a bubble sort algorithm) the matrix of attribute
	 * value -> class value.
	 * 
	 * @param matrix the matrix to be sorted.
	 */
	protected void sort(double[][] matrix)
	{
		for (int i = 0; i < matrix.length; i++)
		{
			double[] min = matrix[i];
			int index = i;

			for (int j = i; j < matrix.length; j++)
			{
				if (matrix[j][0] < min[0])
				{
					min = matrix[j];
					index = j;
				}
			}

			matrix[index] = matrix[i];
			matrix[i] = min;
		}
	}

	/**
	 * Returns the entropy of the specified range of examples.
	 * 
	 * @param matrix the (values,class) distribution.
	 * @param index the start index.
	 * @param length the number of position to evaluate.
	 * @param k the total number of class values.
	 * 
	 * @return the entropy of the specified range of examples.
	 */
	protected double entropy(double[][] matrix, int index, int length, int k)
	{
		int[] frequency = new int[k];
		Arrays.fill(frequency, 0);

		for (int i = index; i < length; i++)
		{
			frequency[(int) matrix[i][1]]++;
		}

		double entropy = 0.0;

		for (int i = 0; i < frequency.length; i++)
		{
			if (frequency[i] > 0)
			{
				double p = frequency[i] / (double) (length - index);
				entropy -= (p * (Math.log(p) / Math.log(2.0)));
			}
		}

		return entropy;
	}

	/**
	 * Returns the number of classes (diversity) in the specified range of
	 * examples.
	 * 
	 * @param matrix the (values,class) distribution.
	 * @param index the start index.
	 * @param length the number of position to evaluate.
	 * @param c the number of classes.
	 * 
	 * @return the number of classes within the specified range.
	 */
	protected int diversity(double[][] matrix, int index, int length, int c)
	{
		int[] values = new int[c];

		for (int i = index; i < length; i++)
		{
			values[(int) matrix[i][1]] = 1;
		}

		int size = 0;

		for (int i = 0; i < values.length; i++)
		{
			if (values[i] > 0)
			{
				size++;
			}
		}

		return size;
	}

	protected double weightedLength(double[][] matrix, int from, int to)
	{
		double length = 0.0;

		for (int i = from; i <= to; i++)
		{
			length += matrix[i][2];
		}

		return length;
	}

	protected double average(double value1, double value2)
	{
		double average = (value1 + value2) / 2.0;

		if (average == value2)
		{
			return value1;
		}

		return average;
	}

	/**
	 * Returns candidate cut point(s) given the data distribution.
	 * 
	 * @param matrix the (values,class) distribution.
	 * @param classes the total number of class values.
	 * 
	 * @return candidate cut point(s) given the data distribution.
	 */
	protected CutPoint[] create(double[][] matrix, int classes)
	{
		return create(matrix, classes, true);
	}

	/**
	 * Returns candidate cut point(s) given the data distribution.
	 * 
	 * @param matrix the (values,class) distribution.
	 * @param classes the total number of class values.
	 * @param filter indicated if the cut points should be filtered. When
	 *        filtering cut points, it is guarenteed that this method will
	 *        return 1 or 2 values and when 2 values is returned, the interval
	 *        to be selected should be the one between the two values.
	 * 
	 * @return candidate cut point(s) given the data distribution.
	 */
	protected abstract CutPoint[] create(double[][] matrix, int classes,
		boolean filter);
}