/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining.data;

import java.io.Serializable;

/**
 * This class represents an attribute of the dataset.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 * 
 * @see Metadata
 */
public abstract class Attribute
	implements Comparable<Attribute>, Cloneable, Serializable
{
	private static final long serialVersionUID = -6243465497305316114L;

	/**
	 * The attribute's name.
	 */
	private String name;

	/**
	 * Indicates if the attribute is the target attribute.
	 */
	private boolean target;

	/**
	 * Creates a new <code>Attribute</code> instance.
	 * 
	 * @param name the attribute's name.
	 */
	public Attribute(String name)
	{
		this.name = name;
	}

	/**
	 * Returns the attribute's name.
	 * 
	 * @return the attribute's name.
	 */
	public String getName()
	{
		return name;
	}

	/**
	 * Sets the attribute's name.
	 * 
	 * @param name the name to set.
	 */
	public void setName(String name)
	{
		this.name = name;
	}

	/**
	 * Indicates if this attribute is the target attribute.
	 * 
	 * @return <code>true</code> if this attribute is the target attribute;
	 *         <code>false</code> otherwise.
	 */
	public boolean isTarget()
	{
		return target;
	}

	/**
	 * Sets the target flag of this attribute.
	 * 
	 * @param target the value to set.
	 */
	void setTarget(boolean target)
	{
		this.target = target;
	}

	/**
	 * Checks of the current value belongs to the attribute's domain.
	 * 
	 * @param value to value to check.
	 */
	public abstract void validate(String value);

	/**
	 * Checks if the specified value corresponds to a missing value.
	 * 
	 * @param value the value to check.
	 * 
	 * @return <code>true</code> if the specified value corresponds to a
	 *         missing value; <code>false</code> otherwise.
	 */
	public abstract boolean isMissing(double value);

	@Override
	public boolean equals(Object o)
	{
		return equals((Attribute) o);
	}

	public boolean equals(Attribute o)
	{
		return (o == null) ? false : compareTo(o) == 0;
	}

	public int compareTo(Attribute o)
	{
		return name.compareTo(o.name);
	}

	@Override
	public int hashCode()
	{
		return name.hashCode();
	}

	@Override
	public Attribute clone()
	{
		try
		{
			return (Attribute) super.clone();
		}
		catch (CloneNotSupportedException e)
		{
			throw new UnknownError(getClass() + ": clone()");
		}
	}
}