/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining.data;

import java.io.Serializable;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.Graph;
import net.sf.myra.framework.Vertex;
import net.sf.myra.framework.Info.Void;

/**
 * This class represents a class hierarchy.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class ClassHierarchy
	implements Cloneable, Serializable
{
	private static final long serialVersionUID = 2956554292809481680L;

	/**
	 * The property key under which the weight value is stored.
	 */
	private static final String WEIGHT =
		"net.sf.myra.datamining.hierarchy.weight";

	/**
	 * The default weight value.
	 */
	private static final double DEFAULT_WEIGHT = 0.75;

	/**
	 * Relationship separator between nodes of the hierarchy.
	 */
	public final static String RELATION = "/";

	/**
	 * The root of the hierarchy.
	 */
	private Node root = null;

	/**
	 * The set of nodes (improves the performance).
	 */
	private Map<String, Node> nodes = new HashMap<String, Node>();

	/**
	 * Tests if the hierarchy has no nodes.
	 * 
	 * @return <code>true</code> if the hierarchy has no nodes;
	 *         <code>false</code> otherwise.
	 */
	public boolean isEmpty()
	{
		return (root == null);
	}

	/**
	 * @return the root
	 */
	public Node getRoot()
	{
		return root;
	}

	/**
	 * Adds a new node to the class hierarchy.
	 * 
	 * @param label the label of the node to be added.
	 * 
	 * @return the node that was added to the class hierarchy.
	 */
	public Node add(String label)
	{
		return add(label, null);
	}

	/**
	 * Adds a new node to the class hierarchy as a child of the specified
	 * parent node.
	 * 
	 * @param label the label of the node to be added.
	 * @param parent the label of the parent node.
	 * 
	 * @return the node that was added to the class hierarchy.
	 */
	public Node add(String label, String parent)
	{
		if (parent == null)
		{
			// checks if we already have a root node
			if (root == null)
			{
				return (root = add(new Node(label)));
			}
			else
			{
				// the class hierarchy already have a root node
				throw new IllegalArgumentException(
					"There is already a root class: " + root);
			}
		}
		else
		{
			Node p = find(parent);
			Node c = add(new Node(label));
			link(p, c);

			return c;
		}
	}

	/**
	 * Adds the specified node to the node set.
	 * 
	 * @param node the node to be added.
	 * 
	 * @return the node instance if it was added to the node set.
	 */
	private Node add(Node node)
	{
		if (!nodes.containsKey(node.getLabel()))
		{
			nodes.put(node.getLabel(), node);
			return node;
		}

		throw new IllegalArgumentException("Duplicated node: " + node);
	}

	/**
	 * Tests if the hierarchy contains a node with the specified label.
	 * 
	 * @param label the node label.
	 * 
	 * @return <code>true</code> is the hierarchy contains a node with the
	 *         specified label; <code>false</code> otherwise.
	 */
	public boolean contains(String label)
	{
		return nodes.containsKey(label);
	}

	/**
	 * Returns the number of levels of the class hierarchy. Note that level
	 * information is only available for tree structures.
	 * 
	 * @return the number of levels of the class hierarchy.
	 */
	public int getLevels()
	{
		if (!isTree())
		{
			throw new IllegalArgumentException(
				"No level information for DAG structures.");
		}

		int levels = 0;

		for (Node node : nodes.values())
		{
			levels = Math.max(getLevel(node.getLabel()), levels);
		}

		return levels;
	}

	/**
	 * Returns the hierarchical level of the specified node. Note that level
	 * information is only available for tree structures.
	 * 
	 * @param label the label of the node.
	 * 
	 * @return the hierarchical level of the specified node.
	 */
	public int getLevel(String label)
	{
		if (!isTree())
		{
			throw new IllegalArgumentException(
				"No level information for DAG structures.");
		}

		Node node = nodes.get(label);
		int level = 0;

		while (!node.isRoot())
		{
			level++;
			node = node.getParents().iterator().next();
		}

		return level;
	}

	/**
	 * Links the specified parent-child nodes.
	 * 
	 * @param parent the label of the parent node.
	 * @param child the label of the child node.
	 */
	public void link(String parent, String child)
	{
		link(find(parent), find(child));
	}

	/**
	 * Links the specified parent-child nodes.
	 * 
	 * @param parent the parent node.
	 * @param child the child node.
	 */
	public void link(Node parent, Node child)
	{
		parent.getChildren().add(child);
		child.getParents().add(parent);
	}

	/**
	 * Removes the specified node from the class hierarchy. The node must be
	 * a leaft in order to be removed.
	 * 
	 * @param node the label of the node to be removed.
	 */
	public void remove(String node)
	{
		remove(find(node));
	}

	/**
	 * Removes the specified node from the class hierarchy. The node must be
	 * a leaft in order to be removed.
	 * 
	 * @param node the node to be removed.
	 */
	public void remove(Node node)
	{
		if (!node.isLeaf())
		{
			throw new IllegalArgumentException("Node is not a leaf: "
				+ node.getLabel());
		}

		for (Node parent : node.getParents())
		{
			parent.getChildren().remove(node);
		}

		if (nodes.remove(node.getLabel()).isRoot())
		{
			root = null;
		}
	}

	/**
	 * Returns the node instance with the specified label.
	 * 
	 * @param label the label of the node.
	 * 
	 * @return the node instance with the specified label.
	 */
	public Node find(String label)
	{
		if (isEmpty())
		{
			throw new IllegalArgumentException("Class hierarchy is empty");
		}

		Node node = nodes.get(label);

		if (node == null)
		{
			throw new IllegalArgumentException("Class not found: " + label);
		}

		return node;
	}

	/**
	 * Returns the number of nodes.
	 * 
	 * @return the number of nodes.
	 */
	public int getSize()
	{
		return nodes.size();
	}

	/**
	 * Returns the collection of all nodes labels.
	 * 
	 * @return the collection of all nodes labels.
	 */
	public Collection<Node> flat()
	{
		return nodes.values();
	}

	/**
	 * Swaps the root node for its single child node.
	 * 
	 * @return <code>true</code> if the swap was performed; <code>false</code>
	 *         otherwise.
	 */
	public boolean shrink()
	{
		if (root.getChildren().size() == 1)
		{
			nodes.remove(root.getLabel());
			root = root.getChildren().iterator().next();
			root.getParents().clear();

			return true;
		}

		return false;
	}

	/**
	 * Computes the weights of all nodes of the hierarchy.
	 */
	public void weigh()
	{
		HashSet<String> weighted = new HashSet<String>(nodes.size(), 1.0f);
		LinkedList<Node> evaluation = new LinkedList<Node>();

		if (!isEmpty())
		{
			double w = Configuration.getDoubleProperty(WEIGHT, DEFAULT_WEIGHT);
			root.setWeight(w);
			weighted.add(root.getLabel());

			for (Node node : root.getChildren())
			{
				node.setWeight(w);
				weighted.add(node.getLabel());
				evaluation.addAll(node.getChildren());
			}

			while (!evaluation.isEmpty())
			{
				boolean computed = true;
				double sum = 0.0;
				Node node = evaluation.removeFirst();

				if (!weighted.contains(node.getLabel()))
				{
					for (Node parent : node.getParents())
					{
						if (weighted.contains(parent.getLabel()))
						{
							sum += parent.getWeight();
						}
						else
						{
							evaluation.addLast(node);
							computed = false;
							break;
						}
					}

					if (computed)
					{
						node.setWeight(w * (sum / node.getParents().size()));
						weighted.add(node.getLabel());
						evaluation.addAll(node.getChildren());
					}
				}
			}
		}
	}

	/**
	 * Returns the label corresponding to the leaf node of the specified
	 * collection of class labels. Note that the label does not necessarily
	 * corresponds to leaf labels of the class hierarchy.
	 * 
	 * @param labels the collection of labels.
	 * 
	 * @return the label corresponding to the leaf node of the specified
	 * collection of class labels.
	 */
	public String leaf(Collection<String> labels)
	{
		return leaf(labels.toArray(new String[labels.size()]));
	}

	/**
	 * Returns the label corresponding to the leaf node of the specified
	 * array of class labels. Note that the label does not necessarily
	 * corresponds to leaf labels of the class hierarchy.
	 * 
	 * @param labels the array of labels.
	 * 
	 * @return the label corresponding to the leaf node of the specified
	 * collection of class labels.
	 */
	public String leaf(String... labels)
	{
		Collection<String> leaves = leaves(labels);

		// although we only support one leaf, the code can find more than one
		// (for multi-label support)
		if (leaves.size() > 1)
		{
			throw new IllegalArgumentException("More than one leaf found: "
				+ leaves);
		}

		return leaves.iterator().next();
	}

	/**
	 * Rerturns the collection of leaves of this class hierarchy.
	 * 
	 * @return the collection of leaves of this class hierarchy.
	 */
	public Collection<Node> getLeaves()
	{
		ArrayList<Node> leaves = new ArrayList<Node>();

		for (Node node : nodes.values())
		{
			if (node.isLeaf())
			{
				leaves.add(node);
			}
		}

		return leaves;
	}

	/**
	 * Returns the leaves label of the specified collection of class labels.
	 * Note that the labels do not necessarily corresponds to leaf labels of
	 * the class hierarchy.
	 * 
	 * @param labels the collection of labels.
	 * 
	 * @return the leaves label of the specified collection of class labels.
	 */
	public Collection<String> leaves(Collection<String> labels)
	{
		return leaves(labels.toArray(new String[labels.size()]));
	}

	/**
	 * Returns the leaves labels of the specified array of class labels. Note
	 * that the labels do not necessarily corresponds to leaf labels of the
	 * class hierarchy.
	 * 
	 * @param labels the array of labels.
	 * 
	 * @return the leaves label of the specified array of class labels.
	 */
	public Collection<String> leaves(String... labels)
	{
		Arrays.sort(labels);
		HashSet<String> leaves = new HashSet<String>();

		for (int i = 0; i < labels.length; i++)
		{
			boolean found = false;

			for (Node child : nodes.get(labels[i]).getChildren())
			{
				if (Arrays.binarySearch(labels, child.getLabel()) >= 0)
				{
					found = true;
					break;
				}
			}

			// if we did not find a child node, therefore this is a leaf node
			if (!found)
			{
				leaves.add(labels[i]);
			}
		}

		return leaves;
	}

	/**
	 * Checks if the class hierarchy has unique class label (i.e. each class
	 * label does not include the ancestor labels).
	 * 
	 * @return <code>true</code> if the class hierarchy contains unique class
	 *         labels; <code>false</code> otherwise.
	 */
	public boolean unique()
	{
		boolean unique = true;

		for (String label : nodes.keySet())
		{
			if (label.contains(RELATION))
			{
				unique = false;
				break;
			}
		}

		return unique;
	}

	/**
	 * Returns <code>true</code> if the class hierarchy is tree structured.
	 * 
	 * @return <code>true</code> if the class hierarchy is tree structured;
	 *         <code>false</code> otherwise.
	 */
	public boolean isTree()
	{
		boolean isTree = true;

		for (Node node : nodes.values())
		{
			if (node.getParents().size() > 1)
			{
				isTree = false;
				break;
			}
		}

		return isTree;
	}

	/**
	 * Returns a <code>Graph</code> representation of the class hierarchy.
	 * 
	 * @return a <code>Graph</code> representation of the class hierarchy.
	 */
	public Graph<Void> toGraph()
	{
		Graph<Void> graph = new Graph<Void>();

		if (!isEmpty())
		{
			// creates the graph nodes
			for (Node node : flat())
			{
				graph.add(new Vertex<Void>(node.getLabel()));
			}

			// creates the hierarchy connection
			for (Node node : flat())
			{
				Vertex<Void> from = graph.get(node.getLabel());

				for (Node child : node.getChildren())
				{
					Vertex<Void> to = graph.get(child.getLabel());
					graph.connect(from, to);
				}
			}
		}

		return graph;
	}

	@Override
	public ClassHierarchy clone()
	{
		try
		{
			ClassHierarchy clone = (ClassHierarchy) super.clone();
			clone.nodes = new HashMap<String, Node>();
			clone.root = null;

			if (!isEmpty())
			{
				for (Node node : nodes.values())
				{
					String label = node.getLabel();
					Node n = new Node(label);
					n.setWeight(node.getWeight());
					clone.nodes.put(label, n);
				}

				clone.root = clone.nodes.get(root.getLabel());

				for (Node node : nodes.values())
				{
					Node parent = clone.nodes.get(node.getLabel());

					for (Node child : node.getChildren())
					{
						clone.link(parent, clone.nodes.get(child.getLabel()));
					}
				}
			}

			return clone;
		}
		catch (CloneNotSupportedException e)
		{
			throw new InternalError(e.getMessage());
		}
	}

	@Override
	public String toString()
	{
		if (isEmpty())
		{
			return "Class Hierarchy: (empty)";
		}

		NumberFormat formatter = new DecimalFormat("0.000");
		Set<String> lines = new TreeSet<String>();

		for (Node node : nodes.values())
		{
			StringBuffer buffer = new StringBuffer();
			buffer.append(node.getLabel());
			buffer.append(" [");
			buffer.append(formatter.format(node.getWeight()));
			buffer.append("]");

			Set<String> parents = new TreeSet<String>();

			for (Node parent : node.getParents())
			{
				parents.add(parent.getLabel());
			}

			if (!parents.isEmpty())
			{
				buffer.append(" {");

				for (String parent : parents)
				{
					buffer.append(parent);
					buffer.append(',');
				}

				buffer.setCharAt(buffer.length() - 1, '}');
			}

			lines.add(buffer.toString());
		}

		StringBuffer buffer = new StringBuffer();
		buffer.append("Class Hierarchy (");
		buffer.append(isTree() ? "TREE" : "DAG");
		buffer.append("):");

		for (String line : lines)
		{
			buffer.append('\n');
			buffer.append(line);
		}

		return buffer.toString();
	}
}