/*
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2012 Fernando Esteban Barril Otero
 */

package net.sf.myra.datamining.data;

import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import net.sf.myra.datamining.data.IntervalBuilder.Interval;

/**
 * This class represents a continuous attribute.
 * 
 * @author Fernando Esteban Barril Otero
 */
public class ContinuousAttribute
	extends Attribute
{
	private static final long serialVersionUID = 4228467855153811092L;

	/**
	 * The initial cut point of this continuous attribute (i.e. the cut
	 * point of the entire dataset).
	 */
	private Interval initial;

	/**
	 * The array of values in the domain of the attribute. This array is
	 * empty unless the method {@link #domain(Dataset)} is called.
	 */
	private Double[] domain = new Double[0];

	/**
	 * Creates a new <code>ContinuousAttribute</code> instance.
	 * 
	 * @param name the attribute name.
	 */
	public ContinuousAttribute(String name)
	{
		super(name);
	}

	/**
	 * Returns the interval previously calculated.
	 * 
	 * @return the interval previously calculated.
	 */
	public Interval getInitial()
	{
		return initial;
	}

	/**
	 * @param initial the initial to set
	 */
	public void setInitial(Interval initial)
	{
		this.initial = initial;
	}

	public String toString()
	{
		return getName();
	}

	@Override
	public void validate(String value)
	{
		if (!Metadata.MISSING_VALUE.equals(value))
		{
			try
			{
				new Double(value);
			}
			catch (NumberFormatException e)
			{
				throw new IllegalArgumentException("Not a numeric value: "
					+ value, e);
			}
		}
	}

	@Override
	public boolean isMissing(double value)
	{
		return Double.isNaN(value);
	}

	/**
	 * Returns the (expected) entropy of this attribute. Note that this method
	 * will call {@link #setInitial(IntervalBuilder.Interval)} with the calculated interval.
	 * 
	 * @param instances the instances used to compute the entropy.
	 * 
	 * @return the (expected) entropy.
	 */
	public double entropy(List<Instance> instances)
	{
		if (instances.isEmpty())
		{
			throw new IllegalArgumentException("Empty instances list.");
		}

		Interval interval =
			IntervalBuilder.getInstance(
				instances.iterator().next().getMetadata()).createSingle(
				instances, this);
		
		// cache the calculated interval
		setInitial(interval);

		// the entropy of the continuous attribute is the same as the
		// lowest entropy in the intervals when the attribute is discretised
		return interval.getImpurity();
	}

	/**
	 * Returns the mean value of this attribute.
	 * 
	 * @param instances the instances used to compute the mean.
	 * 
	 * @return the mean value of this attribute.
	 */
	public double mean(List<Instance> instances)
	{
		if (instances.isEmpty())
		{
			throw new IllegalArgumentException("Empty instances list.");
		}

		int index =
			instances.iterator().next().getMetadata().getIndex(getName());
		double mean = 0.0;
		int count = 0;

		for (Instance instance : instances)
		{
			double value = instance.values()[index];

			if (!Double.isNaN(value))
			{
				mean += value;
				count++;
			}
		}

		return mean / count;
	}

	/**
	 * Returns the domain values of this attribute.
	 * 
	 * @return the domain values of this attribute.
	 */
	public Double[] domain()
	{
		return domain;
	}

	/**
	 * Populates the domain values of this attribute.
	 * 
	 * @param dataset the current dataset.
	 */
	public void domain(Dataset dataset)
	{
		Set<Double> values = new HashSet<Double>(dataset.getSize());
		int index = dataset.getMetadata().getIndex(getName());

		for (Instance instance : dataset)
		{
			double v = instance.values()[index];

			if (!isMissing(v))
			{
				values.add(v);
			}
		}

		domain = values.toArray(new Double[values.size()]);
		Arrays.sort(domain);
	}

	@Override
	public ContinuousAttribute clone()
	{
		ContinuousAttribute clone = (ContinuousAttribute) super.clone();

		if (initial != null)
		{
			clone.initial = initial.clone();
		}

		if (domain != null)
		{
			clone.domain = domain.clone();
		}

		return clone;
	}
}