/*
 * $Id: Dataset.java 2326 2011-01-25 10:54:47Z febo2 $
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining.data;

import static net.sf.myra.datamining.data.Instance.DEFAULT_WEIGHT;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

/**
 * This class represents a dataset.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision: 2326 $ $Date:: 2011-01-25 10:54:47#$
 */
public final class Dataset
	implements Iterable<Instance>, Cloneable, Serializable
{
	private static final long serialVersionUID = -6217067049678702427L;

	/**
	 * Default class separator for multi-label class value.
	 */
	public static final String CLASS_SEPARATOR = "@";

	/**
	 * The list of instances.
	 */
	private ArrayList<Instance> instances;

	/**
	 * The metadata information.
	 */
	private Metadata metadata;

	/**
	 * The dataset filename. This property is set when the dataset has been
	 * read from a file.
	 */
	private String filename;

	/**
	 * Creates a new <code>Dataset</code> instance.
	 * 
	 * @param metadata the metadata information.
	 */
	public Dataset(Metadata metadata)
	{
		this.metadata = metadata;
		instances = new ArrayList<Instance>();
	}

	/**
	 * Returns the dataset filename.
	 * 
	 * @return the dataset filename.
	 */
	public String getFilename()
	{
		return filename;
	}

	/**
	 * Sets the dataset filename.
	 * 
	 * @param filename the filename to set.
	 */
	public void setFilename(String filename)
	{
		this.filename = filename;
	}

	/**
	 * Adds a new instance to the dataset.
	 * 
	 * @param values the instance values.
	 */
	public void add(String[] values)
	{
		Instance instance = newInstance();

		try
		{
			instance.set(values);
		}
		catch (IllegalArgumentException e)
		{
			remove(instance);
			throw e;
		}
	}

	public Iterator<Instance> iterator()
	{
		return instances.iterator();
	}

	/**
	 * Checks if the dataset has duplicated instances (instances with the
	 * same values for every attribute).
	 * 
	 * @return <code>true</code> if the dataset contains duplicated instances;
	 *         <code>false</code> otherwise.
	 */
	public boolean hasDuplicates()
	{
		Set<Instance> s = new HashSet<Instance>(instances);
		return s.size() != instances.size();
	}

	/**
	 * Checks of the dataset contains instances with missing values.
	 * 
	 * @return <code>true</code> if the dataset contains instances with missing
	 *         values; <code>false</code> otherwise.
	 */
	public boolean hasMissing()
	{
		for (Attribute attribute : metadata.getPredictor())
		{
			int index = metadata.getIndex(attribute.getName());

			for (Instance instance : instances)
			{
				if (attribute.isMissing(instance.values()[index]))
				{
					return true;
				}
			}
		}

		return false;
	}

	/**
	 * Returns a new <code>Instance</code> associated with this dataset.
	 * 
	 * @return a new <code>Instance</code> associated with this dataset.
	 */
	public Instance newInstance()
	{
		Instance instance = new Instance(this, instances.size());
		instances.add(instance);

		return instance;
	}

	/**
	 * Returns the <code>Metadata</code> object associated with this
	 * dataset.
	 * 
	 * @return the <code>Metadata</code> object associated with this
	 *         dataset.
	 */
	public final Metadata getMetadata()
	{
		return metadata;
	}

	/**
	 * Returns the size (number of instances) of the dataset.
	 * 
	 * @return the size (number of instances) of the dataset.
	 */
	public int getSize()
	{
		return instances.size();
	}

	/**
	 * Returns the number of instances that belong to the specified class.
	 * 
	 * @param c the class value.
	 * 
	 * @return the number of instances that belong to the specified class found
	 *          in the dataset.
	 */
	public int getSize(String c)
	{
		int index = metadata.getTarget().getIndex(c);
		int size = 0;

		for (Instance instance : instances)
		{
			if (instance.labels()[index])
			{
				size++;
			}
		}

		return size;
	}

	/**
	 * Returns the list of instances.
	 * 
	 * @return  the list of instances.
	 */
	public List<Instance> getInstances()
	{
		return Collections.unmodifiableList(instances);
	}

	/**
	 * Returns the list of instances that belongs to the specified class. This
	 * method returns all instances that either have the specified class as
	 * its most specific class or as an ancestor class.
	 * 
	 * @param c the class value.
	 * 
	 * @return the list of instances that belongs to the specified class.
	 */
	public List<Instance> getInstances(String... c)
	{
		NominalAttribute target = metadata.getTarget();
		int[] indexes = new int[c.length];

		for (int i = 0; i < c.length; i++)
		{
			indexes[i] = target.getIndex(c[i]);
		}

		ArrayList<Instance> filtered = new ArrayList<Instance>();

		for (Instance instance : instances)
		{
			boolean add = true;

			for (int i = 0; i < indexes.length; i++)
			{
				if (!instance.labels()[indexes[i]])
				{
					add = false;
					break;
				}
			}

			if (add)
			{
				filtered.add(instance);
			}
		}

		return filtered;
	}

	/**
	 * Returns the list of instances that have the specified (attribute, value)
	 * combination.
	 * 
	 * @param attribute the attribute to check.
	 * @param value the value to check.
	 * 
	 * @return the list of instances that have the specified (attribute, value)
	 *         combination.
	 */
	public List<Instance> getInstances(Attribute attribute, String value)
	{
		if (attribute instanceof NominalAttribute)
		{
			ArrayList<Instance> filtered = new ArrayList<Instance>();
			NominalAttribute nominal = (NominalAttribute) attribute;

			int aIndex = metadata.getIndex(attribute.getName());
			int vIndex = nominal.getIndex(value);

			for (Instance instance : instances)
			{
				if (instance.values()[aIndex] == vIndex)
				{
					filtered.add(instance);
				}
			}

			return filtered;
		}
		// if it is a continuous attribute, returns the whole list of instances
		else if (attribute instanceof ContinuousAttribute)
		{
			return getInstances();
		}

		throw new IllegalArgumentException("Invalid attribute type: "
			+ attribute.getClass().getName());
	}
	
	/**
	 * Returns the list of instances that satisfy the specified term.
	 * 
	 * @param term the term to be used.
	 * 
	 * @return the list of instances that satisfy the specified term.
	 */
	public List<Instance> getInstances(Term term)
	{
		ArrayList<Instance> filtered = new ArrayList<Instance>();
		int index = metadata.getIndex(term.getAttribute());
		
		for (Instance instance : instances)
		{
			if (term.satisfies(instance.values()[index]))
			{
				filtered.add(instance);
			}
		}
		
		return filtered;
	}

	/**
	 * Returns the list of instances that have the specified label. Note that
	 * this is different from the {@link #getInstances(String...)} method,
	 * since it only returns the instances that have the exact label.
	 * 
	 * @param label the label of the instances.
	 * 
	 * @return the list of instances that have the specified label.
	 */
	public List<Instance> getInstances(Label label)
	{
		ArrayList<Instance> filtered = new ArrayList<Instance>();

		for (Instance instance : instances)
		{
			if (instance.getLabel().equals(label))
			{
				filtered.add(instance);
			}
		}

		return filtered;
	}

	/**
	 * Returns the list of instances that belongs to any of the specified class
	 * labels.
	 * 
	 * @param c the specified class labels. 
	 * 
	 * @return the list of instances that belongs to any of the specified class
	 *         labels.
	 */
	public List<Instance> filter(String... c)
	{
		NominalAttribute target = metadata.getTarget();
		int[] indexes = new int[c.length];

		for (int i = 0; i < c.length; i++)
		{
			indexes[i] = target.getIndex(c[i]);
		}

		ArrayList<Instance> filtered = new ArrayList<Instance>();

		for (Instance instance : instances)
		{
			for (int i = 0; i < indexes.length; i++)
			{
				if (instance.labels()[indexes[i]])
				{
					filtered.add(instance);
					break;
				}
			}
		}

		return filtered;
	}

	/**
	 * Returns <code>true</code> if the dataset contains no instances.
	 * 
	 * @return <code>true</code> if the dataset contains no instances;
	 *         <code>false</code> otherwise.
	 */
	public boolean isEmpty()
	{
		return instances.isEmpty();
	}

	/**
	 * Verifies if this is a multi-label dataset (i.e. contains at least one
	 * multi-label instance).
	 * 
	 * @return <code>true</code> if this is a multi-label dataset;
	 *         <code>false</code> otherwise.
	 */
	public boolean isMultilabel()
	{
		boolean multilabel = false;

		for (Instance instance : instances)
		{
			if (instance.isMultilabel())
			{
				multilabel = true;
				break;
			}
		}

		return multilabel;
	}

	/**
	 * Sets the dataset instances. Note that any instance contained in the
	 * dataset will be removed.
	 * 
	 * @param instances the new dataset instances.
	 */
	public void setInstances(List<Instance> instances)
	{
		this.instances.clear();

		for (Instance instance : instances)
		{
			Instance i = newInstance();

			for (Attribute attribute : metadata.getPredictor())
			{
				i.set(attribute, instance.getValue(attribute));
			}

			i.set(metadata.getTarget(), instance.getLabel().toString());
			i.setWeight(instance.getWeight());
		}
	}

	/**
	 * Checks if the dataset contains an instance with the (attribute,value)
	 * pair.
	 * 
	 * @param attribute the attribute to check.
	 * @param value the value to check.
	 * 
	 * @return <code>true</code> if the dataset contains an instance with the
	 *         specified (attribute,value) pair; <code>false</code> otherwise.
	 */
	public boolean contains(Attribute attribute, String value)
	{
		if (attribute instanceof NominalAttribute)
		{
			NominalAttribute nominal = (NominalAttribute) attribute;

			int aIndex = metadata.getIndex(attribute.getName());
			int vIndex = nominal.getIndex(value);

			for (Instance instance : instances)
			{
				if (instance.values()[aIndex] == vIndex)
				{
					return true;
				}
			}

			return false;
		}

		return true;
	}

	/**
	 * Removes the first occurrence in this dataset of the specified instance.
	 * 
	 * @param instance the instance to be removed, if present.
	 * 
	 * @return <code>true</code> if the dataset contained the specified
	 *         instance; <code>false</code> otherwise.
	 */
	public boolean remove(Instance instance)
	{
		if (instances.remove(instance))
		{
			refresh();
			return true;
		}
		return false;
	}

	/**
	 * Removes the instance with the specified index.
	 * 
	 * @param index the instance index.
	 * 
	 * @return <code>true</code> if the dataset contained the specified
	 *         instance; <code>false</code> otherwise.
	 */
	public boolean remove(int index)
	{
		if (instances.remove(index) != null)
		{
			refresh();
			return true;
		}
		return false;
	}

	/**
	 * Updates the instances' index values.
	 */
	private void refresh()
	{
		for (int i = 0; i < instances.size(); i++)
		{
			instances.get(i).setIndex(i);
		}
	}

	/**
	 * Removes the specified attribute from the dataset. This operation will
	 * remove the attribute from the associated <code>Metadata</code> instance. 
	 * 
	 * @param attribute the attribute to be removed.
	 */
	public void remove(Attribute attribute)
	{
		if (metadata.isTarget(attribute))
		{
			throw new IllegalArgumentException(
				"Target attribute cannot be deleted: " + attribute.getName());
		}

		int index = metadata.getIndex(attribute.getName());
		metadata.remove(attribute);

		for (Instance instance : instances)
		{
			instance.remove(index);
		}
	}

	/**
	 * Removes all this dataset's instances that are contained in the
	 * specified collection.
	 *     
	 * @param instances instances to be removed from this dataset.
	 * 
	 * @return <code>true</code> if this dataset changed as a result of the
	 *         call; <code>false</code> otherwise.
	 */
	public boolean remove(Collection<Instance> instances)
	{
		if (this.instances.removeAll(instances))
		{
			refresh();
			return true;
		}
		return false;
	}

	/**
	 * Returns the class count vector.
	 * 
	 * @return the class count vector.
	 */
	public int[] count()
	{
		int length = metadata.getLabel().length();
		int[] count = new int[length];

		for (Instance instance : instances)
		{
			boolean[] labels = instance.labels();

			for (int i = 0; i < length; i++)
			{
				if (labels[i])
				{
					count[i] += 1;
				}
			}
		}

		return count;
	}

	/**
	 * Returns the class frequency vector.
	 * 
	 * @return the class frequency vector.
	 */
	public double[] frequency()
	{
		int length = metadata.getLabel().length();
		double[] frequency = new double[length];

		for (Instance instance : instances)
		{
			boolean[] labels = instance.labels();

			for (int i = 0; i < length; i++)
			{
				if (labels[i])
				{
					frequency[i] += 1;
				}
			}
		}

		if (!isEmpty())
		{
			int size = instances.size();

			for (int i = 0; i < length; i++)
			{
				frequency[i] /= size;
			}
		}

		return frequency;
	}

	/**
	 * Returns the instance specified by the index.
	 * 
	 * @param index the instance's index.
	 * 
	 * @return the instance specified by the index.
	 */
	public Instance get(int index)
	{
		return instances.get(index);
	}

	/**
	 * Returns a new dataset instance without duplicated instances.
	 * 
	 * @return a new dataset instance without duplicated instances.
	 */
	public Dataset withoutDuplicates()
	{
		Metadata m = new Metadata();

		for (Attribute attribute : metadata.getAttributes())
		{
			if (attribute instanceof ContinuousAttribute)
			{
				m.add(new ContinuousAttribute(attribute.getName()));
			}
			else if (attribute instanceof NominalAttribute)
			{
				NominalAttribute n = (NominalAttribute) attribute;
				m.add(new NominalAttribute(n.getName(), n.getValues()));
			}
		}

		m.setTarget((NominalAttribute) m.get(metadata.getTarget().getName()));
		m.setName(metadata.getName());
		m.setClassHierarchy(metadata.getClassHierarchy());

		Dataset ds = new Dataset(m);
		Set<Instance> instanceSet = new LinkedHashSet<Instance>(instances);

		for (Instance instance : instanceSet)
		{
			Instance i = ds.newInstance();

			for (Attribute attribute : m.getAttributes())
			{
				i.set(attribute, instance.getValue(attribute));
			}
		}

		return ds;
	}

	/**
	 * Computes the continuous attributes' domain values.
	 */
	public void computeDomains()
	{
		for (Attribute attribute : metadata.getAttributes())
		{
			if (attribute instanceof ContinuousAttribute)
			{
				((ContinuousAttribute) attribute).domain(this);
			}
		}
	}

	/**
	 * Resets the instances' weights.
	 */
	public void reset()
	{
		for (Instance instance : instances)
		{
			instance.setWeight(DEFAULT_WEIGHT);
		}
	}

	@Override
	public String toString()
	{
		return instances.toString();
	}

	@Override
	public Dataset clone()
	{
		try
		{
			Dataset clone = (Dataset) super.clone();
			clone.metadata = metadata.clone();
			clone.instances = new ArrayList<Instance>(getSize());
			Label prototype = clone.getMetadata().getLabel();

			for (Instance instance : getInstances())
			{
				Instance i = clone.newInstance();
				i.setValues(instance.values().clone());
				i.setLabels(instance.labels().clone());
				i.setLabel(prototype.toLabel(instance.getLabel().flags()));
			}

			return clone;
		}
		catch (CloneNotSupportedException e)
		{
			throw new UnknownError(getClass() + ": clone()");
		}
	}
}