/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2009  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining.data;

import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadFactory;

import net.sf.myra.datamining.data.AbstractDistanceBasedBuilder.Candidate;
import net.sf.myra.datamining.data.IntervalBuilder.Interval;
import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.ParallelScheduler;

/**
 * This class provides the distance information content for a
 * <code>Dataset</code> instance.
 * 
 * @see Dataset
 * @see EntropyContent
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class DistanceContent
{
	/**
	 * The variance values mapping per term.
	 */
	private Map<String, Double> variances;

	/**
	 * The maximum (normalisation) value.
	 */
	private double maximum;

	/**
	 * The instance builder helper instance.
	 */
	private DistanceIntervalBuilder builder;

	/**
	 * Default constructor.
	 * 
	 * @param dataset the dataset reference.
	 */
	public DistanceContent(Dataset dataset)
	{
		Metadata metadata = dataset.getMetadata();
		this.builder = new DistanceIntervalBuilder(metadata);

		Collection<Attribute> predictor = metadata.getPredictor();
		this.variances = new HashMap<String, Double>(predictor.size(), 1.0f);

		double min = Double.MAX_VALUE;
		double max = Double.MIN_VALUE;

		for (Attribute attribute : predictor)
		{
			double variance = Double.NEGATIVE_INFINITY;

			if (attribute instanceof NominalAttribute)
			{
				NominalAttribute nominal = (NominalAttribute) attribute;

				for (String value : nominal.getValues())
				{
					variances.put(nominal.toString(value), variance);
				}
			}
			else
			{
				variances.put(attribute.getName(), variance);
			}
		}

		// concurrent code to calculate the heuristic-related distance measure

		int available =
			Configuration.getIntProperty(ParallelScheduler.WORKERS, Runtime
				.getRuntime().availableProcessors());

		ExecutorService executor =
			Executors.newFixedThreadPool(available, new ThreadFactory()
			{
				private int id = 0;

				public Thread newThread(Runnable r)
				{
					return new Thread(r, "DistanceContent::worker" + (id++));
				}
			});
		CountDownLatch latch = new CountDownLatch(predictor.size());

		for (Attribute attribute : predictor)
		{
			executor.execute(new Task(dataset, attribute, latch));
		}

		try
		{
			// waits for the calculation of the variance
			latch.await();
		}
		catch (InterruptedException e)
		{
			throw new RuntimeException(e);
		}

		executor.shutdown();

		for (double variance : variances.values())
		{
			// updates the maximum and minimum values
			if (!Double.isInfinite(variance) && (variance != 0.0))
			{
				if (min > variance)
				{
					min = variance;
				}

				if (max < variance)
				{
					max = variance;
				}
			}
		}

		// the maximum value is the minumim + maximum
		this.maximum = min + max;
	}

	/**
	 * Returns the variance associated with the specified term.
	 * 
	 * @param label the label of the term.
	 * 
	 * @return the variance associated with the specified term.
	 */
	public double variance(String label)
	{
		if (!variances.containsKey(label))
		{
			throw new IllegalArgumentException("Term not found: " + label);
		}

		return variances.get(label);
	}

	/**
	 * Returns the maximum (worst case) variance value.
	 *  
	 * @return the maximum (worst case) variance value.
	 */
	public double maximum()
	{
		return maximum;
	}

	/**
	 * Returns <code>true</code> if the specified label corresponds to a term.
	 * 
	 * @param label the label to check.
	 * 
	 * @return <code>true</code> if the specified label corresponds to a term;
	 *         <code>false</code> otherwise.
	 */
	public boolean contains(String label)
	{
		return variances.containsKey(label);
	}

	//--------------------------------------------------------- utility methods

	/**
	 * Returns the variance of the list of instances' class labels.
	 * 
	 * @param instances the list of instances.
	 * 
	 * @return  the variance of the list of instances' class labels.
	 */
	public static double variance(List<Instance> instances)
	{
		if (instances.isEmpty())
		{
			throw new IllegalArgumentException("Empty instance list.");
		}

		Candidate[] candidates = new Candidate[instances.size()];
		Metadata metadata = null;
		int index = 0;

		for (Instance instance : instances)
		{
			candidates[index] = new Candidate();
			candidates[index].flags = instance.getLabel().flags();
			index++;

			if (metadata == null)
			{
				metadata = instance.getMetadata();
			}
		}

		return (new DistanceIntervalBuilder(metadata)).variance(candidates, 0,
			candidates.length);
	}

	/**
	 * Returns the variance of the list of instances and the mean class
	 * label array.
	 * 
	 * @param instances the list of instances.
	 * @param mean the mean class label array.
	 * 
	 * @return the variance of the list of instances and the mean class
	 * label array.
	 */
	public static double variance(List<Instance> instances, double[] mean)
	{
		if (instances.isEmpty())
		{
			throw new IllegalArgumentException("Empty instance list.");
		}

		Candidate[] candidates = new Candidate[instances.size()];
		Metadata metadata = null;
		int index = 0;

		for (Instance instance : instances)
		{
			candidates[index] = new Candidate();
			candidates[index].flags = instance.getLabel().flags();
			index++;

			if (metadata == null)
			{
				metadata = instance.getMetadata();
			}
		}

		return (new DistanceIntervalBuilder(metadata)).variance(candidates, 0,
			candidates.length, mean);
	}

	/**
	 * This class represents the task of calculating the variance of each
	 * predictor attribute.
	 */
	private class Task
		implements Runnable
	{
		private Dataset dataset;

		private Attribute attribute;

		private CountDownLatch latch;

		Task(Dataset dataset, Attribute attribute, CountDownLatch latch)
		{
			this.dataset = dataset;
			this.attribute = attribute;
			this.latch = latch;
		}

		public void run()
		{
			if (attribute instanceof NominalAttribute)
			{
				NominalAttribute nominal = (NominalAttribute) attribute;

				for (String value : nominal.getValues())
				{
					List<Instance> instances =
						dataset.getInstances(attribute, value);

					if (!instances.isEmpty())
					{
						Candidate[] candidates =
							new Candidate[instances.size()];
						int index = 0;

						for (Instance instance : instances)
						{
							candidates[index] = new Candidate();
							candidates[index].flags =
								instance.getLabel().flags();
							index++;
						}

						double variance =
							builder.variance(candidates, 0, candidates.length);
						variances.put(nominal.toString(value), variance);
					}
				}
			}
			else if (attribute instanceof ContinuousAttribute)
			{
				ContinuousAttribute c = (ContinuousAttribute) attribute;
				Interval interval =
					builder.createSingle(dataset.getInstances(), c);
				c.setInitial(interval);

				double variance = interval.getImpurity();
				variances.put(c.getName(), variance);
			}

			latch.countDown();
		}
	}
}