/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining.data;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;

/**
 * This class provides the entropy information content for a
 * <code>Dataset</code> instance.
 * 
 * @see Dataset
 * @see DistanceContent
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class EntropyContent
{
	private Dataset dataset;

	private HashMap<String, Term> terms;

	public EntropyContent(Dataset dataset)
	{
		this.dataset = dataset;
		terms = new HashMap<String, Term>();

		frequency();
		entropy();
	}

	/**
	 * Calculates the frequency of each <em>A<sub>i</sub>=V<sub>ij</sub></em>
	 * term.
	 */
	private void frequency()
	{
		Metadata metadata = dataset.getMetadata();
		Label classes = metadata.getLabel();

		for (Attribute attribute : metadata.getPredictor())
		{
			if (attribute instanceof NominalAttribute)
			{
				NominalAttribute nominal = (NominalAttribute) attribute;
				int aIndex = metadata.getIndex(attribute.getName());

				for (String value : nominal.getValues())
				{
					Term term =
						new Term(attribute.getName(), value, classes.length());
					int vIndex = nominal.getIndex(value);

					for (Instance instance : dataset)
					{
						if (instance.values()[aIndex] == vIndex)
						{
							boolean[] flags = instance.getLabel().flags();

							for (int i = 0; i < flags.length; i++)
							{
								if (flags[i])
								{
									term.increment(i);
								}
							}
						}
					}

					terms.put(nominal.toString(value), term);
				}
			}
		}
	}

	/**
	 * Calculates the entropy of each <em>A<sub>i</sub>=V<sub>ij</sub></em>
	 * term.
	 */
	private void entropy()
	{
		Metadata metadata = dataset.getMetadata();
		Label classes = metadata.getLabel();

		for (Term term : terms.values())
		{
			if (term.getFrequency() > 0)
			{
				double entropy = 0.0;

				for (int i = 0; i < classes.length(); i++)
				{
					if (term.getFrequency(i) > 0)
					{
						double p =
							term.getFrequency(i) / (double) term.getFrequency();

						entropy -= (p * (Math.log(p) / Math.log(2.0)));
					}
				}

				term.setEntropy(entropy);
			}
		}

		// adds terms for the continuous attributes

		for (Attribute attribute : metadata.getPredictor())
		{
			if (attribute instanceof ContinuousAttribute)
			{
				ContinuousAttribute c = (ContinuousAttribute) attribute;
				Term term =
					new Term(attribute.getName(), null, classes.length());

				// to void 0 frequencies
				for (int i = 0; i < classes.length(); i++)
				{
					term.increment(i);
				}

				term.setEntropy(c.entropy(dataset.getInstances()));
				terms.put(c.toString(), term);
			}
		}
	}

	/**
	 * Returns the entropy associated with the term specified by the label.
	 * 
	 * @param label the label that represents the term.
	 * 
	 * @return the entropy associated with the term specified by the label.
	 */
	public double entropy(String label)
	{
		Term term = terms.get(label);

		if (term == null)
		{
			throw new IllegalArgumentException("Term not found: " + label);
		}

		return term.getEntropy();
	}

	/**
	 * Returns the frequency associated with the term specified by the label.
	 * 
	 * @param label the label that represents the term.
	 * 
	 * @return the frequency associated with the term specified by the label.
	 */
	public int frequency(String label)
	{
		Term term = terms.get(label);

		if (term == null)
		{
			throw new IllegalArgumentException("Term not found: " + label);
		}

		return term.getFrequency();
	}

	/**
	 * Returns the collection of terms of the dataset.
	 * 
	 * @return the collection of terms of the dataset.
	 */
	public Collection<Term> getTerms()
	{
		return Collections.unmodifiableCollection(terms.values());
	}

	/**
	 * Returns <code>true</code> if the specified label corresponds to a term.
	 * 
	 * @param label the label to check.
	 * 
	 * @return <code>true</code> if the specified label corresponds to a term;
	 *         <code>false</code> otherwise.
	 */
	public boolean contains(String label)
	{
		return terms.containsKey(label);
	}

	/**
	 * This class represents a <em>A<sub>i</sub>=V<sub>ij</sub></em> term.
	 */
	private static class Term
	{
		/**
		 * The term name.
		 */
		private String name;

		/**
		 * The term value.
		 */
		private String value;

		/**
		 * The frequency of this term for each class attribute value.
		 */
		private int[] frequencies;

		/**
		 * The entropy value.
		 */
		private double entropy;

		/**
		 * Creates a new <code>Term</code> instance.
		 *
		 * @param name the term name.
		 * @param value the term value.
		 */
		Term(String name, String value, int length)
		{
			this.name = name;
			this.value = value;
			frequencies = new int[length];
		}

		/**
		 * Returns the term name.
		 * 
		 * @return the term name.
		 */
		@SuppressWarnings("unused")
		public String getName()
		{
			return name;
		}

		/**
		 * Returns the term value.
		 * 
		 * @return the term value.
		 */
		@SuppressWarnings("unused")
		public String getValue()
		{
			return value;
		}

		/**
		 * Returns the entropy value.
		 * 
		 * @return the entropy value.
		 */
		public double getEntropy()
		{
			return entropy;
		}

		/**
		 * Sets the entropy value.
		 * 
		 * @param entropy the entropy value to set.
		 */
		void setEntropy(double entropy)
		{
			this.entropy = entropy;
		}

		/**
		 * Increments the frequency of this term for the specified class
		 * attribute value.
		 * 
		 * @param value the class attribute value.
		 */
		void increment(int index)
		{
			frequencies[index]++;
		}

		/**
		 * Returns the frequency of this term for the specified class
		 * attribute value.
		 * 
		 * @param value the class attribute value.
		 * 
		 * @return the frequency of this term for the specified class
		 *         attribute value.
		 */
		public int getFrequency(int index)
		{
			return frequencies[index];
		}

		/**
		 * Returns the total frequency of this term.
		 * 
		 * @return the total frequency of this term.
		 */
		public int getFrequency()
		{
			int total = 0;

			for (int i = 0; i < frequencies.length; i++)
			{
				total += frequencies[i];
			}

			return total;
		}

		@Override
		public int hashCode()
		{
			return toString().hashCode();
		}

		@Override
		public boolean equals(Object o)
		{
			return toString().equals(((Term) o).toString());
		}

		@Override
		public String toString()
		{
			return "(" + name + "," + value + ")";
		}
	}
}