/*
 * $Id: EntropyIntervalBuilder.java 2360 2011-04-14 15:23:54Z febo2 $
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining.data;

import java.util.Arrays;

/**
 * Entropy interval builder.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision: 2360 $ $Date:: 2011-04-14 16:23:54#$
 */
public class EntropyIntervalBuilder
	extends AbstractEntropyBasedBuilder
{
	/**
	 * Default constructor.
	 * 
	 * @param metadata the dataset metadata.
	 */
	public EntropyIntervalBuilder(Metadata metadata)
	{
		super(metadata);
	}

	@Override
	protected CutPoint[] create(double[][] matrix, int classes, boolean filter)
	{
		// the minimum number of instances of each partition (similar to
		// J48 algorithm)
		double minimum =
			0.1 * ((double) matrix.length)
				/ ((double) metadata.getTarget().getValues().size());

		if (minimum <= minimumLimit)
		{
			minimum = minimumLimit;
		}
		else if (minimum > 25.0)
		{
			minimum = 25.0;
		}

		CutPoint point = new CutPoint();
		point.gain = Double.NEGATIVE_INFINITY;
		//point.impurity_s = entropy(matrix, 0, matrix.length, classes);
		double size_s = weightedLength(matrix, 0, matrix.length - 1);
		point.impurity_s = entropy(matrix, 0, matrix.length, classes, size_s);

		if (matrix.length > 1 && (size_s >= (minimum * 2)))
		{
			boolean force = false;

			// selects the best cut point

			for (int i = 0; i < (matrix.length - 1); i++)
			{
				if (matrix[i][1] != matrix[i + 1][1] || force)
				{
					if (weightedLength(matrix, 0, i) >= minimum)
					{
						force = false;
						int t = i + 1;

						// if the previous value is equal to the threshold value
						// we skip forward
						while (t < (matrix.length - 1)
							&& matrix[t - 1][0] == matrix[t][0])
						{
							t++;
						}

						// if the upper partition is smaller than the limit, we
						// move backwards
						while ((weightedLength(matrix, t, matrix.length - 1) < minimum)
							|| (t > 0 && matrix[t][0] == matrix[t - 1][0]))
						{
							t--;
						}

						if (weightedLength(matrix, 0, t - 1) < minimum)
						{
							force = true;
						}
						else if (t < (matrix.length - 1))
						{
							int k1 = diversity(matrix, 0, t, classes);
							int k2 =
								diversity(matrix, t, matrix.length, classes);
							/*
							double entropy_s1 = entropy(matrix, 0, t, classes);
							double entropy_s2 =
								entropy(matrix, t, matrix.length, classes);

							double gain =
								point.impurity_s
									- ((t / (double) matrix.length) * entropy_s1)
									- (((matrix.length - t) / (double) matrix.length) * entropy_s2);
							*/
							double size_s1 = weightedLength(matrix, 0, t - 1);
							double size_s2 =
								weightedLength(matrix, t, matrix.length - 1);

							double entropy_s1 =
								entropy(matrix, 0, t, classes, size_s1);
							double entropy_s2 =
								entropy(matrix, t, matrix.length, classes,
									size_s2);

							double gain =
								point.impurity_s
									- ((size_s1 / size_s) * entropy_s1)
									- ((size_s2 / size_s) * entropy_s2);

							if (gain > point.gain && (matrix.length - t) > 1)
							{
								point.gain = gain;
								point.threshold = t;
								point.k1 = k1;
								point.k2 = k2;
								point.impurity_s1 = entropy_s1;
								point.impurity_s2 = entropy_s2;
								point.value = matrix[t][0];
							}
						}
					}
					else
					{
						force = true;
					}
				}
			}
		}

		return new CutPoint[] { point };
	}

	/**
	 * Returns the entropy of the specified range of examples using the
	 * weighted size of the instances.
	 * 
	 * @param matrix the (values,class) distribution.
	 * @param index the start index.
	 * @param length the number of position to evaluate.
	 * @param k the total number of class values.
	 * @param size the weighted size of the instances.
	 * 
	 * @return the entropy of the specified range of examples.
	 */
	private double entropy(double[][] matrix, int index, int length, int k,
			double size)
	{
		int[] frequency = new int[k];
		Arrays.fill(frequency, 0);

		for (int i = index; i < length; i++)
		{
			frequency[(int) matrix[i][1]] += matrix[i][2];
		}

		double entropy = 0.0;

		for (int i = 0; i < frequency.length; i++)
		{
			if (frequency[i] > 0)
			{
				double p = frequency[i] / size;
				entropy -= (p * (Math.log(p) / Math.log(2.0)));
			}
		}

		return entropy;
	}
}