/*
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2012 Fernando Esteban Barril Otero
 */

package net.sf.myra.datamining.data;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Experimental interval builder based on the Laplace function. <b>Note:</b>
 * only works if the class is pre-selected.
 * 
 * @author Fernando Esteban Barril Otero
 */
public class FunctionBasedBuilder
	extends IntervalBuilder
{
	public FunctionBasedBuilder(Metadata metadata)
	{
		super(metadata);
	}

	@Override
	public int count(List<Instance> instances, ContinuousAttribute attribute)
	{
		throw new UnsupportedOperationException();
	}

	@Override
	public Interval[] create(List<Instance> instances,
		ContinuousAttribute attribute)
	{
		throw new UnsupportedOperationException();
	}

	@Override
	public Interval createSingle(List<Instance> instances,
		ContinuousAttribute attribute)
	{
		throw new UnsupportedOperationException();
	}

	@Override
	public Interval createSingle(List<Instance> instances,
		ContinuousAttribute attribute, String label)
	{
		Map<Double, Integer> mapping = new HashMap<Double, Integer>();
		double[][] matrix = prepare(instances, attribute, mapping, label);
		CutPoint point = create(matrix);

		if (!Double.isNaN(point.gain))
		{
			double value = point.value;

			// calculates the average value for the threshold (only if it is
			// not the first value of the array)
			if (point.threshold > 0)
			{
				value = average(matrix[point.threshold - 1][0], value);
			}

			return new Interval(Math.min(point.impurity_s1, point.impurity_s2),
				(point.impurity_s1 > point.impurity_s2) ? Operator.GREATER_THAN
					: Operator.LESS_THAN_EQUAL_TO, value);
		}

		return new Interval(Double.POSITIVE_INFINITY, Operator.LESS_THAN,
			Double.NaN);
	}

	protected CutPoint create(double[][] matrix)
	{
		// the minimum number of instances of each partition (similar to J48)
		double minimum =
			0.1 * ((double) matrix.length)
				/ ((double) metadata.getTarget().getValues().size());

		if (minimum <= minimumLimit)
		{
			minimum = minimumLimit;
		}
		else if (minimum > 25.0)
		{
			minimum = 25.0;
		}

		CutPoint point = new CutPoint();
		point.gain = Double.NEGATIVE_INFINITY;

		if (weightedLength(matrix, 0, matrix.length - 1) >= (minimum * 2))
		{
			boolean force = false;

			// selects the best cut point

			for (int i = 0; i < (matrix.length - 1); i++)
			{
				if (matrix[i][1] != matrix[i + 1][1] || force)
				{
					if (weightedLength(matrix, 0, i) >= minimum)
					{
						force = false;
						int t = i + 1;

						// if the previous value is equal to the threshold
						// value we skip forward
						while (t < (matrix.length - 1)
							&& matrix[t - 1][0] == matrix[t][0])
						{
							t++;
						}

						// if the upper partition is smaller than the limit,
						// we move backwards
						while (weightedLength(matrix, t, matrix.length - 1) < minimum
							|| (t > 0 && matrix[t][0] == matrix[t - 1][0]))
						{
							t--;
						}

						if (weightedLength(matrix, 0, t - 1) < minimum)
						{
							force = true;
						}
						else if (t < (matrix.length - 1))
						{
							double lower = cost(0, t - 1, matrix);
							double upper = cost(t, matrix.length - 1, matrix);

							if (lower > point.gain || upper > point.gain)
							{
								point.gain = Math.max(upper, lower);
								point.threshold = t;
								point.impurity_s1 = 1 - lower;
								point.impurity_s2 = 1 - upper;
								point.value = matrix[t][0];
							}
						}
					}
					else
					{
						force = true;
					}
				}
			}
		}

		return point;
	}

	private double cost(int from, int to, double[][] values)
	{
		double count = 0;

		for (int i = from; i <= to; i++)
		{
			if (values[i][1] == 1)
			{
				count++;
			}
		}

		int k = metadata.getTarget().getValues().size();

		return (count + 1) / ((to - from + 1) + k);
	}

	private double[][] prepare(List<Instance> instances,
		ContinuousAttribute attribute, Map<Double, Integer> mapping,
		String label)
	{
		if (instances.isEmpty())
		{
			throw new IllegalArgumentException("Empty instance list");
		}

		Metadata metadata = instances.iterator().next().getMetadata();
		int klass = metadata.getIndex(metadata.getTarget().getName());
		int attributeIndex = metadata.getIndex(attribute.getName());

		double[][] matrix = new double[instances.size()][3];
		int index = 0;

		for (Instance instance : instances)
		{
			double[] indexes = instance.values();

			if (!Double.isNaN(indexes[attributeIndex]))
			{
				Integer value = mapping.get(indexes[klass]);

				if (value == null)
				{
					value = mapping.size();
					mapping.put(indexes[klass], value);
				}

				matrix[index][0] = indexes[attributeIndex];
				matrix[index][1] = (instance.belongs(label)) ? 1.0 : 0.0;
				matrix[index][2] = instance.getWeight();

				index++;
			}
			else
			{
				double[][] m = new double[matrix.length - 1][3];
				System.arraycopy(matrix, 0, m, 0, index);
				matrix = m;
			}
		}

		// sort the matrix based on the attribute values
		sort(matrix);

		return matrix;
	}

	/**
	 * Sorts (based on a bubble sort algorithm) the matrix of attribute
	 * value -> class value.
	 * 
	 * @param matrix the matrix to be sorted.
	 */
	private void sort(double[][] matrix)
	{
		for (int i = 0; i < matrix.length; i++)
		{
			double[] min = matrix[i];
			int index = i;

			for (int j = i; j < matrix.length; j++)
			{
				if (matrix[j][0] < min[0])
				{
					min = matrix[j];
					index = j;
				}
			}

			matrix[index] = matrix[i];
			matrix[i] = min;
		}
	}

	private double weightedLength(double[][] matrix, int from, int to)
	{
		double length = 0.0;

		for (int i = from; i <= to; i++)
		{
			length += matrix[i][2];
		}

		return length;
	}

	private double average(double value1, double value2)
	{
		double average = (value1 + value2) / 2.0;

		if (average == value2)
		{
			return value1;
		}

		return average;
	}
}