/*
 * $Id: Instance.java 2298 2010-08-04 15:28:20Z febo2 $
 */

package net.sf.myra.datamining.data;

import java.io.Serializable;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
import java.util.StringTokenizer;

/**
 * This class represents a instance of the dataset.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision: 2298 $ $Date:: 2010-08-04 16:28:20#$
 * 
 * @see Dataset
 */
public final class Instance
	implements Serializable
{
	private static final long serialVersionUID = -6293021735863810845L;
	
	/**
	 * The default weight of an instance.
	 */
	public static final double DEFAULT_WEIGHT = 1.0;

	/**
	 * The enclosing dataset instance.
	 */
	private final Dataset dataset;

	/**
	 * The attributes' values indexes.
	 */
	private double[] values;

	/**
	 * The attributes' class values flags.
	 */
	private boolean[] labels;

	/**
	 * The instance label.
	 */
	private Label label;

	/**
	 * The dataset instance index.
	 */
	private int index;

	/**
	 * The instance weight (default to 1.0).
	 */
	private double weight = DEFAULT_WEIGHT;

	/**
	 * Creates a new <code>Instance</code> instance.
	 * 
	 * @param index the dataset instance index.
	 * @param dataset TODO
	 */
	Instance(Dataset dataset, int index)
	{
		this.dataset = dataset;
		this.index = index;

		values = new double[dataset.getMetadata().getAttributes().size()];
		Arrays.fill(values, -1);

		try
		{
			labels =
				new boolean[dataset.getMetadata().getTarget().getValues()
					.size()];
			Arrays.fill(labels, false);
		}
		// in case the target (class) attribute was not set, we still
		// allow the creation of the dataset
		catch (IllegalStateException e)
		{
			labels = new boolean[0];
		}
	}

	/**
	 * Sets the value of the specified attribute.
	 * 
	 * @param attribute the attribute to set the value.
	 * @param value the value to set.
	 */
	public void set(Attribute attribute, String value)
	{
		if (dataset.getMetadata().isTarget(attribute))
		{
			StringTokenizer t =
				new StringTokenizer(value, Dataset.CLASS_SEPARATOR);

			Set<String> labelSet = new HashSet<String>();
			ClassHierarchy hierarchy =
				dataset.getMetadata().getClassHierarchy();

			while (t.hasMoreTokens())
			{
				String v = t.nextToken();
				attribute.validate(v);

				// are we dealing with an hierarchical problem?
				if (hierarchy != null)
				{
					Node node = hierarchy.find(v);
					labelSet.add(node.getLabel());

					for (Node ancestor : node.getAncestors())
					{
						labelSet.add(ancestor.getLabel());
					}
				}
				else
				{
					labelSet.add(v);
				}
			}

			label = dataset.getMetadata().getLabel().toLabel(labelSet);
			this.labels = label.flags();

			// this will give a different value for each combination of
			// labels (useful for computing entropy values)
			int index = dataset.getMetadata().getIndex(attribute.getName());
			values[index] = label.hashCode();
		}
		else
		{
			attribute.validate(value);
			String name = attribute.getName();
			double v = 0.0;

			if (attribute instanceof NominalAttribute)
			{
				v = ((NominalAttribute) attribute).getIndex(value);
			}
			else if (attribute instanceof ContinuousAttribute)
			{
				if (Metadata.MISSING_VALUE.equals(value))
				{
					v = Double.NaN;
				}
				else
				{
					v = new Double(value);
				}
			}

			values[dataset.getMetadata().getIndex(name)] = v;
		}
	}

	/**
	 * Sets the value of the specified attribute.
	 * 
	 * @param attribute the attribute name to set the value.
	 * @param value the value to set.
	 */
	public void set(String attribute, String value)
	{
		set(dataset.getMetadata().get(attribute), value);
	}

	/**
	 * Sets all attributes values.
	 * 
	 * @param values the array containing all attribute values.
	 */
	public void set(String[] values)
	{
		if (values.length != (dataset.getMetadata().getAttributes().size()))
		{
			throw new IllegalArgumentException(
				"Unexpected number of attributes: expected "
					+ dataset.getMetadata().getAttributes().size() + ", found "
					+ values.length);
		}

		for (int index = 0; index < values.length; index++)
		{
			set(dataset.getMetadata().get(index), values[index]);
		}
	}

	/**
	 * Returns the instance index.
	 * 
	 * @return the instance index.
	 */
	public int getIndex()
	{
		return index;
	}

	/**
	 * Returns the instance value for the specified attribute.
	 * 
	 * @param attribute the attribute object.
	 * 
	 * @return the instance value for the specified attribute.
	 */
	public String getValue(Attribute attribute)
	{
		int index = dataset.getMetadata().getIndex(attribute.getName());

		if (dataset.getMetadata().isTarget(attribute))
		{
			return label.toString();
		}
		else if (attribute instanceof NominalAttribute)
		{
			return ((NominalAttribute) attribute).getValue((int) values[index]);
		}

		return Double.isNaN(values[index]) ? Metadata.MISSING_VALUE : Double
			.toString(values[index]);
	}

	/**
	 * Returns the instance value for the specified attribute.
	 * 
	 * @param name the attribute name.
	 * 
	 * @return the instance value for the specified attribute.
	 */
	public String getValue(String name)
	{
		return getValue(dataset.getMetadata().get(name));
	}

	/**
	 * Returns the instance values.
	 * 
	 * @return the instance values.
	 */
	public String[] getValues()
	{
		String[] v = new String[values.length];
		int i = 0;

		for (Attribute attribute : getMetadata().getPredictor())
		{
			v[i] = getValue(attribute);
			i++;
		}

		v[i] = getLabel().toString();

		return v;
	}

	/**
	 * Return the (indexed) value of the specified attribute.
	 * 
	 * @param index the attribute index.
	 * 
	 * @return the (indexed) value of the specified attribute.
	 */
	public double value(int index)
	{
		return values[index];
	}

	/**
	 * Returns the array of indexed values.
	 * 
	 * @return the array of indexed values.
	 */
	public final double[] values()
	{
		return values;
	}

	/**
	 * Returns the array of indexed class values (labels).
	 * 
	 * @return the array of indexed class values (labels).
	 */
	public final boolean[] labels()
	{
		return labels;
	}

	/**
	 * Returns the class label.
	 * 
	 * @return the class label.
	 */
	public Label getLabel()
	{
		return label;
	}

	/**
	 * Checks if the instance has the specified class label.
	 * 
	 * @param label the label to check.
	 * 
	 * @return <code>true</code> if the instance has the specified label;
	 *         <code>false</code> otherwise.
	 */
	public boolean belongs(String label)
	{
		NominalAttribute target = dataset.getMetadata().getTarget();
		return labels[target.getIndex(label)];
	}

	/**
	 * Returns the <code>Metadata</code> object associated with this
	 * instance.
	 * 
	 * @return the <code>Metadata</code> object associated with this
	 *         instance.
	 */
	public Metadata getMetadata()
	{
		return dataset.getMetadata();
	}

	/**
	 * Verifies if this is a multi-label instance.
	 * 
	 * @return <code>true</code> if this is a multi-label instance;
	 *         <code>false</code> otherwise.
	 */
	public boolean isMultilabel()
	{
		ClassHierarchy hierarchy = dataset.getMetadata().getClassHierarchy();
		Collection<String> leaves = label.getValues();

		if (hierarchy != null)
		{
			leaves = hierarchy.leaves(leaves);
		}

		return (leaves.size() > 1);
	}
	
	/**
	 * Verifies if this instance is associated with a leaf class label.
	 * 
	 * @return <code>true</code> if this instance is associated with a leaf
	 *         class label; <code>false</code> otherwise.
	 */
	public boolean isLeaf()
	{
		ClassHierarchy hierarchy = dataset.getMetadata().getClassHierarchy();
		Collection<String> leaves = label.getValues();
		boolean isLeaf = true;

		if (hierarchy != null)
		{
			for (String leaf : hierarchy.leaves(leaves))
			{
				if (!hierarchy.find(leaf).isLeaf())
				{
					isLeaf = false;
				}
			}
		}

		return isLeaf;
	}

	/**
	 * Removes the value from the specified attribute.
	 * 
	 * @param index the index of the attribute.
	 */
	void remove(int index)
	{
		double[] v = new double[values.length - 1];
		System.arraycopy(values, 0, v, 0, index);

		if (v.length > index)
		{
			System.arraycopy(values, index + 1, v, index, v.length - index);
		}

		values = v;
	}

	@Override
	public boolean equals(Object o)
	{
		return equals((Instance) o);
	}

	/**
	 * Compares the specified instance with this instance for equality.
	 * Returns <code>true</code> if the two instances have the same
	 * <code>Metadata</code> object and if the given instance have the
	 * same values.
	 *  
	 * @param o the instance to be compared for equality with this instance.
	 * 
	 * @return <code>true</code> if the specified instance is equal to this
	 *         instance; <code>false</code> otherwise.
	 */
	public boolean equals(Instance o)
	{
		return Arrays.equals(values, o.values);
	}

	/**
	 * Sets the instance index.
	 * 
	 * @param index the index value to set.
	 */
	void setIndex(int index)
	{
		this.index = index;
	}

	@Override
	public int hashCode()
	{
		return Arrays.hashCode(values);
	}

	@Override
	public String toString()
	{
		StringBuffer buffer = new StringBuffer();

		for (Attribute attribute : getMetadata().getPredictor())
		{
			buffer.append(getValue(attribute));
			buffer.append(",");
		}

		buffer.append(getLabel());

		return buffer.toString();
	}

	/**
	 * @param values the values to set
	 */
	void setValues(double[] values)
	{
		this.values = values;
	}

	/**
	 * @param labels the labels to set
	 */
	void setLabels(boolean[] labels)
	{
		this.labels = labels;
	}

	/**
	 * @param label the label to set
	 */
	void setLabel(Label label)
	{
		this.label = label;
	}

	/**
	 * @return the weight
	 */
	public double getWeight()
	{
		return weight;
	}

	/**
	 * @param weight the weight to set
	 */
	public void setWeight(double weight)
	{
		this.weight = weight;
	}
}