/*
 * $Id: IntervalBuilder.java 2304 2010-08-06 22:58:15Z febo2 $
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining.data;

import java.util.List;

import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.ObjectFactory;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision: 2304 $ $Date:: 2010-08-06 23:58:15#$
 */
public abstract class IntervalBuilder
{
	/**
	 * The property key under which the interval builder class name is stored.
	 */
	public static final String BUILDER =
		"net.sf.myra.datamining.interval.builder";
	
	/**
	 * The property key under which the minimum number of exemples per 
	 * interval is stored.
	 */
	public static final String MINIMUM =
		"net.sf.myra.datamining.interval.minimum";

	/**
	 * Default interval builder (entropy).
	 */
	public static final String DEFAULT_BUILDER =
		EntropyIntervalBuilder.class.getName();

	/**
	 * The dataset metadata.
	 */
	protected Metadata metadata;
	
	/**
	 * The minimum number of instance per interval.
	 */
	protected final double minimumLimit;

	/**
	 * Default constructor.
	 * 
	 * @param metadata the dataset metadata.
	 */
	public IntervalBuilder(Metadata metadata)
	{
		this.metadata = metadata;
		minimumLimit = Configuration.getDoubleProperty(MINIMUM, 2.0);
	}

	/**
	 * Returns the default interval builder instance.
	 * 
	 * @return the default interval builder instance.
	 */
	public static IntervalBuilder getInstance(Metadata metadata)
	{
		return (IntervalBuilder) ObjectFactory.create(Configuration
			.getProperty(BUILDER, DEFAULT_BUILDER),
			new Class<?>[] { Metadata.class }, new Object[] { metadata });
	}

	/**
	 * Returns an interval for the specified continuous attribute tailored
	 * for the specified instances.
	 * 
	 * @param instances the list of instances.
	 * @param attribute the continuous attribute.
	 * 
	 * @return an interval for the specified continuous attribute tailored
	 *         for the specified instance.
	 */
	public abstract Interval createSingle(List<Instance> instances,
			ContinuousAttribute attribute);

	/**
	 * Returns a discrete interval for the specified continuous attribute
	 * tailored for the specified instances and class value. The instances
	 * are arranged into a binary distribution (i.e. instances that belong to
	 * the specified class and instances that do not belong to the specified
	 * class). Note that this operation may not be supporter by all interval
	 * builders.
	 * 
	 * @param instances the list of instances.
	 * @param attribute the continuous attribute.
	 * @param label the class label.
	 * 
	 * @return a discrete interval for the specified continuous attribute
	 *         tailored for the specified instances and class value.
	 */
	public abstract Interval createSingle(List<Instance> instances,
			ContinuousAttribute attribute, String label);

	/**
	 * Returns the discrete intervals for the specified continuous attribute.
	 * 
	 * @param instances the list of instances.
	 * @param attribute the continuous attribute.
	 * 
	 * @return the discrete intervals for the specified continuous attribute.
	 */
	public abstract Interval[] create(List<Instance> instances,
			ContinuousAttribute attribute);

	/**
	 * Returns the number of candidate threshold values.
	 * 
	 * @param instances the list of instances.
	 * @param attribute the continuous attribute.
	 * 
	 * @return the number of candidate threshold values.
	 */
	public abstract int count(List<Instance> instances,
			ContinuousAttribute attribute);

	/**
	 * Struct-like class that represents a cut point.
	 */
	protected static class CutPoint
		implements Comparable<CutPoint>
	{
		/**
		 * The threshold value index.
		 */
		protected int threshold = -1;

		/**
		 * The information gain of the threshold.
		 */
		protected double gain;

		/**
		 * Entropy value of the lower partition.
		 */
		protected double impurity_s1;

		/**
		 * Entropy value of the upper partition.
		 */
		protected double impurity_s2;

		/**
		 * Entropy value of the distribution.
		 */
		protected double impurity_s;

		/**
		 * Number of different class values of the lower partition.
		 */
		protected int k1;

		/**
		 * Number of different class values of the upper partition.
		 */
		protected int k2;

		/**
		 * The threshold value.
		 */
		protected double value = Double.NaN;

		public int compareTo(CutPoint o)
		{
			return Double.compare(value, o.value);
		}

		@Override
		public String toString()
		{
			return Double.toString(value);
		}
	}

	/**
	 * This class represents a continuous interval.
	 */
	public static class Interval
		implements Cloneable
	{
		/**
		 * The interval values (ascendent order).
		 */
		private double[] values;

		/**
		 * The interval entropy.
		 */
		private double impurity;

		/**
		 * The operator associated with the interval.
		 */
		private Operator operator;

		/**
		 * Default constructor.
		 * 
		 * @param entropy the interval entropy.
		 * @param values the interval boudary values.
		 */
		Interval(double impurity, Operator operator, double... values)
		{
			this.impurity = impurity;
			this.operator = operator;
			this.values = values;
		}

		/**
		 * Returns the interval entropy value.
		 * 
		 * @return the interval entropy value.
		 */
		public double getImpurity()
		{
			return impurity;
		}

		/**
		 * Returns the interval boundary values.
		 * 
		 * @return the interval boundary values.
		 */
		public double[] getValues()
		{
			return values;
		}

		/**
		 * Returns the interval operator.
		 * 
		 * @return the interval operator.
		 */
		public Operator getOperator()
		{
			return operator;
		}

		@Override
		public Interval clone()
		{
			try
			{
				Interval clone = (Interval) super.clone();
				clone.values = values.clone();
				return clone;
			}
			catch (CloneNotSupportedException e)
			{
				throw new UnknownError(getClass() + ": clone()");
			}
		}
	}
}