/*
 * $Id$ 
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining.data;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;

/**
 * This class represents a label for an instance. The label is represented
 * as an array of all possible labels. Note that for hierarchical and
 * multi-label problems, where more than one label can be associated with
 * an instance, the array can have more than one position active.
 *  
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class Label
	implements Cloneable, Serializable
{
	private static final long serialVersionUID = 5250690155426861649L;

	/**
	 * Name of each (potential) label.
	 */
	private String[] names;

	/**
	 * Active flag for each (potential) label.
	 */
	private boolean[] flags;

	/**
	 * Creates a new <code>Label</code> instance.
	 * 
	 * @param target the target attribute.
	 */
	Label(NominalAttribute target)
	{
		Collection<String> values = target.getValues();

		names = new String[values.size()];
		flags = new boolean[values.size()];

		for (int i = 0; i < names.length; i++)
		{
			// sanity check: the labels order should not change
			names[i] = target.getValue(i);
			flags[i] = false;
		}
	}

	/**
	 * Returns the label length.
	 * 
	 * @return the label length.
	 */
	public int length()
	{
		return flags.length;
	}

	/**
	 * Returns a new <code>Label</code> instance representing the specified
	 * array of labels.
	 * 
	 * @param values the array of labels.
	 * 
	 * @return a new <code>Label</code> instance representing the specified
	 * array of labels.
	 */
	public Label toLabel(String... values)
	{
		return toLabel(Arrays.asList(values));
	}

	/**
	 * Returns a new <code>Label</code> instance representing the specified
	 * list of labels.
	 * 
	 * @param labels the list of labels.
	 * 
	 * @return a new <code>Label</code> instance representing the specified
	 * list of labels.
	 */
	public Label toLabel(Collection<String> labels)
	{
		boolean[] flags = new boolean[this.flags.length];

		for (int i = 0; i < flags.length; i++)
		{
			flags[i] = labels.contains(names[i]);
		}

		return toLabel(flags);
	}

	/**
	 * Returns a new <code>Label</code> instance representing the specified
	 * array of flags.
	 *  
	 * @param flags the array of flags.
	 * 
	 * @return a new <code>Label</code> instance representing the specified
	 * array of flags.
	 */
	public Label toLabel(boolean[] flags)
	{
		if (this.flags.length == flags.length)
		{
			try
			{
				Label clone = (Label) super.clone();
				clone.flags = flags.clone();
				return clone;
			}
			catch (CloneNotSupportedException e)
			{
				throw new UnknownError(getClass() + ": clone()");
			}
		}

		throw new IllegalArgumentException("Invalid array size: <"
			+ this.flags.length + "> expected, <" + flags.length + "> found");
	}

	@Override
	public boolean equals(Object o)
	{
		return equals((Label) o);
	}

	/**
	 * Indicates whether the specified <code>Label</code> instance is equal to
	 * this one.
	 * 
	 * @param o the <code>Label</code> instance with which to compare.
	 * 
	 * @return <code>true</code> if this instance is the same as the
	 *         <code>o</code> argument; <code>false</code> otherwise.
	 */
	public boolean equals(Label o)
	{
		return Arrays.equals(flags, o.flags);
	}

	@Override
	public String toString()
	{
		StringBuffer buffer = new StringBuffer();

		for (int i = 0; i < names.length; i++)
		{
			if (flags[i])
			{
				if (buffer.length() > 0)
				{
					buffer.append(Dataset.CLASS_SEPARATOR);
				}

				buffer.append(names[i]);
			}
		}

		return buffer.toString();
	}

	/**
	 * Returns the collection of class label values that are set for this
	 * instance.
	 * 
	 * @return the collection of class label values that are set for this
	 * instance.
	 */
	public Collection<String> getValues()
	{
		ArrayList<String> values = new ArrayList<String>(flags.length);

		for (int i = 0; i < names.length; i++)
		{
			if (flags[i])
			{
				values.add(names[i]);
			}
		}

		return values;
	}

	/**
	 * Returns <code>true</code> is the specified class label index is active.
	 * 
	 * @param index the class label index.
	 * 
	 * @return <code>true</code> if class label is active; <code>false</code>
	 *         otherwise.
	 */
	public boolean active(int index)
	{
		return flags[index];
	}
	
	/**
	 * Returns <code>true</code> is the specified class label is active.
	 * 
	 * @param value the class label.
	 * 
	 * @return <code>true</code> if class label is active; <code>false</code>
	 *         otherwise.
	 */
	public boolean active(String value)
	{
		return flags[getIndex(value)];
	}

	/**
	 * Returns the class label name specified by the index.
	 * 
	 * @param index the class label index.
	 * 
	 * @return the class label name specified by the index.
	 */
	public String get(int index)
	{
		return names[index];
	}

	/**
	 * Returns the index of the specified class value.
	 * 
	 * @param value the desired class value.
	 * 
	 * @return the index of the specified class value.
	 */
	public int getIndex(String value)
	{
		for (int i = 0; i < names.length; i++)
		{
			if (names[i].equals(value))
			{
				return i;
			}
		}

		throw new IllegalArgumentException("Value not found: " + value);
	}

	/**
	 * Returns the number of class labels that are present in both
	 * <code>Label</code> instances.
	 * 
	 * @param another the label to check against.
	 * 
	 * @return the number of class labels that are present in both
	 * <code>Label</code> instances.
	 */
	public int intersect(Label another)
	{
		if (flags.length != another.flags.length)
		{
			throw new IllegalArgumentException("Invalid label size: " + "<"
				+ flags + "> expected, <" + another.flags.length + "> found");
		}

		int count = 0;

		for (int i = 0; i < flags.length; i++)
		{
			if (flags[i] && another.flags[i])
			{
				count++;
			}
		}

		return count;
	}

	/**
	 * Returns the number of class labels present in this <code>Label</code>
	 * instance.
	 * 
	 * @return the number of class labels present in this <code>Label</code>
	 * instance.
	 */
	public int cardinality()
	{
		int count = 0;

		for (int i = 0; i < flags.length; i++)
		{
			if (flags[i])
			{
				count++;
			}
		}

		return count;
	}

	/**
	 * Returns <code>true</code> if this label contains all values of the
	 * specified label.
	 * 
	 * @param label the label to get the values from.
	 * 
	 * @return <code>true</code> if this label contains all values of the
	 *         specified label; <code>false</code> otherwise.
	 */
	public boolean contains(Label label)
	{
		boolean found = true;

		for (int i = 0; i < flags.length; i++)
		{
			if (label.flags[i] && !flags[i])
			{
				found = false;
			}
		}

		return found;
	}

	@Override
	public int hashCode()
	{
		return Arrays.hashCode(flags);
	}

	/**
	 * Returns the array representation of this label.
	 * 
	 * @return the array representation of this label.
	 */
	public boolean[] flags()
	{
		// the array is not protected from modifications
		return flags;
	}

	/**
	 * Returns the names array of this label.
	 * 
	 * @return the names array of this label.
	 */
	public String[] names()
	{
		// the array is not protected from modifications
		return names;
	}
}