/*
 * $Id$ 
 */

package net.sf.myra.datamining.data;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                    #$
 */
public class MDLDistanceIntervalBuilder
	extends AbstractDistanceBasedBuilder
{
	private DistanceIntervalBuilder builder;

	/**
	 * Default constructor.
	 * 
	 * @param metadata the dataset metadata.
	 */
	public MDLDistanceIntervalBuilder(Metadata metadata)
	{
		super(metadata);
		builder = new DistanceIntervalBuilder(metadata);
	}

	/**
	 * Returns a list of cut points using the MDLPC criterion.
	 * 
	 * @param candidates the candidates points.
	 * 
	 * @return a list of cut points using the MDLPC criterion.
	 */
	@Override
	protected CutPoint[] create(Candidate[] candidates, boolean filter)
	{
		List<CutPoint> points = mdlpc(candidates, 0, false);
		Collections.sort(points);

		if (filter)
		{
			points = filter(candidates, points);
		}

		return points.toArray(new CutPoint[points.size()]);
	}

	/**
	 * Returns a list of cut points using the (recursive) MDLPC criterion.
	 * 
	 * @param candidates the candidates distribution.
	 * @param delta the offset from the original array
	 * @param nested indicates if this is a nested call
	 * 
	 * @return a list of cut points using the (recursive) MDLPC criterion.
	 */
	private List<CutPoint> mdlpc(Candidate[] candidates, int delta,
			boolean nested)
	{
		ArrayList<CutPoint> intervals = new ArrayList<CutPoint>();

		// select the best cut point based on the entropy

		CutPoint point = builder.create(candidates, true)[0];

		// decides if the threshold should be accepted

		int k = count(candidates, 0, candidates.length);

		// we use an aproximation of log2(3^k - 2) since k can be very large
		double criterion =
			((Math.log(candidates.length - 1) / Math.log(2)) / candidates.length)
				+ (((k * (Math.log(3) / Math.log(2))) - ((k * point.impurity_s)
					- (point.k1 * point.impurity_s1) - (point.k2 * point.impurity_s2))) / candidates.length);

		if (!Double.isInfinite(point.gain))
		{
			// if is the first (not nested) call or if the cut point
			// satisfies the MDL critetion
			if (!nested || point.gain > criterion)
			{
				intervals.add(point);

				// recursively tries to partition the intervals

				Candidate[] s1 = new Candidate[point.threshold];
				Candidate[] s2 =
					new Candidate[candidates.length - point.threshold];

				for (int i = 0; i < candidates.length; i++)
				{
					if (i < point.threshold)
					{
						s1[i] = candidates[i];
					}
					else
					{
						s2[i - point.threshold] = candidates[i];
					}
				}

				// corrects the threshold value
				point.threshold += delta;

				intervals.addAll(mdlpc(s1, delta, true));
				intervals.addAll(mdlpc(s2, point.threshold, true));
			}
		}
		else if (!nested)
		{
			intervals.add(point);
		}

		return intervals;
	}

	/**
	 * Returns the best cut point(s).
	 * 
	 * @param matrix the (values,class) distribution.
	 * @param points the list of cut points to filter.
	 * @param classes the number of classes.
	 * 
	 * @return the best cut point(s).
	 */
	private List<CutPoint> filter(Candidate[] candidates, List<CutPoint> points)
	{
		double[] variance = new double[points.size() + 1];
		int[] size = new int[points.size() + 1];
		int threshold = 0;
		int i = 0;

		// calculates the entropy for each interval

		for (CutPoint point : points)
		{
			variance[i] = variance(candidates, threshold, point.threshold);
			size[i] = point.threshold - threshold;
			threshold = point.threshold;
			i++;
		}

		// calculates the entropy for the last interval

		variance[i] = variance(candidates, threshold, candidates.length);
		size[i] = candidates.length - threshold;

		// find the best interval index

		double lowest = Double.MAX_VALUE;
		int bigger = Integer.MIN_VALUE;
		int index = -1;

		for (i = 0; i < variance.length; i++)
		{
			if (variance[i] < lowest
				|| ((variance[i] == lowest) && (size[i] > bigger)))
			{
				lowest = variance[i];
				bigger = size[i];
				index = i;
			}
		}

		ArrayList<CutPoint> filter = new ArrayList<CutPoint>(points.size());

		if (index == 0)
		{
			CutPoint p = points.get(index);
			// makes sure the upper partition has a higher value
			p.impurity_s2 = Double.MAX_VALUE;
			filter.add(p);
		}
		else if (index == (variance.length - 1))
		{
			CutPoint p = points.get(index - 1);
			// makes sure the lower partition has a higher value
			p.impurity_s1 = Double.MAX_VALUE;
			filter.add(p);
		}
		else
		{
			CutPoint p1 = points.get(index - 1);
			p1.impurity_s2 = variance[index];
			filter.add(p1);

			CutPoint p2 = points.get(index);
			p2.impurity_s1 = variance[index];
			filter.add(p2);
		}

		return filter;
	}
}