/*
 * $Id: MDLEntropyIntervalBuilder.java 2070 2009-03-16 16:09:23Z febo $
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining.data;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * Minimum Description Length Principle (MDLP) interval builder. In essence,
 * this class uses the entropy builder in combination of a MDL stopping
 * criteria to create (potentially) multiple intervals.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision: 2070 $ $Date:: 2009-03-16 16:09:23#$
 */
public class MDLEntropyIntervalBuilder
	extends AbstractEntropyBasedBuilder
{
	private EntropyIntervalBuilder builder;

	/**
	 * Default constructor.
	 * 
	 * @param metadata the dataset metadata.
	 */
	public MDLEntropyIntervalBuilder(Metadata metadata)
	{
		super(metadata);
		builder = new EntropyIntervalBuilder(metadata);
	}

	/**
	 * Returns a list of cut points using the MDLPC criterion.
	 * 
	 * @param matrix the (values,class) distribution.
	 * @param classes the total number of class values.
	 * 
	 * @return a list of cut points using the MDLPC criterion.
	 */
	protected CutPoint[] create(double[][] matrix, int classes, boolean filter)
	{
		List<CutPoint> points = mdlpc(matrix, classes, 0, false);
		Collections.sort(points);

		if (!points.isEmpty() && filter)
		{
			points = filter(matrix, points, classes);
		}

		return points.toArray(new CutPoint[0]);
	}

	/**
	 * Returns the best cut point(s).
	 * 
	 * @param matrix the (values,class) distribution.
	 * @param points the list of cut points to filter.
	 * @param classes the number of classes.
	 * 
	 * @return the best cut point(s).
	 */
	private List<CutPoint> filter(double[][] matrix, List<CutPoint> points,
			int classes)
	{
		double[] entropy = new double[points.size() + 1];
		int[] size = new int[points.size() + 1];
		int k = diversity(matrix, 0, matrix.length, classes);
		int threshold = 0;
		int i = 0;

		// calculates the entropy for each interval

		for (CutPoint point : points)
		{
			entropy[i] = entropy(matrix, threshold, point.threshold, k);
			size[i] = point.threshold - threshold;
			threshold = point.threshold;
			i++;
		}

		// calculates the entropy for the last interval

		entropy[i] = entropy(matrix, threshold, matrix.length, k);
		size[i] = matrix.length - threshold;

		// find the best interval index

		double lowest = Double.MAX_VALUE;
		int bigger = Integer.MIN_VALUE;
		int index = -1;

		for (i = 0; i < entropy.length; i++)
		{
			if (entropy[i] < lowest
				|| ((entropy[i] == lowest) && (size[i] > bigger)))
			{
				lowest = entropy[i];
				bigger = size[i];
				index = i;
			}
		}

		ArrayList<CutPoint> filter = new ArrayList<CutPoint>(points.size());

		if (index == 0)
		{
			CutPoint p = points.get(index);
			// makes sure the upper partition has a higher value
			p.impurity_s2 = Double.MAX_VALUE;
			filter.add(p);
		}
		else if (index == (entropy.length - 1))
		{
			CutPoint p = points.get(index - 1);
			// makes sure the lower partition has a higher value
			p.impurity_s1 = Double.MAX_VALUE;
			filter.add(p);
		}
		else
		{
			CutPoint p1 = points.get(index - 1);
			p1.impurity_s2 = entropy[index];
			filter.add(p1);

			CutPoint p2 = points.get(index);
			p2.impurity_s1 = entropy[index];
			filter.add(p2);
		}

		return filter;
	}

	/**
	 * Returns a list of cut points using the (recursive) MDLPC criterion.
	 * 
	 * @param matrix the (values,class) distribution.
	 * @param classes the total number of class values.
	 * @param delta the offset from the original array
	 * @param nested indicates if this is a nested call
	 * 
	 * @return a list of cut points using the (recursive) MDLPC criterion.
	 */
	private List<CutPoint> mdlpc(double[][] matrix, int classes, int delta,
			boolean nested)
	{
		ArrayList<CutPoint> intervals = new ArrayList<CutPoint>();

		// select the best cut point based on the entropy

		CutPoint point = builder.create(matrix, classes)[0];

		// decides if the threshold should be accepted

		int k = diversity(matrix, 0, matrix.length, classes);

		double criterion =
			((Math.log(matrix.length - 1) / Math.log(2)) / matrix.length)
				+ (((Math.log(Math.pow(3, k) - 2) / Math.log(2)) - ((k * point.impurity_s)
					- (point.k1 * point.impurity_s1) - (point.k2 * point.impurity_s2))) / matrix.length);

		// if is the first (not nested) call or if the cut point
		// satisfies the MDL critetion
		if (!Double.isNaN(point.value) && (!nested || point.gain > criterion))
		{
			intervals.add(point);

			// recursively tries to partition the intervals

			double[][] s1 = new double[point.threshold][3];
			double[][] s2 = new double[matrix.length - point.threshold][3];

			for (int i = 0; i < matrix.length; i++)
			{
				if (i < point.threshold)
				{
					s1[i][0] = matrix[i][0];
					s1[i][1] = matrix[i][1];
					s1[i][2] = matrix[i][2];
				}
				else
				{
					s2[i - point.threshold][0] = matrix[i][0];
					s2[i - point.threshold][1] = matrix[i][1];
					s2[i - point.threshold][2] = matrix[i][2];
				}
			}

			// corrects the threshold value
			point.threshold += delta;

			intervals.addAll(mdlpc(s1, classes, delta, true));
			intervals.addAll(mdlpc(s2, classes, point.threshold, true));
		}

		return intervals;
	}
}