/*
 * $Id: Metadata.java 2326 2011-01-25 10:54:47Z febo2 $
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining.data;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.sf.myra.framework.Graph;
import net.sf.myra.framework.Vertex;

/**
 * This class provides attribute metadata information.
 *  
 * @author Fernando Esteban Barril Otero
 * @version $Revision: 2326 $ $Date:: 2011-01-25 10:54:47#$
 */
public class Metadata
	implements Cloneable, Serializable
{
	private static final long serialVersionUID = 7058834992412922585L;

	/**
	 * Constant value for missing attribute values.
	 */
	public static final String MISSING_VALUE = "?";

	/**
	 * The list of attributes.
	 */
	private List<Attribute> attributes;

	/**
	 * The (attribute name,index) mapping.
	 */
	private Map<String, Integer> mapping;

	/**
	 * The class attribute reference.
	 */
	private NominalAttribute target;

	/**
	 * The prototype label instance.
	 */
	private Label label;

	/**
	 * The relation name.
	 */
	private String name;

	/**
	 * The class hierarchy (when the class values are organized in an
	 * hierarchy).
	 */
	private ClassHierarchy hierarchy;

	/**
	 * Default constructor.
	 */
	public Metadata()
	{
		attributes = new ArrayList<Attribute>();
		mapping = new HashMap<String, Integer>();
	}

	/**
	 * Adds an attribute.
	 * 
	 * @param attribute the attribute to add.
	 */
	public void add(Attribute attribute)
	{
		if (mapping.containsKey(attribute.getName()))
		{
			throw new IllegalArgumentException("Duplicated attribute: "
				+ attribute.getName());
		}

		mapping.put(attribute.getName(), attributes.size());
		attributes.add(attribute);
	}

	/**
	 * Returns the relation name.
	 * 
	 * @return the relation name.
	 */
	public String getName()
	{
		return name;
	}

	/**
	 * Sets the relation name.
	 * 
	 * @param name the name to set.
	 */
	public void setName(String name)
	{
		this.name = name;
	}

	/**
	 * Sets the class attribute.
	 * 
	 * @param attribute the class attribute.
	 */
	public void setTarget(NominalAttribute attribute)
	{
		if (target != null)
		{
			target.setTarget(false);
		}

		target = (NominalAttribute) get(attribute.getName());
		target.setTarget(true);
		label = new Label(target);
	}

	/**
	 * Sets the class attribute.
	 * 
	 * @param index the class attribute index.
	 */
	public void setTarget(int index)
	{
		setTarget((NominalAttribute) get(index));
	}

	/**
	 * Returns the class attribute.
	 * 
	 * @return the class attribute.
	 */
	public NominalAttribute getTarget()
	{
		if (target == null)
		{
			throw new IllegalStateException("Class attribute not set.");
		}

		return target;
	}

	/**
	 * Returns the label prototype instance.
	 * 
	 * @return the label prototype instance.
	 */
	public Label getLabel()
	{
		return label;
	}

	/**
	 * Returns <code>true</code> if the specified attribute is the class
	 * attribute.
	 * 
	 * @param attribute thee attribute to test.
	 * 
	 * @return <code>true</code> if the specified attribute is the class
	 *         attribute; <code>false</code> otherwise.
	 */
	public boolean isTarget(Attribute attribute)
	{
		return (target != null && getTarget().equals(attribute));
	}

	/**
	 * Returns <code>true</code> if this metadata instance contains an
	 * attribute with the specified name.
	 * 
	 * @param name the attribute's name.
	 * 
	 * @return <code>true</code> if there is an attribute with the specified
	 *         name; <code>false</code> otherwise.
	 */
	public boolean contains(String name)
	{
		return mapping.containsKey(name);
	}

	/**
	 * Returns the <code>Attribute</code> instance given its name.
	 * 
	 * @param attributeName the attribute name.
	 * 
	 * @return the <code>Attribute</code> instance given its name.
	 */
	public Attribute get(String attributeName)
	{
		if (mapping.containsKey(attributeName))
		{
			return attributes.get(mapping.get(attributeName));
		}

		throw new IllegalArgumentException("Attribute not found: "
			+ attributeName);
	}

	public Attribute get(int index)
	{
		return attributes.get(index);
	}

	/**
	 * Returns the attribute index.
	 * 
	 * @param attribute the attribute name.
	 * 
	 * @return the attribute index.
	 */
	public int getIndex(String attribute)
	{
		Integer index = mapping.get(attribute);

		if (index != null)
		{
			return index;
		}

		throw new IllegalArgumentException("Attribute not found: " + attribute);
	}

	/**
	 * Returns the collection attributes.
	 * 
	 * @return the collection attributes.
	 */
	public Collection<Attribute> getAttributes()
	{
		return Collections.unmodifiableCollection(attributes);
	}

	/**
	 * Returns the collection of predictor attributes.
	 * 
	 * @return the collection of predictor attributes.
	 */
	public Collection<Attribute> getPredictor()
	{
		Collection<Attribute> predictor = new ArrayList<Attribute>();

		for (Attribute attribute : attributes)
		{
			// we ensure that this test can be performed (see setTarget()) 
			if (target != attribute)
			{
				predictor.add(attribute);
			}
		}

		return predictor;
	}

	/**
	 * @return the hierarchy
	 */
	public ClassHierarchy getClassHierarchy()
	{
		return hierarchy;
	}

	/**
	 * @param hierarchy the hierarchy to set
	 */
	public void setClassHierarchy(ClassHierarchy hierarchy)
	{
		this.hierarchy = hierarchy;
	}

	@Override
	public boolean equals(Object o)
	{
		return equals((Metadata) o);
	}

	/**
	 * Compares the specified metadata object with this object for equality.
	 * Returns <code>true</code> if the two instances have the same
	 * attribute list and if the given instance have the class attribute
	 * value.
	 *  
	 * @param o the metadata object to be compared for equality with this
	 *        object.
	 * 
	 * @return <code>true</code> if the specified metadata object is equal to
	 *         this object; <code>false</code> otherwise.
	 */
	public boolean equals(Metadata o)
	{
		if (attributes.equals(o.attributes))
		{
			return (target == null && o.target == null)
				|| target.equals(o.target);
		}

		return false;
	}

	/**
	 * Returns the <code>Graph</code> representation of this metadata instance.
	 * The original Ant-Miner version only supports nominal attributes, so an
	 * <code>UnsupportedOperationException</code> is thrown if a
	 * <code>ContinuousAttribute</code> instance is found. For every nominal
	 * attribute value, a node is added to the graph in the form <em>attribute
	 * = value<sub>i</sub></em>.
	 * 
	 * Note that the class attribute is not included in the graph.
	 * 
	 * @return the <code>Graph</code> representation of this metadata instance.
	 * 
	 * @see Metadata
	 * @see Graph
	 */
	public Graph<Term> toGraph()
	{
		int size = 0;

		for (Attribute attribute : getPredictor())
		{
			if (attribute instanceof ContinuousAttribute)
			{
				size += 1;
			}
			else if (attribute instanceof NominalAttribute)
			{
				NominalAttribute n = (NominalAttribute) attribute;
				size += n.getValues().size();
			}
		}

		Graph<Term> graph = new Graph<Term>(size);

		for (Attribute attribute : getPredictor())
		{
			String name = attribute.getName();
			int index = getIndex(attribute.getName());

			if (attribute instanceof ContinuousAttribute)
			{
				Term term = new Term(name, index);

				Vertex<Term> v = new Vertex<Term>(name);
				v.setInfo(term);

				graph.add(v);
			}
			else if (attribute instanceof NominalAttribute)
			{
				NominalAttribute n = (NominalAttribute) attribute;

				for (String value : n.getValues())
				{
					Term term = new Term(name, index);
					term.setValues(n.getIndex(value));
					term.setDescription(value);
					term.setOperator(Operator.EQUAL_TO);

					Vertex<Term> v = new Vertex<Term>(n.toString(value));
					v.setInfo(term);

					graph.add(v);
				}
			}
		}

		graph.connect();

		return graph;
	}

	@Override
	public Metadata clone()
	{
		try
		{
			Metadata clone = (Metadata) super.clone();
			clone.attributes = new ArrayList<Attribute>();

			for (Attribute attribute : attributes)
			{
				clone.attributes.add(attribute.clone());
			}

			clone.mapping = new HashMap<String, Integer>(mapping);
			clone.setTarget(target);

			if (hierarchy != null)
			{
				clone.hierarchy = hierarchy.clone();
			}

			return clone;
		}
		catch (CloneNotSupportedException e)
		{
			throw new UnknownError(getClass() + ": clone()");
		}
	}

	/**
	 * Expands the specified class label to include all ancestor class labels
	 * according to the class hierarchy.
	 *  
	 * @param label the class label to be expanded.
	 * 
	 * @return the expanded class label instance.
	 */
	public Label expand(Label label)
	{
		if (!isHierarchical())
		{
			return label;
		}

		Set<String> pSet = new HashSet<String>(this.label.length());

		for (String value : label.getValues())
		{
			Node node = hierarchy.find(value);
			pSet.add(node.getLabel());

			for (Node ancestor : node.getAncestors())
			{
				pSet.add(ancestor.getLabel());
			}
		}

		return this.label.toLabel(pSet);
	}

	/**
	 * Tests if this metadata has a class hierarchy associated.
	 * 
	 * @return <code>true</code> if there is a class hierarchy associated;
	 *         <code>false</code> otherwise.
	 */
	public boolean isHierarchical()
	{
		return hierarchy != null;
	}
	
	/**
	 * Returns <code>true</code> if the specified value represents a
	 * missing value.
	 * 
	 * @param index the attribute index.
	 * @param value the value to check.
	 * 
	 * @return <code>true</code> if the specified value represents a
	 *         missing value; otherwise <code>false</code>.
	 */
	public boolean isMissing(int index, double value)
	{
		return get(index).isMissing(value);
	}

	/**
	 * Removes the specified attribute.
	 * 
	 * @param attribute the attribute to be removed.
	 */
	void remove(Attribute attribute)
	{
		attributes.remove(attribute);

		// updates the attributes indexes
		mapping.clear();
		int index = 0;

		for (Attribute a : attributes)
		{
			mapping.put(a.getName(), index);
			index++;
		}
	}
}