
package net.sf.myra.datamining.data;

import java.io.Serializable;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.Set;

/**
 * This class represents a node of the hierarchy.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 * 
 * @see ClassHierarchy
 */
public class Node
	implements Serializable
{
	private static final long serialVersionUID = 5890410285554275885L;

	/**
	 * References to the parent nodes.
	 */
	private Collection<Node> parents;

	/**
	 * References of the children nodes.
	 */
	private Collection<Node> children;

	/**
	 * The node label.
	 */
	private String label;

	/**
	 * The hierarchical node weight.
	 */
	private double weight;

	/**
	 * Default constructor.
	 * 
	 * @param label the label of the node.
	 */
	public Node(String label)
	{
		this.label = label;
		parents = new LinkedList<Node>();
		children = new LinkedList<Node>();
	}

	public boolean isLeaf()
	{
		return children.isEmpty();
	}

	public boolean isRoot()
	{
		return parents.isEmpty();
	}

	void setWeight(double weight)
	{
		this.weight = weight;
	}

	public double getWeight()
	{
		return weight;
	}

	public boolean isAncestor(String label)
	{
		HashSet<String> labels = new HashSet<String>();
		LinkedList<Node> parentList = new LinkedList<Node>(parents);

		while (!parentList.isEmpty())
		{
			Node n = parentList.removeFirst();
			labels.add(n.getLabel());
			parentList.addAll(n.getParents());
		}

		return labels.contains(label);
	}

	public boolean isDescendant(String label)
	{
		HashSet<String> labels = new HashSet<String>();
		LinkedList<Node> childList = new LinkedList<Node>(children);

		while (!childList.isEmpty())
		{
			Node n = childList.removeFirst();
			labels.add(n.getLabel());
			childList.addAll(n.getChildren());
		}

		return labels.contains(label);
	}

	public String getLabel()
	{
		return label;
	}

	public Collection<Node> getParents()
	{
		return parents;
	}

	public Collection<Node> getChildren()
	{
		return children;
	}

	public Node getChild(String label)
	{
		for (Node child : children)
		{
			if (child.getLabel().equals(label))
			{
				return child;
			}
		}

		return null;
	}

	public Collection<Node> getAncestors()
	{
		Set<Node> ancestors = new LinkedHashSet<Node>();
		LinkedList<Node> parents = new LinkedList<Node>(getParents());

		while (!parents.isEmpty())
		{
			Node node = parents.removeFirst();
			ancestors.add(node);
			parents.addAll(node.getParents());
		}

		return ancestors;
	}

	public Collection<Node> getDescendants()
	{
		Set<Node> descendants = new LinkedHashSet<Node>();
		LinkedList<Node> children = new LinkedList<Node>(getChildren());

		while (!children.isEmpty())
		{
			Node node = children.removeFirst();
			descendants.add(node);
			children.addAll(node.getChildren());
		}

		return descendants;
	}

	@Override
	public boolean equals(Object o)
	{
		if (o instanceof Node)
		{
			Node node = (Node) o;
			return label.equals(node.label);
		}

		return false;
	}

	@Override
	public int hashCode()
	{
		return label.hashCode();
	}

	@Override
	public String toString()
	{
		return label;
	}
}