/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining.data;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 * This class represents a nominal attribute.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class NominalAttribute
	extends Attribute
{
	private static final long serialVersionUID = 5486974079601980428L;

	/**
	 * The index for missing values.
	 */
	public static final int MISSING_VALUE_INDEX = -1;

	/**
	 * The mapping of values -> indexes.
	 */
	private List<String> values;

	/**
	 * The mapping of indexes -> values.
	 */
	private Map<String, Integer> indexes;

	/**
	 * Creates a new <code>NominalAttribute</code> instance.
	 * 
	 * @param name the attribute name.
	 */
	public NominalAttribute(String name)
	{
		this(name, new String[0]);
	}

	/**
	 * Creates a new <code>NominalAttribute</code> instance.
	 * 
	 * @param name the attribute name.
	 * @param values the attribute values.
	 */
	public NominalAttribute(String name, String[] values)
	{
		this(name, Arrays.asList(values));
	}

	/**
	 * Creates a new <code>NominalAttribute</code> instance.
	 * 
	 * @param name the attribute name.
	 * @param values the attribute values.
	 */
	public NominalAttribute(String name, Collection<String> values)
	{
		super(name);
		this.values = new ArrayList<String>(values);
		this.indexes = new HashMap<String, Integer>(values.size(), 1.0f);
		update();
	}

	/**
	 * Adds a value to the attrube domain. The values are stored in alphabetic
	 * order.
	 * 
	 * @param value the value to add.
	 */
	public void add(String value)
	{
		values.add(value);
		update();
	}

	/**
	 * Removes the specified value from the attribute domain.
	 * 
	 * @param value the value to remove.
	 */
	public void remove(String value)
	{
		indexes.remove(values.remove(value));
		update();
	}

	/**
	 * Returns the values of the attribute.
	 * 
	 * @return the values of the attribute.
	 */
	public Collection<String> getValues()
	{
		return Collections.unmodifiableList(values);
	}

	/**
	 * Returns the index of the specified value.
	 * 
	 * @param value the nominal attribute's value.
	 * 
	 * @return the index of the specified value.
	 */
	public int getIndex(String value)
	{
		if (Metadata.MISSING_VALUE.equals(value))
		{
			return MISSING_VALUE_INDEX;
		}

		if (indexes.containsKey(value))
		{
			return indexes.get(value);
		}

		throw new IllegalArgumentException(
			"Value not found in attribute domain: [attribute=" + getName()
				+ ",value=" + value + "]");
	}

	/**
	 * Returns a value given its index.
	 * 
	 * @param index the value's index.
	 * 
	 * @return a value given its index.
	 */
	public String getValue(int index)
	{
		return (index == MISSING_VALUE_INDEX) ? Metadata.MISSING_VALUE : values
			.get(index);
	}

	@Override
	public String toString()
	{
		StringBuffer buffer = new StringBuffer();
		buffer.append(getName());
		buffer.append(" [");

		if (!values.isEmpty())
		{
			Iterator<String> iterator = values.iterator();
			buffer.append(iterator.next());

			while (iterator.hasNext())
			{
				buffer.append(",");
				buffer.append(iterator.next());
			}
		}

		buffer.append("]");

		return buffer.toString();
	}

	public String toString(String value)
	{
		validate(value);
		return getName() + " = " + value;
	}

	@Override
	public void validate(String value)
	{
		if (!(indexes.containsKey(value) || Metadata.MISSING_VALUE
			.equals(value)))
		{
			throw new IllegalArgumentException("Value not found in attribute '"
				+ getName() + "' domain: " + value);
		}
	}

	@Override
	public boolean isMissing(double value)
	{
		return (MISSING_VALUE_INDEX == value);
	}

	@Override
	public NominalAttribute clone()
	{
		NominalAttribute clone = (NominalAttribute) super.clone();
		clone.values = new ArrayList<String>(values);
		clone.indexes = new HashMap<String, Integer>(indexes);
		return clone;
	}

	@Override
	public boolean equals(Attribute o)
	{
		return (o instanceof NominalAttribute) && equals((NominalAttribute) o);
	}

	public boolean equals(NominalAttribute o)
	{
		return super.equals(o) && values.equals(values)
			&& indexes.equals(indexes);
	}

	private void update()
	{
		Collections.sort(values);
		int index = 0;

		for (String v : values)
		{
			indexes.put(v, index);
			index++;
		}
	}

	/**
	 * Returns the most frequent (modal) value.
	 * 
	 * @param instances the instances used to compute the modal.
	 * 
	 * @return the most frequent (modal) value.
	 */
	public String modal(List<Instance> instances)
	{
		if (instances.isEmpty())
		{
			throw new IllegalArgumentException("Empty instances list.");
		}

		int index =
			instances.iterator().next().getMetadata().getIndex(getName());
		int[] frequency = new int[values.size()];

		for (Instance instance : instances)
		{
			int v = (int) instance.values()[index];

			// checks if it is not a missing value
			if (v != MISSING_VALUE_INDEX)
			{
				frequency[v]++;
			}
		}

		int value = -1;
		int max = -1;

		for (int i = 0; i < frequency.length; i++)
		{
			if (frequency[i] > max)
			{
				value = i;
			}
		}

		return getValue(value);
	}
}