/*
 * $Id: Operator.java 2359 2011-04-14 15:21:10Z febo2 $
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining.data;

/**
 * This enum represents an operator in a rule's antecedent term.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision: 2359 $ $Date:: 2011-04-14 16:21:10#$
 */
public enum Operator
	implements Cloneable
{
	EQUAL_TO
	{
		@Override
		public boolean evaluate(double attribute, double... values)
		{
			return (attribute == values[0]);
		}

		@Override
		public String toString(String attribute, double[] values,
				String[] description)
		{
			return attribute + " = " + description[0];
		}
		
		@Override
		public String toString()
		{
			return "=";
		}
	},
	NOT_EQUAL_TO
	{
		@Override
		public boolean evaluate(double attribute, double... values)
		{
			return (attribute != values[0]);
		}

		@Override
		public String toString(String attribute, double[] values,
				String[] description)
		{
			return attribute + " != " + description[0];
		}
		
		@Override
		public String toString()
		{
			return "!=";
		}
	},
	LESS_THAN
	{
		@Override
		public boolean evaluate(double attribute, double... values)
		{
			return (attribute < values[0]);
		}

		@Override
		public String toString(String attribute, double[] values,
				String[] description)
		{
			return attribute + " < " + precision(values[0]);
		}
		
		@Override
		public String toString()
		{
			return "<";
		}
	},
	LESS_THAN_EQUAL_TO
	{
		@Override
		public boolean evaluate(double attribute, double... values)
		{
			return (attribute <= values[0]);
		}

		@Override
		public String toString(String attribute, double[] values,
				String[] description)
		{
			return attribute + " <= " + precision(values[0]);
		}
		
		@Override
		public String toString()
		{
			return "<=";
		}
	},
	GREATER_THAN
	{
		@Override
		public boolean evaluate(double attribute, double... values)
		{
			return (attribute > values[0]);
		}

		@Override
		public String toString(String attribute, double[] values,
				String[] description)
		{
			return attribute + " > " + precision(values[0]);
		}
		
		@Override
		public String toString()
		{
			return ">";
		}
	},
	GREATER_THAN_EQUAL_TO
	{
		@Override
		public boolean evaluate(double attribute, double... values)
		{
			return (attribute >= values[0]);
		}

		@Override
		public String toString(String attribute, double[] values,
				String[] description)
		{
			return attribute + " >= " + precision(values[0]);
		}
		
		@Override
		public String toString()
		{
			return ">=";
		}
	},
	IN_RANGE
	{
		@Override
		public boolean evaluate(double attribute, double... values)
		{
			return (values[0] < attribute) && (attribute <= values[1]);
		}

		@Override
		public String toString(String attribute, double[] values,
				String[] description)
		{
			return precision(values[0]) + " < " + attribute + " <= "
				+ precision(values[1]);
		}
		
		@Override
		public String toString()
		{
			return "< <=";
		}
	};

	/**
	 * Verifies if the condition represented by this operatos is satisfied
	 * by the specified values.
	 * 
	 * @param attribute the attribute value.
	 * @param values the condition values.
	 * 
	 * @return <code>true</code> is the attribute value satisfies the operator
	 *         condition; <code>false</code> otherwise.
	 */
	public abstract boolean evaluate(double attribute, double... values);

	/**
	 * Returns the string representation of this operator condition.
	 * 
	 * @param attribute the attribute name.
	 * @param values the condition values.
	 * 
	 * @return the string representation of this operator condition.
	 */
	public abstract String toString(String attribute, double[] values,
			String[] description);

	/**
	 * Returns the double value with up to two decimal digits. This is only
	 * used to output the value as a string (for readability). 
	 * 
	 * @param value the value to be converted.
	 * 
	 * @return the double value with up to four decimal digits.
	 */
	protected final double precision(double value)
	{
		int v = (int) (value * 10000);
		return v / 10000.0;
	}
}