/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining.data;

import java.util.ArrayList;
import java.util.Collection;

/**
 * This class represents a hierarchical strategy for determining positive and
 * negative examples in an hierarchical classification problem.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class Strategy
{
	/**
	 * The property key under which the strategy type is stored.
	 */
	public static final String TYPE_KEY =
		"net.sf.myra.datamining.data.strategy";

	public static final String DEFAULT_TYPE = "inclusive";

	public static final String POSITIVE = "yes";

	public static final String NEGATIVE = "no";

	public static Type get(String type)
	{
		for (Type t : Type.values())
		{
			if (type.equals(t.type))
			{
				return t;
			}
		}

		throw new IllegalArgumentException("Enum value not found for type: "
			+ type);
	}

	public enum Type
	{
		INCLUSIVE("inclusive")
		{
			public Dataset transform(Node node, Dataset dataset)
			{
				Metadata metadata = new Metadata();
				metadata.setName(dataset.getMetadata().getName());

				for (Attribute attribute : dataset.getMetadata().getPredictor())
				{
					metadata.add(attribute);
				}

				NominalAttribute c =
					new NominalAttribute(node.getLabel(), new String[] {
						POSITIVE, NEGATIVE });
				metadata.add(c);
				metadata.setTarget(c);

				Dataset localDS = new Dataset(metadata);

				for (Instance instance : dataset)
				{
					String[] values = instance.getValues();
					String label = NEGATIVE;

					for (String v : instance.getLabel().getValues())
					{
						if (node.getLabel().equals(v) || node.isDescendant(v))
						{
							label = POSITIVE;
							break;
						}
					}

					values[values.length - 1] = label;

					boolean include = true;

					if (NEGATIVE.equals(label))
					{
						for (String v : instance.getLabel().getValues())
						{
							include = !node.isAncestor(v);
						}
					}

					if (include)
					{
						localDS.add(values);
					}
				}

				return localDS;
			}
		},
		LESS_INCLUSIVE("less-inclusive")
		{
			public Dataset transform(Node node, Dataset dataset)
			{
				Metadata metadata = new Metadata();
				metadata.setName(dataset.getMetadata().getName());

				for (Attribute attribute : dataset.getMetadata().getPredictor())
				{
					metadata.add(attribute);
				}

				NominalAttribute c =
					new NominalAttribute(node.getLabel(), new String[] {
						POSITIVE, NEGATIVE });
				metadata.add(c);
				metadata.setTarget(c);

				Dataset localDS = new Dataset(metadata);

				for (Instance instance : dataset)
				{
					String[] values = instance.getValues();
					String label = NEGATIVE;

					for (String v : instance.getLabel().getValues())
					{
						if (node.getLabel().equals(v) || node.isDescendant(v))
						{
							label = POSITIVE;
							break;
						}
					}

					values[values.length - 1] = label;
					localDS.add(values);
				}

				return localDS;
			}
		},
		MULTI_CLASS("multi-class")
		{
			public Dataset transform(Node node, Dataset dataset)
			{
				Metadata metadata = new Metadata();
				metadata.setName(dataset.getMetadata().getName());

				for (Attribute attribute : dataset.getMetadata().getPredictor())
				{
					metadata.add(attribute);
				}

				Collection<String> available = new ArrayList<String>();

				for (Node child : node.getChildren())
				{
					available.add(child.getLabel());
				}

				NominalAttribute c =
					new NominalAttribute(node.getLabel(), available);
				metadata.add(c);
				metadata.setTarget(c);

				Dataset localDS = new Dataset(metadata);

				for (Instance instance : dataset)
				{
					String[] values = instance.getValues();
					ArrayList<String> label = new ArrayList<String>();

					for (String v : available)
					{
						if (instance.getLabel().active(v))
						{
							label.add(v);
						}
					}

					if (label.size() != 1)
					{
						throw new IllegalArgumentException(
							"Invalid label transformation: "
								+ "1 value expected, found " + label.size());
					}

					values[values.length - 1] = label.get(0);
					localDS.add(values);
				}

				return localDS;
			}
		};

		private String type;

		private Type(String type)
		{
			this.type = type;
		}

		public abstract Dataset transform(Node node, Dataset dataset);
	}
}