/*
 * $Id: Term.java 2361 2011-04-14 15:24:19Z febo2 $
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining.data;

import java.util.Arrays;
import net.sf.myra.framework.Info;

/**
 * This class represents a term in the antecedent of a rule.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision: 2361 $ $Date:: 2011-04-14 16:24:19#$
 */
public class Term
	implements Info
{
	private static final long serialVersionUID = 7543193663161084872L;

	/**
	 * The attribute name.
	 */
	private String attribute;

	/**
	 * The attribute index.
	 */
	private int index;

	/**
	 * The term operator.
	 */
	private Operator operator;

	/**
	 * The term values.
	 */
	private double[] values;

	/**
	 * The term values description.
	 */
	private String[] description;

	/**
	 * Default constructor.
	 * 
	 * @param attribute the attribute name.
	 * @param index the attribute index.
	 */
	public Term(String attribute, int index)
	{
		this.attribute = attribute;
		this.index = index;
	}

	/**
	 * Returns the term's attribute name.
	 * 
	 * @return the term's attribute name.
	 */
	public String getAttribute()
	{
		return attribute;
	}
	
	/**
	 * Returns the term's operator.
	 * 
	 * @return the term's operator.
	 */
	public Operator getOperator()
	{
		return operator;
	}

	/**
	 * Returns the term's attribute index.
	 * 
	 * @return the term's attribute index.
	 */
	public int getIndex()
	{
		return index;
	}
	
	/**
	 * Returns the term values.
	 * 
	 * @return the term values.
	 */
	public double[] getValues()
	{
		return values.clone();
	}

	/**
	 * Sets the term operator.
	 * 
	 * @param operator the operator to set.
	 */
	public void setOperator(Operator operator)
	{
		this.operator = operator;
	}

	/**
	 * Sets the term values.
	 * 
	 * @param values the values to set.
	 */
	public void setValues(double... values)
	{
		this.values = values;
	}

	/**
	 * Sets the term values description.
	 * 
	 * @param description the array of values description.
	 */
	public void setDescription(String... description)
	{
		this.description = description;
	}

	/**
	 * Verifies if the specified attribute value satisfies the term.
	 * 
	 * @param value the attribute value.
	 * 
	 * @return <code>true</code> if the attribute value satisfies the term
	 *         operator; <code>false</code> otherwise.
	 */
	public boolean satisfies(double value)
	{
		return operator.evaluate(value, values);
	}

	@Override
	public boolean equals(Object o)
	{
		return equals((Term) o);
	}

	/**
	 * Compares this term against the specified term for equality.
	 * 
	 * @param o the term to be compared to.
	 * 
	 * @return <code>true</code> if both terms are equal; <code>false</code>
	 *         otherwise.
	 */
	public boolean equals(Term o)
	{
		return (o == null) ? false : (operator.equals(o.operator) && Arrays
			.equals(values, o.values));
	}

	@Override
	public Term clone()
	{
		try
		{
			Term clone = (Term) super.clone();

			if (values != null)
			{
				clone.values = values.clone();
			}

			if (description != null)
			{
				clone.description = description.clone();
			}

			return clone;
		}
		catch (CloneNotSupportedException e)
		{
			throw new UnknownError(getClass() + ":clone()");
		}
	}

	@Override
	public String toString()
	{
		return operator.toString(attribute, values, description);
	}
}