/*
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2011 Fernando Esteban Barril Otero, Matthew Medland
 */

package net.sf.myra.datamining.function;

import static net.sf.myra.datamining.ConfusionMatrix.FN;
import static net.sf.myra.datamining.ConfusionMatrix.FP;
import static net.sf.myra.datamining.ConfusionMatrix.TN;
import static net.sf.myra.datamining.ConfusionMatrix.TP;
import net.sf.myra.datamining.AbstractObjectiveFunction;
import net.sf.myra.datamining.ConfusionMatrix;
import net.sf.myra.datamining.Evaluator;
import net.sf.myra.datamining.Rule;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.Cost;
import net.sf.myra.framework.NumericCost;
import net.sf.myra.framework.Trail;

/**
 * This class implements the Klosgen rule evaluation function.
 * 
 * @author Matthew Medland
 */
public class KlosgenFunction
	extends AbstractObjectiveFunction
{
	public static final String OMEGA = "net.sf.myra.datamining.klosgen.omega";

	private double omega;

	/**
	 * Default constructor.
	 * 
	 * @param dataset the dataset used in the evaluation.
	 */
	public KlosgenFunction(Dataset dataset)
	{
		this(dataset, Configuration.getDoubleProperty(OMEGA, 0.5));
	}
	
	/**
	 * Construction with parameter.
	 * 
	 * @param dataset the dataset used in the evaluation.
	 * @param omega the omega value to be used
	 */
	public KlosgenFunction(Dataset dataset, double omega)
	{
		super(dataset);
		this.omega = omega;
	}

	public Cost evaluate(Trail trail)
	{
		Rule rule = (Rule) trail;
		Dataset dataset = getDataset();

		double klosgen = 0.0;

		if (rule.getConsequent().length() > 0)
		{
			ConfusionMatrix matrix =
				Evaluator.getConfusionMatrix(rule, dataset);

			double P = matrix.get(TP) + matrix.get(FN);
			double N = matrix.get(FP) + matrix.get(TN);

			double precision = matrix.getPrecision();
			double coverage = (matrix.get(TP) + matrix.get(FP)) / (P + N);

			klosgen = Math.pow(coverage, omega) * (precision - (P / (P + N)));
		}

		return new NumericCost(klosgen);
	}
}