/*
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2011 Fernando Esteban Barril Otero, Matthew Medland
 */

package net.sf.myra.datamining.function;

import static net.sf.myra.datamining.ConfusionMatrix.FN;
import static net.sf.myra.datamining.ConfusionMatrix.FP;
import static net.sf.myra.datamining.ConfusionMatrix.TN;
import static net.sf.myra.datamining.ConfusionMatrix.TP;
import net.sf.myra.datamining.AbstractObjectiveFunction;
import net.sf.myra.datamining.ConfusionMatrix;
import net.sf.myra.datamining.Evaluator;
import net.sf.myra.datamining.Rule;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.Cost;
import net.sf.myra.framework.NumericCost;
import net.sf.myra.framework.Trail;

/**
 * This class implements the m-estimate rule evaluation function.
 * 
 * @author Matthew Medland
 */
public class MEstimateFunction
	extends AbstractObjectiveFunction
{
	/**
	 * The property key under which the estimated number of covered examples
	 * of a rule value is stored.
	 */
	public static final String C = "net.sf.myra.datamining.mestimate.m";

	/**
	 * The estimated number of covered examples of a rule.
	 */
	private double m;

	/**
	 * Default constructor.
	 * 
	 * @param dataset the dataset used in the evaluation.
	 */
	public MEstimateFunction(Dataset dataset)
	{
		// the default value 2 is the special case where the m-estimate is
		// the same as the laplace function
		this(dataset, Configuration.getDoubleProperty(C, 2));
	}

	/**
	 * Construction with parameter.
	 * 
	 * @param dataset the dataset used in the evaluation.
	 * @param m the m value to be used
	 */
	public MEstimateFunction(Dataset dataset, double m)
	{
		super(dataset);
		this.m = m;
	}

	public Cost evaluate(Trail trail)
	{
		Rule rule = (Rule) trail;
		Dataset dataset = getDataset();

		double mEstimate = 0.0;

		if (rule.getConsequent().length() > 0)
		{
			ConfusionMatrix matrix =
				Evaluator.getConfusionMatrix(rule, dataset);
			double p = matrix.get(TP);
			double P = matrix.get(TP) + matrix.get(FN);
			double n = matrix.get(FP);
			double N = matrix.get(FP) + matrix.get(TN);

			mEstimate = (p + m * (P / (P + N))) / (p + n + m);
		}

		return new NumericCost(Double.isNaN(mEstimate) ? 0.0 : mEstimate);
	}
}