/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2009  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining.function.hierarchical;

import java.util.Collection;
import java.util.List;
import net.sf.myra.datamining.AbstractObjectiveFunction;
import net.sf.myra.datamining.ProbabilisticRule;
import net.sf.myra.datamining.data.ClassHierarchy;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Label;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.statistics.Curve;
import net.sf.myra.datamining.statistics.CurveFactory.PRCurveFactory;
import net.sf.myra.framework.Cost;
import net.sf.myra.framework.NumericCost;
import net.sf.myra.framework.Trail;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class PooledPRCurveFunction
	extends AbstractObjectiveFunction
{
	/**
	 * Default constructor.
	 * 
	 * @param dataset the dataset used in the evaluation.
	 */
	public PooledPRCurveFunction(Dataset dataset)
	{
		super(dataset);
	}

	public Cost evaluate(Trail trail)
	{
		ProbabilisticRule rule = (ProbabilisticRule) trail;
		Dataset dataset = getDataset();
		Metadata metadata = dataset.getMetadata();

		double[] probabilities = rule.getProbabiblities();
		String[] labels =
			labels(metadata.getLabel(), probabilities, metadata
				.getClassHierarchy());
		String[] filtered = filtered(labels);

		List<Instance> instances = dataset.filter(filtered);
		double[][] values = new double[instances.size() * filtered.length][2];
		int index = 0;

		for (Instance instance : instances)
		{
			boolean[] original = instance.getLabel().flags();
			boolean covered = rule.isCovered(instance);

			for (int i = 0; i < labels.length; i++)
			{
				if (labels[i] != null)
				{
					values[index][0] = (covered ? probabilities[i] : 0.0);
					values[index][1] = (original[i] ? 1 : 0);
					index++;
				}
			}
		}

		Curve curve = PRCurveFactory.PRECISION_RECALL.create(values);
		return new NumericCost(curve.area());
	}

	/**
	 * Returns an array with all class labels that have a probability
	 * associated. Any position of the array with a <code>null</code>
	 * indicates that the class label is not present.
	 * 
	 * @param prototype
	 * @param probabilities
	 * @param root
	 * @return
	 */
	private String[] labels(Label prototype, double[] probabilities,
			ClassHierarchy hierarchy)
	{
		String[] labels = new String[probabilities.length];
		String[] names = prototype.names();
		String root = hierarchy.getRoot().getLabel();

		for (int i = 0; i < names.length; i++)
		{
			if (!root.equals(names[i]) && probabilities[i] > 0.0)
			{
				labels[i] = names[i];
			}
		}

		Collection<String> leaves = hierarchy.leaves(filtered(labels));

		for (int i = 0; i < labels.length; i++)
		{
			if (!leaves.contains(labels[i]))
			{
				labels[i] = null;
			}
		}

		return labels;
	}

	/**
	 * Returns an array which represent only the not <code>null</code> values.
	 * 
	 * @param labels
	 * 
	 * @return
	 */
	private String[] filtered(String[] labels)
	{
		int count = 0;

		for (int i = 0; i < labels.length; i++)
		{
			if (labels[i] != null)
			{
				count++;
			}
		}

		String[] filtered = new String[count];
		count = 0;

		for (int i = 0; i < labels.length; i++)
		{
			if (labels[i] != null)
			{
				filtered[count] = labels[i];
				count++;
			}
		}

		return filtered;
	}
}