/*
 * $Id: HierarchicalExporter.java 2276 2010-04-15 11:37:24Z febo $
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2009  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining.io;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.text.DateFormat;
import java.util.Calendar;
import net.sf.myra.datamining.Model;
import net.sf.myra.datamining.data.ClassHierarchy;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Label;
import net.sf.myra.datamining.data.Metadata;

/**
 * This class is used to output the individual predictions of each test
 * instance.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision: 2276 $ $Date:: 2010-04-15 12:37:24#$
 */
public class Exporter
{
	/**
	 * The output writer.
	 */
	protected PrintWriter writer;

	/**
	 * The class hierarchy reference.
	 */
	protected ClassHierarchy hierarchy;

	/**
	 * Default constructor.
	 * 
	 * @param file the output file.
	 * @param metadata the dataset metadata information.
	 */
	public void initialize(File file, Metadata metadata)
		throws IOException
	{
		boolean append = file.exists();
		this.writer =
			new PrintWriter(new BufferedWriter(new FileWriter(file, append)));

		if (!append)
		{
			writeHeader(metadata);
		}
	}

	/**
	 * Writes the predictions to the file.
	 * 
	 * @param testSet the set of test instances.
	 * @param discovered the discovered model.
	 */
	public void write(Dataset testSet, Model discovered)
	{
		for (Instance instance : testSet)
		{
			export(instance, discovered.classify(instance));
		}
	}

	/**
	 * Writes the output file header.
	 * 
	 * @param metadata the dataset metadata information.
	 */
	protected void writeHeader(Metadata metadata)
	{
		writer.println("% ARFF file for " + metadata.getName());
		writer.println("% Generated on "
			+ DateFormat.getDateTimeInstance(DateFormat.LONG, DateFormat.LONG)
				.format(Calendar.getInstance().getTime()));
		writer.println("%");
		writer.println("@RELATION " + metadata.getName() + "-predictions");
		writer.println();

		Label label = metadata.getLabel();

		// original class information
		writer.println("@ATTRIBUTE class-a string");

		for (int i = 0; i < label.length(); i++)
		{
			String l = label.names()[i];
			writer.println("@ATTRIBUTE class-a-" + l + " {0,1}");
		}

		for (int i = 0; i < label.length(); i++)
		{
			String l = label.names()[i];
			writer.println("@ATTRIBUTE Original-p-" + l + " numeric");
		}

		writer.println();
		writer.println("@DATA");
	}

	/**
	 * Writes the instance informations.
	 * 
	 * @param instance the instance reference.
	 * @param predicted the predicted label. 
	 */
	protected void export(Instance instance, Label predicted)
	{
		StringBuffer buffer = new StringBuffer();

		Label original = instance.getLabel();
		buffer.append(original.toString());

		for (int i = 0; i < original.length(); i++)
		{
			buffer.append(",");
			buffer.append(original.active(i) ? "1" : "0");
		}

		for (int i = 0; i < predicted.length(); i++)
		{
			buffer.append(",");
			buffer.append(predicted.active(i) ? "1" : "0");
		}

		writer.println(buffer.toString());
	}

	/**
	 * Closes the output file.
	 */
	public void close()
	{
		if (writer != null)
		{
			writer.close();
		}
	}
}
