/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining.io;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.Reader;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Set;
import java.util.StringTokenizer;

import net.sf.myra.datamining.data.Attribute;
import net.sf.myra.datamining.data.ClassHierarchy;
import net.sf.myra.datamining.data.ContinuousAttribute;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.data.Node;
import net.sf.myra.datamining.data.NominalAttribute;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class ExtendedArffHelper
	extends ArffHelper
{
	@Override
	protected void writeTarget(Metadata metadata, PrintWriter writer)
	{
		ClassHierarchy hierarchy = metadata.getClassHierarchy();

		if (hierarchy != null)
		{
			boolean unique = hierarchy.unique();

			StringBuffer values = new StringBuffer();
			values.append("hierarchical ");

			LinkedList<Node> parents = new LinkedList<Node>();
			Set<String> relations = new HashSet<String>();
			boolean comma = true;

			if (unique)
			{
				for (Node child : hierarchy.getRoot().getChildren())
				{
					values.append("root/");
					values.append(child.getLabel());
					values.append(",");

					parents.add(child);
				}

				values.deleteCharAt(values.length() - 1);
			}
			else
			{
				comma = false;
				parents.add(hierarchy.getRoot());
			}

			while (!parents.isEmpty())
			{
				Node parent = parents.removeFirst();

				for (Node child : parent.getChildren())
				{
					String value =
						(unique ? parent.getLabel() + "/" + child.getLabel()
							: child.getLabel());

					if (!relations.contains(value))
					{
						if (comma)
						{
							values.append(",");
						}
						else
						{
							comma = true;
						}

						values.append(value);
						relations.add(value);
					}

					parents.add(child);
				}
			}

			NominalAttribute target = metadata.getTarget();
			writer.println("@attribute "
				+ target.getName().replaceAll("(\\s)+", "_") + " "
				+ values.toString().replace(":", ""));
		}
		else
		{
			throw new IllegalArgumentException("Class hierarchy is null");
		}
	}

	@Override
	protected void writeData(Dataset dataset, PrintWriter writer)
	{
		Node root = dataset.getMetadata().getClassHierarchy().getRoot();

		for (Instance instance : dataset)
		{
			String line = instance.toString();
			line = line.substring(0, line.lastIndexOf(',') + 1);
			StringBuffer buffer = new StringBuffer();

			// we only output the most specific class labels
			Collection<String> leaves =
				dataset.getMetadata().getClassHierarchy().leaves(
					instance.getLabel().getValues());

			for (String label : leaves)
			{
				if (!label.equals(root.getLabel()))
				{
					if (buffer.length() > 0)
					{
						buffer.append(Dataset.CLASS_SEPARATOR);
					}

					buffer.append(label.replace(":", ""));
				}
			}

			buffer.insert(0, line);
			writer.println(buffer.toString());
		}
	}

	@Override
	public Dataset read(Reader reader)
		throws IOException
	{
		BufferedReader input = null;

		try
		{
			input = new BufferedReader(reader);
			Metadata metadata = new Metadata();
			ClassHierarchy hierarchy = new ClassHierarchy();

			String line = null;
			Attribute last = null;

			// reads the metadata information
			while ((line = input.readLine()) != null)
			{
				line = line.trim();

				if ("@data".equals(line.toLowerCase()))
				{
					// we reached the data section
					break;
				}

				if (!(isComment(line) || isBlank(line)))
				{
					// reads the attribute information
					if (line.toLowerCase().startsWith("@attribute"))
					{
						String[] fields = line.split("\\s+");

						if (fields.length < 3)
						{
							throw new IllegalArgumentException(
								"Illegal attribute definition: " + line);
						}

						String name = trim(fields[1]);
						String type = trim(fields[2]).toLowerCase();

						if ("string".equals(type) || "date".equals(type))
						{
							last = new UnsupportedAttribute(name);
						}
						else if (line.indexOf("{") != -1)
						{
							StringTokenizer values =
								new StringTokenizer(line.substring(
									line.indexOf("{") + 1, line.indexOf("}"))
									.trim(), ",");

							NominalAttribute attribute =
								new NominalAttribute(name);

							while (values.hasMoreTokens())
							{
								attribute.add(trim(values.nextToken()));
							}

							last = attribute;
						}
						else if (line.indexOf("hierarchical ") != -1)
						{
							int start = line.indexOf("hierarchical ") + 13;
							int end = line.indexOf(',');
							String root = line.substring(start, end);

							// indicates if each node in the hierarchy has a
							// unique label or not
							boolean unique = false;

							if (root.indexOf('/') != -1)
							{
								// finds the root label
								Set<String> parents = new HashSet<String>();
								Set<String> children = new HashSet<String>();

								for (String element : line.substring(start)
									.trim().split(","))
								{
									String[] labels = element.split("/");
									parents.add(labels[0]);
									children.add(labels[1]);
								}

								parents.removeAll(children);

								if (parents.size() != 1)
								{
									throw new IllegalArgumentException(
										"Invalid hierarchy structure: "
											+ parents);
								}

								root = parents.iterator().next();
								unique = true;
							}
							else
							{
								root = "root";
							}

							Set<String> labels = new HashSet<String>();
							labels.add(root);

							LinkedList<String> values =
								new LinkedList<String>(Arrays.asList(line
									.substring(start).trim().split(",")));
							hierarchy.add(root);

							while (!values.isEmpty())
							{
								String value = values.removeFirst();
								int split = value.lastIndexOf('/');

								if (split != -1)
								{
									String[] relationship = new String[2];
									relationship[0] = value.substring(0, split);
									relationship[1] =
										(unique ? value.substring(split + 1)
											: value);

									if (labels.contains(relationship[0]))
									{
										if (hierarchy.contains(relationship[1]))
										{
											hierarchy.link(relationship[0],
												relationship[1]);
										}
										else
										{
											hierarchy.add(relationship[1],
												relationship[0]);
											labels.add(relationship[1]);
										}
									}
									else
									{
										values.addLast(value);
									}
								}
								// must be a child of the root node
								else
								{
									hierarchy.add(value, root);
									labels.add(value);
								}
							}

							if (hierarchy.shrink())
							{
								labels.remove(root);
							}

							NominalAttribute target =
								new NominalAttribute(name);

							for (String label : labels)
							{
								target.add(label);
							}

							last = target;
						}
						else
						{
							last = new ContinuousAttribute(name);
						}

						// this is here as a quick fix for duplicated
						// attributes but duplicated attributes should be
						// considered as an error, so if there is more than
						// two attributes with the same name, an exception
						// will be thrown
						if (metadata.contains(name))
						{
							last.setName(name + "_2");
						}

						// adds the attribute
						metadata.add(last);
					}
					else if (line.toLowerCase().startsWith("@relation"))
					{
						int firstSpace = line.indexOf(" ");
						metadata.setName(line.substring(firstSpace).trim());
					}
				}
			}

			if (!hierarchy.isEmpty())
			{
				metadata.setClassHierarchy(hierarchy);
				hierarchy.weigh();
			}

			if (last instanceof NominalAttribute)
			{
				metadata.setTarget((NominalAttribute) last);
			}

			Dataset dataset = new Dataset(metadata);

			int total = 0;
			int ignored = 0;

			// reads the data information
			while ((line = input.readLine()) != null)
			{
				if (!(isComment(line) || isBlank(line)))
				{
					StringTokenizer tokens = new StringTokenizer(line, ",");
					String[] values = new String[tokens.countTokens()];

					for (int i = 0; tokens.hasMoreTokens(); i++)
					{
						values[i] = trim(tokens.nextToken());
					}

					try
					{
						dataset.add(values);
					}
					catch (IllegalArgumentException e)
					{
						ignored++;
					}

					total++;
				}
			}

			if (ignored > 0)
			{
				//TODO use a logger instance
				System.err.println("[" + getClass().getSimpleName() + ": "
					+ ignored + " of " + total + " record(s) ignored]");
			}

			return dataset;
		}
		finally
		{
			if (input != null)
			{
				input.close();
			}
		}
	}

	/**
	 * Returns the default file extension.
	 * 
	 * @return the default file extension.
	 */
	@Override
	protected String getExtension()
	{
		return "arffx";
	}
}