/*
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2012 Fernando Esteban Barril Otero
 */

package net.sf.myra.datamining.io;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.Writer;

import net.sf.myra.datamining.data.Dataset;

/**
 * Base class for dataset file helper.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public abstract class Helper
{
	/**
	 * Flag to indicate that the output should be compressed.
	 */
	private boolean compressed;

	/**
	 * Tests if the output should be compressed.
	 */
	public boolean isCompressed()
	{
		return compressed;
	}

	/**
	 * Sets the value of the compressed flag.
	 * 
	 * @param compressed the value to set.
	 */
	public void setCompressed(boolean compressed)
	{
		this.compressed = compressed;
	}

	/**
	 * Writes a file representing the specified dataset.
	 * 
	 * @param directory the output directory.
	 * @param dataset the dataset to be written.
	 * 
	 * @throws IOException
	 */
	public abstract void write(File directory, Dataset dataset)
		throws IOException;

	/**
	 * Writes a file representing the specified dataset.
	 * 
	 * @param writer the output writer.
	 * @param dataset the dataset to be written.
	 * 
	 * @throws IOException
	 */
	public abstract void write(Writer writer, Dataset dataset)
		throws IOException;

	/**
	 * Returns a <code>Dataset</code> instance that represents the data of
	 * the specified file.
	 * 
	 * @param filename the name of the file to read the data from.
	 * 
	 * @return a <code>Dataset</code> instance.
	 * 
	 * @throws IOException if any file operation fails.
	 */
	public Dataset read(String filename)
		throws IOException
	{
		return read(new File(filename));
	}

	/**
	 * Returns a <code>Dataset</code> instance that represents the data of
	 * the specified file.
	 * 
	 * @param file the file to read the data from.
	 * 
	 * @return a <code>Dataset</code> instance.
	 * 
	 * @throws IOException if any file operation fails.
	 */
	public abstract Dataset read(File file)
		throws IOException;
	
	/**
	 * Returns a <code>Helper</code> instance to read the specified file.
	 * 
	 * @param file the file to be read.
	 * 
	 * @return a <code>Helper</code> instance to read the specified file.
	 */
	public static Helper getHelper(String filename)
	{
		return getHelper(new File(filename));
	}
	
	/**
	 * Returns a <code>Helper</code> instance to read the specified file.
	 * 
	 * @param file the file to be read.
	 * 
	 * @return a <code>Helper</code> instance to read the specified file.
	 */
	public static Helper getHelper(File file)
	{
		String name = file.getName().toLowerCase();
		int index = file.getName().lastIndexOf('.');
		String ext = file.getName().substring(index + 1);

		if ("arffx".equals(ext) || name.contains("arffx"))
		{
			return new ExtendedArffHelper();
		}
		if ("arff".equals(ext) || name.contains("arff"))
		{
			return new ArffHelper();
		}
		else if ("data".equals(ext) || "test".equals(ext))
		{
			return new C45Helper();
		}
		
		throw new IllegalArgumentException("Could not find a Helper for: "
			+ file);
	}

	/**
	 * Returns a <code>Dataset</code> instance that represents the data of
	 * the specified file.
	 * 
	 * @param filename the filename to read the data from.
	 * 
	 * @return a <code>Dataset</code> instance.
	 * 
	 * @throws IOException if any file operation fails.
	 */
	public static Dataset open(String filename)
		throws IOException
	{
		return open(new File(filename));
	}
	
	/**
	 * Returns a <code>Dataset</code> instance that represents the data of
	 * the specified file.
	 * 
	 * @param file the file to read the data from.
	 * 
	 * @return a <code>Dataset</code> instance.
	 * 
	 * @throws IOException if any file operation fails.
	 */
	public static Dataset open(File file)
		throws IOException
	{
		if (!file.exists())
		{
			throw new FileNotFoundException(file.getName());
		}
		else if (!file.canRead())
		{
			throw new IOException("Cannot read file: " + file.getName());
		}
		
		try
		{
			ExtendedArffHelper arffx = new ExtendedArffHelper();
			Dataset dataset = arffx.read(file);

			if (dataset.getMetadata().isHierarchical())
			{
				return dataset;
			}
		}
		catch (Exception e)
		{
			// not in an extended ARFF format
		}
		
		try
		{
			ArffHelper arff = new ArffHelper();
			Dataset dataset = arff.read(file);

			if (!dataset.isEmpty())
			{
				return dataset;
			}
		}
		catch (Exception e)
		{
			// not in an ARFF format
		}

		try
		{
			C45Helper c45 = new C45Helper();
			Dataset dataset = c45.read(file);

			if (!dataset.isEmpty())
			{
				return dataset;
			}
		}
		catch (Exception e)
		{
			// not in a C45 format
		}

		throw new IllegalArgumentException("Could not find a Helper for: "
			+ file);
	}
}