/*
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2009 Fernando Esteban Barril Otero
 */

package net.sf.myra.datamining.measure.hierarchical;

import java.text.DecimalFormat;
import java.text.NumberFormat;
import net.sf.myra.datamining.Measure;
import net.sf.myra.datamining.Model;
import net.sf.myra.datamining.Unit;
import net.sf.myra.datamining.UnitAveragingMode;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.model.ProbabilisticRuleModel;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class PRCurveMeasure
	implements Measure
{
	static
	{
		UnitAveragingMode.register(CurveArea.class,
			new CurveAreaUnitAveragingMode());
	}

	/**
	 * The pooled PR curve measure.
	 */
	private PooledPRCurveMeasure pMeasure = new PooledPRCurveMeasure();

	/**
	 * The averaged PR curve measure.
	 */
	private AveragedPRCurveMeasure aMeasure = new AveragedPRCurveMeasure();

	/**
	 * The weighted PR curve measure.
	 */
	private WeightedPRCurveMeasure wMeasure = new WeightedPRCurveMeasure();

	public Unit<CurveArea> evaluate(Dataset dataset, Model model)
	{
		return evaluate(dataset, (ProbabilisticRuleModel) model);
	}

	public Unit<CurveArea> evaluate(Dataset dataset,
			ProbabilisticRuleModel model)
	{
		CurveArea area =
			new CurveArea(pMeasure.evaluate(dataset, model).getValue(),
				aMeasure.evaluate(dataset, model).getValue(), wMeasure
					.evaluate(dataset, model).getValue());

		return new Unit<CurveArea>(area);
	}

	/**
	 * This class represents the different PR curve measures.
	 */
	public static class CurveArea
		implements Comparable<CurveArea>
	{
		private static final NumberFormat FORMATTER =
			new DecimalFormat("#,##0.0000000000");

		/**
		 * Area under the pooled curve value.
		 */
		private double pooled;

		/**
		 * Area under the averaged curve value.
		 */
		private double averaged;

		/**
		 * Area under the weighted averaged curve value.
		 */
		private double weighted;

		/**
		 * Default constructor.
		 * 
		 * @param pooled the area under the pooled curve value.
		 * @param averaged the area under the averaged curve value.
		 * @param weighted the area under the weighted averaged curve value.
		 */
		CurveArea(double pooled, double averaged, double weighted)
		{
			this.pooled = pooled;
			this.averaged = averaged;
			this.weighted = weighted;
		}

		public int compareTo(CurveArea o)
		{
			double sum1 = pooled + averaged + weighted;
			double sum2 = o.pooled + o.averaged + o.weighted;

			return Double.compare(sum1, sum2);
		}

		public double pooled()
		{
			return pooled;
		}

		public double averaged()
		{
			return averaged;
		}

		public double weighted()
		{
			return weighted;
		}

		@Override
		public String toString()
		{
			StringBuffer output = new StringBuffer();
			output.append("{");
			output.append(FORMATTER.format(pooled));
			output.append(",");
			output.append(FORMATTER.format(averaged));
			output.append(",");
			output.append(FORMATTER.format(weighted));
			output.append("}");

			return output.toString();
		}
	}

	public static class CurveAreaUnitAveragingMode
		extends UnitAveragingMode<CurveArea>
	{
		protected String getAverage(Unit<CurveArea>[] values)
		{
			StringBuffer output = new StringBuffer();

			double[] pooled = new double[values.length];
			double[] averaged = new double[values.length];
			double[] weighted = new double[values.length];

			for (int i = 0; i < values.length; i++)
			{
				pooled[i] = values[i].getValue().pooled();
				averaged[i] = values[i].getValue().averaged();
				weighted[i] = values[i].getValue().weighted();
			}

			output.append("{");
			output.append(UnitAveragingMode.stdev(pooled));
			output.append(",");
			output.append(UnitAveragingMode.stdev(averaged));
			output.append(",");
			output.append(UnitAveragingMode.stdev(weighted));
			output.append("}");

			return output.toString();
		}
	}
}