/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2009  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining.measure.hierarchical;

import java.util.Arrays;
import java.util.List;

import net.sf.myra.datamining.Measure;
import net.sf.myra.datamining.Model;
import net.sf.myra.datamining.Unit;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Label;
import net.sf.myra.datamining.model.ProbabilisticRuleModel;
import net.sf.myra.datamining.statistics.Curve;
import net.sf.myra.datamining.statistics.CurveFactory;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class WeightedPRCurveMeasure
	implements Measure
{
	private static final List<String> IGNORED = Arrays.asList(new String[] {
		"GO0003674", "GO0005575", "GO0008150" });

	public Unit<?> evaluate(Dataset dataset, Model model)
	{
		return evaluate(dataset, (ProbabilisticRuleModel) model);
	}

	public Unit<Double> evaluate(Dataset dataset, ProbabilisticRuleModel model)
	{
		Label label = dataset.getMetadata().getLabel();
		String root =
			dataset.getMetadata().getClassHierarchy().getRoot().getLabel();

		// the set of ignored classes (no positive examples on the test set)
		boolean[] enabled = new boolean[label.length()];
		Arrays.fill(enabled, true);

		for (int i = 0; i < enabled.length; i++)
		{
			if (root.equals(label.names()[i])
				|| dataset.getInstances(label.names()[i]).size() == 0
				|| IGNORED.contains(label.names()[i]))
			{
				enabled[i] = false;
			}
		}

		double[][] values = new double[dataset.getSize()][2];
		double[] area = new double[label.length()];
		double[] frequency = new double[label.length()];

		double total = 0.0;

		for (int i = 0; i < enabled.length; i++)
		{
			if (enabled[i])
			{
				int index = 0;
				int positives = 0;

				for (Instance instance : dataset)
				{
					boolean[] original = instance.getLabel().flags();
					double[] probabilities = model.probabilities(instance);

					values[index][0] = probabilities[i];
					values[index][1] = (original[i] ? 1 : 0);
					positives += values[index][1];
					index++;
				}

				Curve curve = CurveFactory.PRECISION_RECALL.create(values);
				area[i] = curve.area();
				frequency[i] = positives;
				total += positives;
			}
		}

		double average = 0.0;

		for (int i = 0; i < enabled.length; i++)
		{
			if (enabled[i])
			{
				average += ((frequency[i] / total) * area[i]);
			}
		}

		return new Unit<Double>(average);
	}
}