/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining.model;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;

import net.sf.myra.datamining.Gatherer;
import net.sf.myra.datamining.Model;
import net.sf.myra.datamining.Rule;
import net.sf.myra.datamining.UnitAveragingMode;
import net.sf.myra.datamining.data.Label;

/**
 * Base class for (flat) rule models.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public abstract class AbstractRuleModel
	implements Model, Serializable
{
	private static final long serialVersionUID = -1203066342226016985L;

	/**
	 * The collection of discovered rules.
	 */
	private Collection<Rule> rules;

	/**
	 * The majority class observed in the training set. This is used to make
	 * the default prediction (when none of the rules cover a test example).
	 */
	private Label majority;

	/**
	 * Default constructor.
	 * 
	 * @param rules the collection of discovered rules.
	 */
	public AbstractRuleModel(Collection<Rule> rules)
	{
		this.rules = rules;
	}

	public void add(Rule rule)
	{
		rules.add(rule);
	}

	public void remove(Rule rule)
	{
		rules.remove(rule);
	}

	public Collection<Rule> getRules()
	{
		return rules;
	}

	/**
	 * @return the majority
	 */
	public Label getMajority()
	{
		return majority;
	}

	/**
	 * @param majority the majority to set
	 */
	public void setMajority(Label majority)
	{
		this.majority = majority;
	}

	public String toString()
	{
		StringBuffer output = new StringBuffer();

		if (rules.isEmpty())
		{
			output.append("(empty)\n");
		}
		else
		{
			for (Rule rule : rules)
			{
				output.append(rule);
				output.append("\n");
			}
		}

		return output.toString();
	}

	/**
	 * Checks if the discovered rule set has a default rule (rule with no
	 * terms).
	 * 
	 * @return <code>true</code> if the discovered rule set contains a
	 *         default rule; <code>false</code> otherwise.
	 */
	public boolean hasDefault()
	{
		for (Rule rule : rules)
		{
			if (rule.getSize() == 0)
			{
				return true;
			}
		}

		return false;
	}

	/**
	 * Returns the default rule.
	 * 
	 * @return the default rule.
	 */
	public Rule getDefault()
	{
		for (Rule rule : rules)
		{
			if (rule.getSize() == 0)
			{
				return rule;
			}
		}

		throw new IllegalArgumentException("Default rule not found.");
	}

	/**
	 * Returns the number of rules of the model.
	 * 
	 * @return the number of rules of the model.
	 */
	public int size()
	{
		return rules.size();
	}

	public Gatherer getGatherer()
	{
		return new DefaultGatherer();
	}

	public static class DefaultGatherer
		implements Gatherer
	{
		/**
		 * The total number of terms.
		 */
		private ArrayList<Integer> length = new ArrayList<Integer>();

		/**
		 * The number of rules.
		 */
		private ArrayList<Integer> size = new ArrayList<Integer>();

		/**
		 * The average number of terms per rule.
		 */
		private ArrayList<Double> terms = new ArrayList<Double>();

		public void add(Model model)
		{
			if (model instanceof AbstractRuleModel)
			{
				Collection<Rule> rules = ((AbstractRuleModel) model).getRules();

				int count = countTerms(rules);
				length.add(count);
				size.add(rules.size());
				terms.add(count / (double) rules.size());
			}
			else
			{
				throw new IllegalArgumentException("Invalid model class: "
					+ model.getClass().getName());
			}
		}

		/**
		 * Returns the number of term of the discovered rule set.
		 * 
		 * @param discovered the discovered rule set.
		 * 
		 * @return the number of term of the discovered rule set.
		 */
		protected int countTerms(Collection<Rule> discovered)
		{
			int count = 0;

			for (Rule rule : discovered)
			{
				count += rule.getSize();
			}

			return count;
		}

		@Override
		public String toString()
		{
			StringBuffer buffer = new StringBuffer();

			buffer.append("Model size: ");
			buffer.append(UnitAveragingMode.stdev(length) + "\n");

			buffer.append("Number of terms per rule: ");
			buffer.append(UnitAveragingMode.stdev(terms) + "\n");

			buffer.append("Number of rules: ");
			buffer.append(UnitAveragingMode.stdev(size) + "\n");

			return buffer.toString();
		}
	}
}