/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining.model;

import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import net.sf.myra.datamining.DefaultConfidenceFactor;
import net.sf.myra.datamining.Rule;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Label;
import net.sf.myra.framework.Cost;

/**
 * This class represents a rule set model.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class RuleSet
	extends AbstractRuleModel
{
	private static final long serialVersionUID = -2026094459692477471L;

	private ConflictResolutionStrategy strategy;

	public RuleSet()
	{
		this(ConflictResolutionStrategy.CONFIDENCE);
	}

	public RuleSet(ConflictResolutionStrategy strategy)
	{
		super(new HashSet<Rule>());
		this.strategy = strategy;
	}

	@Override
	public Set<Rule> getRules()
	{
		return (Set<Rule>) super.getRules();
	}

	public Label classify(Instance instance)
	{
		LinkedList<Rule> pool = new LinkedList<Rule>();

		for (Rule rule : getRules())
		{
			if (rule.isCovered(instance))
			{
				pool.add(rule);
			}
		}

		if (pool.isEmpty())
		{
			return instance.getLabel().toLabel();
		}
		else
		{
			return (pool.size() == 1) ? pool.getFirst().getConsequent()
				: strategy.resolve(pool);
		}
	}

	public enum ConflictResolutionStrategy
	{
		RULE_QUALITY()
		{
			@Override
			public Label resolve(Collection<Rule> rules)
			{
				Rule best = null;

				for (Rule rule : rules)
				{
					if ((best == null) || (best.getCost() == null))
					{
						best = rule;
					}
					else
					{
						Cost cost = rule.getCost();

						if ((cost != null)
							&& (cost.compareTo(best.getCost()) == 1))
						{
							best = rule;
						}
					}
				}

				return best.getConsequent();
			}
		},
		FREQUENT_CLASS()
		{
			@Override
			public Label resolve(Collection<Rule> rules)
			{
				HashMap<String, Integer> occurences =
					new HashMap<String, Integer>();
				Label template = null;

				for (Rule rule : rules)
				{
					if (template == null)
					{
						template = rule.getConsequent();
					}

					DefaultConfidenceFactor confidence =
						(DefaultConfidenceFactor) rule.getConfidence();

					for (String value : rule.getConsequent().names())
					{
						Integer frequency = occurences.get(value);

						if (frequency == null)
						{
							frequency = 0;
						}

						occurences
							.put(value, frequency + confidence.get(value));
					}
				}

				String majority = null;
				int frequency = 0;

				for (String value : occurences.keySet())
				{
					if (occurences.get(value) > frequency)
					{
						majority = value;
						frequency = occurences.get(value);
					}
				}

				return template.toLabel(majority);
			}
		},
		CONFIDENCE()
		{
			@Override
			public Label resolve(Collection<Rule> rules)
			{
				SortedSet<Rule> sorted =
					new TreeSet<Rule>(new Comparator<Rule>()
						{
							public int compare(Rule o1, Rule o2)
							{
								return o1.getConfidence().compareTo(
									o2.getConfidence());
							}
						});
				sorted.addAll(rules);

				return sorted.last().getConsequent();
			}
		},
		UNSUPPORTED()
		{
			@Override
			public Label resolve(Collection<Rule> rules)
			{
				throw new UnsupportedOperationException();
			}
		};

		public abstract Label resolve(Collection<Rule> rules);
	}
}