/*
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2011 Fernando Esteban Barril Otero
 */

package net.sf.myra.datamining.seco;

import static net.sf.myra.datamining.Main.MEASURE;
import static net.sf.myra.datamining.data.IntervalBuilder.BUILDER;

import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;

import net.sf.myra.datamining.AbstractObjectiveFunction;
import net.sf.myra.datamining.ClassAssignator;
import net.sf.myra.datamining.Classifier;
import net.sf.myra.datamining.CoveredClassAssignator;
import net.sf.myra.datamining.Evaluator;
import net.sf.myra.datamining.HierarchicalPruner;
import net.sf.myra.datamining.Model;
import net.sf.myra.datamining.Rule;
import net.sf.myra.datamining.data.Attribute;
import net.sf.myra.datamining.data.ContinuousAttribute;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.IntervalBuilder;
import net.sf.myra.datamining.data.IntervalBuilder.Interval;
import net.sf.myra.datamining.data.NominalAttribute;
import net.sf.myra.datamining.data.Term;
import net.sf.myra.datamining.function.SensitivitySpecificityFunction;
import net.sf.myra.datamining.model.RuleList;
import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.Graph;
import net.sf.myra.framework.ObjectFactory;
import net.sf.myra.framework.Vertex;

/**
 * @author Fernando Esteban Barril Otero
 */
public class SeCo
	extends Classifier
{
	public static final String OPTIONS_FILE = "myra-seco.messages";

	/**
	 * The property key under which the class assignator name is stored.
	 */
	public static final String ASSIGNATOR =
		"net.sf.myra.datamining.seco.assignator";

	/**
	 * The property key under which the minimum covered cases by a rule is
	 * stored.
	 */
	public static final String MINIMUM = "net.sf.myra.datamining.seco.minimum";

	/**
	 * Default constructor.
	 */
	public SeCo()
	{
		super(OPTIONS_FILE);
	}

	@Override
	public Properties filter(Properties properties)
	{
		Iterator<Object> keys = properties.keySet().iterator();
		List<String> valid =
			Arrays.asList(new String[] { RULE_QUALITY_FUNCTION, MEASURE,
				ASSIGNATOR, BUILDER, MINIMUM });

		while (keys.hasNext())
		{
			if (!valid.contains(keys.next()))
			{
				keys.remove();
			}
		}

		return properties;
	}

	@Override
	public Model run(Dataset dataset)
	{
		RuleList list = new RuleList();

		while (!dataset.isEmpty())
		{
			Rule rule = learnRule(dataset);
			dataset.remove(Evaluator.findCoveredCases(rule, dataset));
			list.add(rule.clone());
		}

		return list;
	}

	/**
	 * Returns a rule for the given dataset.
	 * 
	 * @param dataset the current dataset.
	 * 
	 * @return a rule for the given dataset.
	 */
	private Rule learnRule(Dataset dataset)
	{
		Graph<Term> graph = dataset.getMetadata().toGraph();
		Collection<Vertex<Term>> refinements = graph.getVertices();

		AbstractObjectiveFunction function =
			(AbstractObjectiveFunction) ObjectFactory.create(Configuration
				.getProperty(RULE_QUALITY_FUNCTION,
					SensitivitySpecificityFunction.class.getName()),
				new Class<?>[] { Dataset.class }, new Object[] { dataset });
		ClassAssignator assignator =
			(ClassAssignator) ObjectFactory.create(Configuration.getProperty(
				ASSIGNATOR, CoveredClassAssignator.class.getName()));
		final int minimum = Configuration.getIntProperty(MINIMUM, 0);

		// we start with an empty rule
		Rule rule = new Rule(dataset.getMetadata());
		assignator.assign(rule, dataset);
		rule.setCost(function.evaluate(rule));

		Dataset covered = dataset.clone();

		while (!refinements.isEmpty())
		{
			double best = rule.getCost().getValue();
			Rule current = rule;
			Vertex<Term> selected = null;

			for (Vertex<Term> vertex : refinements)
			{
				Rule clone = rule.clone();
				clone.add(process(vertex, covered));
				assignator.assign(clone, covered);

				// if the rule with the refinement (term) covers any example
				// (rules that do not cover examples have an empty consequent)
				if (!clone.getConsequent().getValues().isEmpty())
				{
					HierarchicalPruner pruner =
						new HierarchicalPruner(function, null);
					clone = (Rule) pruner.explore(clone);

					if (clone.getConsequent().cardinality() > 0
						&& Evaluator.findCoveredCases(clone, covered).size() >= minimum)
					{
						clone.setCost(function.evaluate(clone));
						double value = clone.getCost().getValue();

						if (value > best)
						{
							best = value;
							current = clone;
							selected = vertex;
						}
					}
				}
			}

			if (current == rule)
			{
				// we could not find any improvements
				break;
			}
			else
			{
				rule = current;
				covered.setInstances(Evaluator.findCoveredCases(rule, covered));

				// if we added a nominal attribute term, we remove all
				// occurences of that attribute in the list of potential
				// refinements

				Attribute attribute =
					dataset.getMetadata().get(selected.getInfo().getIndex());

				if (attribute instanceof NominalAttribute)
				{
					int index = selected.getInfo().getIndex();
					List<Vertex<Term>> terms = new LinkedList<Vertex<Term>>();

					for (Vertex<Term> vertex : refinements)
					{
						if (vertex.getInfo().getIndex() == index)
						{
							terms.add(vertex);
						}
					}

					refinements.removeAll(terms);
				}
			}
		}

		return rule;
	}

	/**
	 * Returns a term ready to be added to the antecedent of a rule. In case
	 * of nominal attribute terms, the specified term is returned; for
	 * continuous attribute terms, a new term repreting the dynamically
	 * created threshold is returned.
	 * 
	 * @param vertex the vertex representing the term.
	 * @param dataset the current dataset.
	 * 
	 * @return a term ready to be added to the antecedent of a rule.
	 */
	private Vertex<Term> process(Vertex<Term> vertex, Dataset dataset)
	{
		Attribute attribute =
			dataset.getMetadata().get(vertex.getInfo().getIndex());

		if (attribute instanceof ContinuousAttribute)
		{
			IntervalBuilder builder =
				IntervalBuilder.getInstance(dataset.getMetadata());

			Interval interval =
				builder.createSingle(dataset.getInstances(),
					(ContinuousAttribute) attribute);

			Vertex<Term> clone = (Vertex<Term>) vertex.clone();
			Term info = clone.getInfo();
			info.setValues(interval.getValues());
			info.setOperator(interval.getOperator());

			return clone;
		}

		return vertex;
	}
}