/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2009  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining.statistics;

import java.util.ArrayList;
import java.util.List;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class Curve
{
	/**
	 * The list of (x,y) coordinates pairs (points).
	 */
	private ArrayList<Point> coordinates;

	/**
	 * Default constructor.
	 */
	public Curve()
	{
		coordinates = new ArrayList<Point>();
	}

	/**
	 * Adds the specified point to this curve.
	 * 
	 * @param point the point to be added.
	 */
	public void add(Point point)
	{
		coordinates.add(point);
	}

	/**
	 * Adds a point with the specified (x,y) coordinates to this curve.
	 * 
	 * @param x the x axis value of the point.
	 * @param y the y axis value of the point.
	 */
	public void add(double x, double y)
	{
		add(new Point(x, y));
	}

	/**
	 * Returns the last point of the curve.
	 * 
	 * @return the last point of the curve.
	 */
	public Point getLast()
	{
		return getLast(1);
	}

	/**
	 * Returns the <code>n</code>-<i>th</i> last point of the curve.
	 * 
	 * @param n the index of the previous point.
	 * 
	 * @return the <code>n</code>-<i>th</i> last point of the curve.
	 */
	public Point getLast(int n)
	{
		if (coordinates.size() < n)
		{
			return null;
		}

		return coordinates.get(coordinates.size() - n);
	}

	/**
	 * Returns the list of points of the curve.
	 * 
	 * @return the list of points of the curve.
	 */
	@SuppressWarnings("unchecked")
	public List<Point> getPoints()
	{
		return (List<Point>) coordinates.clone();
	}

	/**
	 * Returns <code>true</code> if this curve contains no coordinates.
	 * 
	 * @return <code>true</code> if this curve contains no coordinates;
	 *         <code>false</code> otherwise.
	 */
	public boolean isEmpty()
	{
		return coordinates.isEmpty();
	}

	/**
	 * Returns the estimated area under this curve.
	 * 
	 * @return the estimated area under this curve.
	 */
	public double area()
	{
		if (coordinates.size() < 2)
		{
			return 0.0;
		}

		double area = 0.0;
		Point from = coordinates.get(0);

		for (int i = 1; i < coordinates.size(); i++)
		{
			Point to = coordinates.get(i);
			double base = to.getX() - from.getX();
			double height = (to.getY() + from.getY()) / 2.0;
			area += (base * height);
			from = to;
		}

		return area;
	}
	
	/**
	 * Returns the number of coordinaties (points) of this curve.
	 * 
	 * @return the number of coordinaties (points) of this curve.
	 */
	public int size()
	{
		return coordinates.size();
	}
}