/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2009  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining.statistics;

import java.util.Arrays;
import java.util.Comparator;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public abstract class CurveFactory
{
	/**
	 * The negative value constant.
	 */
	public static final int NEGATIVE = 0;

	/**
	 * The positive value constant.
	 */
	public static final int POSITIVE = 1;

	/**
	 * The precision-recall curve factory.
	 */
	public static final PRCurveFactory PRECISION_RECALL = new PRCurveFactory();

	/**
	 * The ROC curve factory.
	 */
	public static final ROCCurveFactory ROC = new ROCCurveFactory();

	/**
	 * Returns a curve based on the specified (predicted,class) values.
	 * 
	 * @param values
	 * 
	 * @return a curve based on the specified (predicted,class) values.
	 */
	public abstract Curve create(double[][] values);

	/**
	 * Returns the number of occurences of the specified class value.
	 * 
	 * @param values the bidimensional array of values.
	 * @param value the class value.
	 * 
	 * @return the number of occurences of the specified class value.
	 */
	protected static int count(double[][] values, int value)
	{
		int counter = 0;

		for (int i = 0; i < values.length; i++)
		{
			if (values[i][1] == value)
			{
				counter++;
			}
		}

		return counter;
	}

	/**
	 * Sorts (based on a bubble sort algorithm) a bidimensional array
	 * (predicted value,class value).
	 * 
	 * @param values the array to be sorted.
	 */
	protected static void sort(double[][] values)
	{
		Arrays.sort(values, new Comparator<double[]>()
		{
			public int compare(double[] o1, double[] o2)
			{
				return Double.compare(o2[0], o1[0]);
			}
		});
	}

	/**
	 * This class represents the Precision-Recall curve factory.
	 */
	public static class ROCCurveFactory
		extends CurveFactory
	{
		@Override
		public Curve create(double[][] values)
		{
			Curve curve = new Curve();

			sort(values);
			double threshold = values[0][0];
			double pTotal = count(values, POSITIVE);
			double nTotal = count(values, NEGATIVE);

			int tpCounter = 0;
			int fpCounter = 0;

			for (int i = 0; i < values.length; i++)
			{
				if (values[i][0] != threshold && (tpCounter + fpCounter) != 0)
				{
					if (curve.isEmpty())
					{
						curve.add(0.0, 0.0);
					}

					double x = (fpCounter == 0 ? 0.0 : fpCounter / nTotal);
					double y = (tpCounter == 0 ? 0.0 : tpCounter / pTotal);
					curve.add(x, y);
					threshold = values[i][0];
				}

				if (values[i][1] == POSITIVE)
				{
					tpCounter++;
				}
				else
				{
					fpCounter++;
				}
			}

			// adds the last point to the curve

			double x = (fpCounter == 0 ? 0.0 : fpCounter / nTotal);
			double y = (tpCounter == 0 ? 0.0 : tpCounter / pTotal);
			curve.add(x, y);
			curve.add(1.0, 1.0);

			return curve;
		}
	}

	/**
	 * This class represents the Precision-Recall curve factory.
	 */
	public static class PRCurveFactory
		extends CurveFactory
	{
		/**
		 * The minumum distance between two points in order to be considered
		 * different values.
		 */
		private static final double DELTA = 1e-15;

		/**
		 * Returns a precision-recall curve based on the specified
		 * (predicted,class) values distribution.
		 *  
		 * @param values the values distribution.
		 *  
		 * @return a precision-recall <code>Curve</code> instance. 
		 */
		public Curve create(double[][] values)
		{
			Curve curve = new Curve();

			sort(values);
			double threshold = values[0][0];
			int pTotal = count(values, POSITIVE);

			int tpCounter = 0;
			int fpCounter = 0;
			int previousTp = 0;
			int previousFp = 0;

			for (int i = 0; i < values.length; i++)
			{
				if (values[i][0] != threshold && (tpCounter + fpCounter) != 0)
				{
					double precision =
						tpCounter / (double) (tpCounter + fpCounter);
					double recall = tpCounter / (double) pTotal;

					if (!curve.isEmpty())
					{
						// interpolating precision-recall points
						interpolate(tpCounter, fpCounter, previousTp,
							previousFp, pTotal, curve);
					}

					normalize(curve, precision, recall);
					previousTp = tpCounter;
					previousFp = fpCounter;

					threshold = values[i][0];
				}

				if (values[i][1] == POSITIVE)
				{
					tpCounter++;
				}
				else
				{
					fpCounter++;
				}
			}

			// add the last point to the curve

			interpolate(tpCounter, fpCounter, previousTp, previousFp, pTotal,
				curve);

			double p = tpCounter / (double) (tpCounter + fpCounter);
			double r = tpCounter / (double) pTotal;
			normalize(curve, p, r);

			return curve;
		}

		/**
		 * Returns a precision-recall curve based on the specified
		 * (predicted,class) values distribution.
		 *  
		 * @param values the values distribution.
		 * @param thresholds the thresholds values.
		 *  
		 * @return a precision-recall <code>Curve</code> instance. 
		 */
		public Curve create(double[][] values, double[] thresholds)
		{
			Curve curve = new Curve();

			sort(values);
			Arrays.sort(thresholds);

			int pTotal = count(values, POSITIVE);

			int tpCounter = 0;
			int fpCounter = 0;
			int previousTp = 0;
			int previousFp = 0;
			int index = 0;

			for (int i = (thresholds.length - 1); i >= 0; i--)
			{
				while ((index < values.length)
					&& (values[index][0] >= thresholds[i]))
				{
					if (values[index][1] == POSITIVE)
					{
						tpCounter++;
					}
					else
					{
						fpCounter++;
					}

					index++;
				}

				if ((previousTp != tpCounter) || (previousFp != fpCounter))
				{
					double precision =
						tpCounter / (double) (tpCounter + fpCounter);
					double recall = tpCounter / (double) pTotal;

					if (curve.isEmpty())
					{
						curve.add(0.0, precision);
					}
					else
					{
						// interpolating precision-recall points
						interpolate(tpCounter, fpCounter, previousTp,
							previousFp, pTotal, curve);
					}

					normalize(curve, precision, recall);
					previousTp = tpCounter;
					previousFp = fpCounter;
				}
			}

			// add the last point to the curve

			interpolate(tpCounter, fpCounter, previousTp, previousFp, pTotal,
				curve);

			double p = tpCounter / (double) (tpCounter + fpCounter);
			double r = tpCounter / (double) pTotal;
			normalize(curve, p, r);

			return curve;
		}

		/**
		 * Interpolates precision-recall values.
		 * 
		 * @param tpCounter the current TP counter.
		 * @param fpCounter the current FP counter.
		 * @param previousTp the previous TP counter.
		 * @param previousFp the previous FP counter.
		 * @param pTotal the total positive count.
		 * @param prCurve the curve object.
		 */
		private void interpolate(int tpCounter, int fpCounter, int previousTp,
				int previousFp, int pTotal, Curve curve)
		{
			for (int x = 1; x < (tpCounter - previousTp); x++)
			{
				int tpX = (previousTp + x);
				double skew =
					(fpCounter - previousFp)
						/ (double) (tpCounter - previousTp);

				double p = tpX / (tpX + previousFp + (skew * x));
				double r = tpX / (double) pTotal;

				normalize(curve, p, r);
			}
		}

		/**
		 * Adds a new point to the specified curve, normalized by the default
		 * <code>DELTA</code> value.
		 * 
		 * @param curve the curve to add the point.
		 * @param precision the y axis value. 
		 * @param recall the x axis value.
		 */
		private void normalize(Curve curve, double precision, double recall)
		{
			if (curve.isEmpty())
			{
				curve.add(0.0, precision);
			}

			Point p1 = curve.getLast(1);
			Point p2 = curve.getLast(2);

			if (p1 != null && p2 != null
				&& !(p1.getX() == recall && p1.getY() == recall))
			{
				if (Math.abs(p1.getY() - precision) < DELTA
					&& Math.abs(p2.getY() - precision) < DELTA)
				{
					p1.setX(recall);
				}
				else if (Math.abs(p1.getX() - recall) < DELTA
					&& Math.abs(p2.getX() - recall) < DELTA)
				{
					p1.setY(precision);
				}
				else
				{
					curve.add(new Point(recall, precision));
				}
			}
			else
			{
				curve.add(new Point(recall, precision));
			}
		}
	}
}