/*
 * $Id$ 
 */

package net.sf.myra.datamining.statistics;

import static net.sf.myra.datamining.statistics.CurveFactory.NEGATIVE;
import static net.sf.myra.datamining.statistics.CurveFactory.POSITIVE;
import static net.sf.myra.datamining.statistics.Parser.ENABLED;
import static net.sf.myra.datamining.statistics.Parser.ORIGINAL_PREFIX;
import static net.sf.myra.datamining.statistics.Parser.PREDICTED_PREFIX;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Map;
import java.util.TreeMap;
import net.sf.myra.datamining.data.Attribute;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.io.Helper;

/**
 * This class exports the precision and recall values for each class.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class Exporter
{
	/**
	 * Default number formatter.
	 */
	private static final NumberFormat FORMATTER = new DecimalFormat("0.####");

	/**
	 * The output object.
	 */
	private static PrintWriter out = null;

	/**
	 * Runs the exporter.
	 * 
	 * @param args command-line arguments.
	 */
	public static void main(String[] args)
		throws Exception
	{
		if (args.length < 2)
		{
			StringBuffer usage = new StringBuffer();
			usage.append("usage: java ");
			usage.append(Exporter.class.getName());
			usage.append(" <predictions file> <output file>");

			System.out.println(usage.toString());
			System.exit(1);
		}

/*
		Parser parser = new Parser();
		Information info = parser.read(args[0], CurveFactory.PRECISION_RECALL);
		PrintWriter out = new PrintWriter(new FileWriter(args[1]));

		for (Point point : info.getCurve().getPoints())
		{
			out.print(point.getX());
			out.print(",");
			out.println(point.getY());
		}
*/
		out = new PrintWriter(new FileWriter(args[1]));
		print(load(args[0]), thresholds());
		out.close();
	}

	/**
	 * Loads the class-wise predictions.
	 * 
	 * @param file the file to be loaded.
	 * 
	 * @return the class-wise predictions mapping.
	 */
	private static Map<String, ClassInfo> load(String file)
		throws IOException
	{
		Dataset dataset = Helper.open(file);
		Metadata metadata = dataset.getMetadata();

		Map<String, ClassInfo> matrix = new TreeMap<String, ClassInfo>();

		for (Attribute attribute : metadata.getAttributes())
		{
			String name = attribute.getName();

			if (name.startsWith(ORIGINAL_PREFIX))
			{
				name = name.split(ORIGINAL_PREFIX)[1];
				ClassInfo info = new ClassInfo();
				info.instances = new InstanceInfo[dataset.getSize()];
				info.values = new double[dataset.getSize()][2];
				matrix.put(name, info);
			}
		}

		// the instance counter
		int i = 0;

		for (Instance instance : dataset)
		{
			for (Attribute attribute : metadata.getAttributes())
			{
				String name = attribute.getName();

				if (name.startsWith(ORIGINAL_PREFIX))
				{
					name = name.split(ORIGINAL_PREFIX)[1];
					InstanceInfo[] instances = matrix.get(name).instances;
					instances[i] = new InstanceInfo();
					instances[i].index = i;

					boolean flag = ENABLED.equals(instance.getValue(attribute));
					instances[i].values[1] = (flag ? POSITIVE : NEGATIVE);
					matrix.get(name).values[i][1] = instances[i].values[1];

					int index = metadata.getIndex(PREDICTED_PREFIX + name);
					instances[i].values[0] = instance.values()[index];
					matrix.get(name).values[i][0] = instances[i].values[0];
				}
			}

			i++;
		}

		for (String key : matrix.keySet())
		{
			ClassInfo info = matrix.get(key);
			sort(info.instances);
			info.positives = CurveFactory.count(info.values, POSITIVE);
			info.frequency = info.positives / (double) info.values.length;
		}

		return matrix;
	}

	/**
	 * Prints the class-wise precision and recall values for each threshold
	 * value.
	 * 
	 * @param matrix the class-wise prediction mapping.
	 * @param threshold the threshold values to be evaluated.
	 */
	private static void print(Map<String, ClassInfo> matrix, double[] threshold)
	{
		print("Testing error");
		print("-------------");
		print("");
		print("Number of examples: "
			+ matrix.values().iterator().next().values.length);
		print("Coverage:");
		print("   Default: 0");
		print("   Original: 0");

		StringBuffer buffer = new StringBuffer();
		buffer.append("Hierarchical accuracy by class\n");

		Map<Double, Integer> coverage = new TreeMap<Double, Integer>();

		int total = 0;

		for (String key : matrix.keySet())
		{
			ClassInfo info = matrix.get(key);
			total += info.positives;
		}

		for (int i = 0; i < threshold.length; i++)
		{
			buffer
				.append("   T("
					+ (threshold[i] * 100)
					+ ")         : precision: 0, recall: 0, coverage: 0, TP: 0, FP: 0, nbPos: "
					+ total + "\n");
			HashSet<Integer> indexes = new HashSet<Integer>();

			for (String key : matrix.keySet())
			{
				ClassInfo info = matrix.get(key);

				if (info.positives > 0)
				{
					int tp = 0;
					int fp = 0;

					for (int j = 0; j < info.instances.length; j++)
					{
						if (info.instances[j].values[0] < threshold[i])
						{
							break;
						}
						else if (info.instances[j].values[1] == POSITIVE)
						{
							tp++;
						}
						else
						{
							fp++;
						}

						indexes.add(info.instances[j].index);
					}

					if ((tp + fp) > 0)
					{
						double precision = tp / (double) (tp + fp);
						double recall = tp / (double) info.positives;

						buffer.append("      " + key + ", def: "
							+ FORMATTER.format(info.frequency) + ", prec: "
							+ FORMATTER.format(precision) + ", rec: "
							+ FORMATTER.format(recall) + ", TP: " + tp
							+ ", FP: " + fp + ", nbPos: " + info.positives
							+ "\n");
					}
				}
			}

			coverage.put((double) ((int) (threshold[i] * 100)), indexes.size());
		}

		for (Double key : coverage.keySet())
		{
			print("   T(" + key + "): " + coverage.get(key));
		}

		print(buffer.toString());
	}

	/**
	 * Returns the default threshold values.
	 * 
	 * @return the default threshold values.
	 */
	private static double[] thresholds()
	{
		double[] threshold = new double[51];

		for (int i = 0; i < threshold.length; i++)
		{
			threshold[i] = (i * 2) / 100.0;
		}

		return threshold;
	}

	/**
	 * Prints the specified line.
	 * 
	 * @param line the line to be printed.
	 */
	private static void print(String line)
	{
		out.println(line);
	}

	/**
	 * Sorts a bidimensional array (predicted value,class value).
	 * 
	 * @param values the array to be sorted.
	 */
	private static void sort(InstanceInfo[] instances)
	{
		Arrays.sort(instances, new Comparator<InstanceInfo>()
		{
			public int compare(InstanceInfo o1, InstanceInfo o2)
			{
				return Double.compare(o2.values[0], o1.values[0]);
			}
		});
	}

	/**
	 * Struct-like class to hold the instance prediction information.
	 */
	private static class InstanceInfo
	{
		double[] values = new double[2];

		int index;
	}

	/**
	 * Struct-like class to hold the class prediction information.
	 */
	private static class ClassInfo
	{
		double[][] values;

		InstanceInfo[] instances;

		double frequency;

		int positives;
	}
}