/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2009  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining.statistics;

import static net.sf.myra.datamining.statistics.CurveFactory.POSITIVE;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.Collection;
import java.util.Map;
import java.util.TreeMap;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class Information
{
	/**
	 * The key for the avereaged curve.
	 */
	public static final String AVERAGE_CURVE = "Averaged Curve";

	/**
	 * The <code>PropertyChangeSupport</code> object.
	 */
	private PropertyChangeSupport bean = new PropertyChangeSupport(this);

	/**
	 * The average area under the PR curve over all classes.
	 */
	private double average;

	/**
	 * The weighted (based on individual class frequency)  average area under
	 * the PR curve over all classes.
	 */
	private double weighted;

	/**
	 * The average PR curve.
	 */
	private Curve curve;

	/**
	 * The PR curve for each class value.
	 */
	private Map<String, Curve> curves;

	/**
	 * The curve factory reference.
	 */
	private CurveFactory factory;

	/**
	 * Default constructor.
	 */
	Information(CurveFactory factory)
	{
		this.factory = factory;
	}

	/**
	 * Default constructor.
	 * 
	 * @param curves the mapping (class value, doube[][]).
	 * @param dataset the predictions dataset.
	 */
	void load(Map<String, double[][]> mapping)
	{
		int size = mapping.values().iterator().next().length;

		double[][] total = new double[0][0];
		int index = 0;

		double average = 0.0;
		double weighted = 0.0;
		double frequencies = 0.0;
		int count = 0;

		curves = new TreeMap<String, Curve>();

		for (String c : mapping.keySet())
		{
			double[][] v = mapping.get(c);
			double positives = CurveFactory.count(v, POSITIVE);

			if (positives > 0)
			{
				double[][] t = new double[total.length + size][2];
				System.arraycopy(total, 0, t, 0, total.length);
				total = t;

				System.arraycopy(v, 0, total, index, size);
				index += size;

				Curve curve = factory.create(v);
				double area = curve.area();
				average += area;

				double f = positives / v.length;
				weighted += (f * area);
				frequencies += f;

				curves.put(c, curve);
				count++;
			}

			// tries to free some memory
			mapping.put(c, null);

			bean.firePropertyChange("load", null, c);
		}

		this.curve = factory.create(total);
		this.average = (average / count);
		this.weighted = (weighted / frequencies);
	}

	/**
	 * Returns the curve for the specified class label.
	 * 
	 * @param c the class label.
	 * 
	 * @return the curve for the specified class label.
	 */
	public Curve get(String c)
	{
		if (AVERAGE_CURVE.equals(c))
		{
			return curve;
		}

		Curve curve = curves.get(c);

		if (c == null)
		{
			throw new IllegalArgumentException("Cruve not found: " + c);
		}

		return curve;
	}

	/**
	 * Returns the average area under the PR curve over all classes.
	 * 
	 * @return the average area under the PR curve over all classes.
	 */
	public double getAverage()
	{
		return average;
	}

	/**
	 * Returns the weighted (based on individual class frequency)  average area
	 * under the PR curve over all classes.
	 * 
	 * @return the weighted average area under the PR curve
	 */
	public double getWeighted()
	{
		return weighted;
	}

	/**
	 * Returns the average PR curve.
	 * 
	 * @return the average PR curve.
	 */
	public Curve getCurve()
	{
		return curve;
	}

	/**
	 * Returns the collection of available class labels.
	 * 
	 * @return the collection of available class labels.
	 */
	public Collection<String> getAvailable()
	{
		return curves.keySet();
	}

	/**
	 * Adds a property change listener.
	 * 
	 * @param listener the listener to add.
	 */
	void addListener(PropertyChangeListener listener)
	{
		bean.addPropertyChangeListener(listener);
	}
}