/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.StringTokenizer;
import java.util.TreeMap;

import net.sf.myra.datamining.data.Attribute;
import net.sf.myra.datamining.data.ContinuousAttribute;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.data.NominalAttribute;
import net.sf.myra.datamining.io.Helper;
import weka.classifiers.trees.J48;
import weka.core.FastVector;
import weka.core.Instances;

/**
 * This class represents the C45Disc discretization procedure.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class C45Disc
{
	/**
	 * Discretizes the specified dataset.
	 * 
	 * @param dataset the dataset to be discretized.
	 * 
	 * @return the discretized dataset instance.
	 * 
	 * @throws Exception if any problem is encountered.
	 */
	public static Dataset discretize(Dataset dataset)
		throws Exception
	{
		return discretize(dataset, null)[0];
	}

	/**
	 * Discretizes the specified training and test datasets. Note that the
	 * discretization rules are determined using only the training dataset.
	 * 
	 * @param trainingSet the training dataset to be discretized.
	 * @param testSet the test dataset to be discretized.
	 * 
	 * @return a 2-dimensional array, where the first index represents the
	 *         training dataset and the second index represents the test
	 *         dataset.
	 * 
	 * @throws Exception if any problem is encountered.
	 */
	public static Dataset[] discretize(Dataset trainingSet, Dataset testSet)
		throws Exception
	{
		Metadata metadata = trainingSet.getMetadata();
		Map<Attribute, List<Rule>> mapping =
			new LinkedHashMap<Attribute, List<Rule>>();

		// generates the rules for discretization using only the training
		// set information

		for (Attribute attribute : metadata.getPredictor())
		{
			if (attribute instanceof ContinuousAttribute)
			{
				mapping.put(attribute, discretize(trainingSet,
					(ContinuousAttribute) attribute,
					(NominalAttribute) metadata.getTarget()));
			}
		}

		// generates the discretized metadata information

		Metadata discrete = new Metadata();
		discrete.setName(metadata.getName());

		for (Attribute attribute : metadata.getAttributes())
		{
			if (attribute instanceof ContinuousAttribute)
			{
				String[] values = new String[mapping.get(attribute).size()];

				for (int i = 0; i < values.length; i++)
				{
					values[i] = Integer.toString(i);
				}

				discrete.add(new NominalAttribute(attribute.getName(), values));
			}
			else
			{
				discrete.add(attribute);
			}
		}

		discrete.setTarget(metadata.getTarget());
		Dataset[] result = new Dataset[2];

		result[0] = apply(metadata, trainingSet, mapping, discrete);

		if (testSet != null)
		{
			result[1] = apply(metadata, testSet, mapping, discrete.clone());
			result[1].getMetadata().setName(testSet.getMetadata().getName());
		}

		return result;
	}

	/**
	 * Creates discretization rules for the specified continuous attribute.
	 * 
	 * @param dataset the dataset to be discretized.
	 * @param attribute the continuous attribute to be discretized.
	 * @param classAttribute the dataset class attribute.
	 * 
	 * @return a list of discretization rules.
	 * 
	 * @throws Exception if any problem is encountered.
	 */
	private static List<Rule> discretize(Dataset dataset,
			ContinuousAttribute attribute, NominalAttribute classAttribute)
		throws Exception
	{
		// metadata information

		FastVector attributes = new FastVector();
		FastVector classValues = new FastVector();

		attributes.addElement(new weka.core.Attribute(attribute.getName()));

		for (String value : classAttribute.getValues())
		{
			classValues.addElement(value);
		}

		attributes.addElement(new weka.core.Attribute(classAttribute.getName(),
			classValues));

		// instance information

		Instances instances =
			new Instances(dataset.getMetadata().getName(), attributes, dataset
				.getSize());
		instances.setClassIndex(instances.numAttributes() - 1);

		for (Instance instance : dataset)
		{
			weka.core.Instance i = new weka.core.Instance(attributes.size());
			i.setDataset(instances);

			int index = dataset.getMetadata().getIndex(attribute.getName());

			if (instance.values()[index] == -1)
			{
				i.setMissing(instances.attribute(attribute.getName()));
			}
			else
			{
				i.setValue(instances.attribute(attribute.getName()), instance
					.values()[index]);
			}

			if (instance.getLabel().getValues().size() > 1)
			{
				throw new UnsupportedOperationException(
					"Multi-label instace found.");
			}

			index = dataset.getMetadata().getIndex(classAttribute.getName());
			i.setClassValue(instance.getLabel().toString());

			instances.add(i);
		}

		// runs the classifier

		J48 classifier = new J48();
		classifier.setConfidenceFactor(0.01f);

		classifier.buildClassifier(instances);

		return parse(classifier.toString(), attribute.getName());
	}

	/**
	 * Parsers the output of C4.5 to extract discretization rules for the
	 * specified attribute.
	 * 
	 * @param output the C4.5 output.
	 * @param attribute the attribute to be discretized.
	 * 
	 * @return a list of discretization rules.
	 * 
	 * @throws IOException if any problem is encountered.
	 */
	private static List<Rule> parse(String output, String attribute)
		throws IOException
	{
		BufferedReader in = new BufferedReader(new StringReader(output));
		ArrayList<Rule> tree = new ArrayList<Rule>();
		RuleParser parser = new RuleParser();

		String line = null;
		boolean copy = false;

		int index = 0;
		LinkedList<String> terms = new LinkedList<String>();

		while ((line = in.readLine()) != null)
		{
			if (!copy)
			{
				copy = line.startsWith(attribute);
			}

			if (copy)
			{
				if (line.trim().equals(""))
				{
					// end of the decision tree
					break;
				}

				int start = line.indexOf(attribute);
				int end = line.indexOf(":", start);

				if (end == -1)
				{
					if (level(line) == 0)
					{
						terms.clear();
					}

					// it is a partial rule
					terms.add(line.substring(start));
				}
				else
				{
					int level = level(line);

					while (terms.size() != level)
					{
						terms.removeLast();
					}

					StringBuffer rule = new StringBuffer();
					rule.append("IF ");

					if (level > 0)
					{
						for (String term : terms)
						{
							rule.append(term);
							rule.append(" AND ");
						}
					}

					rule.append(line.substring(start, end));
					rule.append(" THEN ");
					rule.append(index);

					tree.add(parser.parse(rule.toString()));
					index++;
				}
			}
		}

		// if we can not find any rule, we add a default rule
		if (tree.isEmpty())
		{
			Rule rule = new Rule();
			rule.setConsequent(0);
			tree.add(rule);
		}

		return tree;
	}

	/**
	 * Determines the level of the specified line in relation to the C4.5
	 * output tree.
	 * 
	 * @param line the line of the output tree.
	 * 
	 * @return the level of the line in relation to the output tree.
	 */
	private static int level(String line)
	{
		int level = 0;

		for (char c : line.toCharArray())
		{
			if (c == '|')
			{
				level++;
			}
		}

		return level;
	}

	/**
	 * Applies the discretization rules to the specified dataset.
	 * 
	 * @param metadata the discrete metadata information.
	 * @param dataset the dataset to be discretized.
	 * @param mapping the discretization rules.
	 * 
	 * @return a discretized dataset instance.
	 */
	private static Dataset apply(Metadata metadata, Dataset dataset,
			Map<Attribute, List<Rule>> mapping, Metadata discrete)
	{
		Dataset discreteDS = new Dataset(discrete);

		for (Instance instance : dataset)
		{
			Instance instanceDS = discreteDS.newInstance();

			for (Attribute attribute : metadata.getAttributes())
			{
				if (attribute instanceof ContinuousAttribute)
				{
					boolean found = false;

					if (!Metadata.MISSING_VALUE.equals(instance
						.getValue(attribute)))
					{
						for (Rule rule : mapping.get(attribute))
						{
							if (rule.isCovered(instance))
							{
								instanceDS.set(attribute.getName(), Integer
									.toString(rule.getConsequent()));
								found = true;
								break;
							}
						}
					}

					if (!found)
					{
						instanceDS.set(attribute.getName(),
							Metadata.MISSING_VALUE);
					}
				}
				else
				{
					instanceDS.set(attribute.getName(), instance
						.getValue(attribute));
				}
			}
		}

		return discreteDS;
	}

	/**
	 * C4.5Disc entry point.
	 * 
	 * @param args command line arguments.
	 * 
	 * @throws Exception if any problem is encountered.
	 */
	public static void main(String[] args)
		throws Exception
	{
		if (args.length < 1)
		{
			StringBuffer usage = new StringBuffer();
			usage.append("usage: java ");
			usage.append(C45Disc.class.getName());
			usage.append(" <input directory> <output directory>");

			System.out.println(usage.toString());
		}
		else
		{
			File directory = new File(args[0]);
			String[] filenames = directory.list(new FilenameFilter()
			{
				public boolean accept(File dir, String name)
				{
					return name.endsWith(".arff");
				}
			});

			Map<String, String> mapping = new TreeMap<String, String>();

			for (int i = 0; i < filenames.length; i++)
			{
				if (filenames[i].startsWith("TR"))
				{
					for (int j = 0; j < filenames.length; j++)
					{
						if (filenames[j].startsWith("TS")
							&& filenames[j].contains(filenames[i].substring(2,
								4)))
						{
							mapping.put(filenames[i], filenames[j]);
						}
					}
				}
			}

			for (Entry<String, String> entry : mapping.entrySet())
			{
				Dataset trainingSet =
					Helper.open(
						new File(args[0], entry.getKey()));
				Dataset testSet =
					Helper.open(
						new File(args[0], entry.getValue()));

				Dataset[] discrete = C45Disc.discretize(trainingSet, testSet);
				trainingSet = discrete[0];

				Helper.getHelper(entry.getKey()).write(new File(args[1]),
					trainingSet);

				testSet = discrete[1];

				for (Instance instance : trainingSet)
				{
					testSet.remove(instance);
				}

				Helper.getHelper(entry.getValue()).write(new File(args[1]),
					testSet);

				System.out.println(entry.getKey() + " " + trainingSet.getSize()
					+ " " + testSet.getSize());
			}
		}
	}

	/**
	 * Discretization rule parser.
	 */
	public static class RuleParser
	{
		private static final String BEGIN = "IF";

		private static final String END = "THEN";

		private static final String CONDITION = "AND";

		public Rule parse(String text)
		{
			StringTokenizer tokens = new StringTokenizer(text);
			boolean consequent = false;
			Rule rule = null;
			ArrayList<String> slices = null;

			while (tokens.hasMoreTokens())
			{
				String token = tokens.nextToken();

				if (consequent)
				{
					rule.setConsequent(Integer.parseInt(trim(token)));
				}
				if (token.equals(BEGIN))
				{
					rule = new Rule();
					slices = new ArrayList<String>();
				}
				else if (token.equals(END) || token.equals(CONDITION))
				{
					if (slices.size() == 3)
					{
						Condition condition = new Condition();
						condition.setAttribute(trim(slices.get(0)));
						condition.setType(ConditionType.convert(slices.get(1)));
						condition.setValue(trim(slices.get(2)));

						slices.clear();
						rule.add(condition);
					}
					else if (slices.size() != 0)
					{
						throw new RuntimeException("Invalid rule input");
					}

					consequent = token.equals(END);
				}
				else
				{
					slices.add(token);
				}
			}

			return rule;
		}

		private String trim(String s)
		{
			return s.replace('\'', ' ').trim();
		}
	}

	/**
	 * This class represents a discretization rule.
	 */
	public static class Rule
	{
		private ArrayList<Condition> antecedent = new ArrayList<Condition>();

		private int consequent = -1;

		public void add(Condition condition)
		{
			antecedent.add(condition);
		}

		public int getConsequent()
		{
			return consequent;
		}

		public void setConsequent(int consequent)
		{
			this.consequent = consequent;
		}

		public boolean isCovered(Instance instance)
		{
			boolean covered = true;

			for (Condition condition : antecedent)
			{
				ContinuousAttribute c =
					(ContinuousAttribute) instance.getMetadata().get(
						condition.getAttribute());
				int index = instance.getMetadata().getIndex(c.getName());

				if ((instance.values()[index] == -1)
					|| !condition.getType().eval(
						instance.getValue(c.getName()), condition.getValue()))
				{
					covered = false;
					break;
				}
			}

			return covered;
		}

		@Override
		public String toString()
		{
			StringBuffer buffer = new StringBuffer();

			for (Condition condition : antecedent)
			{
				buffer.append(condition.toString());
				buffer.append(" ");
				buffer.append("AND");
				buffer.append(" ");
			}

			buffer.delete(buffer.length() - 5, buffer.length());
			return buffer.toString();
		}
	}

	/**
	 * This class represents a rule condition.
	 */
	private static class Condition
	{
		private String attribute;

		private String value;

		private ConditionType type;

		/**
		 * @return the attribute
		 */
		public String getAttribute()
		{
			return attribute;
		}

		/**
		 * @param attribute the attribute to set
		 */
		public void setAttribute(String attribute)
		{
			this.attribute = attribute;
		}

		/**
		 * @return the type
		 */
		public ConditionType getType()
		{
			return type;
		}

		/**
		 * @param type the type to set
		 */
		public void setType(ConditionType type)
		{
			this.type = type;
		}

		/**
		 * @return the value
		 */
		public String getValue()
		{
			return value;
		}

		/**
		 * @param value the value to set
		 */
		public void setValue(String value)
		{
			this.value = value;
		}

		@Override
		public String toString()
		{
			return attribute + " " + type.symbol + " " + value;
		}
	}

	/**
	 * The list of condition types that can be found in a rule.
	 */
	enum ConditionType
	{
		GREATER_THAN(">")
		{
			boolean eval(String x, String y)
			{
				double xValue = new Double(x);
				double yValue = new Double(y);

				return xValue > yValue;
			}
		},
		LESS_EQUAL("<=")
		{
			boolean eval(String x, String y)
			{
				double xValue = new Double(x);
				double yValue = new Double(y);

				return xValue <= yValue;
			}
		};

		private String symbol;

		private ConditionType(String symbol)
		{
			this.symbol = symbol;
		}

		abstract boolean eval(String x, String y);

		public static ConditionType convert(String symbol)
		{
			for (ConditionType type : ConditionType.values())
			{
				if (symbol.equals(type.symbol))
				{
					return type;
				}
			}

			throw new IllegalArgumentException(
				"Enum value not found for symbol: " + symbol);
		}
	}
}