/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining.util;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.data.NominalAttribute;
import net.sf.myra.datamining.io.Helper;
import net.sf.myra.framework.RandomGenerator;

/**
 * This class represents the cross-validation procedure for flat classification
 * datasets.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date: 2008-02-23 18:50:58#$
 */
public class FlatCrossValidation
{
	/**
	 * Splits the dataset into <code>folds</code> partitions.
	 * 
	 * @param dataset the dataset to split.
	 * @param folds the number of folds.
	 * 
	 * @return an array containing <code>folds</code> partitions.
	 */
	public static Partition[] split(Dataset dataset, int folds, Mode mode)
	{
		Partition[] partitions = new Partition[folds];

		for (int i = 0; i < folds; i++)
		{
			partitions[i] = new Partition(i);
		}

		return (mode == Mode.NORMAL) ? normalSplit(dataset, partitions)
			: stratifiedSplit(dataset, partitions);
	}

	/**
	 * Splits the dataset into <code>folds</code> partitions.
	 * 
	 * @param dataset the dataset to split.
	 * @param partitions the partitions to fill up.
	 * 
	 * @return an array containing <code>folds</code> partitions.
	 */
	private static Partition[] normalSplit(Dataset dataset,
			Partition[] partitions)
	{
		RandomGenerator random = RandomGenerator.getInstance();

		while (!dataset.isEmpty())
		{
			for (int i = 0; i < partitions.length; i++)
			{
				if (dataset.isEmpty())
				{
					break;
				}

				int index = random.nextInt(dataset.getSize());
				Instance instance = dataset.get(index);
				dataset.remove(index);
				partitions[i].add(instance);
			}
		}

		return partitions;
	}

	/**
	 * Splits the dataset into <code>folds</code> stratified partitions.
	 * 
	 * @param dataset the dataset to split.
	 * @param partitions the partitions to fill up.
	 * 
	 * @return an array containing <code>folds</code> stratified partitions.
	 */
	private static Partition[] stratifiedSplit(Dataset dataset,
			Partition[] partitions)
	{
		Map<String, Integer> frequencies = new HashMap<String, Integer>();
		NominalAttribute klass =
			(NominalAttribute) dataset.getMetadata().getTarget();

		for (String value : klass.getValues())
		{
			frequencies.put(value, 0);
		}

		for (Instance instance : dataset)
		{
			String label = instance.getLabel().toString();

			int count = frequencies.get(label);
			frequencies.put(label, count + 1);
		}

		RandomGenerator random = RandomGenerator.getInstance();
		int current = 0;

		for (String value : frequencies.keySet())
		{
			int count = frequencies.get(value);

			while (count > 0)
			{
				for (int i = current; i < partitions.length; i++)
				{
					while (count > 0)
					{
						int index = random.nextInt(dataset.getSize());
						Instance instance = dataset.get(index);

						if (value.equals(instance.getLabel().toString()))
						{
							dataset.remove(index);
							partitions[i].add(instance);
							count--;
							current = i;

							break;
						}
					}
				}

				current = ((current + 1) < partitions.length) ? current + 1 : 0;
			}
		}

		return partitions;
	}

	/**
	 * Exports the dataset into <code>folds</code> partitions. For each
	 * partiton, a training and test file is generated.
	 * 
	 * @param dataset the dataset to export.
	 * @param folds the number of folds.
	 * @param directory the target directory.
	 * 
	 * @throws IOException If an I/O error has occurred.
	 */
	public static void export(Dataset dataset, int folds, File directory,
			Mode mode, Helper helper)
		throws IOException
	{
		Partition[] partitions = split(dataset, folds, mode);
		Metadata metadata = dataset.getMetadata();
		String name = metadata.getName();

		for (int i = 0; i < partitions.length; i++)
		{
			// writes the test file for the current partition

			metadata.setName("TS" + (i < 10 ? "0" + i : i) + "_" + name);
			Dataset ts = new Dataset(metadata);
			ts.setInstances(partitions[i].getInstances());

			// removes the test examples which may occur in the training set

			//for (int j = 0; j < partitions.length; j++)
			//{
			//	if (i != j)
			//	{
			//		for (Instance instance : partitions[j].getInstances())
			//		{
			//			ts.remove(instance);
			//		}
			//	}
			//}

			helper.write(directory, ts);

			// writes the training file for the current partition

			List<Instance> instances = new ArrayList<Instance>();

			for (int j = 0; j < partitions.length; j++)
			{
				if (i != j)
				{
					instances.addAll(partitions[j].getInstances());
				}
			}

			metadata.setName("TR" + (i < 10 ? "0" + i : i) + "_" + name);
			Dataset tr = new Dataset(metadata);
			tr.setInstances(instances);

			helper.write(directory, tr);
		}
	}

	/**
	 * CrossValidation entry point.
	 * 
	 * @param args an array of command line arguments.
	 */
	public static void main(String[] args)
		throws Exception
	{
		if (args.length < 2)
		{
			StringBuffer usage = new StringBuffer();
			usage.append("usage: java ");
			usage.append(FlatCrossValidation.class.getName());
			usage.append(" <dataset file> <output directory>");

			System.out.println(usage.toString());
		}
		else
		{
			Helper helper = Helper.getHelper(args[0]);
			Dataset dataset = helper.read(args[0]);

			FlatCrossValidation.export(dataset, 10, new File(args[1]),
				Mode.STRATIFIED, helper);
		}
	}

	/**
	 * Represents the cross-validation mode.
	 */
	public enum Mode
	{
		NORMAL, STRATIFIED;
	}

	/**
	 * This class represents a partition (fold) of the cross-validation.
	 */
	public static class Partition
	{
		/**
		 * The partition fold number.
		 */
		private int fold;

		/**
		 * The partition instance list.
		 */
		private ArrayList<Instance> instances;

		/**
		 * Creates a new partition.
		 * 
		 * @param fold the partition fold number.
		 */
		public Partition(int fold)
		{
			this.fold = fold;
			instances = new ArrayList<Instance>();
		}

		/**
		 * Adds the specified instance to this partition.
		 * 
		 * @param instance the instance to be added.
		 */
		public void add(Instance instance)
		{
			instances.add(instance);
		}

		/**
		 * Returns the partition size.
		 * 
		 * @return the partition size.
		 */
		public int getSize()
		{
			return instances.size();
		}

		/**
		 * Returns the partition instance list.
		 * 
		 * @return the partition instance list.
		 */
		public List<Instance> getInstances()
		{
			return Collections.unmodifiableList(instances);
		}

		/**
		 * Returns the partition fold number.
		 * 
		 * @return the partition fold number.
		 */
		public int getFold()
		{
			return fold;
		}

		@Override
		public String toString()
		{
			return "Partition " + fold + ": size=" + getSize();
		}
	}
}