/*
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2012 Fernando Esteban Barril Otero
 */

package net.sf.myra.datamining.util;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import net.sf.myra.datamining.data.ClassHierarchy;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.data.Node;
import net.sf.myra.datamining.io.Helper;
import net.sf.myra.datamining.util.FlatCrossValidation.Partition;
import net.sf.myra.framework.RandomGenerator;

/**
 * This class represents the cross-validation procedure for hierarchical
 * classification problems.
 * 
 * @author Fernando Esteban Barril Otero
 */
public class HierarchicalCrossValidation
{
	/**
	 * Splits the dataset into <code>folds</code> stratified partitions.
	 * 
	 * @param dataset the dataset to split.
	 * @param folds the number of cross-validation folds.
	 * 
	 * @return an array containing <code>folds</code> stratified partitions.
	 */
	public static Partition[] partition(Dataset dataset, int folds)
	{
		Partition[] partitions = new Partition[folds];

		for (int i = 0; i < folds; i++)
		{
			partitions[i] = new Partition(i);
		}

		Set<String> leaves = new HashSet<String>();
		ClassHierarchy hierarchy = dataset.getMetadata().getClassHierarchy();

		for (Node node : hierarchy.flat())
		{
			if (node.isLeaf())
			{
				leaves.add(node.getLabel());
			}
		}

		HashMap<String, Integer> frequencies = new HashMap<String, Integer>();

		for (Instance instance : dataset.filter(leaves.toArray(new String[0])))
		{
			StringBuffer multilabel = new StringBuffer();

			for (String label : instance.getLabel().getValues())
			{
				if (leaves.contains(label))
				{
					if (multilabel.length() > 0)
					{
						multilabel.append(",");
					}

					multilabel.append(label);
				}
			}

			String label = multilabel.toString();
			Integer count = frequencies.get(label);

			if (count == null)
			{
				count = 0;
			}

			frequencies.put(label, count + 1);
		}

		for (Iterator<String> i = frequencies.keySet().iterator(); i.hasNext();)
		{
			String label = i.next();

			if (frequencies.get(label) < folds)
			{
				i.remove();
			}
		}

		RandomGenerator random = RandomGenerator.getInstance();
		List<String> ordered = new LinkedList<String>(frequencies.keySet());
		Collections.sort(ordered, new Comparator<String>()
		{
			public int compare(String o1, String o2)
			{
				return Double.compare(o2.length(), o1.length());
			}
		});

		for (String label : ordered)
		{
			int limit = frequencies.get(label) / folds;
			List<Instance> instances = dataset.getInstances(label.split(","));

			for (int i = 0; i < partitions.length; i++)
			{
				int counter = 0;

				while (counter < limit)
				{
					int index = random.nextInt(instances.size());
					counter++;

					Instance instance = instances.remove(index);
					dataset.remove(instance);
					partitions[i].add(instance);
				}
			}
		}

		while (!dataset.isEmpty())
		{
			for (int i = 0; i < partitions.length; i++)
			{
				if (dataset.isEmpty())
				{
					break;
				}

				int index = random.nextInt(dataset.getSize());
				partitions[i].add(dataset.get(index));
				dataset.remove(index);
			}
		}

		return partitions;
	}

	/**
	 * Splits the dataset into <code>folds</code> stratified partitions.
	 * 
	 * @param filename the dataset filename to split.
	 * @param folds the number of cross-validation folds.
	 * 
	 * @return an array containing <code>folds</code> stratified
	 *         <code>Dataset</code> instances.
	 *         
	 * @see Dataset
	 */
	public static Dataset[] split(String filename, int folds)
		throws IOException
	{
		return split(Helper.open(filename), folds);
	}

	/**
	 * Splits the dataset into <code>folds</code> stratified partitions.
	 * 
	 * @param dataset the dataset to split.
	 * @param folds the number of cross-validation folds.
	 * 
	 * @return an array containing <code>folds</code> stratified
	 *         <code>Dataset</code> instances.
	 *         
	 * @see Dataset
	 */
	public static Dataset[] split(Dataset dataset, int folds)
		throws IOException
	{
		Metadata metadata = dataset.getMetadata();

		Partition[] partitions = partition(dataset, folds);
		Dataset[] datasets = new Dataset[folds];

		for (int i = 0; i < folds; i++)
		{
			datasets[i] = new Dataset(metadata.clone());
			datasets[i].setInstances(partitions[i].getInstances());
		}

		return datasets;
	}

	/**
	 * Exports the dataset into <code>folds</code> partitions. For each
	 * partiton, a training and test file is generated.
	 * 
	 * @param dataset the dataset to export.
	 * @param directory the target directory.
	 * @param helper the file helper.
	 * @param folds the number of folds.
	 * 
	 * @throws IOException If an I/O error has occurred.
	 */
	public static void export(Dataset dataset, File directory, Helper helper,
			int folds, boolean validation)
		throws IOException
	{
		Partition[] partitions = partition(dataset, folds);
		Metadata metadata = dataset.getMetadata();
		String name = metadata.getName();

		for (int i = 0; i < partitions.length; i++)
		{
			String fold = "" + (i < 10 ? "0" + i : i);

			// writes the test file for the current partition

			metadata.setName("TS" + fold + "_" + name);
			Dataset ts = new Dataset(metadata);
			ts.setInstances(partitions[i].getInstances());
			helper.write(directory, ts);

			// writes the training file for the current partition

			List<Instance> instances = new ArrayList<Instance>();

			for (int j = 0; j < partitions.length; j++)
			{
				if (i != j)
				{
					instances.addAll(partitions[j].getInstances());
				}
			}

			metadata.setName("TR" + fold + "_" + name);
			Dataset tr = new Dataset(metadata);
			tr.setInstances(instances);
			helper.write(directory, tr);

			// writes the training/validation files (if 'validation' flag)
			if (validation)
			{
				Partition[] triple = partition(tr, 3);

				metadata.setName("VS" + fold + "_" + name);
				Dataset va = new Dataset(metadata);
				va.setInstances(triple[0].getInstances());
				helper.write(directory, va);

				instances = new ArrayList<Instance>();
				instances.addAll(triple[1].getInstances());
				instances.addAll(triple[2].getInstances());

				metadata.setName("VR" + fold + "_" + name);
				tr = new Dataset(metadata);
				tr.setInstances(instances);
				helper.write(directory, tr);
			}
		}
	}

	/**
	 * Exports the dataset into training, validation and test set.
	 * 
	 * @param dataset the dataset to export.
	 * @param directory the target directory.
	 * @param helper the file helper.
	 * @param folds the number of folds.
	 * 
	 * @throws IOException If an I/O error has occurred.
	 */
	public static void export(Dataset dataset, File directory, Helper helper,
			int folds)
		throws IOException
	{
		export(dataset, directory, helper, folds, false);
	}

	/**
	 * CrossValidation entry point.
	 * 
	 * @param args an array of command line arguments.
	 */
	public static void main(String[] args)
		throws Exception
	{
		if (args.length < 2)
		{
			StringBuffer usage = new StringBuffer();
			usage.append("usage: java ");
			usage.append(HierarchicalCrossValidation.class.getName());
			usage.append(" <dataset file> <output directory> [-validation]");

			System.out.println(usage.toString());
		}
		else
		{
			Helper helper = Helper.getHelper(args[0]);
			helper.setCompressed(args[0].contains(".zip"));

			Dataset dataset = helper.read(args[0]);
			//dataset.getMetadata().setName(args[3]);

			//if (dataset.hasDuplicates())
			//{
			//	int size = dataset.getSize();
			//	dataset = dataset.withoutDuplicates();

			//	System.out.println("Removing " + (size - dataset.getSize())
			//		+ " duplicated instance(s).");
			//}

			boolean validation =
				(args.length > 2) ? "-validation".equals(args[2]) : false;

			HierarchicalCrossValidation.export(dataset, new File(args[1]),
				helper, 10, validation);
		}
	}
}