/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining.util;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;
import java.util.LinkedList;

import net.sf.myra.datamining.data.Attribute;
import net.sf.myra.datamining.data.ClassHierarchy;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.data.Node;
import net.sf.myra.datamining.data.NominalAttribute;
import net.sf.myra.datamining.io.ArffHelper;
import net.sf.myra.datamining.io.ExtendedArffHelper;

/**
 * Utility class to transform an ARFF file in Holden's format (hierarchical).
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class HoldenARFFConverter
{
	/**
	 * The default class attribute name.
	 */
	private static final String CLASS_ATTRIBUTE = "class";

	/**
	 * The default class separator.
	 */
	private static final String CLASS_SEPARATOR = "_";

	/**
	 * The root class label.
	 */
	private static final String ROOT = "ROOT";

	/**
	 * Returns the dataset instance specified by the <code>filename</code>.
	 * 
	 * @param filename the dataset file.
	 * 
	 * @return the dataset instance specified by the <code>filename</code>.
	 */
	public Dataset read(String filename)
		throws IOException
	{
		return read(new FileReader(filename));
	}

	/**
	 * Returns the dataset instance specified by the <code>reader</code>.
	 * 
	 * @param reader the reader instance.
	 * 
	 * @return the dataset instance specified by the <code>filename</code>.
	 */
	public Dataset read(Reader reader)
		throws IOException
	{
		ArffHelper helper = new ArffHelper();
		Dataset dataset = helper.read(reader);

		LinkedList<String> values =
			new LinkedList<String>(dataset.getMetadata().getTarget()
				.getValues());
		NominalAttribute target = new NominalAttribute(CLASS_ATTRIBUTE);

		ClassHierarchy hierarchy = new ClassHierarchy();
		hierarchy.add(ROOT);

		while (!values.isEmpty())
		{
			String value = values.removeFirst();
			String[] split = value.split(CLASS_SEPARATOR);
			Node parent = hierarchy.getRoot();
			String label = "";

			for (int i = 0; i < split.length; i++)
			{
				label += (i == 0 ? "" : ".") + split[i];
				Node node = parent.getChild(label);

				if (node == null)
				{
					hierarchy.add(label, parent.getLabel());
					parent = hierarchy.find(label);
					target.add(label);
				}
				else
				{
					parent = node;
				}
			}
		}

		// creates the transformed dataset

		Metadata metadata = new Metadata();

		for (Attribute attribute : dataset.getMetadata().getPredictor())
		{
			metadata.add(attribute);
		}

		metadata.add(target);
		metadata.setTarget(target);
		metadata.setName(dataset.getMetadata().getName());
		metadata.setClassHierarchy(hierarchy);

		int index = dataset.getMetadata().getIndex(CLASS_ATTRIBUTE);
		Dataset ds = new Dataset(metadata);

		while (!dataset.isEmpty())
		{
			String[] instance = dataset.get(0).getValues();
			dataset.remove(0);
			instance[index] = instance[index].replace('_', '.');
			ds.add(instance);

		}

		// removes the instance that are not associated with a leaf label
		/*
		Iterator<Instance> iterator = ds.getInstances().iterator();
		Collection<Instance> toRemove = new ArrayList<Instance>();
		
		while (iterator.hasNext())
		{
			Instance instance = iterator.next();
			
			if (!instance.isLeaf())
			{
				toRemove.add(instance);
			}
		}
		
		for (Instance instance : toRemove)
		{
			ds.remove(instance);
		}
		*/

		return ds;
	}

	/**
	 * Runs the converter from the command-line.
	 * 
	 * @param args command-line arguments.
	 */
	public static void main(String[] args)
		throws Exception
	{
		if (args.length < 2)
		{
			System.out.println("usage: java "
				+ HoldenARFFConverter.class.getName() + " <input> <output>");
			System.exit(1);
		}

		Dataset dataset = new HoldenARFFConverter().read(args[0]);
		File file = new File(args[0]);
		String name = file.getName().substring(0, file.getName().indexOf('.'));
		dataset.getMetadata().setName(name);

		new ExtendedArffHelper().write(new File(args[1]), dataset);
	}
}