/*
 * $Id: RuleParser.java 32 2008-03-27 21:00:01Z fernando $
 */

package net.sf.myra.datamining.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

import net.sf.myra.datamining.Evaluator;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.data.NominalAttribute;
import net.sf.myra.datamining.data.Operator;
import net.sf.myra.datamining.data.Term;
import net.sf.myra.datamining.function.LaplaceFunction;
import net.sf.myra.datamining.io.Helper;
import net.sf.myra.framework.Cost;
import net.sf.myra.framework.Vertex;

/**
 * Utility (experimental) to parse rules from the output file.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision: 32 $ $Date:: 2008-03-27 21:00:01#$
 */
public class RuleParser
{
	private static final String BEGIN = "IF";

	private static final String END = "THEN";

	private static final String CONDITION = "AND";

	private static final String RELATION = "Relation:";

	private static final String END_RULES = "Measure";

	public Rule parse(String text)
	{
		StringTokenizer tokens = new StringTokenizer(text);
		boolean consequent = false;
		Rule rule = null;
		ArrayList<String> slices = null;

		while (tokens.hasMoreTokens())
		{
			String token = tokens.nextToken();

			if (consequent)
			{
				rule.setConsequent(trim(token));
				break;
			}
			if (token.equals(BEGIN))
			{
				rule = new Rule();
				slices = new ArrayList<String>();
			}
			else if (token.equals(END) || token.equals(CONDITION))
			{
				if (slices.size() == 3)
				{
					Condition condition = new Condition();
					condition.setAttribute(trim(slices.get(0)));
					condition.setType(ConditionType.convert(slices.get(1)));
					condition.setValue(trim(slices.get(2)));

					slices.clear();
					rule.add(condition);
				}
				else if (slices.size() == 5)
				{
					Condition condition = new Condition();
					condition.setAttribute(trim(slices.get(2)));
					condition
						.setValue(trim(slices.get(0)), trim(slices.get(4)));
					condition.setType(ConditionType.IN_RANGE);

					slices.clear();
					rule.add(condition);
				}
				else if (slices.size() != 0)
				{
					throw new RuntimeException("Invalid rule input");
				}

				consequent = token.equals(END);
			}
			else if ("<empty>".equals(trim(token)))
			{
				// it is ok, we reached the default rule
			}
			else
			{
				slices.add(token);
			}
		}

		return rule;
	}

	private String trim(String s)
	{
		return s.replace('\'', ' ').trim();
	}

	public static void main(String[] args)
		throws Exception
	{
		BufferedReader reader = new BufferedReader(new FileReader(args[0]));
		String line = null;

		RuleParser parser = new RuleParser();
		ArrayList<net.sf.myra.datamining.Rule> rules =
			new ArrayList<net.sf.myra.datamining.Rule>();

		Dataset training = null;
		String relation = null;
		LaplaceFunction function = null;

		double average = 0.0;
		int count = 0;

		while ((line = reader.readLine()) != null)
		{
			if (line.startsWith(RELATION))
			{
				relation = line.substring(RELATION.length() + 1).trim();
				training = Helper.open(new File(args[1], relation + ".arff"));
				function = new LaplaceFunction(training);

				rules.clear();
			}
			else if (line.startsWith(BEGIN))
			{
				net.sf.myra.datamining.Rule rule =
					parser.parse(line).convert(training.getMetadata());
				rule.setConfidence(Evaluator
					.getConfidenceFactor(rule, training));
				rule.setCost(function.evaluate(rule));

				rules.add(rule);
			}
			else if (line.startsWith(END_RULES) && !rules.isEmpty())
			{
				Dataset test =
					Helper.open(new File(args[1], relation.replace("TR", "TS")
						+ ".arff"));
				
				int type = Integer.parseInt(args[2]);
				
				switch (type)
				{
					case 0:
						average += averageOrderedList(test, rules);
						break;
						
					case 1:
						average += averageConfidenceSet(test, rules);
						break;
						
					case 2:
						average += averageFrequencySet(test, rules);
						break;
				}

				count++;	
			}
		}
		
		System.out.println(average / count);
		reader.close();

		/*
		String rule =
			"IF length = 2103.0000-3111.0000 AND weight = 2103.0000-3111.0000 THEN GO:0005215";
		//String rule = "IF <empty> THEN GO:5234";
		System.out.println(new RuleParser().parse(rule));
		ZipFile zip = new ZipFile(new File(args[0]), ZipFile.OPEN_READ);
		Enumeration entries = zip.entries();

		while (entries.hasMoreElements())
		{
			ZipEntry entry = (ZipEntry) entries.nextElement();

			if (entry.getName().endsWith(".out"))
			{
				File temp = File.createTempFile("PARSER", ".zip");
				temp.deleteOnExit();

				PrintStream writer =
					new PrintStream(new FileOutputStream(temp));
				BufferedInputStream reader =
					new BufferedInputStream(zip.getInputStream(entry));
				byte[] buffer = new byte[4096];
				int read = -1;

				while ((read = reader.read(buffer)) != -1)
				{
					writer.write(buffer, 0, read);
				}

				writer.close();
				reader.close();

				BufferedReader outfile =
					new BufferedReader(new FileReader(temp));
				String line = null;
				RuleParser parser = new RuleParser();

				int rules = 0;
				int terms = 0;

				while ((line = outfile.readLine()) != null)
				{
					if (line.startsWith("IF"))
					{
						terms += parser.parse(line).getSize();
						rules++;
					}
				}

				System.out.println(entry.getName() + " :: " + rules + " - "
					+ terms);
			}
		}*/
	}

	public static double averageOrderedList(Dataset test,
		ArrayList<net.sf.myra.datamining.Rule> rules)
	{
		double terms = 0.0;

		for (Instance instance : test)
		{
			boolean found = false;

			for (net.sf.myra.datamining.Rule rule : rules)
			{
				terms += rule.getSize();

				if (rule.isCovered(instance))
				{
					found = true;
					break;
				}
			}

			if (!found)
			{
				throw new RuntimeException("Example not covered.");
			}
		}

		return terms / test.getSize();
	}
	
	public static double averageConfidenceSet(Dataset test,
		ArrayList<net.sf.myra.datamining.Rule> rules)
	{
		double terms = 0.0;

		for (Instance instance : test)
		{
			net.sf.myra.datamining.Rule best = null;

			for (net.sf.myra.datamining.Rule rule : rules)
			{
				if (rule.isCovered(instance))
				{
					if ((best == null) || (best.getCost() == null))
					{
						best = rule;
					}
					else
					{
						Cost cost = rule.getCost();

						if ((cost != null)
							&& (cost.compareTo(best.getCost()) == 1))
						{
							best = rule;
						}
					}
				}
			}

			if (best == null)
			{
				//throw new RuntimeException("Example not covered.");
			}
			else
			{
				terms += best.getSize();
			}
		}

		return terms / test.getSize();
	}
	
	public static double averageFrequencySet(Dataset test,
		ArrayList<net.sf.myra.datamining.Rule> rules)
	{
		double terms = 0.0;

		for (Instance instance : test)
		{
			boolean found = false;

			for (net.sf.myra.datamining.Rule rule : rules)
			{
				if (rule.isCovered(instance))
				{
					terms += rule.getSize();
					found = true;
				}
			}
			
			if (!found)
			{
				//throw new RuntimeException("Example not covered.");
			}
		}

		return terms / test.getSize();
	}

	/**
	 * This class represents a classification rule.
	 */
	public static class Rule
	{
		private ArrayList<Condition> antecedent = new ArrayList<Condition>();

		private String consequent;

		public void add(Condition condition)
		{
			antecedent.add(condition);
		}

		public List<Condition> getAntecedent()
		{
			return antecedent;
		}

		public String getConsequent()
		{
			return consequent;
		}

		public void setConsequent(String consequent)
		{
			this.consequent = consequent;
		}

		public int getSize()
		{
			return antecedent.size();
		}

		/*
				public boolean covered(Instances dataset, Instance instance)
				{
					boolean covered = true;

					for (Condition condition : antecedent)
					{
						Attribute attribute =
							dataset.attribute(condition.getAttribute());
						double value = instance.value(attribute);

						if (attribute.isNominal())
						{
							covered = condition.getType().eval(attribute.value((int) value),
								condition.getValue());
						}
						else
						{
							covered = condition.getType().eval(Double.toString(value),
								condition.getValue());
						}
						
						if (!covered)
						{
							// the instance is not covered by one of the conditions
							break;
						}
					}

					return covered;
				}
		*/
		public String toString()
		{
			return Integer.toString(antecedent.size());
		}

		public net.sf.myra.datamining.Rule convert(Metadata metadata)
		{
			net.sf.myra.datamining.Rule rule =
				new net.sf.myra.datamining.Rule(metadata);

			for (Condition condition : getAntecedent())
			{
				int index = metadata.getIndex(condition.getAttribute());
				Term term = new Term(condition.getAttribute(), index);
				Vertex<Term> vertex = null;

				if (condition.getType() == ConditionType.EQUAL)
				{
					NominalAttribute n = (NominalAttribute) metadata.get(index);

					term.setValues(n.getIndex(condition.getValue()[0]));
					term.setDescription(condition.getValue());
					term.setOperator(Operator.EQUAL_TO);

					vertex =
						new Vertex<Term>(n.toString(condition.getValue()[0]));
					vertex.setInfo(term);

				}
				else if (condition.getType() == ConditionType.IN_RANGE)
				{
					term.setValues(Double.parseDouble(condition.getValue()[0]),
						Double.parseDouble(condition.getValue()[1]));
					term.setOperator(Operator.IN_RANGE);

					vertex = new Vertex<Term>(condition.getAttribute());
					vertex.setInfo(term);
				}
				else
				{
					term.setValues(Double.parseDouble(condition.getValue()[0]));
					Operator operator = null;

					switch (condition.getType())
					{
						case GREATER_THAN:
							operator = Operator.GREATER_THAN;
							break;

						case GREATER_THAN_EQUAL_TO:
							operator = Operator.GREATER_THAN_EQUAL_TO;
							break;

						case LESS_THAN:
							operator = Operator.LESS_THAN;
							break;

						case LESS_THAN_EQUAL_TO:
							operator = Operator.LESS_THAN_EQUAL_TO;
							break;

						case EQUAL:
							operator = Operator.EQUAL_TO;
							break;

						case IN_RANGE:
							operator = Operator.IN_RANGE;
							break;
					}

					term.setOperator(operator);

					vertex = new Vertex<Term>(condition.getAttribute());
					vertex.setInfo(term);
				}

				rule.add(vertex);
			}

			rule.setConsequent(metadata.getLabel().toLabel(consequent));

			return rule;
		}
	}

	/**
	 * This class represents a rule condition.
	 */
	public static class Condition
	{
		private String attribute;

		private String[] value;

		private ConditionType type;

		/**
		 * @return the attribute
		 */
		public String getAttribute()
		{
			return attribute;
		}

		/**
		 * @param attribute the attribute to set
		 */
		public void setAttribute(String attribute)
		{
			this.attribute = attribute;
		}

		/**
		 * @return the type
		 */
		public ConditionType getType()
		{
			return type;
		}

		/**
		 * @param type the type to set
		 */
		public void setType(ConditionType type)
		{
			this.type = type;
		}

		/**
		 * @return the value
		 */
		public String[] getValue()
		{
			return value;
		}

		/**
		 * @param value the value to set
		 */
		public void setValue(String... value)
		{
			this.value = value;
		}
	}

	/**
	 * The list of condition types that can be found in a rule.
	 */
	enum ConditionType
	{
		GREATER_THAN(">")
		{
			boolean eval(String x, String y)
			{
				double xValue = new Double(x);
				double yValue = new Double(y);

				return xValue > yValue;
			}
		},
		GREATER_THAN_EQUAL_TO(">=")
		{
			boolean eval(String x, String y)
			{
				double xValue = new Double(x);
				double yValue = new Double(y);

				return xValue >= yValue;
			}
		},
		LESS_THAN("<")
		{
			boolean eval(String x, String y)
			{
				double xValue = new Double(x);
				double yValue = new Double(y);

				return xValue <= yValue;
			}
		},
		LESS_THAN_EQUAL_TO("<=")
		{
			boolean eval(String x, String y)
			{
				double xValue = new Double(x);
				double yValue = new Double(y);

				return xValue <= yValue;
			}
		},
		EQUAL("=")
		{
			boolean eval(String x, String y)
			{
				try
				{
					double value = new Double(x);
					return range(value, y);
				}
				catch (NumberFormatException e)
				{
					return (x.compareToIgnoreCase(y) == 0);
				}
			}
		},
		IN_RANGE("< <=")
		{
			boolean eval(String x, String y)
			{
				throw new UnsupportedOperationException();
			}
		};

		private String symbol;

		private ConditionType(String symbol)
		{
			this.symbol = symbol;
		}

		abstract boolean eval(String x, String y);

		boolean range(double value, String range)
		{
			try
			{
				new Double(range);
				throw new RuntimeException("Invalid range value: " + range);
			}
			catch (NumberFormatException e)
			{
				// we may be dealing with a range values
				double l = new Double(range.substring(0, range.indexOf('-')));
				double h = new Double(range.substring(range.indexOf('-') + 1));

				return (l <= value) && (value <= h);
			}
		}

		public static ConditionType convert(String symbol)
		{
			for (ConditionType type : ConditionType.values())
			{
				if (symbol.equals(type.symbol))
				{
					return type;
				}
			}

			throw new IllegalArgumentException(
				"Enum value not found for symbol: " + symbol);
		}
	}
}