/*
 * $Id: HoldenARFFConverter.java 2148 2009-06-14 17:56:57Z febo $
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.datamining.util;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;
import java.util.LinkedList;

import net.sf.myra.datamining.data.Attribute;
import net.sf.myra.datamining.data.ClassHierarchy;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.data.Node;
import net.sf.myra.datamining.data.NominalAttribute;
import net.sf.myra.datamining.io.ArffHelper;
import net.sf.myra.datamining.io.ExtendedArffHelper;

/**
 * Utility class to transform an ARFF file in Silla's format (hierarchical).
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision: 2148 $ $Date:: 2009-06-14 18:56:57#$
 */
public class SillaARFFConverter
{
	/**
	 * The default class attribute name.
	 */
	private static final String CLASS_ATTRIBUTE = "class";

	/**
	 * The default class separator.
	 */
	private static final String CLASS_SEPARATOR = "/";

	/**
	 * The root class label.
	 */
	private static final String ROOT = "root";

	/**
	 * Returns the dataset instance specified by the <code>filename</code>.
	 * 
	 * @param filename the dataset file.
	 * 
	 * @return the dataset instance specified by the <code>filename</code>.
	 */
	public Dataset read(String filename)
		throws IOException
	{
		return read(new FileReader(filename));
	}

	/**
	 * Returns the dataset instance specified by the <code>reader</code>.
	 * 
	 * @param reader the reader instance.
	 * 
	 * @return the dataset instance specified by the <code>filename</code>.
	 */
	public Dataset read(Reader reader)
		throws IOException
	{
		ArffHelper helper = new ArffHelper();
		Dataset dataset = helper.read(reader);

		LinkedList<String> values =
			new LinkedList<String>(dataset.getMetadata().getTarget()
				.getValues());
		NominalAttribute target = new NominalAttribute(CLASS_ATTRIBUTE);

		ClassHierarchy hierarchy = new ClassHierarchy();
		hierarchy.add(ROOT);

		while (!values.isEmpty())
		{
			String value = values.removeFirst();
			String[] split = value.split("\\" + ".");
			Node parent = hierarchy.getRoot();

			if (!"R".equals(split[0]))
			{
				throw new IllegalArgumentException("Invalid class value: "
					+ value);
			}

			String label = "";

			for (int i = 1; i < split.length; i++)
			{
				label += (label.length() > 0 ? CLASS_SEPARATOR : "") + split[i];
				Node node = parent.getChild(label);

				if (node == null)
				{
					hierarchy.add(label, parent.getLabel());
					parent = hierarchy.find(label);
					target.add(label);
				}
				else
				{
					parent = node;
				}
			}
		}

		// creates the transformed dataset

		Metadata metadata = new Metadata();

		for (Attribute attribute : dataset.getMetadata().getPredictor())
		{
			metadata.add(attribute);
		}

		metadata.add(target);
		metadata.setTarget(target);
		metadata.setName(dataset.getMetadata().getName());
		metadata.setClassHierarchy(hierarchy);

		//int index = dataset.getMetadata().getIndex(CLASS_ATTRIBUTE);
		Dataset ds = new Dataset(metadata);

		while (!dataset.isEmpty())
		{
			String[] instance = dataset.get(0).getValues();
			dataset.remove(0);
			//String label = instance[index];
			//instance[index] =
			//	label.substring(label.lastIndexOf(CLASS_SEPARATOR) + 1);
			String label = instance[instance.length - 1];
			label = label.substring(label.indexOf("R.") + 2);
			instance[instance.length - 1] = label.replace(".", CLASS_SEPARATOR);

			ds.add(instance);
		}

		return ds;
	}

	/**
	 * Runs the converter from the command-line.
	 * 
	 * @param args command-line arguments.
	 */
	public static void main(String[] args)
		throws Exception
	{
		// args[0] = source directory
		// args[1] = pattern
		// args[2] = target directory

		if (args.length < 3)
		{
			System.out.println("usage: java "
				+ SillaARFFConverter.class.getName()
				+ " <source> <pattern> <target>");
			System.exit(1);
		}

		if (args[1].contains("fold"))
		{
			args[1] = args[1].substring(0, args[1].indexOf("-fold"));
		}

		final String PATTERN = args[1] + "-fold-";
		final String SUFFIX = ".harff";

		for (int i = 0;; i++)
		{
			String fold = Integer.toString(i);
			String fullFold = (i < 10) ? "0" + i : Integer.toString(i);

			File tr = new File(args[0], PATTERN + fold + "-train" + SUFFIX);
			File vr = new File(args[0], PATTERN + fold + "-subTrain" + SUFFIX);
			File vs = new File(args[0], PATTERN + fold + "-subValid" + SUFFIX);
			File ts = new File(args[0], PATTERN + fold + "-test" + SUFFIX);

			if (tr.exists() && ts.exists())
			{
				Dataset trSet =
					new SillaARFFConverter().read(tr.getAbsolutePath());
				Dataset tsSet =
					new SillaARFFConverter().read(ts.getAbsolutePath());

				File directory = new File(args[2]);

				trSet.getMetadata().setName("TR" + fullFold + "_" + args[1]);
				new ExtendedArffHelper().write(directory, trSet);

				tsSet.getMetadata().setName("TS" + fullFold + "_" + args[1]);
				new ExtendedArffHelper().write(directory, tsSet);

				if (vr.exists() && vs.exists())
				{
					Dataset vrSet =
						new SillaARFFConverter().read(vr.getAbsolutePath());
					Dataset vsSet =
						new SillaARFFConverter().read(vs.getAbsolutePath());

					vrSet.getMetadata()
						.setName("VR" + fullFold + "_" + args[1]);
					new ExtendedArffHelper().write(directory, vrSet);

					vsSet.getMetadata()
						.setName("VS" + fullFold + "_" + args[1]);
					new ExtendedArffHelper().write(directory, vsSet);
				}
			}
			else
			{
				break;
			}
		}
	}
}