/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.framework;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

/**
 * This class represents a colony of artificial ants.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class Colony
{
	/**
	 * The property key under which the colony size value is stored.
	 */
	public static final String COLONY_SIZE = "net.sf.myra.framework.colony";

	/**
	 * The default colony size.
	 */
	public static final int DEFAULT_COLONY_SIZE = 1;

	/**
	 * The colony size.
	 */
	private int size;

	/**
	 * The colony population.
	 */
	private List<Ant> population;

	/**
	 * The problem trail factory.
	 */
	private TrailFactory factory;

	/**
	 * Default constructor. 
	 */
	public Colony()
	{
		this(Configuration.getIntProperty(COLONY_SIZE, DEFAULT_COLONY_SIZE));
	}

	/**
	 * Creates a new colony.
	 *  
	 *  @param size the colony size.
	 */
	public Colony(int size)
	{
		this.size = size;

		// initializes the propulation
		initialize();
	}

	/**
	 * Initializes the population of ants.
	 */
	protected void initialize()
	{
		population = new LinkedList<Ant>();

		for (int i = 0; i < size; i++)
		{
			population.add(new Ant(this));
		}
	}

	/**
	 * Returns the colony size.
	 * 
	 * @return the colony size.
	 */
	public int getSize()
	{
		return size;
	}

	/**
	 * Returns the trail factory.
	 * 
	 * @return the trail factory.
	 */
	public TrailFactory getTrailFactory()
	{
		return factory;
	}

	/**
	 * Sets the trail factory.
	 * 
	 * @param factory the <code>TrailFactory</code> to set.
	 */
	public void setTrailFactory(TrailFactory factory)
	{
		this.factory = factory;
	}

	/**
	 * Returns the population of ants.
	 * 
	 * @return the population of ants.
	 */
	public List<Ant> getPopulation()
	{
		return Collections.unmodifiableList(population);
	}

	/**
	 * Returns the best ant (the ant with the highest trail) of the colony.
	 * 
	 * @return the best ant of the colony.
	 */
	public Ant getBest()
	{
		Ant best = null;

		for (Ant ant : getPopulation())
		{
			if ((best == null) || (ant.compareTo(best) == 1))
			{
				best = ant;
			}
		}

		return best;
	}

	/**
	 * Returns the current best ant (the ant with the highest trail) of the
	 * colony.
	 * 
	 * @return the current best ant of the colony.
	 */
	public Ant getCurrent()
	{
		Ant best = null;

		for (Ant ant : getPopulation())
		{
			if (best == null)
			{
				best = ant;
			}
			else if (ant.getTrail() != null)
			{
				if ((best.getTrail() != null)
					&& (ant.getTrail().getCost().compareTo(
						best.getTrail().getCost()) == 1))
				{
					best = ant;
				}
			}
		}

		return best;
	}

	@Override
	public String toString()
	{
		StringWriter output = new StringWriter();
		PrintWriter w = new PrintWriter(output);

		for (Ant ant : population)
		{
			w.println(ant);
		}

		w.close();
		return output.toString();
	}
}