/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.framework;

import java.util.LinkedList;
import java.util.List;

/**
 * This class represents a stop condition made by the combination of several
 * stop conditions.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date: 2008-02-08 15:54:20#$
 */
public class CompoundStopCondition
	implements StopCondition
{
	/**
	 * Represents the type of the test applied. When <code>Type.ANY</code> is
	 * used, the {@link #isReached(Environment)} method returns
	 * <code>true</code> if any condition is reached. When
	 * <code>Type.ALL</code>, the {@link #isReached(Environment)} method
	 * returns <code>true</code> only if all conditions are reached. Default
	 * type used is <code>Type.ANY</code>.
	 */
	public enum Type
	{
		ANY, ALL
	};

	private Type type;

	private List<StopCondition> conditions;

	/**
	 * Creates a new <code>CompoundStopCondition</code> instance.
	 */
	public CompoundStopCondition()
	{
		this(Type.ANY);
	}

	/**
	 * Creates a new <code>CompoundStopCondition</code> instance.
	 * 
	 * @param type the stoping criteria type.
	 */
	public CompoundStopCondition(Type type)
	{
		this.type = type;
		conditions = new LinkedList<StopCondition>();
	}

	/**
	 * Adds the specified condition.
	 * 
	 * @param condition the condition to add.
	 */
	public void add(StopCondition condition)
	{
		conditions.add(condition);
	}

	/**
	 * Removes the specified condition.
	 * 
	 * @param condition the condition to remove.
	 */
	public void remove(StopCondition condition)
	{
		conditions.remove(condition);
	}

	public boolean isReached(Environment environment)
	{
		int reached = 0;
		
		for (StopCondition condition : conditions)
		{
			if (condition.isReached(environment))
			{
				if (type.equals(Type.ANY))
				{
					return true;
				}
				
				reached++;
			}
		}

		return (reached == conditions.size());
	}
}