/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.framework;

import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.LogManager;
import java.util.logging.Logger;

/**
 * Helper class to get configuration property values.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date: 2008-02-08 15:54:20#$
 */
public final class Configuration
{
	/**
	 * The property key under which the config file name is stored.
	 */
	public static final String CONFIG_FILE = "net.sf.myra.framework.config";

	/**
	 * The default configuration file name.
	 */
	private static final String DEFAULT_CONFIG_FILE = "/myra.properties";

	/**
	 * The message (help) file name.
	 */
	private static final String MESSAGE_FILE = "myra-framework.messages";

	/**
	 * The log configuration file.
	 */
	private static final String LOG_FILE = "/logging.properties";

	// initialises the logging configuration
	static
	{
		InputStream configuration =
			Configuration.class.getResourceAsStream(LOG_FILE);

		if (configuration == null)
		{
			// turns off logging
			LogManager.getLogManager().getLogger("").setLevel(Level.OFF);
		}
		else
		{
			try
			{
				LogManager.getLogManager().readConfiguration(configuration);
			}
			catch (IOException e)
			{
				throw new RuntimeException(e);
			}
		}
	}

	/**
	 * The singleton configuration instance.
	 */
	private static final Configuration singleton = new Configuration();

	/**
	 * The configuration file properties.
	 */
	private Properties properties;

	/**
	 * The runtime properties.
	 */
	private HashMap<String, String> runtime;

	/**
	 * Default private constructor.
	 */
	private Configuration()
	{
		String name = System.getProperty(CONFIG_FILE, DEFAULT_CONFIG_FILE);

		if (!name.startsWith("/"))
		{
			name = "/" + name;
		}

		runtime = new HashMap<String, String>();
		properties = new Properties();

		try
		{
			properties.load(Configuration.class.getResourceAsStream(name));
		}
		catch (Exception e)
		{
			Logger.getLogger(getClass().getName()).warning(
				"Error reading configuration file: " + name);
		}
	}

	/**
	 * Returns the singleton instance.
	 * 
	 * @return the singleton instance.
	 */
	private static Configuration getInstance()
	{
		return singleton;
	}

	/**
	 * Returns the property indicated by the specified key.
	 * 
	 * @param key the name of the system property.
	 * 
	 * @return the property indicated by the specified key, or <code>null</code>
	 *         if there is no property with that key.
	 */
	public static String getProperty(String key)
	{
		// (1) checks the runtime properties
		String value = getInstance().runtime.get(key);

		if (value == null)
		{
			// (2) checks the system properties 
			value = System.getProperty(key);

			// (3) checks the properties file
			if (value == null)
			{
				return getInstance().properties.getProperty(key);
			}
		}

		return value;
	}

	/**
	 * Returns the property indicated by the specified key as an integer value.
	 * 
	 * @param key the name of the system property.
	 * 
	 * @return the property indicated by the specified key, or <code>null</code>
	 *         if there is no property with that key.
	 */
	public static int getIntProperty(String key)
	{
		return Integer.parseInt(getProperty(key));
	}

	/**
	 * Returns the property indicated by the specified key as an long value.
	 * 
	 * @param key the name of the system property.
	 * 
	 * @return the property indicated by the specified key, or <code>null</code>
	 *         if there is no property with that key.
	 */
	public static long getLongProperty(String key)
	{
		return Long.parseLong(getProperty(key));
	}

	/**
	 * Returns the property indicated by the specified key as a double value.
	 * 
	 * @param key the name of the system property.
	 * 
	 * @return the property indicated by the specified key.
	 */
	public static double getDoubleProperty(String key)
	{
		return Double.parseDouble(getProperty(key));
	}

	/**
	 * Returns the property indicated by the specified key as a boolean value.
	 * 
	 * @param key the name of the system property.
	 * 
	 * @return the property indicated by the specified key.
	 */
	public static boolean getBooleanProperty(String key)
	{
		return Boolean.parseBoolean(getProperty(key));
	}

	/**
	 * Returns the property indicated by the specified key.
	 * 
	 * @param key the name of the system property.
	 * @param defaultValue a default value.
	 * 
	 * @return the property indicated by the specified key, or the default value
	 *         if there is no property with that key.
	 */
	public static String getProperty(String key, String defaultValue)
	{
		String value = getProperty(key);
		return (value == null ? defaultValue : value);
	}

	/**
	 * Returns the property indicated by the specified key as an integer value.
	 * 
	 * @param key the name of the system property.
	 * @param defaultValue a default value.
	 * 
	 * @return the property indicated by the specified key, or the default value
	 *         if there is no property with that key.
	 */
	public static int getIntProperty(String key, int defaultValue)
	{
		return Integer
			.parseInt(getProperty(key, Integer.toString(defaultValue)));
	}

	/**
	 * Returns the property indicated by the specified key as an long value.
	 * 
	 * @param key the name of the system property.
	 * @param defaultValue a default value.
	 * 
	 * @return the property indicated by the specified key, or the default value
	 *         if there is no property with that key.
	 */
	public static long getLongProperty(String key, int defaultValue)
	{
		return Long.parseLong(getProperty(key, Long.toString(defaultValue)));
	}

	/**
	 * Returns the property indicated by the specified key as an double value.
	 * 
	 * @param key the name of the system property.
	 * @param defaultValue a default value.
	 * 
	 * @return the property indicated by the specified key, or the default value
	 *         if there is no property with that key.
	 */
	public static double getDoubleProperty(String key, double defaultValue)
	{
		return Double.parseDouble(getProperty(key,
			Double.toString(defaultValue)));
	}

	/**
	 * Returns the property indicated by the specified key as a boolean value.
	 * 
	 * @param key the name of the system property.
	 * @param defaultValue a default value.
	 * 
	 * @return the property indicated by the specified key, or the default value
	 *         if there is no property with that key.
	 */
	public static boolean getBooleanProperty(String key, boolean defaultValue)
	{
		return Boolean.parseBoolean(getProperty(key,
			Boolean.toString(defaultValue)));
	}

	/**
	 * Checks if the property has a value associated.
	 * 
	 * @param key the property key.
	 * 
	 * @return <code>true</code> if the property has a value associated;
	 *         otherwise <code>false</code>.
	 *         
	 * @see #getProperty(String)
	 */
	public static boolean isPresent(String key)
	{
		return getProperty(key) != null;
	}

	/**
	 * Returns the configuration property options.
	 * 
	 * @return the configuration property options.
	 */
	public static Properties getOptions()
	{
		try
		{
			Properties messages = new Properties();
			messages
				.load(Configuration.class.getResourceAsStream(MESSAGE_FILE));

			return messages;
		}
		catch (Exception e)
		{
			throw new RuntimeException("Messages file not found: "
				+ MESSAGE_FILE);
		}
	}

	public static void setProperty(String property, String value)
	{
		getInstance().runtime.put(property, value);
	}

	public static void setProperty(String property, boolean value)
	{
		getInstance().runtime.put(property, Boolean.toString(value));
	}

	public static void setProperty(String property, Double value)
	{
		getInstance().runtime.put(property, Double.toString(value));
	}

	public static void setProperty(String property, int value)
	{
		getInstance().runtime.put(property, Integer.toString(value));
	}

	public static void setProperty(String property, long value)
	{
		getInstance().runtime.put(property, Long.toString(value));
	}
}