/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.framework;

import java.util.Collection;
import java.util.Collections;
import java.util.LinkedList;

/**
 * Default implementation of the {@link TrailFactory} interface. The factory
 * creates a trail using the <code>TRAIL_CLASS_KEY</code> and select a vertex
 * to add to the trail according to the <code>DecisionPolicy</code> instance.
 * The returned vertex is then evaluated by the constraints. If no constraint
 * is violated, it is added to the trail; otherwise, a new vertex is selected.
 * 
 * This process goes until a {@link NoAvailableVertexException} is thrown by
 * the <code>DecisionPolicy</code> instance.
 * 
 * @see DecisionPolicy
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class DefaultTrailFactory
	implements TrailFactory
{
	/**
	 * The property key under which the trail class value is stored.
	 */
	private static final String TRAIL_CLASS_KEY = "net.sf.myra.framework.trail";

	/**
	 * The default trail class name.
	 */
	private static final String DEFAULT_TRAIL_CLASS =
		DefaultTrail.class.getName();

	/**
	 * The trail construction policy.
	 */
	private DecisionPolicy policy;

	/**
	 * The problem graph.
	 */
	private Graph<?> graph;

	/**
	 * The trail construction constraints.
	 */
	private Collection<Constraint> constraints;

	/**
	 * Default constructor.
	 * 
	 * @param policy the trail construction policy.
	 */
	public DefaultTrailFactory(Graph<?> graph, DecisionPolicy policy)
	{
		this.graph = graph;
		this.policy = policy;
		constraints = new LinkedList<Constraint>();
	}

	/**
	 * Adds a trail construction constraint. 
	 * 
	 * @param constraint the constraint to be added.
	 */
	public void addConstraint(Constraint constraint)
	{
		constraints.add(constraint);
	}

	/**
	 * Returns the list of trail construction constraints.
	 * 
	 * @return the list of trail construction constraints.
	 */
	public Collection<Constraint> getConstraints()
	{
		return Collections.unmodifiableCollection(constraints);
	}

	/**
	 * Returns the trail construction policy.
	 * 
	 * @return the trail construction policy.
	 */
	public DecisionPolicy getDecisionPolicy()
	{
		return policy;
	}

	/**
	 * Returns the problem graph.
	 * 
	 * @return the problem graph.
	 */
	public Graph<?> getGraph()
	{
		return graph;
	}

	public Trail createTrail()
	{
		Trail trail =
			(Trail) ObjectFactory.create(Configuration.getProperty(
				TRAIL_CLASS_KEY, DEFAULT_TRAIL_CLASS));

		try
		{
			// iteratively selects a node based on the partial trail, checks
			// it against the problem constraints and adds it to the trail
			while (true)
			{
				Vertex<?> vertex = policy.select(getGraph(), trail);

				try
				{
					for (Constraint constraint : constraints)
					{
						constraint.check(trail, vertex);
					}

					trail.add(vertex);
				}
				catch (UnsatisfiedConstraintException e)
				{
					// the vertex cannot be added to the trail
				}
			}
		}
		catch (NoAvailableVertexException e)
		{
			// there aren't available vertices to select
		}

		return trail;
	}
}