/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.framework;

/**
 * This class represents an edge (connection) of the problem construction
 * graph.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date: 2008-04-22 10:15:14#$
 */
public class Edge<V extends Info>
	implements Comparable<Edge<V>>, Weighable
{
	/**
	 * The edge's start.
	 */
	private Vertex<V> from;

	/**
	 * The edge's end.
	 */
	private Vertex<V> to;

	/**
	 * The wrapped weighted component.
	 */
	private Weighable component;

	/**
	 * Creates a new edge with the specified start and end nodes.
	 * 
	 * @param from the start node.
	 * @param to the end node.
	 */
	Edge(Vertex<V> from, Vertex<V> to, Weighable component)
	{
		this.from = from;
		this.to = to;
		this.component = component;
	}

	/**
	 * Returns the edge's start node.
	 * 
	 * @return the edge's start node.
	 */
	public Vertex<V> getFrom()
	{
		return from;
	}

	/**
	 * Returns the edge's end node.
	 * 
	 * @return the edge's end node.
	 */
	public Vertex<V> getTo()
	{
		return to;
	}

	public int compareTo(Edge<V> other)
	{
		int c = from.compareTo(other.from);

		if (c == 0)
		{
			return to.compareTo(other.to);
		}

		return c;
	}

	public int hashCode()
	{
		return from.hashCode() + to.hashCode();
	}

	@SuppressWarnings("unchecked")
	@Override
	public boolean equals(Object o)
	{
		return equals((Edge<V>) o);
	}

	public boolean equals(Edge<V> edge)
	{
		return compareTo(edge) == 0;
	}

	public HeuristicValue getHeuristicValue()
	{
		return component.getHeuristicValue();
	}

	public HeuristicValue getHeuristicValue(int level)
	{
		return component.getHeuristicValue(level);
	}

	public void setHeuristicValue(HeuristicValue heuristicValue)
	{
		component.setHeuristicValue(heuristicValue);
	}

	public void setHeuristicValue(HeuristicValue heuristicValue, int level)
	{
		component.setHeuristicValue(heuristicValue, level);
	}

	public Pheromone getPheromone()
	{
		return component.getPheromone();
	}

	public Pheromone getPheromone(int level)
	{
		return component.getPheromone(level);
	}

	public void setPheromone(Pheromone pheromone)
	{
		component.setPheromone(pheromone);
	}

	public void setPheromone(Pheromone pheromone, int level)
	{
		component.setPheromone(pheromone, level);
	}

	@Override
	public String toString()
	{
		return "edge(" + from.getLabel() + "->" + to.getLabel() + ") "
			+ component.toString();
	}
}