/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.framework;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * This class represents the problem construction graph. Note that this
 * implementation is not synchronized.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class Graph<V extends Info>
{
	/**
	 * The edge pheromone normalization transformation.
	 */
	public static final Transform EDGE_EVAPORATION = new Transform()
		{
			@Override
			public void apply(Graph<?> graph)
			{
				for (int i = 0; i < graph.edges.length; i++)
				{
					// computes the total pheromone value for th i-th vertex

					ArrayList<Double> total = new ArrayList<Double>();

					for (int j = 0; j < graph.edges[i].length; j++)
					{
						if (graph.edges[i][j] != null)
						{
							int levels = graph.edges[i][j].levels();

							for (int k = 0; k < levels; k++)
							{
								double value =
									graph.edges[i][j].getPheromone(k)
										.getValue();

								if (k >= total.size())
								{
									total.add(value);
								}
								else
								{
									total.set(k, total.get(k) + value);
								}
							}
						}
					}

					// normalizes the pheromone values	

					for (int j = 0; j < graph.edges[i].length; j++)
					{
						if (graph.edges[i][j] != null)
						{
							int levels = graph.edges[i][j].levels();

							for (int k = 0; k < levels; k++)
							{
								double v =
									graph.edges[i][j].getPheromone(k)
										.getValue();
								graph.edges[i][j].getPheromone(k).setValue(
									v / total.get(k));
							}
						}
					}
				}
			}
		};

	/**
	 * The edge pheromone initialisation transformation.
	 */
	public static final Transform EDGE_INITIALIZATION = new Transform()
		{
			@Override
			public void apply(Graph<?> graph)
			{
				for (int i = 0; i < graph.edges.length; i++)
				{
					int count = 0;

					for (int j = 0; j < graph.edges[i].length; j++)
					{
						if (graph.edges[i][j] != null)
						{
							count++;
						}
					}

					if (count > 0)
					{
						Pheromone pheromone = new Pheromone(1.0 / count);

						for (int j = 0; j < graph.edges[i].length; j++)
						{
							if (graph.edges[i][j] != null)
							{
								graph.edges[i][j].setPheromone(pheromone);
							}
						}
					}
				}
			}
		};

	/**
	 * The matrix of edges. Only the <code>Element</code> object is stored
	 * and <code>Edge</code> instances are created when needed. This improves
	 * the memory consumption for highly connected graphs.
	 */
	private Element[][] edges;

	/**
	 * The mapping (label, vertice).
	 */
	private Map<String, Vertex<V>> mapping;

	/**
	 * The array of vertices.
	 */
	private Vertex<V>[] vertices;

	/**
	 * Default constructor.
	 */
	public Graph()
	{
		this(0);
	}

	/**
	 * Creates a new <code>Graph</code> instance with the specified initial
	 * capacity.
	 * 
	 * @param capacity the initial capacity.
	 */
	@SuppressWarnings("unchecked")
	public Graph(int capacity)
	{
		// 0 (zero) initial capacity will be slow down the creation of the
		// graph, but will guarantee a better memory utilisation during the
		// run of the algorithm
		//
		// (non-generic vertex instantiation to avoid the compiler error)

		mapping = new HashMap<String, Vertex<V>>(capacity, 1.0f);
		vertices = new Vertex[capacity];
		edges = new Element[capacity][capacity];
	}

	/**
	 * Adds a new node to the graph.
	 * 
	 * @param vertex the node to add.
	 */
	@SuppressWarnings("unchecked")
	public void add(Vertex<V> vertex)
	{
		vertex.setIndex(mapping.size());
		mapping.put(vertex.getLabel(), vertex);

		if (!(vertex.getIndex() < vertices.length))
		{
			// grows the vertices array to create room for the new vertex

			Vertex<V>[] verticesX = new Vertex[vertices.length + 1];
			System.arraycopy(vertices, 0, verticesX, 0, vertices.length);
			verticesX[vertices.length] = vertex;
			vertices = verticesX;

			// grows the edges matrix to create room for the new vertex

			Element[][] edgesX = new Element[edges.length + 1][];

			for (int i = 0; i < edges.length; i++)
			{
				edgesX[i] = new Element[edges[i].length + 1];
				System.arraycopy(edges[i], 0, edgesX[i], 0, edges[i].length);
			}

			edgesX[edges.length] = new Element[edges.length + 1];
			edges = edgesX;
		}
		else
		{
			vertices[vertex.getIndex()] = vertex;
		}
	}

	/**
	 * Applies the specified tranformation in this graph.
	 * 
	 * @param transform the transformation to be applied.
	 */
	public void apply(Transform transform)
	{
		transform.apply(this);
	}

	/**
	 * Creates an edge (link) between the specified nodes.
	 * 
	 * @param from the start node.
	 * @param to the end node.
	 */
	public void connect(Vertex<V> from, Vertex<V> to)
	{
		if (!mapping.containsKey(from.getLabel()))
		{
			throw new IllegalArgumentException(
				"The source node does not belong to the graph: " + from);
		}

		if (!mapping.containsKey(to.getLabel()))
		{
			throw new IllegalArgumentException(
				"The target node does not belong to the graph: " + to);
		}

		//if (from.equals(to))
		//{
		//	throw new IllegalArgumentException(
		//		"The source and target nodes cannot be the same.");
		//}

		edges[from.getIndex()][to.getIndex()] = new Element();
	}

	/**
	 * Fully connects all nodes of the graph.
	 */
	public void connect()
	{
		for (int i = 0; i < vertices.length; i++)
		{
			for (int j = 0; j < vertices.length; j++)
			{
				if ((i != j) && (vertices[i] != null) && (vertices[j] != null))
				{
					connect(vertices[i], vertices[j]);
				}
			}
		}
	}

	/**
	 * Merges the specified graph with this graph, by adding an edge from
	 * every vertex on this graph to every vertex of the specified graph.
	 * 
	 * @param other the graph to be merged.
	 */
	public void merge(Graph<V> other)
	{
		// vertices previously in this graph
		Vertex<V>[] original = vertices();

		// adds all nodes to this graph
		for (Vertex<V> vertex : other.vertices)
		{
			if (vertex != null)
			{
				add(vertex);
			}
		}

		// re-created the connections
		for (int i = 0; i < other.edges.length; i++)
		{
			for (int j = 0; j < other.edges[i].length; j++)
			{
				if (other.edges[i][j] != null)
				{
					connect(other.vertices[i], other.vertices[j]);
				}
			}
		}

		// connects all nodes of this graph to the nodes
		// of the merged graph
		for (Vertex<V> vertex : original)
		{
			for (Vertex<V> neighbor : other.vertices)
			{
				if (neighbor != null)
				{
					connect(vertex, get(neighbor.getLabel()));
				}
			}
		}
	}

	/**
	 * Returns the list of neighbors of the specified node.
	 * 
	 * @param vertex the node in the graph.
	 * 
	 * @return the list of neighbors of the node.
	 */
	public Collection<Vertex<V>> getNeighbours(Vertex<V> vertex)
	{
		if (!mapping.containsKey(vertex.getLabel()))
		{
			throw new IllegalArgumentException("Node not found on graph: "
				+ vertex);
		}

		Element[] elements = edges[vertex.getIndex()];
		Set<Vertex<V>> neighbors =
			new HashSet<Vertex<V>>(elements.length, 1.0f);

		for (int i = 0; i < elements.length; i++)
		{
			if (elements[i] != null)
			{
				neighbors.add(vertices[i]);
			}
		}

		return neighbors;
	}

	/**
	 * Returns the list of edges of the specified vertex.
	 * 
	 * @param vertex the node in the graph.
	 * 
	 * @return the list of edges of the vertex.
	 */
	public Collection<Edge<V>> getEdges(Vertex<V> vertex)
	{
		if (!mapping.containsKey(vertex.getLabel()))
		{
			throw new IllegalArgumentException("Node not found on graph: "
				+ vertex);
		}

		Element[] elements = edges[vertex.getIndex()];
		Set<Edge<V>> edges = new HashSet<Edge<V>>(elements.length, 1.0f);

		for (int i = 0; i < elements.length; i++)
		{
			if (elements[i] != null)
			{
				edges.add(new Edge<V>(vertex, vertices[i], elements[i]));
			}
		}

		return edges;
	}

	/**
	 * Returns the edge between the specified vertices.
	 * 
	 * @param from the "from" end of the edge.
	 * @param to the "to" end of the edge.
	 * 
	 * @return the edge between the specified vertices.
	 */
	public Edge<V> getEdge(Vertex<V> from, Vertex<V> to)
	{
		if (!mapping.containsKey(from.getLabel()))
		{
			throw new IllegalArgumentException("Vertex not found on graph: "
				+ from);
		}
		else if (!mapping.containsKey(to.getLabel()))
		{
			throw new IllegalArgumentException("Vertex not found on graph: "
				+ to);
		}

		Element element = edges[from.getIndex()][to.getIndex()];

		if (element != null)
		{
			return new Edge<V>(from, to, element);
		}

		throw new IllegalArgumentException("Edge not found on graph: " + from
			+ " -> " + to);
	}

	/**
	 * Returns the specified vertex given its label.
	 * 
	 * @param label the label of the vertex.
	 * 
	 * @return the specified vertex given its label.
	 */
	public Vertex<V> get(String label)
	{
		Vertex<V> vertex = mapping.get(label);

		if (vertex == null)
		{
			throw new IllegalArgumentException("Vertex not found on graph: "
				+ label);
		}

		return vertex;
	}

	/**
	 * Returns the set of vertices of this graph.
	 * 
	 * @return the set of vertices of this graph.
	 */
	public Collection<Vertex<V>> getVertices()
	{
		ArrayList<Vertex<V>> copy = new ArrayList<Vertex<V>>(vertices.length);

		for (Vertex<V> v : vertices)
		{
			if (v != null)
			{
				copy.add(v);
			}
		}

		return copy;
	}

	/**
	 * Returns the array of vertices. Note that the array may have
	 * <code>null</code> position if the initial capacity is greater than
	 * the number of vertices.
	 * 
	 * @return the array of vertices.
	 */
	@SuppressWarnings("unchecked")
	public Vertex<V>[] vertices()
	{
		Collection<Vertex<V>> copy = getVertices();
		return copy.toArray(new Vertex[copy.size()]);
	}

	@Override
	public String toString()
	{
		StringBuffer output = new StringBuffer();
		output.append("==> Graph (" + vertices.length + " vertices)");

		for (int i = 0; i < vertices.length; i++)
		{
			output.append('\n');
			output.append(vertices[i]);
		}

		return output.toString();
	}

	/**
	 * This class represents an internal modification of the graph.
	 */
	private static abstract class Transform
	{
		public abstract void apply(Graph<?> graph);
	}

	/**
	 * This class represents an element of the graph.
	 */
	public static class Element
		implements Cloneable, Weighable
	{
		/**
		 * The heuristic value associated with each level.
		 */
		private List<HeuristicValue> heuristics;

		/**
		 * The initial pheromone value associated with this component.
		 */
		private Pheromone initial;

		/**
		 * The pheromone value associated with each level.
		 */
		private List<Pheromone> pheromones;

		/**
		 * Default constructor.
		 */
		Element()
		{
			heuristics = new ArrayList<HeuristicValue>(1);
			pheromones = new ArrayList<Pheromone>(1);
			initial = null;
		}

		public HeuristicValue getHeuristicValue()
		{
			return getHeuristicValue(0);
		}

		public HeuristicValue getHeuristicValue(int level)
		{
			if (heuristics.size() > level)
			{
				return heuristics.get(level);
			}

			return null;
		}

		public void setHeuristicValue(HeuristicValue heuristicValue)
		{
			setHeuristicValue(heuristicValue, 0);
		}

		public void setHeuristicValue(HeuristicValue heuristicValue, int level)
		{
			if (level < heuristics.size())
			{
				heuristics.set(level, heuristicValue.clone());
			}
			else
			{
				while (level >= heuristics.size())
				{
					heuristics.add(null);
				}

				heuristics.set(level, heuristicValue.clone());
			}
		}

		public Pheromone getPheromone()
		{
			return getPheromone(0);
		}

		public Pheromone getPheromone(int level)
		{
			if (pheromones.size() > level)
			{
				return pheromones.get(level);
			}
			else if (initial != null)
			{
				return (Pheromone) initial.clone();
			}

			return null;
		}

		public void setPheromone(Pheromone pheromone)
		{
			setPheromone(pheromone, 0);
		}

		public void setPheromone(Pheromone pheromone, int level)
		{
			if (initial == null)
			{
				initial = (Pheromone) pheromone.clone();
			}

			if (level < pheromones.size())
			{
				pheromones.set(level, pheromone.clone());
			}
			else
			{
				while (level >= pheromones.size())
				{
					pheromones.add((Pheromone) initial.clone());
				}

				pheromones.set(level, pheromone.clone());
			}
		}

		/**
		 * Returns the number of pheromone levels of this element.
		 * 
		 * @return the number of pheromone levels of this element.
		 */
		public int levels()
		{
			return pheromones.size();
		}

		/**
		 * Creates and returns a copy of this weighted component instance.
		 *
		 * @return a copy of this weighted component instance.
		 */
		@Override
		protected Object clone()
		{
			try
			{
				Element c = (Element) super.clone();
				c.heuristics = new ArrayList<HeuristicValue>(heuristics.size());
				c.pheromones = new ArrayList<Pheromone>(pheromones.size());

				for (Pheromone pheromone : pheromones)
				{
					c.pheromones.add((Pheromone) pheromone.clone());
				}

				for (HeuristicValue heuristic : heuristics)
				{
					c.heuristics.add((HeuristicValue) heuristic.clone());
				}

				if (initial != null)
				{
					c.initial = (Pheromone) initial.clone();
				}

				return c;
			}
			catch (CloneNotSupportedException e)
			{
				throw new UnknownError(getClass() + ": clone()");
			}
		}

		@Override
		public String toString()
		{
			return "<h=" + heuristics + ", p=" + pheromones + ">";
		}
	}

	/**
	 * This class represents the probabilistic evaporation transform.
	 */
	public static class EdgeEvaporateTransform
		extends Transform
	{
		private double factor;

		public EdgeEvaporateTransform(double factor)
		{
			this.factor = factor;
		}

		@Override
		public void apply(Graph<?> graph)
		{
			for (int i = 0; i < graph.edges.length; i++)
			{
				for (int j = 0; j < graph.edges[i].length; j++)
				{
					if (graph.edges[i][j] != null)
					{
						int levels = graph.edges[i][j].levels();

						for (int k = 0; k < levels; k++)
						{
							double v =
								graph.edges[i][j].getPheromone(k).getValue();
							graph.edges[i][j].getPheromone(k).setValue(
								v * factor);
						}
					}
				}
			}
		}
	}

	/**
	 * This class represents an initialization transform.
	 */
	public static class EdgeInitializationTransform
		extends Transform
	{
		private double initial;

		public EdgeInitializationTransform(double initial)
		{
			this.initial = initial;
		}

		@Override
		public void apply(Graph<?> graph)
		{
			Pheromone pheromone = new Pheromone(initial);

			for (int i = 0; i < graph.edges.length; i++)
			{
				for (int j = 0; j < graph.edges[i].length; j++)
				{
					if (graph.edges[i][j] != null)
					{
						graph.edges[i][j].setPheromone(pheromone);
					}
				}
			}
		}
	}

	/**
	 * This class represents the range-fixed pheromone transform.
	 */
	public static class EdgeRangeTransform
		extends Transform
	{
		private double min;

		private double max;

		public EdgeRangeTransform(double min, double max)
		{
			this.min = min;
			this.max = max;
		}

		@Override
		public void apply(Graph<?> graph)
		{
			for (int i = 0; i < graph.edges.length; i++)
			{
				for (int j = 0; j < graph.edges[i].length; j++)
				{
					if (graph.edges[i][j] != null)
					{
						int levels = graph.edges[i][j].levels();

						for (int k = 0; k < levels; k++)
						{
							double v =
								graph.edges[i][j].getPheromone(k).getValue();

							if (v > max)
							{
								v = max;
							}
							else if (v < min)
							{
								v = min;
							}

							graph.edges[i][j].getPheromone(k).setValue(v);
						}
					}
				}
			}
		}
	}
}