/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.framework;

import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadFactory;
import java.util.logging.Logger;

/**
 * This class represents an extension of the <code>Scheduler</code>, which
 * support parallel trail construction. Note that for this implementation
 * to work, the trail factory, local search operator and objective function
 * implementations must be thread safe.
 *
 * @see Scheduler
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class ParallelScheduler
	extends Scheduler
{
	/**
	 * The property key under which the number of workers threads value is
	 * stored. If not specified, there will be one worker thread per
	 * processor/core. 
	 */
	public static final String WORKERS = "net.sf.myra.framework.workers";

	/**
	 * The logger instance.
	 */
	private Logger logger = Logger.getLogger(ParallelScheduler.class.getName());

	/**
	 * Tasks (trail creation) executor services.
	 */
	private ExecutorService executor;

	/**
	 * Creates a new <code>ParallelScheduler</code> instance.
	 * 
	 * @param colony the colony of ants.
	 * @param condition the stop condition.
	 * @param function the objective function.
	 * @param updater the pheromone update policy.
	 */
	public ParallelScheduler(Colony colony, StopCondition condition,
		ObjectiveFunction function, PheromoneUpdater updater,
		HeuristicInformation information)
	{
		super(colony, condition, function, updater, information);
	}

	@Override
	protected void createSolutions()
	{
		Colony colony = environment.getColony();
		CountDownLatch latch = new CountDownLatch(colony.getSize());

		for (Ant ant : colony.getPopulation())
		{
			executor.execute(new Task(ant, latch));
		}

		try
		{
			// waits for the creation of trails
			latch.await();
		}
		catch (InterruptedException e)
		{
			throw new RuntimeException(e);
		}
	}

	@Override
	public void run()
	{
		int available =
			Configuration.getIntProperty(WORKERS, Runtime.getRuntime()
				.availableProcessors());
		
		logger.fine("Available workers: " + available);

		executor = Executors.newFixedThreadPool(available, new ThreadFactory()
			{
				private int id = 0;

				public Thread newThread(Runnable r)
				{
					return new Thread(r, "ParallelScheduler::worker" + (id++));
				}
			});

		super.run();

		logger.fine("Shutting down thread pool.");

		// finilizes the executor threads
		executor.shutdown();
	}

	/**
	 * This class is responsible for discovering the classification rules.
	 */
	private class Task
		implements Runnable
	{
		/**
		 * The ant reference of this task.
		 */
		private Ant ant;

		/**
		 * The synchronization count down aid.
		 */
		private CountDownLatch latch;

		/**
		 * Default constructor.
		 * 
		 *  @param id the worker id.
		 */
		public Task(Ant ant, CountDownLatch latch)
		{
			this.ant = ant;
			this.latch = latch;
		}

		public void run()
		{
			Trail trail = search.explore(ant.walk());
			ant.setTrail(trail);

			trail.setCost(function.evaluate(trail));
			trail.setIteration(environment.getIteration());

			latch.countDown();
		}
	}
}