/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.framework;

import java.util.Random;

/**
 * This class represents the random number generation source.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public abstract class RandomGenerator
{
	/**
	 * The property key under which the random class value is stored.
	 */
	private static final String RANDOM_CLASS_KEY =
		"net.sf.myra.framework.random";

	/**
	 * The singleton random generator instance.
	 */
	private static RandomGenerator instance;

	static
	{
		String random = Configuration.getProperty(RANDOM_CLASS_KEY);

		if (random != null)
		{
			instance = (RandomGenerator) ObjectFactory.create(random);
		}
		else
		{
			instance = new DefaultRandomGenerator();
		}
	}

	/**
	 * Returns the <code>RandomGenerator</code> instance.
	 * 
	 * @return the <code>RandomGenerator</code> instance.
	 */
	public static RandomGenerator getInstance()
	{
		return instance;
	}

	/**
	 * Returns the next pseudorandom double value between 0.0 and 1.0 from this
	 * random number generator's sequence.
	 * 
	 * @return the next pseudorandom double value between 0.0 and 1.0 from this
	 *         random number generator's sequence.
	 */
	public abstract double nextDouble();

	/**
	 * Returns the next pseudorandom int value from this random number
	 * generator's sequence.
	 * 
	 * @return the next pseudorandom int value from this random number
	 *         generator's sequence.
	 */
	public abstract int nextInt();

	/**
	 * Returns a pseudorandom int value between 0 (inclusive) and the specified
	 * value (exclusive), drawn from this random number generator's sequence.
	 * 
	 * @param n the upper bound on the random number to be returned.
	 * 
	 * @return a pseudorandom int value between 0 (inclusive) and n (exclusive).
	 */
	public abstract int nextInt(int n);

	/**
	 * Returns the seed value which was used to initilise the random number
	 * generator.
	 *  
	 * @return the seed value which was used to initilise the random number
	 *         generator.
	 */
	public abstract long getSeed();

	/**
	 * Default <code>RandomGenerator</code> implementation.
	 */
	private static class DefaultRandomGenerator
		extends RandomGenerator
	{
		/**
		 * The property key under which the random seed value is stored.
		 */
		private static final String RANDOM_SEED_KEY =
			"net.sf.myra.framework.seed";

		/**
		 * The internal random number generator.
		 */
		private Random random;

		/**
		 * The seed value.
		 */
		private long seed;

		/**
		 * Default constructor.
		 */
		public DefaultRandomGenerator()
		{
			try
			{
				seed = Configuration.getLongProperty(RANDOM_SEED_KEY);
			}
			catch (NumberFormatException e)
			{
				seed = System.currentTimeMillis();
			}

			random = new Random(seed);
		}

		@Override
		public double nextDouble()
		{
			return random.nextDouble();
		}

		@Override
		public int nextInt()
		{
			return random.nextInt();
		}

		@Override
		public int nextInt(int n)
		{
			return random.nextInt(n);
		}

		@Override
		public long getSeed()
		{
			return seed;
		}
	}
}