/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.framework;

import java.util.logging.Logger;

/**
 * This class represents the schedule of activities to evolve the colony
 * of ants, in order to perform a search for solutions.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class Scheduler
{
	/**
	 * The termination condition.
	 */
	protected StopCondition condition;

	/**
	 * The running environment.
	 */
	protected Environment environment;

	/**
	 * The objective function used evaluate the ant trails (solutions).
	 */
	protected ObjectiveFunction function;

	/**
	 * The local search operator.
	 */
	protected LocalSearch search;

	/**
	 * The daemon action.
	 */
	protected DaemonAction action;

	/**
	 * The pheromone update policy.
	 */
	protected PheromoneUpdater updater;

	/**
	 * The heuristic information.
	 */
	protected HeuristicInformation information;

	/**
	 * The logger instance.
	 */
	private Logger logger = Logger.getLogger(Scheduler.class.getName());

	/**
	 * Creates a new <code>Scheduler</code> instance.
	 * 
	 * @param colony the colony of ants.
	 * @param condition the stop condition.
	 * @param function the objective function.
	 * @param updater the pheromone update policy.
	 */
	public Scheduler(Colony colony, StopCondition condition,
			ObjectiveFunction function, PheromoneUpdater updater,
			HeuristicInformation information)
	{
		this.condition = condition;
		this.function = function;
		this.updater = updater;
		this.information = information;

		environment = new Environment(colony, this);
		search = new LocalSearch()
		{
			public Trail explore(Trail trail)
			{
				return trail;
			}
		};
		action = new DaemonAction()
		{
			public void perform(Colony colony)
			{
			}
		};
	}

	/**
	 * Evolves the colony 'till the stop condition is reached.
	 */
	public void run()
	{
		initialize();

		while (!condition.isReached(environment))
		{
			environment.increment();

			createSolutions();
			updatePheromones();
			daemonActions();
		}
	}

	/**
	 * Creates the ants' trails (solutions).
	 */
	protected void createSolutions()
	{
		logger.info("iteration start");

		for (Ant ant : environment.getColony().getPopulation())
		{
			logger.fine("creating rule");
			Trail trail = ant.walk();
			logger.fine("rule created");

			logger.fine("pruning rule (size=" + trail.getSize() + ")");
			trail = search.explore(trail);
			logger.fine("rule pruning complete (size=" + trail.getSize() + ")");

			ant.setTrail(trail);
			trail.setCost(function.evaluate(trail));
			trail.setIteration(environment.getIteration());
		}

		logger.info("iteration " + environment.getIteration() + " finished");
	}

	/**
	 * Updates the pheromone values.
	 */
	protected void updatePheromones()
	{
		logger.info("updating pheromones");

		Colony colony = environment.getColony();
		updater.update(colony.getCurrent().getTrail(), colony);
		updater.evaporate(colony);

		logger.info("pheromones updated");
	}

	/**
	 * Executes the daemon actions.
	 */
	protected void daemonActions()
	{
		action.perform(environment.getColony());
	}

	/**
	 * Initialization step.
	 */
	protected void initialize()
	{
		updater.initialize(environment.getColony());
		information.initialize(environment.getColony().getTrailFactory()
			.getGraph());
	}

	/**
	 * Returns the local search operator.
	 * 
	 * @return the local search operator.
	 */
	public LocalSearch getLocalSearch()
	{
		return search;
	}

	/**
	 * Returns the pheromone updater.
	 * 
	 * @return the pheromone updater.
	 */
	public PheromoneUpdater getPheromoneUpdater()
	{
		return updater;
	}

	/**
	 * Sets the local search operator.
	 * 
	 * @param search the local search operator to set.
	 */
	public void setLocalSearch(LocalSearch search)
	{
		this.search = search;
	}

	/**
	 * Returns the daemon action.
	 * 
	 * @return the daemon action.
	 */
	public DaemonAction getDaemonAction()
	{
		return action;
	}

	/**
	 * Sets the daemon action.
	 * 
	 * @param action the daemon action to set.
	 */
	public void setDaemonAction(DaemonAction action)
	{
		this.action = action;
	}
}