/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.framework;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.io.Serializable;
import java.util.Collection;

/**
 * This class represents an ant tour in the problem construction graph.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public abstract class Trail
	implements Cloneable, Serializable
{
	private static final long serialVersionUID = -8048106011605163317L;

	/**
	 * The trail quality (cost).
	 */
	private Cost cost;

	/**
	 * The <code>PropertyChangeSupport</code> object.
	 */
	private PropertyChangeSupport bean = new PropertyChangeSupport(this);

	/**
	 * The iteration number that created this trail.
	 */
	private int iteration;

	/**
	 * Default constructor.
	 */
	public Trail()
	{}

	/**
	 * Adds a node to the trail.
	 * 
	 * @param vertex the vertex to be added.
	 */
	public abstract void add(Vertex<?> vertex);

	/**
	 * Removes the specified vertex.
	 * 
	 * @param vertex the vertex to remove.
	 */
	public abstract void remove(Vertex<?> vertex);

	/**
	 * Checks if this trail is empty.
	 * 
	 * @return <code>true</code> if this trail hasn't vertices associated;
	 *         <code>false</code> otherwise.
	 */
	public abstract boolean isEmpty();

	/** 
	 * Returns the trail cost.
	 * 
	 * @return the cost.
	 */
	public Cost getCost()
	{
		return cost;
	}

	/**
	 * Sets the trail cost.
	 * 
	 * @param cost the cost to set.
	 */
	public void setCost(Cost cost)
	{
		Cost old = this.cost;
		this.cost = cost;

		bean.firePropertyChange("cost", old, cost);
	}

	/**
	 * Returns the iteration number that created this trail.
	 * 
	 * @return the iteration number that created this trail.
	 */
	public int getIteration()
	{
		return iteration;
	}

	/**
	 * Sets the iteration that created this trail.
	 * 
	 * @param iteration the iteration number to set.
	 */
	public void setIteration(int iteration)
	{
		this.iteration = iteration;
	}

	/**
	 * Returns the vertices of this trail.
	 * 
	 * @return the vertices of this trail.
	 */
	public abstract Collection<Vertex<?>> getVertices();

	/**
	 * Returns the last added (visited) vertex. For partial trails, the last
	 * added vertex represents the current ant position.
	 * 
	 * @return the last added (visited) vertex.
	 */
	public abstract Vertex<?> getLast();

	/**
	 * Registers a property change listener.
	 * 
	 * @param listener the listener to add.
	 */
	void addPropertyChangeListener(PropertyChangeListener listener)
	{
		bean.addPropertyChangeListener(listener);
	}

	/**
	 * Removes a property change listener.
	 * 
	 * @param listener the listener to remove.
	 */
	void removePropertyChangeListener(PropertyChangeListener listener)
	{
		bean.removePropertyChangeListener(listener);
	}

	/**
	 * Returns the trail size (number of vertices).
	 * 
	 * @return the trail size (number of vertices).
	 */
	public int getSize()
	{
		return getVertices().size();
	}

	/**
	 * Creates and returns a copy of this trail instance.
	 *
	 * @return a copy of this trail instance.
	 */
	@Override
	public Trail clone()
	{
		try
		{
			Trail trail = (Trail) super.clone();

			if (cost != null)
			{
				trail.cost = (Cost) cost.clone();
			}

			trail.bean = new PropertyChangeSupport(trail);

			return trail;
		}
		catch (CloneNotSupportedException e)
		{
			throw new UnknownError(getClass() + ": clone()");
		}
	}

	@Override
	public String toString()
	{
		StringBuffer buffer = new StringBuffer();

		for (Vertex<?> vertex : getVertices())
		{
			if (buffer.length() > 0)
			{
				buffer.append(" -> ");
			}

			buffer.append(vertex);
		}

		return buffer.toString();
	}
}