/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2009  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.geom.AffineTransform;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.Arrays;
import java.util.HashSet;

import javax.swing.BorderFactory;
import javax.swing.ButtonGroup;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.SwingUtilities;
import net.sf.myra.datamining.statistics.Curve;
import net.sf.myra.datamining.statistics.CurveFactory;
import net.sf.myra.datamining.statistics.Information;
import net.sf.myra.datamining.statistics.Parser;
import net.sf.myra.datamining.statistics.Point;

/**
 * @author Fernando Esteban Barril Otero
 */
@SuppressWarnings("rawtypes")
public class CurvePanel
	extends AbstractToolPanel
	implements PropertyChangeListener
{
	private static final long serialVersionUID = -5194016284802434087L;

	/**
	 * The graph (curve) display area dimension.
	 */
	private static final int DIMENSION = 300;

	/**
	 * The graph start margin.
	 */
	private static final int START = 40;

	/**
	 * The graph length.
	 */
	private static final int END = (DIMENSION - 1) - (2 * START);

	/**
	 * The number of intervals to plot.
	 */
	private static final int INTERVALS = 5;

	/**
	 * The graph (curve) display area.
	 */
	private JPanel graph;

	/**
	 * The statistical information of the selected file.
	 */
	private Information information;

	/**
	 * The available curves (per class).
	 */
	private JComboBox available;

	/**
	 * Progress monitoring bar.
	 */
	private JProgressBar progress;

	/**
	 * The (JVM heap) memory message.
	 */
	private JLabel message;

	/**
	 * The status bar.
	 */
	private JPanel status;

	/**
	 * The selected file (<code>null</code> if no file was selected).
	 */
	private File selected;

	/**
	 * The precision-recall curve type button.
	 */
	private JRadioButton prButton;

	/**
	 * The ROC curve type button.
	 */
	private JRadioButton rocButton;

	/**
	 * The default number format instance.
	 */
	private NumberFormat formatter = new DecimalFormat("0.00000000");

	/**
	 * Default constructor.
	 */
	@SuppressWarnings("serial")
	public CurvePanel()
	{
		super(new BorderLayout());

		// input file

		JPanel input = new JPanel(new BorderLayout());
		input.setBorder(BorderFactory.createTitledBorder(" Predictions "));

		JPanel options = new JPanel(new FlowLayout(FlowLayout.LEFT, 0, 0));

		JButton button = new JButton("File...");
		button.addActionListener(new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					Frame.CHOOSER.setFileSelectionMode(JFileChooser.FILES_ONLY);
					int c = Frame.CHOOSER.showOpenDialog(getParent());

					if (c == JFileChooser.APPROVE_OPTION)
					{
						((JButton) e.getSource()).transferFocusBackward();
						selected = Frame.CHOOSER.getSelectedFile();
						load();
					}
				}
			});

		options.add(button);

		prButton = new JRadioButton("PR");
		prButton.setSelected(true);
		prButton.setFocusable(false);
		prButton.addActionListener(new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					load();
				}
			});
		options.add(prButton);

		rocButton = new JRadioButton("ROC");
		rocButton.setFocusable(false);
		rocButton.addActionListener(new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					load();
				}
			});
		options.add(rocButton);

		ButtonGroup group = new ButtonGroup();
		group.add(prButton);
		group.add(rocButton);

		input.add(options, BorderLayout.WEST);

		available = new JComboBox();
		available.setEnabled(false);
		available.addActionListener(new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					graph.repaint();
				}
			});
		input.add(available, BorderLayout.CENTER);

		add(input, BorderLayout.NORTH);

		// result area

		JPanel output = new JPanel(new FlowLayout(FlowLayout.CENTER));
		graph = new JPanel()
			{
				@Override
				public void paint(Graphics g)
				{
					if (information != null)
					{
						super.paint(g);
						Graphics2D g2 = (Graphics2D) g;
						String type = prButton.isSelected() ? "PRC" : "ROC";

						// plots the graph frame

						Font font = g2.getFont();
						font = new Font("Monospaced", font.getStyle(), 10);
						g2.setFont(font);
						FontMetrics metrics = g2.getFontMetrics(font);
						int height = metrics.getAscent();
						int gap = metrics.getDescent() * 4;

						int width = metrics.stringWidth("0.0");
						float x = START - (width / 2.0f);
						float y = (START + END) + (height / 2.0f + gap);
						text(g2, "0.0", x, y);

						width = metrics.stringWidth("1.0");
						x = START - (width + (gap / 2.0f));
						y = START + (height / 2.0f);
						text(g2, "1.0", x, y);

						g2.draw(new Rectangle2D.Double(START, START, END, END));
						double increment = (1.0 / (double) INTERVALS);

						for (int i = 1; i < INTERVALS; i++)
						{
							double pos = increment * i;
							double point = START + (pos * END);

							line(g2, point, START + END, point, START + END - 5);
							line(g2, point, START, point, START + 5);
							line(g2, START, point, START + 5, point);
							line(g2, START + END, point, START + END - 5, point);

							String s = String.format("%1$1.1f", new Float(pos));
							width = metrics.stringWidth(s);
							x = (float) point - (width / 2.0f);
							y = (START + END) + (height / 2.0f + gap);
							text(g2, s, x, y);

							s = String.format("%1$1.1f", new Float(1 - pos));
							width = metrics.stringWidth(s);
							x = START - (width + (gap / 2.0f));
							y = (float) point + (height / 2.0f);
							text(g2, s, x, y);
						}

						width = metrics.stringWidth("1.0");
						x = (START + END) - (width / 2.0f);
						y = (START + END) + (height / 2.0f + gap);
						text(g2, "1.0", x, y);

						width = metrics.stringWidth("0.0");
						x = START - (width + (gap / 2.0f));
						y = (START + END) + (height / 2.0f);
						text(g2, "0.0", x, y);

						String detail =
							prButton.isSelected() ? "recall"
								: "false positive rate";
						width = metrics.stringWidth(detail);
						x = START + (END / 2.0f) - (width / 2.0f);
						y = DIMENSION - height - (gap / 3.0f);
						text(g2, detail, x, y);

						AffineTransform transform = g2.getTransform();
						g2.translate(0, DIMENSION);
						g2.rotate(-(Math.PI / 2.0));

						detail =
							prButton.isSelected() ? "precision"
								: "true positive rate";
						width = metrics.stringWidth(detail);
						x = START + (END / 2.0f) - (width / 2.0f);
						y = height + (gap / 10.0f);
						text(g2, detail, x, y);

						g2.setTransform(transform);

						// plots the graph curve

						String label = (String) available.getSelectedItem();
						Curve curve = information.get(label);
						Point2D.Double last = null;

						for (Point point : curve.getPoints())
						{
							double px = START + (point.getX() * END);
							double py = START + ((1 - point.getY()) * END);

							if (last == null)
							{
								last = new Point2D.Double(px, py);
							}
							else
							{
								Point2D.Double first =
									new Point2D.Double(px, py);
								g2.draw(new Line2D.Double(first, last));
								last = first;
							}
						}

						// plots the curve information

						if (Information.AVERAGE_CURVE.equals(label))
						{
							int x1 = metrics.stringWidth("AU(") + DIMENSION;
							int x2 = metrics.stringWidth(type) + x1;
							line(g2, x1, START, x2, START);

							String s = "AU(" + type + ") = ";
							width = metrics.stringWidth(s);
							text(g2, s, DIMENSION, START + height);

							String value = formatter.format(curve.area());
							x = DIMENSION + width;
							width = metrics.stringWidth(value);
							text(g2, value, x, START + height);

							width = metrics.stringWidth("AU" + type);
							y = START + (2 * height);
							line(g2, DIMENSION, y, DIMENSION + width, y);

							width = metrics.stringWidth("AU" + type + "   = ");
							y += height;
							text(g2, "AU" + type + "   = ", DIMENSION, y);

							s = formatter.format(information.getAverage());
							x = DIMENSION + width;
							width = metrics.stringWidth(value);
							text(g2, s, x, y);

							width = metrics.stringWidth("AU" + type);
							y = START + (4 * height);
							line(g2, DIMENSION, y, DIMENSION + width, y);

							s = "AU" + type + "w  = ";
							width = metrics.stringWidth(s);
							y += height;
							text(g2, s, DIMENSION, y);

							s = formatter.format(information.getWeighted());
							x = DIMENSION + width;
							width = metrics.stringWidth(value);
							text(g2, s, x, y);
						}
						else
						{
							String s = "AU" + type + " = ";
							width = metrics.stringWidth(s);
							text(g2, s, DIMENSION, START + height);

							String value = formatter.format(curve.area());
							x = DIMENSION + width;
							width = metrics.stringWidth(value);
							text(g2, value, x, START + height);
						}
					}
				}
			};
		graph.setBackground(Color.WHITE);
		graph
			.setPreferredSize(new Dimension((int) (DIMENSION * 1.5), DIMENSION));
		output.add(graph);

		add(output, BorderLayout.CENTER);

		// status area

		status = new JPanel(new BorderLayout());
		message =
			new JLabel(
				"You may need to increase the JVM memory heap when dealing with"
					+ " large files.", JLabel.CENTER);
		message.setIcon(new ImageIcon(getClass()
			.getResource("/information.png")));

		Font font = message.getFont();
		font = new Font(font.getName(), font.getStyle(), 10);
		message.setFont(font);

		status.setBorder(BorderFactory.createEmptyBorder(8, 0, 0, 0));
		status.add(message, BorderLayout.CENTER);

		add(status, BorderLayout.SOUTH);
	}

	/**
	 * Enables the 'plot' button.
	 */
	private void enablePlot(final Information information)
	{
		SwingUtilities.invokeLater(new Runnable()
			{
				public void run()
				{
					CurvePanel.this.information = information;
					status.remove(progress);
					status.add(message, BorderLayout.CENTER);
					available.setEnabled(true);
				}
			});
	}

	/**
	 * Restores the status message.
	 */
	private void restoreStatus()
	{
		SwingUtilities.invokeLater(new Runnable()
			{
				public void run()
				{
					status.remove(progress);
					status.add(message, BorderLayout.CENTER);
					status.repaint();
				}
			});
	}

	public void propertyChange(PropertyChangeEvent event)
	{
		if (Parser.MAPPING_SIZE.equals(event.getPropertyName()))
		{
			progress.setIndeterminate(false);
			progress.setMinimum(0);
			progress.setMaximum((Integer) event.getNewValue());
		}
		else if (!progress.isIndeterminate())
		{
			progress.setValue(progress.getValue() + 1);
		}
	}

	/**
	 * Loads the selected file.
	 */
	private void load()
	{
		if (selected != null)
		{
			progress = new JProgressBar();
			progress.setIndeterminate(true);

			available.setEnabled(false);
			information = null;

			status.remove(message);
			status.add(progress, BorderLayout.CENTER);

			new Thread()
				{
					@SuppressWarnings("unchecked")
					public void run()
					{
						try
						{
							startWaitCursor();

							Parser parser = new Parser();
							parser.addListener(CurvePanel.this);

							CurveFactory factory =
								(prButton.isSelected() ? CurveFactory.PRECISION_RECALL
									: CurveFactory.ROC);

							Information information =
								parser.read(
									selected,
									factory,
									new HashSet<String>(Arrays
										.asList(new String[] { "GO0003674",
											"GO0005575", "GO0008150" })));
							//parser.read(selected, factory);

							available.removeAllItems();
							available.addItem(Information.AVERAGE_CURVE);

							for (String label : information.getAvailable())
							{
								available.addItem(label);
							}

							enablePlot(information);
						}
						catch (Exception exception)
						{
							stopWaitCursor();

							StringWriter out = new StringWriter();
							exception.printStackTrace(new PrintWriter(out));
							JTextArea text = new JTextArea(out.toString());
							text.setColumns(40);
							text.setRows(10);

							JOptionPane.showMessageDialog(getParent(),
								new JScrollPane(text),
								"Invalid predictions file",
								JOptionPane.ERROR_MESSAGE);

							restoreStatus();
						}
						finally
						{
							stopWaitCursor();
						}
					}
				}.start();
		}
	}

	/**
	 * Draws a line on the specified graphics instance.
	 * 
	 * @param g2 the graphics instance to draw on.
	 * @param x1 the x coordinate of the first point.
	 * @param y1 the y coordinate of the first point.
	 * @param x2 the x coordinate of the second point.
	 * @param y2 the y coordinate of the second point.
	 */
	private void line(Graphics2D g2, double x1, double y1, double x2, double y2)
	{
		g2.draw(new Line2D.Double(x1, y1, x2, y2));
	}

	/**
	 * Draws the text on the specified graphics instance.
	 * 
	 * @param g2 the graphics instance to draw on.
	 * @param text the text to be drawn.
	 * @param x the x coordinate of the text.
	 * @param y the y coordinate of the text.
	 */
	private void text(Graphics2D g2, String text, float x, float y)
	{
		g2.drawString(text, x, y);
	}
}