/*
 * $Id: ParserPanel.java 2033 2009-03-03 15:55:32Z febo $
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2009  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.gui;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.text.DecimalFormat;
import java.text.NumberFormat;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.SwingUtilities;
import javax.swing.text.DefaultCaret;

import net.sf.myra.datamining.data.Attribute;
import net.sf.myra.datamining.data.ContinuousAttribute;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.data.NominalAttribute;
import net.sf.myra.datamining.io.Helper;

/**
 * Panel to show the details of a dataset instance.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision: 2033 $ $Date:: 2009-03-03 15:55:32#$
 */
public class DatasetDetailsPanel
	extends AbstractToolPanel
{
	private static final long serialVersionUID = -484622103505571880L;

	/**
	 * The dataset file.
	 */
	private JTextField file = new JTextField();

	/**
	 * The text field to display the dataset details.
	 */
	private JTextArea details;

	/**
	 * Default constructor.
	 * 
	 * @param frame the parent frame.
	 */
	public DatasetDetailsPanel(Frame frame)
	{
		super(new BorderLayout());

		// parser options

		JPanel options = new JPanel(new BorderLayout());
		options.setBorder(BorderFactory.createTitledBorder(" Dataset File "));

		JButton button = new JButton("File...");
		button.addActionListener(new ActionListener()
		{
			public void actionPerformed(ActionEvent e)
			{
				Frame.CHOOSER.setFileSelectionMode(JFileChooser.FILES_ONLY);
				int c = Frame.CHOOSER.showOpenDialog(getParent());

				if (c == JFileChooser.APPROVE_OPTION)
				{
					details.setText("loading...");
					((JButton) e.getSource()).transferFocusBackward();
					file.setText(Frame.CHOOSER.getSelectedFile()
						.getAbsolutePath());
					showDetails();
				}
			}
		});
		options.add(button, BorderLayout.WEST);

		file.setEditable(false);
		options.add(file, BorderLayout.CENTER);

		add(options, BorderLayout.NORTH);

		// result area

		details = new JTextArea();
		details.setEditable(false);
		((DefaultCaret) details.getCaret())
			.setUpdatePolicy(DefaultCaret.NEVER_UPDATE);

		JScrollPane output = new JScrollPane(details);

		JPanel panel = new JPanel(new BorderLayout());
		panel.setBorder(BorderFactory.createEmptyBorder(0, 3, 3, 3));
		panel.add(output, BorderLayout.CENTER);
		add(panel, BorderLayout.CENTER);
	}

	/**
	 * Displays the dataset info.
	 */
	private void showDetails()
	{
		final String name = file.getText();

		new Thread()
		{
			public void run()
			{
				startWaitCursor();

				try
				{

					Dataset dataset = Helper.open(name);
					Metadata metadata = dataset.getMetadata();

					if (dataset.isEmpty())
					{
						setDetails("Unsupported (or invalid) dataset file type: '"
							+ name.substring(name.lastIndexOf('.')) + "'");
						return;
					}

					StringBuffer output = new StringBuffer();

					output.append("Relation: ");
					output.append(metadata.getName());
					output.append("\nInstances: ");
					output.append(dataset.getSize());
					output.append("\nAttributes: ");
					output.append(metadata.getPredictor().size());

					int cAttributes = 0;
					int nAttributes = 0;

					for (Attribute attribute : metadata.getPredictor())
					{
						if (attribute instanceof ContinuousAttribute)
						{
							cAttributes++;
						}
						else if (attribute instanceof NominalAttribute)
						{
							nAttributes++;
						}
					}

					output.append(" (");

					if (cAttributes > 0)
					{
						output.append(cAttributes);
						output.append(" continuous");
					}

					if (nAttributes > 0)
					{
						if (cAttributes > 0)
						{
							output.append(", ");
						}

						output.append(nAttributes);
						output.append(" nominal");
					}

					output.append(")");

					output.append("\nClasses: ");
					output.append(metadata.getTarget().getValues().size());
					output.append("\nDuplicates: ");
					output.append(dataset.hasDuplicates() ? "yes" : "no");
					output.append("\nMissing values: ");
					output.append(dataset.hasMissing() ? "yes" : "no");
					output.append("\nHierarchical: ");
					output.append(metadata.isHierarchical() ? "yes" : "no");

					if (metadata.isHierarchical())
					{
						boolean tree = metadata.getClassHierarchy().isTree();
						output.append(tree ? " (TREE)" : " (DAG)");

						if (tree)
						{
							output.append("\nNumber of levels: ");
							output.append(metadata.getClassHierarchy()
								.getLevels());
						}
					}

					output.append("\nMulti-Label: ");

					double tLabel = 0;
					double sLabel = 0;
					int count = 0;

					for (Instance instance : dataset)
					{
						if (instance.isMultilabel())
						{
							count++;
						}

						if (metadata.isHierarchical())
						{
							// does not count the root label
							tLabel += (instance.getLabel().cardinality() - 1);

							sLabel +=
								metadata.getClassHierarchy().leaves(
									instance.getLabel().getValues()).size();
						}
					}

					if (dataset.isMultilabel())
					{
						output.append("yes (" + count + " instances)");
					}
					else
					{
						output.append("no");
					}

					if (metadata.isHierarchical())
					{
						NumberFormat formatter = new DecimalFormat("0.00");

						tLabel /= dataset.getSize();
						sLabel /= dataset.getSize();

						output.append("\nAverage class labels: "
							+ formatter.format(tLabel));
						output.append("\nAverage specific class labels: "
							+ formatter.format(sLabel));
					}
					
					output.append("\nVirtual class hierarchy: ");
					count = 0;

					for (Instance instance : dataset)
					{
						if (!instance.isLeaf())
						{
							count++;
						}
					}

					if (count == 0)
					{
						output.append("yes");
					}
					else
					{
						output.append("no (" + count + " instances)");
					}

					setDetails(output.toString());
				}
				catch (Exception e)
				{
					StringWriter out = new StringWriter();
					e.printStackTrace(new PrintWriter(out));
					setDetails(out.toString());
				}
				finally
				{
					stopWaitCursor();
				}
			}
		}.start();
	}

	/**
	 * Updates the details field value.
	 * 
	 * @param text the new value.
	 */
	private void setDetails(final String text)
	{
		SwingUtilities.invokeLater(new Runnable()
		{
			public void run()
			{
				details.setText(text);
			}
		});
	}
}