/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.gui;

import static javax.swing.JOptionPane.OK_CANCEL_OPTION;
import static javax.swing.JOptionPane.PLAIN_MESSAGE;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.StringReader;
import java.text.DateFormat;
import java.text.NumberFormat;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JFormattedTextField;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTabbedPane;
import javax.swing.JTable;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.JToggleButton;
import javax.swing.ListCellRenderer;
import javax.swing.SwingUtilities;
import javax.swing.border.EmptyBorder;
import javax.swing.border.TitledBorder;
import javax.swing.table.AbstractTableModel;
import javax.swing.text.DefaultCaret;
import javax.swing.text.NumberFormatter;

import net.sf.myra.datamining.Main;
import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.ObjectFactory;

/**
 * This class represents the main window.
 *
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class Frame
	extends JFrame
{
	private static final long serialVersionUID = -6643624963342441948L;

	/**
	 * The package version.
	 */
	private static final String VERSION = "v3.7";

	/**
	 * The size of the read buffer.
	 */
	private static final int READ_BUFFER = 500;

	/**
	 * An empty string (just for convinience and elegance).
	 */
	private static final String EMPTY = "";

	/**
	 * Logo image file.
	 */
	private static final String LOGO = "/myra-logo.png";

	/**
	 * Website address.
	 */
	private static final String WEBSITE =
		"http://sourceforge.net/projects/myra";

	/**
	 * List of parameters to be filteterd, since they are selected using
	 * the GUI.
	 */
	private static final List<String> FILTER_PARAMETERS =
		Arrays.asList(new String[] { Configuration.CONFIG_FILE,
			Main.CROSS_VALIDATION_FOLDS, Main.BULK_SIZE, Main.CLASSIFIER,
			Main.EXPORTER_FILE });

	/**
	 * The classifiers configuration file.
	 */
	private static final String CLASSIFIER_FILE = "/myra-classifier.properties";

	/**
	 * The classifiers default configuration values file.
	 */
	private static final String DEFAULT_VALUES_FILE = "myra.default";

	/**
	 *The (global) file chooser instance.
	 */
	static final JFileChooser CHOOSER = new JFileChooser();

	/**
	 * The external java executable path used to launch the classifiers.
	 */
	private String external;

	/**
	 * The classifier runner.
	 */
	private Runner runner = new Runner();

	/**
	 * The output area.
	 */
	private JTextArea output = new JTextArea();

	/**
	 * The list of available classifiers.
	 */
	@SuppressWarnings("rawtypes")
	private JComboBox classifiers = new JComboBox();

	/**
	 * Start button instance.
	 */
	private JButton start = new JButton("Start");

	/**
	 * Stop button instance.
	 */
	private JButton stop = new JButton("Stop");

	/**
	 * Training file text field (single-run).
	 */
	private JTextField training = new JTextField();

	/**
	 * Test file text field (single-run).
	 */
	private JTextField test = new JTextField();

	/**
	 * Folds (cross-validation) text field.
	 */
	@SuppressWarnings("serial")
	private JFormattedTextField folds =
		new JFormattedTextField(new NumberFormatter(NumberFormat
			.getIntegerInstance())
		{
			public Object stringToValue(String string)
				throws ParseException
			{
				if (string == null || string.length() == 0)
				{
					return null;
				}

				return super.stringToValue(string);
			}
		});

	/**
	 * Batch directory file text field.
	 */
	private JTextField directory = new JTextField();

	/**
	 * Batch pattern text field.
	 */
	private JTextField pattern = new JTextField();

	/**
	 * Batch bulk text field.
	 */
	@SuppressWarnings("serial")
	private JFormattedTextField bulk =
		new JFormattedTextField(new NumberFormatter(NumberFormat
			.getIntegerInstance())
		{
			public Object stringToValue(String string)
				throws ParseException
			{
				if (string == null || string.length() == 0)
				{
					return null;
				}

				return super.stringToValue(string);
			}
		});

	/**
	 * Export predictions text field.
	 */
	private JTextField predictions = new JTextField();

	/**
	 * The parameter panel.
	 */
	private JPanel parameter = new JPanel();

	/**
	 * Java VM options.
	 */
	private String options = EMPTY;

	/**
	 * The time stamp information of a classifier run.
	 */
	private JLabel timestamp = new JLabel();

	/**
	 * The button to indicated of the output vertical scroll is enabled.
	 */
	private JToggleButton scrollLock = new JToggleButton("Scroll Lock");

	/**
	 * The scrollable output panel.
	 */
	private JScrollPane outputPanel;

	/**
	 * Default constructor.
	 * 
	 * @param external the external java executable path.
	 */
	@SuppressWarnings("unchecked")
	public Frame(String external)
	{
		super("Myra");
		this.external = external;

		output.setEditable(false);
		Font font = output.getFont();
		output.setFont(new Font("Monospaced", Font.PLAIN, font.getSize()));
		start.setEnabled(false);

		// text fields size
		folds.setColumns(3);
		bulk.setColumns(3);
		training.setColumns(15);
		test.setColumns(15);
		directory.setColumns(15);
		pattern.setColumns(15);

		// loads all classifiers available in the classpath
		ClassifierInfo[] available = findClassifiers();

		classifiers.setRenderer(new CustomCellRenderer(classifiers
			.getRenderer(), available));
		classifiers.addItem("");

		for (Object classifier : available)
		{
			classifiers.addItem(classifier);
		}

		classifiers.addActionListener(new ActionListener()
		{
			public void actionPerformed(ActionEvent e)
			{
				Object selected = classifiers.getSelectedItem();

				if (selected instanceof ClassifierInfo)
				{
					initParameterPanel((ClassifierInfo) selected);
					start.setEnabled(true);
				}
				else
				{
					initParameterPanel(null);
					start.setEnabled(false);
				}

				parameter.validate();
			}
		});

		Container root = getContentPane();
		root.setLayout(new BorderLayout());

		JPanel panel = new JPanel(new BorderLayout());
		panel.add(initClassifierPanel(), BorderLayout.NORTH);
		panel.add(initOptionPanel(), BorderLayout.CENTER);

		JPanel configurationPanel = new JPanel(new GridLayout(1, 2));
		configurationPanel.add(panel);
		configurationPanel.add(initParameterPanel(null));

		// main window

		JSplitPane splitPane =
			new JSplitPane(JSplitPane.VERTICAL_SPLIT, configurationPanel,
				initOutputPanel());
		splitPane.setOneTouchExpandable(true);
		splitPane.setDividerLocation(405);

		root.add(splitPane, BorderLayout.CENTER);

		// window behavior

		setDefaultCloseOperation(EXIT_ON_CLOSE);
		setPreferredSize(new Dimension(900, 600));
		Dimension screen = getToolkit().getScreenSize();
		setLocation((screen.width / 2) - 450, (screen.height / 2) - 300);

		pack();
	}

	/**
	 * Returns the content of the output text field.
	 * 
	 * @return the content of the output text field.
	 */
	public String getOutput()
	{
		return (output == null ? EMPTY : output.getText());
	}

	/**
	 * Returns the list of available classifiers.
	 * 
	 * @return the list of available classifiers.
	 */
	private ClassifierInfo[] findClassifiers()
	{
		try
		{
			Properties properties = new Properties();
			properties.load(getClass().getResourceAsStream(CLASSIFIER_FILE));
			Set<ClassifierInfo> list =
				new TreeSet<ClassifierInfo>(new Comparator<ClassifierInfo>()
				{
					public int compare(ClassifierInfo o1, ClassifierInfo o2)
					{
						int c = o1.getType().compareTo(o2.getType());
						return (c == 0) ? o1.getLabel().compareToIgnoreCase(
							o2.getLabel()) : c;
					}
				});

			for (Iterator<Object> i = properties.keySet().iterator(); i
				.hasNext();)
			{
				String label = (String) i.next();

				ClassifierInfo info = createInfo(properties.getProperty(label));
				info.setType(ClassifierType.get(label.substring(0, 4)));
				info.setLabel(label.substring(4));

				list.add(info);
			}

			return list.toArray(new ClassifierInfo[0]);
		}
		catch (IOException e)
		{
			return new ClassifierInfo[0];
		}
	}

	/**
	 * Return the classifier info instance for the specified classifier.
	 * 
	 * @param name the classifier class name.
	 * 
	 * @return the classifier info instance for the specified classifier.
	 */
	private ClassifierInfo createInfo(String name)
	{
		ClassifierInfo info = new ClassifierInfo(name);
		runner.start(info);

		Object classifier = ObjectFactory.create(name);
		Properties pDefault = new Properties();

		try
		{
			InputStream in =
				classifier.getClass().getResourceAsStream(DEFAULT_VALUES_FILE);

			if (in != null)
			{
				pDefault.load(in);
			}
		}
		catch (IOException e)
		{
			// we quietly ignore, since the user will have the change to set
			// the parameter values
		}

		StringReader output = new StringReader(this.output.getText());
		BufferedReader reader = new BufferedReader(output);

		String line = null;
		String property = null;

		try
		{
			while ((line = reader.readLine()) != null)
			{
				if (line.startsWith("-D"))
				{
					if (!FILTER_PARAMETERS.contains(line.substring(2, line
						.indexOf('='))))
					{
						property = line.substring(0, line.indexOf('='));
					}
				}
				else if (property != null)
				{
					line = line.substring(0, line.indexOf('.')).trim();
					info.addParameter(property, line);

					String value = pDefault.getProperty(property.substring(2));

					if (value != null)
					{
						info.setValue(property, value);
					}

					property = null;
				}
			}
		}
		catch (IOException e)
		{
			throw new RuntimeException(e);
		}
		finally
		{
			this.output.setText("");
		}

		return info;
	}

	/**
	 * Validates the run options.
	 * 
	 * @return <code>true</code> if the run options are valid;
	 *         <code>false</code> otherwise.
	 */
	private String validateOptions()
	{
		if (EMPTY.equals(training.getText().trim()))
		{
			// directory and pattern fields must NOT BE null, we don't care
			// about the bulk check

			if (EMPTY.equals(directory.getText().trim())
				|| EMPTY.equals(pattern.getText().trim()))
			{
				return "Either use the single or batch run options.";
			}
		}
		else
		{
			// directory and pattern fields must BE null, we don't care
			// about the test field

			if (!EMPTY.equals(directory.getText().trim())
				|| !EMPTY.equals(pattern.getText().trim()))
			{
				return "Either use the single or batch run options.";
			}
		}

		if (!EMPTY.equals(folds.getText().trim()))
		{
			if (!EMPTY.equals(test.getText().trim()))
			{
				return "Cross-validation cannot be used if a test file is provided.";
			}
		}

		if (!EMPTY.equals(predictions.getText().trim()))
		{
			if (EMPTY.equals(test.getText().trim())
				&& EMPTY.equals(directory.getText().trim())
				&& EMPTY.equals(folds.getText().trim()))
			{
				// test file or batch mode has to be specified

				return "Either specify a test file or use batch run options.";
			}
		}

		// everything seems fine
		return null;
	}

	/**
	 * Updates the output panel view.
	 */
	private void print(final String text)
	{
		// if the outputPanel is null, the window is not visible (it is
		// safe to append the text directly)
		if (outputPanel == null)
		{
			output.append(text);
		}
		// delegates the task to the event dispatching thread
		else
		{
			SwingUtilities.invokeLater(new Runnable()
			{
				public void run()
				{
					output.append(text);
				}
			});
		}
	}

	/**
	 * Returns the classifier panel.
	 * 
	 * @return the classifier panel.
	 */
	private JPanel initClassifierPanel()
	{
		JPanel panel = new JPanel(new BorderLayout());
		panel.setBorder(new TitledBorder(" Algorithm "));
		panel.add(classifiers, BorderLayout.CENTER);

		start.addActionListener(new ActionListener()
		{
			public void actionPerformed(ActionEvent e)
			{
				String message = validateOptions();

				if (message != null)
				{
					JOptionPane.showMessageDialog(Frame.this, message,
						"Invalid Run Options", JOptionPane.INFORMATION_MESSAGE);
				}
				else
				{
					start.setEnabled(false);
					stop.setEnabled(true);

					String stamp =
						((ClassifierInfo) classifiers.getSelectedItem())
							.getLabel();
					stamp +=
						" ("
							+ DateFormat.getDateTimeInstance(DateFormat.LONG,
								DateFormat.LONG).format(
								Calendar.getInstance().getTime()) + ")";

					timestamp.setText(stamp);
					timestamp.setToolTipText("Java Application [" + external
						+ "]");

					Thread thread = new Thread("ClassifierRunner")
					{
						@Override
						public void run()
						{
							runner.start((ClassifierInfo) classifiers
								.getSelectedItem());
						}
					};
					thread.start();
				}
			}
		});

		panel.add(start, BorderLayout.EAST);

		return panel;
	}

	/**
	 * Returns the options panel.
	 * 
	 * @return the options panel.
	 */
	private JPanel initOptionPanel()
	{
		JPanel panel = new JPanel();
		panel.setLayout(new BorderLayout());
		panel.setBorder(new TitledBorder(" Run Options "));

		// single run

		JPanel singlePanel = new JPanel(new GridBagLayout());

		GridBagConstraints c = new GridBagConstraints();
		c.insets = new Insets(0, 4, 0, 0);
		c.anchor = GridBagConstraints.WEST;
		c.weighty = 0;
		c.weightx = 0;
		c.fill = GridBagConstraints.REMAINDER;
		c.gridx = 0;
		c.gridy = 0;
		c.gridwidth = 3;
		singlePanel.add(new JLabel("Training file:", JLabel.LEFT), c);

		c.insets = new Insets(0, 0, 0, 0);
		c.weightx = 1.0;
		c.gridwidth = 2;
		c.fill = GridBagConstraints.BOTH;
		c.gridy = 1;
		singlePanel.add(training, c);

		c.weightx = 0.0;
		c.gridwidth = 1;
		c.gridx = 2;
		JButton button = new JButton("Choose...");
		button.addActionListener(new ActionListener()
		{
			public void actionPerformed(ActionEvent e)
			{
				CHOOSER.setFileSelectionMode(JFileChooser.FILES_ONLY);
				int c = CHOOSER.showOpenDialog(Frame.this);

				if (c == JFileChooser.APPROVE_OPTION)
				{
					training.setText(CHOOSER.getSelectedFile()
						.getAbsolutePath());
				}
			}
		});
		singlePanel.add(button, c);

		c.fill = GridBagConstraints.REMAINDER;
		c.gridx = 0;
		c.gridy = 2;
		c.gridwidth = 3;
		singlePanel.add(Box.createRigidArea(new Dimension(10, 10)), c);

		c.insets = new Insets(0, 4, 0, 0);
		c.anchor = GridBagConstraints.WEST;
		c.weighty = 0;
		c.weightx = 0;
		c.fill = GridBagConstraints.REMAINDER;
		c.gridx = 0;
		c.gridy = 3;
		c.gridwidth = 3;
		singlePanel.add(new JLabel("Test file:", JLabel.LEFT), c);

		c.insets = new Insets(0, 0, 0, 0);
		c.weightx = 1.0;
		c.gridwidth = 2;
		c.fill = GridBagConstraints.BOTH;
		c.gridy = 4;
		singlePanel.add(test, c);

		c.weightx = 0.0;
		c.gridx = 2;
		c.gridwidth = 1;
		button = new JButton("Choose...");
		button.addActionListener(new ActionListener()
		{
			public void actionPerformed(ActionEvent e)
			{
				CHOOSER.setFileSelectionMode(JFileChooser.FILES_ONLY);
				int c = CHOOSER.showOpenDialog(Frame.this);

				if (c == JFileChooser.APPROVE_OPTION)
				{
					test.setText(CHOOSER.getSelectedFile().getAbsolutePath());
				}
			}
		});
		singlePanel.add(button, c);

		c.fill = GridBagConstraints.REMAINDER;
		c.gridx = 0;
		c.gridy = 5;
		c.gridwidth = 3;
		singlePanel.add(Box.createRigidArea(new Dimension(10, 10)), c);

		c.insets = new Insets(0, 4, 0, 0);
		c.anchor = GridBagConstraints.WEST;
		c.weighty = 0;
		c.weightx = 0;
		c.fill = GridBagConstraints.REMAINDER;
		c.gridx = 0;
		c.gridy = 6;
		c.gridwidth = 1;
		singlePanel
			.add(new JLabel("Folds (cross-validation):", JLabel.LEFT), c);

		c.weightx = 0.0;
		c.gridx = 1;
		c.gridwidth = 1;
		singlePanel.add(folds, c);

		c.insets = new Insets(0, 0, 0, 0);
		c.gridx = 0;
		c.gridy = 7;
		c.gridwidth = 2;
		c.weighty = 1.0;
		c.fill = GridBagConstraints.BOTH;
		singlePanel.add(new JPanel(), c);

		// batch run

		JPanel batchPanel = new JPanel(new GridBagLayout());

		c = new GridBagConstraints();
		c.insets = new Insets(0, 4, 0, 0);
		c.anchor = GridBagConstraints.WEST;
		c.weighty = 0;
		c.weightx = 0;
		c.fill = GridBagConstraints.REMAINDER;
		c.gridx = 0;
		c.gridy = 0;
		c.gridwidth = 3;
		batchPanel.add(new JLabel("Directory:", JLabel.LEFT), c);

		c.insets = new Insets(0, 0, 0, 0);
		c.weightx = 1.0;
		c.gridwidth = 1;
		c.fill = GridBagConstraints.BOTH;
		c.gridy = 1;
		c.gridwidth = 2;
		batchPanel.add(directory, c);

		c.weightx = 0.0;
		c.gridx = 2;
		c.gridwidth = 1;
		button = new JButton("Choose...");
		button.addActionListener(new ActionListener()
		{
			public void actionPerformed(ActionEvent e)
			{
				CHOOSER.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
				int c = CHOOSER.showOpenDialog(Frame.this);

				if (c == JFileChooser.APPROVE_OPTION)
				{
					directory.setText(CHOOSER.getSelectedFile()
						.getAbsolutePath());
				}
			}
		});
		batchPanel.add(button, c);

		c.fill = GridBagConstraints.REMAINDER;
		c.gridx = 0;
		c.gridy = 2;
		c.gridwidth = 3;
		batchPanel.add(Box.createRigidArea(new Dimension(10, 10)), c);

		c.insets = new Insets(0, 4, 0, 0);
		c.anchor = GridBagConstraints.WEST;
		c.weighty = 0;
		c.weightx = 0;
		c.fill = GridBagConstraints.REMAINDER;
		c.gridx = 0;
		c.gridy = 3;
		c.gridwidth = 3;
		batchPanel.add(new JLabel("Pattern:", JLabel.LEFT), c);

		c.insets = new Insets(0, 0, 0, 0);
		c.weightx = 1.0;
		c.fill = GridBagConstraints.BOTH;
		c.gridy = 4;
		c.gridwidth = 3;
		batchPanel.add(pattern, c);

		c.fill = GridBagConstraints.REMAINDER;
		c.gridx = 0;
		c.gridy = 5;
		batchPanel.add(Box.createRigidArea(new Dimension(10, 10)), c);

		c.insets = new Insets(0, 0, 0, 0);
		c.gridx = 0;
		c.gridy = 6;
		c.gridwidth = 3;
		c.weighty = 1.0;
		c.fill = GridBagConstraints.BOTH;
		batchPanel.add(new JPanel(), c);

		// put both together

		JTabbedPane tabbedPane = new JTabbedPane();
		tabbedPane.addTab("Single", singlePanel);
		tabbedPane.addTab("Batch", batchPanel);

		panel.add(tabbedPane, BorderLayout.CENTER);

		// advanced options

		JPanel advanced = new JPanel(new BorderLayout());

		JPanel settings = new JPanel(new FlowLayout(FlowLayout.LEFT, 6, 4));
		settings.add(new JLabel("Bulk:", JLabel.LEFT));
		settings.add(bulk);

		advanced.add(settings, BorderLayout.NORTH);

		JPanel predictionPanel = new JPanel(new GridBagLayout());

		c = new GridBagConstraints();
		c.insets = new Insets(0, 4, 0, 0);
		c.anchor = GridBagConstraints.WEST;
		c.weighty = 0;
		c.weightx = 0;
		c.fill = GridBagConstraints.REMAINDER;
		c.gridx = 0;
		c.gridy = 0;
		c.gridwidth = 3;
		predictionPanel.add(new JLabel("Export predictions:", JLabel.LEFT), c);

		c.insets = new Insets(0, 0, 0, 0);
		c.weightx = 1.0;
		c.gridwidth = 2;
		c.fill = GridBagConstraints.BOTH;
		c.gridy = 1;
		predictionPanel.add(predictions, c);

		c.weightx = 0.0;
		c.gridwidth = 1;
		c.gridx = 2;
		button = new JButton("Choose...");
		button.addActionListener(new ActionListener()
		{
			public void actionPerformed(ActionEvent e)
			{
				CHOOSER.setFileSelectionMode(JFileChooser.FILES_ONLY);
				int c = CHOOSER.showSaveDialog(Frame.this);

				if (c == JFileChooser.APPROVE_OPTION)
				{
					predictions.setText(CHOOSER.getSelectedFile()
						.getAbsolutePath());
				}
			}
		});
		predictionPanel.add(button, c);

		predictionPanel.setBorder(BorderFactory.createEmptyBorder(2, 2, 8, 2));
		advanced.add(predictionPanel, BorderLayout.CENTER);

		settings = new JPanel(new FlowLayout(FlowLayout.RIGHT, 2, 2));
		button = new JButton("Advanced...");
		button.addActionListener(new ActionListener()
		{
			public void actionPerformed(ActionEvent e)
			{
				JOptionPane pane =
					new JOptionPane("Java VM options:", PLAIN_MESSAGE);
				pane.setWantsInput(true);
				pane.setInitialSelectionValue(options);
				pane.setOptionType(OK_CANCEL_OPTION);

				JDialog dialog = pane.createDialog(Frame.this, "Advanced");
				dialog.setVisible(true);
				dialog.dispose();

				String value = (String) pane.getInputValue();

				if (value != null)
				{
					options = value;
				}
			}
		});
		settings.add(button);
		settings.add(Box.createRigidArea(new Dimension(2, 2)));
		button = new JButton("Reset");
		button.addActionListener(new ActionListener()
		{
			public void actionPerformed(ActionEvent e)
			{
				training.setText(EMPTY);
				test.setText(EMPTY);
				folds.setText(EMPTY);

				directory.setText(EMPTY);
				pattern.setText(EMPTY);
				bulk.setText(EMPTY);

				predictions.setText(EMPTY);
				options = EMPTY;
			}
		});
		settings.add(button);

		advanced.add(settings, BorderLayout.SOUTH);

		panel.add(advanced, BorderLayout.SOUTH);

		return panel;
	}

	/**
	 * Returns the options panel for the specified classifier.
	 * 
	 * @param info the classifier info instance.
	 * 
	 * @return the options panel for the specified classifier.
	 */
	private JPanel initParameterPanel(ClassifierInfo info)
	{
		parameter.removeAll();
		parameter.setLayout(new BorderLayout());

		if (info == null)
		{
			parameter.setBorder(new EmptyBorder(8, 8, 8, 8));

			JPanel panel = new JPanel();
			BoxLayout layout = new BoxLayout(panel, BoxLayout.Y_AXIS);
			panel.setLayout(layout);

			ImageIcon icon = new ImageIcon(getClass().getResource(LOGO));
			JLabel label = new JLabel(icon);
			label.setAlignmentX(Component.RIGHT_ALIGNMENT);
			panel.add(label);

			label = new JLabel(VERSION);
			label.setAlignmentX(Component.RIGHT_ALIGNMENT);
			panel.add(label);

			label = new JLabel(WEBSITE);
			label.setAlignmentX(Component.RIGHT_ALIGNMENT);
			panel.add(label);

			parameter.add(panel, BorderLayout.CENTER);
		}
		else
		{
			parameter.setBorder(new TitledBorder(" Parameters "));

			JTable table = new JTable(info);
			table.setRowSelectionAllowed(false);

			JScrollPane scrollPane = new JScrollPane(table);
			scrollPane.setBorder(BorderFactory.createLoweredBevelBorder());
			scrollPane.getVerticalScrollBar().setUnitIncrement(10);
			parameter.add(scrollPane, BorderLayout.CENTER);
		}

		return parameter;
	}

	/**
	 * Returns the output panel.
	 * 
	 * @return the output panel.
	 */
	private JPanel initOutputPanel()
	{
		JPanel panel = new JPanel(new BorderLayout());

		// output area

		outputPanel = new JScrollPane(output);
		outputPanel.setBorder(BorderFactory.createLoweredBevelBorder());
		outputPanel.getVerticalScrollBar().setUnitIncrement(10);
		panel.add(outputPanel, BorderLayout.CENTER);

		// action area

		JPanel actions = new JPanel();
		actions.setLayout(new BoxLayout(actions, BoxLayout.X_AXIS));
		actions.add(Box.createRigidArea(new Dimension(4, 0)));

		actions.add(timestamp);
		actions.add(Box.createHorizontalGlue());

		JPanel buttons = new JPanel(new FlowLayout(FlowLayout.RIGHT));
		JButton button = new JButton("Clear");
		button.addActionListener(new ActionListener()
		{
			public void actionPerformed(ActionEvent e)
			{
				output.setText("");
			}
		});
		buttons.add(button);

		stop = new JButton("Stop");
		stop.addActionListener(new ActionListener()
		{
			public void actionPerformed(ActionEvent e)
			{
				stop.setEnabled(false);
				runner.stop();
				start.setEnabled(true);
				timestamp.setText("<terminated> " + timestamp.getText());
				print("");
			}
		});
		stop.setEnabled(false);
		buttons.add(stop);

		actions.add(buttons);

		panel.add(actions, BorderLayout.NORTH);

		// options area

		JPanel status = new JPanel();
		status.setLayout(new BoxLayout(status, BoxLayout.X_AXIS));

		JPanel options = new JPanel(new FlowLayout(FlowLayout.LEFT));

		button = new JButton("Tools...");
		button.addActionListener(new ActionListener()
		{
			public void actionPerformed(ActionEvent e)
			{
				ToolDialog s = new ToolDialog(Frame.this);
				s.setVisible(true);
			}
		});

		options.add(button);

		status.add(options);
		status.add(Box.createHorizontalGlue());

		options = new JPanel(new FlowLayout(FlowLayout.RIGHT));

		scrollLock.setFocusable(false);
		scrollLock.addActionListener(new ActionListener()
		{
			public void actionPerformed(ActionEvent e)
			{
				((DefaultCaret) output.getCaret()).setUpdatePolicy(scrollLock
					.isSelected() ? DefaultCaret.NEVER_UPDATE
					: DefaultCaret.ALWAYS_UPDATE);
			}
		});
		options.add(scrollLock);
		status.add(options);

		panel.add(status, BorderLayout.SOUTH);

		return panel;
	}

	/**
	 * Class responsible for running the classifier process.
	 */
	private class Runner
	{
		/**
		 * The classifier process.
		 */
		private Process process;

		/**
		 * Runs the specified classifier.
		 * 
		 * @param classifier the classifier to run.
		 */
		public void start(ClassifierInfo info)
		{
			List<String> command = new ArrayList<String>(3);
			command.add(external);

			if (!EMPTY.equals(options.trim()))
			{
				for (String option : options.trim().split(" "))
				{
					command.add(option);
				}
			}

			for (String parameter : info.getParameters().keySet())
			{
				String value = info.getValue(parameter);

				if (value != null && !EMPTY.equals(value.trim()))
				{
					command.add(parameter + "=" + value);
				}
			}

			if (!EMPTY.equals(folds.getText().trim()))
			{
				command.add("-D" + Main.CROSS_VALIDATION_FOLDS + "="
					+ folds.getText().trim());
			}

			if (!EMPTY.equals(bulk.getText().trim()))
			{
				command
					.add("-D" + Main.BULK_SIZE + "=" + bulk.getText().trim());
			}

			if (!EMPTY.equals(predictions.getText().trim()))
			{
				command.add("-D" + Main.EXPORTER_FILE + "="
					+ predictions.getText().trim());
			}

			command.add("-cp");
			command.add(System.getProperty("java.class.path"));
			command
				.add("-Dnet.sf.myra.datamining.classifier=" + info.getName());
			command.add(Main.class.getName());

			if (!EMPTY.equals(pattern.getText().trim()))
			{
				// we are running a batch mode

				command.add("-batch");
				command.add(pattern.getText().trim());
				command.add(directory.getText().trim());
			}
			else if (!EMPTY.equals(training.getText().trim()))
			{
				// we are running a single mode

				if (!EMPTY.equals(test.getText().trim()))
				{
					command.add("-test");
					command.add(test.getText().trim());
				}

				command.add(training.getText().trim());
			}

			Thread hook = new Thread("ShutdownHook")
			{
				@Override
				public void run()
				{
					Runner.this.stop();
				}
			};
			Runtime.getRuntime().addShutdownHook(hook);

			ProcessBuilder builder = new ProcessBuilder(command);
			builder.redirectErrorStream(true);

			try
			{
				process = builder.start();

				BufferedReader reader =
					new BufferedReader(new InputStreamReader(process
						.getInputStream()));
				char[] text = new char[READ_BUFFER];
				int size = -1;

				while ((size = reader.read(text)) != -1)
				{
					print(new String(text, 0, size));
				}
			}
			catch (Exception e)
			{
				e.printStackTrace();
				throw new RuntimeException(e);
			}
			finally
			{
				stop.doClick();
				Runtime.getRuntime().removeShutdownHook(hook);
			}
		}

		/**
		 * Stop the execution of the classifier process.
		 */
		public void stop()
		{
			if (process != null)
			{
				process.destroy();
				process = null;
			}
		}
	}

	/**
	 * Class that represents a classifier.
	 */
	private static class ClassifierInfo
		extends AbstractTableModel
	{
		/**
		 * The seriar version UID.
		 */
		private static final long serialVersionUID = -5925991349823272028L;

		/**
		 * The classifier options.
		 */
		private Map<String, String> parameters = new TreeMap<String, String>();

		/**
		 * The classifier options values.
		 */
		private Map<String, String> values = new TreeMap<String, String>();

		/**
		 * The classifier class name.
		 */
		private String name;

		/**
		 * The classifier display name.
		 */
		private String label;

		/**
		 * The classifier type.
		 */
		private ClassifierType type;

		/**
		 * Default constructor.
		 */
		ClassifierInfo(String name)
		{
			this.name = name;
			this.label = name.substring(name.lastIndexOf('.') + 1);
		}

		/**
		 * Returns the classifier name.
		 * 
		 * @return the classifier name.
		 */
		String getName()
		{
			return name;
		}

		/**
		 * Sets the classifier name.
		 * 
		 * @param name the name to set.
		 */
		@SuppressWarnings("unused")
		void setName(String name)
		{
			this.name = name;
		}

		/**
		 * Returns the classifier label.
		 * 
		 * @return the classifier label.
		 */
		String getLabel()
		{
			return label;
		}

		/**
		 * Sets the classifier label.
		 * 
		 * @param label the label to set.
		 */
		void setLabel(String label)
		{
			this.label = label;
		}

		/**
		 * @return the type
		 */
		public ClassifierType getType()
		{
			return type;
		}

		/**
		 * @param type the type to set
		 */
		public void setType(ClassifierType type)
		{
			this.type = type;
		}

		/**
		 * Adds the specified property.
		 * 
		 * @param name the parameter name.
		 * @param description the parameter description.
		 */
		void addParameter(String name, String description)
		{
			parameters.put(name, description);
		}

		/**
		 * Returns the classifier parameters.
		 * 
		 * @return the classifier parameters.
		 */
		Map<String, String> getParameters()
		{
			return parameters;
		}

		/**
		 * Sets the value for the specified name.
		 * 
		 * @param name the parameter name.
		 * @param value the value to set.
		 */
		void setValue(String name, String value)
		{
			values.put(name, value);
		}

		/**
		 * Returns the value of the specified paramter.
		 * 
		 * @param name the parameter name.
		 * 
		 * @return the value of the specified paramter.
		 */
		String getValue(String name)
		{
			return values.get(name);
		}

		/**
		 * Returns the number of columns. This value is used to display this
		 * classifier info in a table.
		 */
		public int getColumnCount()
		{
			return 2;
		}

		/**
		 * Returns the number of rows. This value is used to display this
		 * classifier info in a table.
		 */
		public int getRowCount()
		{
			return parameters.size();
		}

		/**
		 * Returns the value at the specified position.
		 * 
		 * @param rowIndex the row index.
		 * @param columnIndex the column index.
		 */
		public Object getValueAt(int rowIndex, int columnIndex)
		{
			int index = 0;
			Iterator<String> iterator = parameters.keySet().iterator();
			String key = null;

			while (!(index > rowIndex))
			{
				key = iterator.next();
				index++;
			}

			return (columnIndex == 0 ? parameters.get(key) : values.get(key));
		}

		@Override
		public String getColumnName(int column)
		{
			return (column == 0 ? "name" : "value");
		}

		@Override
		public boolean isCellEditable(int rowIndex, int columnIndex)
		{
			return (columnIndex == 1);
		}

		@Override
		public Class<?> getColumnClass(int columnIndex)
		{
			return String.class;
		}

		@Override
		public void setValueAt(Object value, int rowIndex, int columnIndex)
		{
			int index = 0;
			Iterator<String> iterator = parameters.keySet().iterator();
			String key = null;

			while (!(index > rowIndex))
			{
				key = iterator.next();
				index++;
			}

			values.put(key, value.toString());
		}

		@Override
		public String toString()
		{
			return label;
		}
	}

	/**
	 * Custom JComboBox renderer.
	 */
	@SuppressWarnings("rawtypes")
	public static class CustomCellRenderer
		implements ListCellRenderer
	{
		/**
		 * The actual cell renderer.
		 */
		private ListCellRenderer delegate;

		/**
		 * The custom renderer panel.
		 */
		private JPanel custom = new JPanel(new BorderLayout());

		/**
		 * Indexes where the separator should be added.
		 */
		private Map<Integer, String> indexes = new HashMap<Integer, String>();

		public CustomCellRenderer(ListCellRenderer delegate,
				ClassifierInfo[] infos)
		{
			this.delegate = delegate;

			for (int i = 0; i < (infos.length - 1); i++)
			{
				if (infos[i].getType() != infos[i + 1].getType())
				{
					indexes.put(i + 1, infos[i + 1].getType().toString());
				}
			}

			// since the first element is empty, we add a separator to the
			// first classifier info
			if (infos.length != 0)
			{
				indexes.put(0, infos[0].getType().toString());
			}
		}

		public Component getListCellRendererComponent(JList list, Object value,
				int index, boolean isSelected, boolean cellHasFocus)
		{
			@SuppressWarnings("unchecked")
			Component component =
				delegate.getListCellRendererComponent(list, value, index,
					isSelected, cellHasFocus);

			if (index != -1 && indexes.containsKey(index))
			{
				custom.removeAll();
				custom.add(component, BorderLayout.CENTER);
				custom.setBackground(list.getBackground());

				JLabel message = new JLabel(indexes.get(index));
				Font font = message.getFont();
				font = new Font(font.getName(), Font.ITALIC, 12);
				message.setFont(font);
				message.setBorder(BorderFactory.createEmptyBorder(6, 0, 0, 0));

				custom.add(message, BorderLayout.SOUTH);
				return custom;
			}

			return component;
		}
	}

	/**
	 * Enum to represent the different classification algorithms types.
	 */
	private enum ClassifierType
	{
		FLAT_SINGLE_LABEL("fsl_")
		{

			@Override
			public String toString()
			{
				return "flat single-label";
			}
		},
		HIERARCHICAL_SINGLE_LABEL("hsl_")
		{
			@Override
			public String toString()
			{
				return "hierarchical single-label";
			}
		},
		HIERARCHICAL_MULTI_LABEL("hml_")
		{
			@Override
			public String toString()
			{
				return "hierarchical multi-label";
			}
		};

		private String prefix;

		ClassifierType(String prefix)
		{
			this.prefix = prefix;
		}

		public static ClassifierType get(String prefix)
		{
			for (ClassifierType type : ClassifierType.values())
			{
				if (prefix.equals(type.prefix))
				{
					return type;
				}
			}

			throw new IllegalArgumentException(
				"Enum value not found for prefix: " + prefix);
		}
	}
}