/*
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2012 Fernando Esteban Barril Otero
 */

package net.sf.myra.gui.parser;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;

import net.sf.myra.datamining.util.RuleParser;

/**
 * This class represents the parser for the rule list size.
 * 
 * @author Fernando Esteban Barril Otero
 */
public class SizeParser
	extends ResultParser
{
	private final static String START_RULE_LIST = "=== Discovered Model ===";

	private final static String START_RULE = "IF";

	// constants for the bulk option

	private final static String SIZE_TOKEN = "size: ";

	private final static String TERMS_TOKEN = "terms per rule: ";

	private final static String RULES_TOKEN = "of rules: ";

	@Override
	public String parse(String text)
	{
		return (text.contains(START_RULE_LIST) && text.contains(START_RULE)) ? parseList(text)
			: parseBulk(text);
	}

	private String parseList(String text)
	{
		ArrayList<Double> size = new ArrayList<Double>();
		ArrayList<Double> terms = new ArrayList<Double>();
		ArrayList<Double> rules = new ArrayList<Double>();

		RuleParser parser = new RuleParser();
		BufferedReader input = new BufferedReader(new StringReader(text));
		String line = null;

		try
		{
			ArrayList<String> current = null;

			while ((line = input.readLine()) != null)
			{
				if (line.indexOf(START_RULE_LIST) != -1)
				{
					current = new ArrayList<String>();
				}
				else if (line.indexOf(START_RULE) != -1)
				{
					current.add(line);
				}
				// do we completed the previous list of rules?
				else if (current != null && !current.isEmpty())
				{
					double length = 0.0;

					for (String r : current)
					{
						length += parser.parse(r).getSize();
					}

					size.add(length);
					terms.add(length / current.size());
					rules.add((double) current.size());

					current = null;
				}
			}
		}
		catch (IOException e)
		{
			throw new RuntimeException(e);
		}

		return generateOutput(size, terms, rules);
	}

	private String parseBulk(String text)
	{
		ArrayList<Double> size = new ArrayList<Double>();
		ArrayList<Double> terms = new ArrayList<Double>();
		ArrayList<Double> rules = new ArrayList<Double>();

		BufferedReader input = new BufferedReader(new StringReader(text));
		String line = null;

		try
		{
			while ((line = input.readLine()) != null)
			{
				if (line.indexOf(SIZE_TOKEN) != -1)
				{
					int start = line.indexOf(SIZE_TOKEN) + SIZE_TOKEN.length();
					int end = line.indexOf(' ', start);

					if (end == -1)
					{
						size.add(new Double(line.substring(start)));
					}
					else
					{
						size.add(new Double(line.substring(start, end)));
					}
				}
				else if (line.indexOf(TERMS_TOKEN) != -1)
				{
					int start =
						line.indexOf(TERMS_TOKEN) + TERMS_TOKEN.length();
					int end = line.indexOf(' ', start);

					if (end == -1)
					{
						terms.add(new Double(line.substring(start)));
					}
					else
					{
						terms.add(new Double(line.substring(start, end)));
					}
				}
				else if (line.indexOf(RULES_TOKEN) != -1)
				{
					int start =
						line.indexOf(RULES_TOKEN) + RULES_TOKEN.length();
					int end = line.indexOf(' ', start);

					if (end == -1)
					{
						rules.add(new Double(line.substring(start)));
					}
					else
					{
						rules.add(new Double(line.substring(start, end)));
					}
				}
			}

			// if we do not have the size information, we do an estimation
			// based on the number of rules and terms
			if (rules.size() == terms.size() && rules.size() > size.size())
			{
				size.add(rules.get(rules.size() - 1)
					* terms.get(terms.size() - 1));
			}
		}
		catch (IOException e)
		{
			throw new RuntimeException(e);
		}

		return generateOutput(size, terms, rules);
	}

	private String generateOutput(ArrayList<Double> size,
		ArrayList<Double> terms, ArrayList<Double> rules)
	{
		StringBuffer output = new StringBuffer();
		output.append("=== Size Parser ===\n");

		output.append("\n[model size]\n");
		for (Double v : size)
		{
			output.append(v);
			output.append('\n');
		}

		output.append(stdev((Double[]) size.toArray(new Double[0])));

		output.append("\n\n[terms per rule]\n");
		for (Double v : terms)
		{
			output.append(v);
			output.append('\n');
		}

		output.append(stdev((Double[]) terms.toArray(new Double[0])));

		output.append("\n\n[number of rules]\n");
		for (Double v : rules)
		{
			output.append(v);
			output.append('\n');
		}

		output.append(stdev((Double[]) rules.toArray(new Double[0])));

		return output.toString();
	}

	public static void main(String[] args)
		throws Exception
	{
		String dir = args[0];

		for (int i = 1; i < args.length; i++)
		{
			print(dir, args[i]);
		}
	}

	public static void print(final String dir, final String p)
		throws Exception
	{
		final String pattern = p + "-";
		File directory = new File(dir);

		ArrayList<Double> accuracy = new ArrayList<Double>();
		ArrayList<Double> size = new ArrayList<Double>();
		ArrayList<Double> rules = new ArrayList<Double>();
		ArrayList<Double> terms = new ArrayList<Double>();

		AccuracyParser aParser = new AccuracyParser();
		SizeParser sParser = new SizeParser();

		for (String file : directory.list(new FilenameFilter()
			{
				public boolean accept(File dir, String name)
				{
					return name.matches(pattern + "\\d\\d.out");
				}
			}))
		{
			StringBuffer buffer = new StringBuffer();
			BufferedReader reader =
				new BufferedReader(new FileReader(new File(directory, file)));
			String line = null;

			while ((line = reader.readLine()) != null)
			{
				buffer.append(line);
				buffer.append("\n");
			}

			// accuracy

			reader =
				new BufferedReader(new StringReader(aParser.parse(buffer
					.toString())));
			line = null;
			String last = null;

			while ((line = reader.readLine()) != null)
			{
				last = line;
			}

			Double v = new Double(last.substring(0, last.indexOf(' ')));

			if (v > 0)
			{
				accuracy.add(v);
			}

			// size

			reader =
				new BufferedReader(new StringReader(sParser.parse(buffer
					.toString())));
			line = null;
			// 0: model size
			// 1: number of terms
			// 2: number of rules
			double[] values = new double[3];
			int index = 0;

			while ((line = reader.readLine()) != null)
			{
				if (line.contains("+/-"))
				{
					v = new Double(line.substring(0, line.indexOf(' ')));
					values[index++] = v;
				}
			}

			size.add(values[0]);
			terms.add(values[1]);
			rules.add(values[2]);
		}

		//System.out.println(aParser.stdev((Double[]) accuracy
		//	.toArray(new Double[0])));
		System.out
			.println(sParser.stdev((Double[]) size.toArray(new Double[0])));
		//System.out.println(sParser.stdev((Double[]) rules
		//	.toArray(new Double[0])));
		//System.out.println(sParser.stdev((Double[]) terms
		//	.toArray(new Double[0])));
		//System.out.println("(" + accuracy.size() + ")");
	}
}