/*
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2011 Fernando Esteban Barril Otero
 */

package net.sf.myra.gui.util;

import java.util.HashMap;
import java.util.Map;

/**
 * This class represents a candidate configuration undergoing the F-Race.
 */
public class Configuration
	implements Comparable<Configuration>
{
	/**
	 * The configuration name.
	 */
	private String name;

	/**
	 * The mapping of the performance of this configuration on the instances
	 * (tests).
	 */
	private Map<String, Observation> observations =
		new HashMap<String, Observation>();

	/**
	 * The accumulated rank of this configuration on the instances used by
	 * the current step of the race. Note that this rank changes at each
	 * step of the race.
	 */
	private double rank;

	/**
	 * Creates a new <code>Configuration</code>.
	 * 
	 * @param name the configuration name.
	 */
	public Configuration(String name)
	{
		this.name = name;
	}

	/**
	 * Returns the configuration name.
	 * 
	 * @return the configuration name.
	 */
	public String getName()
	{
		return name;
	}

	/**
	 * Adds an observation to this configuration.
	 * 
	 * @param name the name of the observation, which usually correspond to the
	 *        name of the instances (test).
	 * @param observation the observation to add.
	 */
	public void add(String name, Observation observation)
	{
		observations.put(name, observation);
	}

	/**
	 * Returns the obsservation associated with the specified name.
	 * 
	 * @param name the name of the observation.
	 * 
	 * @return the obsservation associated with the specified name.
	 */
	public Observation getObservation(String name)
	{
		return observations.get(name);
	}

	/**
	 * Returns the value of the accumulated rank of this configuration.
	 * 
	 * @return the value of the accumulated rank of this configuration.
	 */
	public double getRank()
	{
		return rank;
	}

	/**
	 * Sets the accumulated rank of this configuration over the instances
	 * used in the current step of the race.
	 * 
	 * @param rank the rank to set.
	 */
	public void setRank(double rank)
	{
		this.rank = rank;
	}

	/**
	 * Compares this configuration against the specified configuration.
	 * 
	 * @return <code>0</code> if both configuration have the same accumulated
	 *         rank; a negative integer if this configuration has a higher
	 *         (worse) rank; a positive integer if this configuration has a
	 *         lower (better) rank.
	 */
	public int compareTo(Configuration o)
	{
		return Double.compare(rank, o.rank);
	}
}