/*
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2011 Fernando Esteban Barril Otero
 */

package net.sf.myra.gui.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FilenameFilter;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import net.sf.myra.gui.parser.AccuracyParser;

/**
 * This class represents the F-Race optimisation method. The basic ideia is
 * to test various configurations (set of parameters) of an algorithm on a
 * limited set of instances (test) in a racing approach, and remove the ones
 * that are less promosing as soon as a statistical difference is found. The
 * ones that survive are allowed to continue to the next step. At the end of
 * the 'race', the one with the highest rank is considered the best
 * configuration.
 */
public class FRace
{
	/**
	 * Chi-square distribution for alpha = 0.05 (95% probability) and degrees
	 * of freedom = 75.
	 */
	private static final double[] CHI_SQUARE = { 0.004, 0.103, 0.352, 0.711,
		1.145, 1.635, 2.167, 2.733, 3.325, 3.940, 4.575, 5.226, 5.892, 6.571,
		7.261, 7.962, 8.672, 9.390, 10.117, 10.851, 11.591, 12.338, 13.091,
		13.848, 14.611, 15.379, 16.151, 16.928, 17.708, 18.493, 19.281, 20.072,
		20.867, 21.664, 22.465, 23.269, 24.075, 24.884, 25.695, 26.509, 27.326,
		28.144, 28.965, 29.787, 30.612, 31.439, 32.268, 33.098, 33.930, 34.764,
		35.600, 36.437, 37.276, 38.116, 38.958, 39.801, 40.646, 41.492, 42.339,
		43.188, 44.038, 44.889, 45.741, 46.595, 47.450, 48.305, 49.162, 50.020,
		50.879, 51.739, 52.600, 53.462, 54.325, 55.189, 56.054 };

	/**
	 * Student's T distributions for alpha = 0.05 (0.025 one tail) and degrees
	 * of freedom = 75.
	 */
	private static final double[] STUDENT_T = { 12.71, 4.303, 3.182, 2.776,
		2.571, 2.447, 2.365, 2.306, 2.262, 2.228, 2.201, 2.179, 2.160, 2.145,
		2.131, 2.120, 2.110, 2.101, 2.093, 2.086, 2.080, 2.074, 2.069, 2.064,
		2.060, 2.056, 2.052, 2.048, 2.045, 2.042, 2.040, 2.037, 2.035, 2.032,
		2.030, 2.028, 2.026, 2.024, 2.023, 2.021, 2.020, 2.018, 2.017, 2.015,
		2.014, 2.013, 2.012, 2.011, 2.010, 2.009, 2.008, 2.007, 2.006, 2.005,
		2.004, 2.003, 2.002, 2.002, 2.001, 2.000, 2.000, 1.999, 1.998, 1.998,
		1.997, 1.997, 1.996, 1.995, 1.995, 1.994, 1.994, 1.993, 1.993, 1.993,
		1.992 };

	/**
	 * Entry point. The only parameter expected is the directory with the
	 * output files for each instance (test). Current implementation expect
	 * that the name of the configuration is part of the output file and
	 * starts with the substring '-c'. Additionally, the output file should
	 * include the number of the fold, since each fold of the cross-validation
	 * should have its own file, e.g.:
	 * 
	 * <ul>
	 *   <li>automobile-0-c10-e0.85-u10.out</li>
	 *   <li>automobile-1-c10-e0.85-u10.out</li>
	 *   <li>...</li>
	 *   <li>automobile-9-c10-e0.85-u10.out</li>
	 * </ul>
	 * 
	 * @param args the command line arguments.
	 * 
	 * @throws Exception if any error occurs when reading the output files.
	 */
	public static void main(String[] args)
		throws Exception
	{
		File directory = new File(args[0]);
		AccuracyParser parser = new AccuracyParser();

		// the output file names
		TreeSet<String> total = new TreeSet<String>();
		// the pattern of the output files, so we can group the files from the
		// cross-validation
		Set<String> stems = new TreeSet<String>();

		Map<String, Configuration> configurations =
			new HashMap<String, Configuration>();

		// (1) reads all the output files

		for (String file : directory.list(new FilenameFilter()
			{
				public boolean accept(File dir, String name)
				{
					return name.endsWith(".out");
				}
			}))
		{
			StringBuffer buffer = new StringBuffer();
			BufferedReader reader =
				new BufferedReader(new FileReader(new File(directory, file)));
			String line = null;

			while ((line = reader.readLine()) != null)
			{
				buffer.append(line);
				buffer.append("\n");
			}

			reader =
				new BufferedReader(new StringReader(parser.parse(buffer
					.toString())));
			line = null;
			String last = null;

			while ((line = reader.readLine()) != null)
			{
				last = line;
			}

			Double v = new Double(last.substring(0, last.indexOf(' ')));

			String[] slices = file.split("-c\\d");
			String observation = slices[0];
			total.add(observation);
			stems.add(observation.substring(0, observation.lastIndexOf('-')));

			slices = file.split("-\\d-");
			String current = slices[1].substring(0, slices[1].lastIndexOf('.'));

			Configuration config = configurations.get(current);

			if (config == null)
			{
				config = new Configuration(current);
				configurations.put(current, config);
			}

			config.add(observation, new Observation(v));
		}

		// (2) now it is time to race
		System.out.println("Starting the F-Race...");

		TreeSet<String> available = new TreeSet<String>();
		List<String> patterns = new ArrayList<String>(stems);
		//Collections.shuffle(patterns);
		Collections.sort(patterns);

		for (String pattern : patterns)
		{
			// files from the same dataset are considered in the same group

			for (String t : total)
			{
				if (t.startsWith(pattern))
				{
					available.add(t);
				}
			}

			System.out.println("processing pattern: " + pattern + " ("
				+ available.size() + " instances, " + configurations.size()
				+ " configurations)");

			for (String instance : available)
			{
				ArrayList<Observation> observations =
					new ArrayList<Observation>();

				for (Configuration config : configurations.values())
				{
					observations.add(config.getObservation(instance));
				}

				computeRanks(observations);
			}

			race(configurations, available);

			if (configurations.size() == 1)
			{
				break;
			}
		}

		System.out.println("\nNon-dominated configurations ("
			+ configurations.size() + "):");
		ArrayList<Configuration> candidates =
			new ArrayList<Configuration>(configurations.values());
		Collections.sort(candidates);

		for (Configuration c : candidates)
		{
			System.out.println((c.getRank() / available.size()) + ": "
				+ c.getName());
		}
	}

	/**
	 * Computes the ranks of each observation by comparing the performance
	 * across all configurations.
	 * 
	 * @param observations the observations of all configuration in the race.
	 */
	private static void computeRanks(ArrayList<Observation> observations)
	{
		Collections.sort(observations);

		double rank = 1;
		int previous = -1;

		for (int i = 0; i < observations.size(); i++)
		{
			Observation o1 = observations.get(i);

			if ((i + 1) < observations.size())
			{
				Observation o2 = observations.get(i + 1);

				if (o1.getValue() > o2.getValue())
				{
					// if the value is unique, we just set the rank
					if (previous == -1)
					{
						o1.setRank(rank);
					}
					// the value is not unique, so the rank should be
					// averaged
					else
					{
						int size = (i - previous) + 1;

						for (int j = 0; j < size; j++)
						{
							observations.get(previous + j).setRank(rank / size);
						}

						previous = -1;
					}
				}
				else if (previous == -1)
				{
					// we are going to average the rank
					previous = i;
				}
			}
			// we are the last one
			else
			{
				// if the value is unique, we just set the rank
				if (previous == -1)
				{
					o1.setRank(rank);
				}
				// the value is not unique, so the rank should be
				// averaged
				else
				{
					int size = (i - previous) + 1;

					for (int j = 0; j < size; j++)
					{
						observations.get(previous + j).setRank(rank / size);
					}

					previous = -1;
				}
			}

			if (previous == -1)
			{
				rank = (i + 2);
			}
			else
			{
				rank += (i + 2);
			}
		}
	}

	/**
	 * Performs the statistical analysis for the current step in the race,
	 * removing the configurations that are statistically worse than the
	 * best one.
	 * 
	 * @param configurations the current configurations.
	 * @param available the instances (tests) used in the current step.
	 */
	private static void race(Map<String, Configuration> configurations,
		Set<String> available)
	{
		final double n = configurations.size();
		final double k = available.size();
		final int df = (int) n - 1;

		// (1) perform a friedman test

		double numerator = 0;

		for (Configuration config : configurations.values())
		{
			double rank = 0;

			for (String instance : available)
			{
				rank += config.getObservation(instance).getRank();
			}

			config.setRank(rank);

			rank = rank - ((k * (n + 1)) / 2);
			numerator += (rank * rank);
		}

		numerator = (n - 1) * numerator;
		double denominator = 0;
		double constant = (k * n * ((n + 1) * (n + 1))) / 4.0;

		for (String instance : available)
		{
			for (Configuration config : configurations.values())
			{
				double rank = config.getObservation(instance).getRank();
				denominator += ((rank * rank));
			}
		}

		double friedman = numerator / (denominator - constant);

		// do we have statistically significant differences?

		if (friedman > CHI_SQUARE[df])
		{
			ArrayList<Configuration> candidates =
				new ArrayList<Configuration>(configurations.values());
			Collections.sort(candidates);
			Configuration best = candidates.get(0);

			double m1 = 2 * k * (1 - (friedman / (k * (n - 1))));
			double pairDenominator =
				(m1 * (denominator - constant)) / ((k - 1) * n - 1);

			// (2) let's perform pairwise comparisons

			for (int i = 1; i < candidates.size(); i++)
			{
				Configuration other = candidates.get(i);

				double pairwise =
					Math.abs(best.getRank() - other.getRank())
						/ Math.sqrt(pairDenominator);

				if (pairwise > STUDENT_T[df])
				{
					// the current configuration is significantly worse than
					// the best one
					configurations.remove(other.getName());
				}
			}
		}
	}
}