/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.hantminer;

import java.util.ArrayList;
import java.util.Collection;
import net.sf.myra.datamining.Rule;
import net.sf.myra.framework.Cost;
import net.sf.myra.framework.Trail;
import net.sf.myra.framework.Vertex;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class HybridRule
	extends Trail
	implements Comparable<HybridRule>
{
	private static final long serialVersionUID = 1665139788722186776L;

	/**
	 * The antecedent of the rule.
	 */
	private Rule antecedent;

	/**
	 * The consequent of the rule.
	 */
	private Trail consequent;

	/**
	 * Default constructor.
	 * 
	 * @param antecedent the rule instance representing the antecedent.
	 * @param consequent the trail instance representing the consequent.
	 */
	public HybridRule(Rule antecedent, Trail consequent)
	{
		this.antecedent = antecedent;
		this.consequent = consequent;

		ArrayList<String> values = new ArrayList<String>();

		for (Vertex<?> vertex : consequent.getVertices())
		{
			values.add(vertex.getLabel());
		}

		antecedent.setConsequent(antecedent.getConsequent().toLabel(values));
	}

	/**
	 * Returns the antecedent part of the rule.
	 * 
	 * @return the antecedent part of the rule.
	 */
	public Rule getAntecedent()
	{
		return antecedent;
	}

	/**
	 * Returns the consequent part of the rule.
	 * 
	 * @return the consequent part of the rule.
	 */
	public Trail getConsequent()
	{
		return consequent;
	}

	/**
	 * Sets the rule cost.
	 * 
	 * @param cost the rule cost to set.
	 */
	public void setCost(Cost cost)
	{
		super.setCost(cost);
		antecedent.setCost(cost);
		consequent.setCost(cost);
	}

	/**
	 * Returns the rule cost.
	 * 
	 * @return the rule cost.
	 */
	public Cost getCost()
	{
		return antecedent.getCost();
	}

	/**
	 * Compares this <code>HybridRule</code> instance agains the specified
	 * <code>o</code> instance.
	 * 
	 * @param o the <code>HybridRule</code> instance to be compared.
	 */
	public int compareTo(HybridRule o)
	{
		Cost cost1 = antecedent.getCost();
		Cost cost2 = o.antecedent.getCost();

		if (cost1 == null)
		{
			return (cost2 == null) ? 0 : -1;
		}

		return (cost2 == null) ? 1 : cost1.compareTo(cost2);
	}

	@Override
	public final void add(Vertex<?> vertex)
	{
		throw new UnsupportedOperationException("Vertices cannot be added");
	}

	@Override
	public final void remove(Vertex<?> vertex)
	{
		throw new UnsupportedOperationException("Vertices cannot be removed");
	}

	@Override
	public final Vertex<?> getLast()
	{
		return antecedent.getLast();
	}

	@Override
	public final Collection<Vertex<?>> getVertices()
	{
		return antecedent.getVertices();
	}

	@Override
	public final boolean isEmpty()
	{
		return antecedent.isEmpty();
	}

	@Override
	public HybridRule clone()
	{
		HybridRule clone = (HybridRule) super.clone();
		clone.antecedent = antecedent.clone();
		clone.consequent = consequent.clone();

		return clone;
	}
}