/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.hantminer;

import net.sf.myra.datamining.Rule;
import net.sf.myra.framework.Cost;
import net.sf.myra.framework.Trail;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class Pruner
{
	/**
	 * The rule evaluator instance.
	 */
	private RuleEvaluator evaluator;

	/**
	 * Defautl constructor.
	 * 
	 * @param evaluator the rule evaluator instance used to evaluate the
	 *        candidate rules.
	 */
	public Pruner(RuleEvaluator evaluator)
	{
		this.evaluator = evaluator;
	}

	/**
	 * Returns a (potentially) improved rule by pruning both antecedent and
	 * consequent of the specified rule.
	 * 
	 * @param rule the rule to be pruned.
	 * 
	 * @return a (potentially) improved rule by pruning both antecedent and
	 *         consequent of the specified rule.
	 */
	public HybridRule explore(HybridRule rule)
	{
		HybridRule rBest = rule;
		Cost qBest = evaluator.evaluate(rBest);

		// prune procedure works on a copy of the original rule
		rule = rule.clone();

		do
		{
			HybridRule candidate = prune(rule, evaluator.evaluate(rule));
			Cost cost = evaluator.evaluate(candidate);

			if (cost.compareTo(qBest) > -1)
			{
				rBest = candidate;
				qBest = cost;
			}
			else
			{
				// we could not improve the best rule
				break;
			}

			// if the rule is already empty, the while test will break the loop
			if (!rule.isEmpty())
			{
				rule.getAntecedent().remove(rule.getAntecedent().getLast());
			}
		}
		while (rule.getAntecedent().getSize() > 0);

		rBest.setCost(qBest);
		return rBest;
	}

	/**
	 * Returns a (potentially) better rule by removing one class label at a
	 * time from the rule consequent.
	 * 
	 * @param candidate the candidate rule to be pruned.
	 *
	 * @return a (potentially) better rule by removing one class label at a
	 * time from the rule consequent.
	 */
	private HybridRule prune(HybridRule candidate, Cost cost)
	{
		// prune procedure works on a copy of the original rule
		candidate = candidate.clone();

		Rule antecedent = candidate.getAntecedent();
		Trail consequent = candidate.getConsequent().clone();

		while (consequent.getSize() > 2)
		{
			consequent.remove(consequent.getLast());

			Rule clone = antecedent.clone();
			HybridRule r = new HybridRule(clone, consequent.clone());
			Cost c = evaluator.evaluate(r);

			// if the rule quality has improved
			if (c.compareTo(cost) > 0)
			{
				candidate = r;
				cost = c;
			}
		}

		return candidate;
	}

	public RuleEvaluator getEvaluator()
	{
		return evaluator;
	}
}