/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.hantminer;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import net.sf.myra.cantminer.EdgePheromoneUpdater;
import net.sf.myra.cantminer.EdgeTrailFactory;
import net.sf.myra.cantminer.cAntMiner2;
import net.sf.myra.datamining.Rule;
import net.sf.myra.datamining.data.ClassHierarchy;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.EntropyHeuristicInformation;
import net.sf.myra.datamining.data.Term;
import net.sf.myra.framework.Ant;
import net.sf.myra.framework.Colony;
import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.DecisionPolicy;
import net.sf.myra.framework.DefaultTrailFactory;
import net.sf.myra.framework.Edge;
import net.sf.myra.framework.Environment;
import net.sf.myra.framework.Graph;
import net.sf.myra.framework.HeuristicInformation;
import net.sf.myra.framework.HeuristicValue;
import net.sf.myra.framework.NoAvailableVertexException;
import net.sf.myra.framework.Pheromone;
import net.sf.myra.framework.RandomGenerator;
import net.sf.myra.framework.StopCondition;
import net.sf.myra.framework.Trail;
import net.sf.myra.framework.Vertex;
import net.sf.myra.framework.Info.Void;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class RuleFactory
{
	/**
	 * The property key under which the number of ants updating the pheromone
	 * is stored.
	 */
	public final static String UPDATE_SIZE = "net.sf.myra.hantminer.update";

	/**
	 * The species responsible for creating rule antecedents.
	 */
	private Species antecedent;

	/**
	 * The species responsible for creating rule consequents.
	 */
	private Species consequent;

	/**
	 * The termination condition.
	 */
	private StopCondition condition;

	/**
	 * The running environment.
	 */
	private Environment environment;

	/**
	 * The objective function used evaluate the ant trails (solutions).
	 */
	private RuleEvaluator evaluator;

	/**
	 * The number of ants updating pheromone levels at each iteration.
	 */
	private int updateSize;

	/**
	 * Default constructor.
	 * 
	 * @param dataset the dataset used to extract rules.
	 */
	public RuleFactory(Dataset dataset, StopCondition condition,
			RuleEvaluator evaluator)
	{
		this.condition = condition;
		this.evaluator = evaluator;
		this.updateSize = Configuration.getIntProperty(UPDATE_SIZE, 1);

		// initializes the antecedent species

		Graph<Term> graph = dataset.getMetadata().toGraph();
		Vertex<Term> start = new Vertex<Term>(cAntMiner2.VIRTUAL_VERTEX);
		graph.add(start);

		// creates an edge between the start vertex to all vertices (excluding
		// the start vertex) of the graph
		Vertex<Term>[] vertices = graph.vertices();

		for (int i = 0; i < vertices.length; i++)
		{
			if (!vertices[i].equals(start))
			{
				graph.connect(start, vertices[i]);
			}
		}

		antecedent =
			new Species(new EdgePheromoneUpdater(),
				new EntropyHeuristicInformation(dataset), new EdgeTrailFactory(
					graph, dataset));
		environment = new Environment(antecedent.getColony(), null);

		// initializes the consequent species

		ClassHierarchy h = dataset.getMetadata().getClassHierarchy();

		consequent =
			new Species(new ConsequentPheromoneUpdater(),
				new ConsequentHeuristicInformation(dataset),
				new DefaultTrailFactory(h.toGraph(),
					new ConsequentDecisionPolicy(h.getRoot().getLabel())));
	}

	private List<HybridRule> evolve()
	{
		List<HybridRule> rules =
			new ArrayList<HybridRule>(environment.getColony().getSize());
		Pruner pruner = new Pruner(evaluator);

		antecedent.iterate();
		consequent.iterate();

		Iterator<Ant> iterA = antecedent.getColony().getPopulation().iterator();
		Iterator<Ant> iterB = consequent.getColony().getPopulation().iterator();

		while (iterA.hasNext() && iterB.hasNext())
		{
			Ant antA = iterA.next();
			Ant antB = iterB.next();

			HybridRule rule =
				new HybridRule((Rule) antA.getTrail(), antB.getTrail());

			rule = pruner.explore(rule);
			antA.setTrail(rule.getAntecedent());
			antB.setTrail(rule.getConsequent());

			rule.getAntecedent().setIteration(environment.getIteration());

			rules.add(rule);
		}

		return rules;
	}

	private void update(List<HybridRule> rules)
	{
		Collections.sort(rules, new Comparator<HybridRule>()
		{
			public int compare(HybridRule o1, HybridRule o2)
			{
				int c = o1.compareTo(o2);
				return (c == 0) ? 0 : -c;
			}
		});

		for (HybridRule rule : rules.subList(0, updateSize))
		{
			antecedent.update(rule.getAntecedent());
			consequent.update(rule.getConsequent());
		}
	}

	public Rule create()
	{
		antecedent.initialize();
		consequent.initialize();

		while (!condition.isReached(environment))
		{
			environment.increment();

			List<HybridRule> rules = evolve();
			update(rules);
		}

		return (Rule) antecedent.getColony().getBest().getBestTrail();
	}

	/**
	 * Decision policy to create the consequent trails.
	 */
	private static class ConsequentDecisionPolicy
		implements DecisionPolicy
	{
		/**
		 * The root vertex label of the hierarchy.
		 */
		private String root;

		/**
		 * Default constructor.
		 * 
		 *  @param root the root vertex label of the hierarchy.
		 */
		public ConsequentDecisionPolicy(String root)
		{
			this.root = root;
		}

		@SuppressWarnings("unchecked")
		public Vertex<?> select(Graph<?> graph, Trail trail)
			throws NoAvailableVertexException
		{
			Graph<Void> g = (Graph<Void>) graph;
			Vertex<Void> last = (Vertex<Void>) trail.getLast();

			if (last == null)
			{
				return g.get(root);
			}

			Collection<Edge<Void>> neighbors = g.getEdges(last);
			Map<Edge<?>, Double> roulette = probabilities(neighbors);

			if (!roulette.isEmpty())
			{
				double probability = RandomGenerator.getInstance().nextDouble();
				double position = 0.0;

				for (Entry<Edge<?>, Double> entry : roulette.entrySet())
				{
					position += entry.getValue();

					if (probability <= position)
					{
						return entry.getKey().getTo();
					}
				}
			}

			throw new NoAvailableVertexException("Empty neighbors list.");
		}

		Map<Edge<?>, Double> probabilities(Collection<Edge<Void>> neighbors)
		{
			Map<Edge<?>, Double> roulette = new HashMap<Edge<?>, Double>();

			double total = 0.0;

			for (Edge<Void> edge : neighbors)
			{
				total +=
					(edge.getPheromone().getValue() * edge.getTo()
						.getHeuristicValue().getValue());
			}

			for (Edge<Void> edge : neighbors)
			{
				double probability =
					(edge.getPheromone().getValue() * edge.getTo()
						.getHeuristicValue().getValue());

				roulette.put(edge, (probability > 0) ? probability / total : 0);
			}

			return roulette;
		}
	}

	/**
	 * Pheromone trails policy for consequent trails.
	 */
	private static class ConsequentPheromoneUpdater
		extends EdgePheromoneUpdater
	{
		@Override
		@SuppressWarnings("unchecked")
		public void update(Trail trail, Colony colony)
		{
			Collection<Vertex<?>> used = trail.getVertices();
			Graph<Void> g = (Graph<Void>) colony.getTrailFactory().getGraph();
			Vertex<Void> from = null;

			// updates the pheromone value of the used terms (vertices)
			for (Vertex<?> to : used)
			{
				if (from != null)
				{
					Edge<?> edge = g.getEdge(from, (Vertex<Void>) to);
					double value = edge.getPheromone().getValue();

					edge.setPheromone(new Pheromone(value
						+ (value * trail.getCost().getValue())));
				}

				from = (Vertex<Void>) to;
			}
		}
	}

	private static class ConsequentHeuristicInformation
		implements HeuristicInformation
	{
		private Dataset dataset;

		ConsequentHeuristicInformation(Dataset dataset)
		{
			this.dataset = dataset;
		}

		public void initialize(Graph<?> graph)
		{
			for (Vertex<?> vertex : graph.getVertices())
			{
				vertex.setHeuristicValue(new HeuristicValue(dataset
					.getInstances(vertex.getLabel()).size()));
			}
		}
	}
}