/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.hantminer;

import java.util.Properties;
import net.sf.myra.datamining.Classifier;
import net.sf.myra.datamining.ConvergenceTest;
import net.sf.myra.datamining.Evaluator;
import net.sf.myra.datamining.IterationTest;
import net.sf.myra.datamining.Model;
import net.sf.myra.datamining.Rule;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Label;
import net.sf.myra.datamining.model.RuleList;
import net.sf.myra.framework.CompoundStopCondition;
import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.ObjectFactory;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class hAntMiner
	extends Classifier
{
	/**
	 * Default options filename.
	 */
	public static final String OPTIONS_FILE = "myra-hantminer.messages";

	/**
	 * The property key under which the rule evaluator class name is stored.
	 */
	public final static String RULE_EVALUATOR =
		"net.sf.myra.hantminer.evaluator";

	/**
	 * Default constructor.
	 */
	public hAntMiner()
	{
		super(OPTIONS_FILE);
	}

	@Override
	public Properties filter(Properties properties)
	{
		properties.remove(RULE_QUALITY_FUNCTION);
		return properties;
	}

	@Override
	public Model run(Dataset dataset)
	{
		if (!dataset.getMetadata().isHierarchical())
		{
			throw new IllegalArgumentException("Not a hierarchical dataset.");
		}

		int uncovered =
			Configuration.getIntProperty(MAXIMUM_UNCOVERED_CASES,
				DEFAULT_MAXIMUM_UNCOVERED_CASES);

		Label majority = Evaluator.findMajorityClass(dataset);

		RuleList discovered = new RuleList();

		while (dataset.getSize() > uncovered)
		{
			Rule rule = evolve(dataset);
			dataset.remove(Evaluator.findCoveredCases(rule, dataset));
			discovered.add(rule.clone());
		}

		if (!discovered.hasDefault())
		{
			if (!dataset.isEmpty())
			{
				// re-calculate the majority class considering only the
				// examples that were not covered by any rule
				majority = Evaluator.findMajorityClass(dataset);
			}

			Rule rule = new Rule(dataset.getMetadata());
			rule.setConsequent(majority);
			discovered.add(rule);
		}

		return discovered;
	}

	/**
	 * Evolves a new <code>Rule</code> based on the specified dataset.
	 * 
	 * @param dataset the dataset to extract the rule.
	 * 
	 * @return a new <code>Rule</code> based on the specified dataset.
	 */
	private Rule evolve(Dataset dataset)
	{
		CompoundStopCondition condition = new CompoundStopCondition();
		condition.add(new ConvergenceTest());
		condition.add(new IterationTest());

		RuleEvaluator evaluator =
			(RuleEvaluator) ObjectFactory.create(Configuration.getProperty(
				RULE_EVALUATOR, hFmeasureEvaluator.class.getName()),
				new Class<?>[] { Dataset.class }, new Object[] { dataset });

		RuleFactory factory = new RuleFactory(dataset, condition, evaluator);

		Rule rule = factory.create();
		rule.setConsequent(dataset.getMetadata().expand(rule.getConsequent()));

		return rule;
	}
}