/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2009  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.hantminer.multilabel;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;
import java.util.concurrent.CyclicBarrier;

import net.sf.myra.cantminer.EdgePheromoneUpdater;
import net.sf.myra.cantminer.cAntMiner2;
import net.sf.myra.datamining.AbstractObjectiveFunction;
import net.sf.myra.datamining.BacktrackPruner;
import net.sf.myra.datamining.ClassAssignator;
import net.sf.myra.datamining.Classifier;
import net.sf.myra.datamining.ConvergenceTest;
import net.sf.myra.datamining.Evaluator;
import net.sf.myra.datamining.FixedClassAssignator;
import net.sf.myra.datamining.IterationTest;
import net.sf.myra.datamining.Main;
import net.sf.myra.datamining.Rule;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Label;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.data.Node;
import net.sf.myra.datamining.data.Term;
import net.sf.myra.datamining.function.FmeasureFunction;
import net.sf.myra.datamining.io.Helper;
import net.sf.myra.datamining.io.HierarchicalExporter;
import net.sf.myra.datamining.model.ProbabilisticRuleSet;
import net.sf.myra.datamining.model.RuleSet;
import net.sf.myra.framework.Colony;
import net.sf.myra.framework.CompoundStopCondition;
import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.Graph;
import net.sf.myra.framework.ObjectFactory;
import net.sf.myra.framework.Scheduler;
import net.sf.myra.framework.TrailFactory;
import net.sf.myra.framework.Vertex;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class cAntMinerHM
	extends Classifier
{
	/**
	 * Default options filename.
	 */
	public static final String OPTIONS_FILE = "myra-cantminer-hm.messages";

	/**
	 * The property key under which the number of workers threads value is
	 * stored. If not specified, there will be one worker thread per
	 * processor/core. 
	 */
	private static final String WORKERS =
		"net.sf.myra.hantminer.multilabel.workers";

	/**
	 * The property key under which the rules file value is stored. If this
	 * proerty is set, the discovered rule set will be serialized to the
	 * specified file. 
	 */
	private static final String RULES_FILE =
		"net.sf.myra.hantminer.multilabel.rules";

	/**
	 * The property key under which the ancestors flag is stored. If this value
	 * is <code>true</code>, the ancestor instances will be used as negative
	 * examples; <code>false</code>, the ancestor instances will not be used
	 * (neither as negative nor positive).
	 */
	private static final String ANCESTORS =
		"net.sf.myra.hantminer.multilabel.ancestors";

	/**
	 * The property key under which the descendants flag is stored. If this
	 * value is <code>true</code>, the ancestor instances will be used as
	 * negative examples; <code>false</code>, the ancestor instances will not
	 * be used (neither as negative nor positive).
	 */
	private static final String DESCENDANTS =
		"net.sf.myra.hantminer.multilabel.descendants";

	/**
	 * Default temp file prefix.
	 */
	private static final String PREFIX = cAntMinerHM.class.getSimpleName();

	/**
	 * Default temp file suffix.
	 */
	private static final String SUFFIX = ".ser";

	/**
	 * The synchronization barrier for worker threads.
	 */
	private CyclicBarrier barrier;

	/**
	 * The list of class label of the target classification problem.
	 */
	private LinkedList<String> labels;

	/**
	 * The current dataset.
	 */
	private Dataset dataset;

	/**
	 * The maximum number of uncovered examples (per class).
	 */
	private int uncovered;

	/**
	 * Default constructor.
	 */
	public cAntMinerHM()
	{
		super(OPTIONS_FILE);
	}

	@Override
	public Properties filter(Properties properties)
	{
		properties.remove(RULE_QUALITY_FUNCTION);
		return properties;
	}

	@Override
	@SuppressWarnings("unchecked")
	public ProbabilisticRuleSet run(Dataset dataset)
	{
		if (!dataset.getMetadata().isHierarchical())
		{
			throw new IllegalArgumentException("Not a hierarchical dataset.");
		}

		// ensures that the default temporary directory does not contain
		// files from a previous execution
		clean();

		initialize(dataset);

		// starts the worker threads

		for (int i = 0; i < (barrier.getParties() - 1); i++)
		{
			new Worker(i).start();
		}

		try
		{
			barrier.await();
		}
		catch (Exception e)
		{
			throw new RuntimeException(e);
		}

		// calculates the confidence factor of each rule

		ProbabilisticRuleSet discovered = new ProbabilisticRuleSet();

		try
		{
			for (File file : getPredictions())
			{
				ObjectInputStream in =
					new ObjectInputStream(new FileInputStream(file));
				List<Rule> rules = (List<Rule>) in.readObject();

				for (Rule rule : rules)
				{
					if (!rule.isEmpty())
					{
						discovered.add(rule);
					}
				}

				in.close();
			}
		}
		catch (Exception e)
		{
			throw new RuntimeException(e);
		}

		// adds a default rule (when needed)

		Rule rule = new Rule(dataset.getMetadata());
		rule.setConsequent(Evaluator.findMajorityClass(dataset));
		rule.setCost(new FmeasureFunction(dataset).evaluate(rule));
		discovered.add(rule);

		// serializes the discovered rule set

		serialize(discovered);

		return discovered;
	}

	/**
	 * Initialize the run parameters.
	 * 
	 * @param dataset the dataset to be classified.
	 */
	private void initialize(Dataset dataset)
	{
		this.uncovered =
			Configuration.getIntProperty(MAXIMUM_UNCOVERED_CASES,
				DEFAULT_MAXIMUM_UNCOVERED_CASES);

		this.barrier =
			new CyclicBarrier(Configuration.getIntProperty(WORKERS, Runtime
				.getRuntime().availableProcessors()) + 1);

		this.labels = new LinkedList<String>();
		this.dataset = dataset;

		for (Node node : dataset.getMetadata().getClassHierarchy().flat())
		{
			if (!node.isRoot())
			{
				labels.add(node.getLabel());
			}
		}
	}

	/**
	 * Returns the next class label to be processed, removing its reference
	 * from the label list.
	 * 
	 * @return the next class label to be processed; <code>null</code> when
	 *         the label list is empty.
	 */
	private synchronized String next()
	{
		return labels.isEmpty() ? null : labels.removeFirst();
	}

	/**
	 * Returns the prediction filenames found in the default temporary
	 * directory.
	 * 
	 * @return the prediction filenames.
	 */
	private File[] getPredictions()
	{
		File tempDir = new File(System.getProperty("java.io.tmpdir"));
		return tempDir.listFiles(new FilenameFilter()
		{
			/* (non-Javadoc)
			 * @see java.io.FilenameFilter#accept(java.io.File, java.lang.String)
			 */
			public boolean accept(File dir, String name)
			{
				return name.startsWith(PREFIX) && name.endsWith(SUFFIX);
			}
		});
	}

	/**
	 * Removes any prediction file from the default temporary directory. In a
	 * normal execution, the temporary files are removed at the end of the
	 * execution by the JVM. This method guarantees that at the begining of
	 * a new execution, there are no existent files from a previous execution.
	 * Hence, it is not possible to run the algorithm more than once at a time
	 * on the same machine.
	 */
	private void clean()
	{
		for (File file : getPredictions())
		{
			if (!file.delete())
			{
				throw new RuntimeException("Could not delete file: "
					+ file.getName());
			}
		}
	}

	/**
	 * Serializes the discovered rule set to a file, if the
	 * <code>RULES_FILE</code> was set.
	 * 
	 * @param rules the discovered rule set.
	 */
	private void serialize(ProbabilisticRuleSet rules)
	{
		String filename = Configuration.getProperty(RULES_FILE);

		if (filename != null)
		{
			File file = new File(filename);

			try
			{
				ObjectOutputStream out =
					new ObjectOutputStream(new BufferedOutputStream(
						new FileOutputStream(file)));
				out.writeObject(rules);
				out.close();
			}
			catch (IOException e)
			{
				if (file.exists())
				{
					file.delete();
				}

				throw new RuntimeException(
					"Error serializing discovered rule set.", e);
			}
		}
	}

	/**
	 * Utility main method to export predictions file.
	 * 
	 * @param args the command-line arguments.
	 */
	public static void main(String[] args)
		throws Exception
	{
		if (args.length < 2)
		{
			StringBuffer usage = new StringBuffer();
			usage.append("usage: java ");
			usage.append(cAntMinerHM.class.getName());
			usage.append(" <test file> <rules file>");

			System.out.println(usage.toString());
		}
		else
		{
			Dataset testSet = Helper.open(args[0]);

			ObjectInputStream in =
				new ObjectInputStream(new FileInputStream(args[1]));
			RuleSet rules = (RuleSet) in.readObject();
			in.close();

			HierarchicalExporter exporter =
				(HierarchicalExporter) ObjectFactory.create(Configuration
					.getProperty(Main.EXPORTER, HierarchicalExporter.class
						.getName()));

			exporter.initialize(new File(Configuration
				.getProperty(Main.EXPORTER_FILE)), testSet.getMetadata());
			exporter.write(testSet, rules);
			exporter.close();
		}
	}

	/**
	 * This class is responsible for discovering the classification rules.
	 */
	private class Worker
		extends Thread
	{
		/**
		 * Default constructor.
		 * 
		 *  @param id the worker id.
		 */
		public Worker(int id)
		{
			super("hAnt-MinerMC::worker" + id);
		}

		@Override
		public void run()
		{
			// should use ancestor instances during training?
			boolean ancestors =
				"true".equals(Configuration.getProperty(ANCESTORS, "true"));

			// should use ancestor instances during training?
			boolean descendants =
				"true".equals(Configuration.getProperty(DESCENDANTS, "true"));

			// the target class label
			String target = null;

			while (!isInterrupted() && (target = next()) != null)
			{
				Metadata metadata = dataset.getMetadata();
				Label label = metadata.getLabel();
				label = metadata.expand(label.toLabel(target));

				List<Rule> rules = new ArrayList<Rule>();
				Dataset local = dataset.clone();

				if (!ancestors)
				{
					removeAncestorInstances(local, target);
				}

				if (!descendants)
				{
					removeDescendantInstances(local, target);
				}

				if (local.getSize(target) > 0)
				{
					do
					{
						Rule rule = evolve(local, label, target);
						local.remove(Evaluator.findPositiveCases(rule, local));
						rules.add(rule.clone());
					}
					while (local.getSize(target) > uncovered);
				}

				try
				{
					// writes the discovered rules to a temp file
					File predictions = File.createTempFile(PREFIX, SUFFIX);
					predictions.deleteOnExit();

					ObjectOutputStream output =
						new ObjectOutputStream(
							new FileOutputStream(predictions));
					output.writeObject(rules);
					output.close();
				}
				catch (IOException e)
				{
					throw new RuntimeException(
						"Error writing the discovered rules for label: "
							+ target, e);
				}
			}

			try
			{
				// waits until all workers have finished
				barrier.await();
			}
			catch (Exception e)
			{
				throw new RuntimeException(e);
			}
		}

		/**
		 * Ant Colony Optimization (ACO) procedure to evolve a rule.
		 * 
		 * @param dataset
		 * @return
		 */
		private Rule evolve(Dataset dataset, Label label, String c)
		{
			Graph<Term> graph = dataset.getMetadata().toGraph();
			Vertex<Term> start = new Vertex<Term>(cAntMiner2.VIRTUAL_VERTEX);
			graph.add(start);

			// creates an edge between the start vertex to all vertices (excluding
			// the start vertex) of the graph
			Vertex<Term>[] vertices = graph.vertices();

			for (int i = 0; i < vertices.length; i++)
			{
				if (!vertices[i].equals(start))
				{
					graph.connect(start, vertices[i]);
				}
			}

			TrailFactory factory =
				new ClassAwareTrailFactory(graph, dataset, c);

			Colony colony = new Colony();
			colony.setTrailFactory(factory);

			CompoundStopCondition condition = new CompoundStopCondition();
			condition.add(new ConvergenceTest());
			condition.add(new IterationTest());

			ClassAssignator assignator = new FixedClassAssignator(label);
			AbstractObjectiveFunction function = new FmeasureFunction(dataset);

			Scheduler scheduler =
				new Scheduler(colony, condition, function, new EdgePheromoneUpdater(),
					new ClassAwareHeuristicInformation(dataset, c));
			scheduler.setLocalSearch(new BacktrackPruner(function, assignator));

			scheduler.run();

			// get the best ant trail constructed
			return (Rule) colony.getBest().getBestTrail();
		}

		/**
		 * Removes the ancestor instances of the specified class label.
		 * 
		 * @param dataset the dataset instances.
		 * @param target the class label.
		 */
		private void removeAncestorInstances(Dataset dataset, String target)
		{
			Metadata metadata = dataset.getMetadata();
			Node node = metadata.getClassHierarchy().find(target);

			for (Node ancestor : node.getAncestors())
			{
				Label label = metadata.getLabel().toLabel(ancestor.getLabel());
				dataset.remove(dataset.getInstances(metadata.expand(label)));
			}
		}

		/**
		 * Removes the descendant instances of the specified class label.
		 * 
		 * @param dataset the dataset instances.
		 * @param target the class label.
		 */
		private void removeDescendantInstances(Dataset dataset, String target)
		{
			Metadata metadata = dataset.getMetadata();
			Node node = metadata.getClassHierarchy().find(target);

			for (Node descendant : node.getDescendants())
			{
				Label l = metadata.getLabel().toLabel(descendant.getLabel());
				dataset.remove(dataset.getInstances(metadata.expand(l)));
			}
		}
	}
}