/*
 * $Id$ 
 */

package net.sf.myra.hantminer.pittsburgh;

import static net.sf.myra.cantminer.cAntMiner2.VIRTUAL_VERTEX;
import static net.sf.myra.datamining.ConvergenceTest.CONVERGENCE_TEST_SIZE;
import static net.sf.myra.datamining.ConvergenceTest.DEFAULT_CONVERGENCE_TEST_SIZE;
import static net.sf.myra.datamining.IterationTest.DEFAULT_MAXIMUM_ITERATIONS;
import static net.sf.myra.datamining.IterationTest.MAXIMUM_ITERATIONS;
import static net.sf.myra.framework.Graph.EDGE_EVAPORATION;
import static net.sf.myra.framework.Graph.EDGE_INITIALIZATION;
import static net.sf.myra.framework.ParallelScheduler.WORKERS;

import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadFactory;

import net.sf.myra.cantminer.EdgeTrailFactory;
import net.sf.myra.datamining.AbstractObjectiveFunction;
import net.sf.myra.datamining.BacktrackPruner;
import net.sf.myra.datamining.ClassAssignator;
import net.sf.myra.datamining.ClassFrequencyAssignator;
import net.sf.myra.datamining.Classifier;
import net.sf.myra.datamining.Evaluator;
import net.sf.myra.datamining.Measure;
import net.sf.myra.datamining.Model;
import net.sf.myra.datamining.ProbabilisticRule;
import net.sf.myra.datamining.Rule;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.DistanceContent;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Term;
import net.sf.myra.datamining.function.hierarchical.VarianceGainFunction;
import net.sf.myra.datamining.io.Helper;
import net.sf.myra.datamining.measure.hierarchical.PooledPRCurveMeasure;
import net.sf.myra.datamining.model.ProbabilisticRuleList;
import net.sf.myra.framework.Ant;
import net.sf.myra.framework.Colony;
import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.Edge;
import net.sf.myra.framework.Graph;
import net.sf.myra.framework.HeuristicValue;
import net.sf.myra.framework.LocalSearch;
import net.sf.myra.framework.ObjectFactory;
import net.sf.myra.framework.ObjectiveFunction;
import net.sf.myra.framework.Pheromone;
import net.sf.myra.framework.Trail;
import net.sf.myra.framework.Vertex;

/**
 * This class represents the hmAnt-MinerPB classification algorithm.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class hmAntMinerPB
	extends Classifier
{
	/**
	 * Default options filename.
	 */
	public static final String OPTIONS_FILE = "myra-hmantminer-pb.messages";

	/**
	 * The property key under which the niches value is stored.
	 */
	public static final String NICHES =
		"net.sf.myra.hantminer.pittsburgh.niches";

	/**
	 * The property key under which the confidence measure class name is stored.
	 * This measure is used to compute the quality of a complete list of rules.
	 */
	public static final String CONFIDENCE_MEASURE =
		"net.sf.myra.hantminer.pittsburgh.confidence.measure";

	/**
	 * The property key under which the early stopping validation set is stored.
	 */
	private static final String VALIDATION_SET =
		"net.sf.myra.hantminer.pittsburgh.validation";

	/**
	 * The measure used to set the confidence to rule lists.
	 *
	 * @see net.sf.myra.datamining.Main
	 */
	private final Measure measure =
		(Measure) ObjectFactory.create(Configuration.getProperty(
			CONFIDENCE_MEASURE, PooledPRCurveMeasure.class.getName()));

	/**
	 * The maximum number of uncovered cases.
	 */
	private final int uncovered =
		Configuration.getIntProperty(MAXIMUM_UNCOVERED_CASES,
			DEFAULT_MAXIMUM_UNCOVERED_CASES);

	/**
	 * The maximum number of iterations.
	 */
	private final int iterations =
		Configuration.getIntProperty(MAXIMUM_ITERATIONS,
			DEFAULT_MAXIMUM_ITERATIONS);

	/**
	 * The number of worker threads.
	 */
	private final int workers =
		Configuration.getIntProperty(WORKERS, Runtime.getRuntime()
			.availableProcessors());

	/**
	 * The number of niches (rule lists) per iteration. The default value is 1.
	 */
	private final int niches = Configuration.getIntProperty(NICHES, 1);

	/**
	 * The size of the convergence test. If a best-so-far rule list quality
	 * does not improve over <code>convergence</code> iterations, the search
	 * is finilised.
	 */
	private final int convergence =
		Configuration.getIntProperty(CONVERGENCE_TEST_SIZE,
			DEFAULT_CONVERGENCE_TEST_SIZE);

	/**
	 * Default constructor.
	 */
	public hmAntMinerPB()
	{
		super(OPTIONS_FILE);
	}

	@Override
	public Model run(Dataset dataset)
	{
		if (!dataset.getMetadata().isHierarchical())
		{
			throw new IllegalArgumentException("Not a hierarchical dataset.");
		}

		Dataset validation = null;
		double previous = Double.NaN;

		if (Configuration.isPresent(VALIDATION_SET))
		{
			try
			{
				String filename = Configuration.getProperty(VALIDATION_SET);
				validation = Helper.open(filename);
			}
			catch (IOException e)
			{
				throw new IllegalArgumentException(e);
			}
		}

		Graph<Term> graph = dataset.getMetadata().toGraph();
		Vertex<Term> start = new Vertex<Term>(VIRTUAL_VERTEX);
		graph.add(start);

		// creates an edge between the start vertex to all vertices (excluding
		// the start vertex) of the graph. Note that the construction graph
		// does not change during the search for rule lists 
		Vertex<Term>[] vertices = graph.vertices();

		for (int i = 0; i < vertices.length; i++)
		{
			if (!vertices[i].equals(start))
			{
				graph.connect(start, vertices[i]);
			}
		}

		graph.apply(EDGE_INITIALIZATION);
		ProbabilisticRuleList best = null;
		int test = 0;

		ExecutorService executor =
			Executors.newFixedThreadPool(workers, new ThreadFactory()
			{
				private int id = 0;

				public Thread newThread(Runnable r)
				{
					return new Thread(r, "hmAntMinerPB::worker" + (id++));
				}
			});

		for (int i = 0; i < iterations; i++)
		{
			ProbabilisticRuleList current = null;

			for (int j = 0; j < niches; j++)
			{
				ProbabilisticRuleList model = build(graph, dataset, executor);
				model.setConfidence(measure.evaluate(dataset, model));

				if (current == null || model.compareTo(current) > 0)
				{
					current = model;
				}
			}

			pheromone(graph, current);
			
			if (best == null || current.compareTo(best) > 0)
			{
				best = current;
				test = 0;
			}
			// convergence test (if we cannot improve the rule list quality
			// for 'convergence' iterations, we stop the search)
			else
			{
				test++;

				if (test == convergence)
				{
					break;
				}
			}

			if (validation != null)
			{
				double quality =
					measure.evaluate(validation, best).doubleValue();

				if (!Double.isNaN(previous) && (quality < previous))
				{
					// time to stop the search, quality on the validation set
					// has decreased
					break;
				}
				else
				{
					previous = quality;
				}
			}
		}

		// finilizes the executor threads
		executor.shutdown();

		return best;
	}

	private ProbabilisticRuleList build(Graph<Term> graph, Dataset dataset,
			ExecutorService executor)
	{
		ProbabilisticRuleList model = new ProbabilisticRuleList();
		Dataset local = dataset.clone();
		EdgeTrailFactory factory = new EdgeTrailFactory(graph, local);

		ClassAssignator assignator = new ClassFrequencyAssignator();
		ProbabilisticRule rule = new ProbabilisticRule(dataset.getMetadata());
		assignator.assign(rule, dataset);

		while (local.getSize() > uncovered)
		{
			AbstractObjectiveFunction function =
				(AbstractObjectiveFunction) ObjectFactory.create(Configuration
					.getProperty(RULE_QUALITY_FUNCTION,
						VarianceGainFunction.class.getName()),
					new Class<?>[] { Dataset.class }, new Object[] { local });

			LocalSearch pruner = new BacktrackPruner(function, assignator);
			Colony colony = new Colony();
			colony.setTrailFactory(factory);

			// updates the heuristic information
			heuristic(graph, local);

			CountDownLatch latch = new CountDownLatch(colony.getSize());

			for (Ant ant : colony.getPopulation())
			{
				executor.execute(new Task(ant, pruner, function, latch));
			}

			try
			{
				// waits for the creation of trails
				latch.await();
			}
			catch (InterruptedException e)
			{
				throw new RuntimeException(e);
			}

			Rule best = (Rule) colony.getCurrent().getTrail();

			if (best.getCost().getValue() > 0)
			{
				List<Instance> list = Evaluator.findCoveredCases(best, local);
				local.remove(list);
				model.add(best.clone());
			}
			else
			{
				break;
			}
		}

		if (!model.hasDefault())
		{
			if (!local.isEmpty())
			{
				assignator.assign(rule, local);
			}

			model.add(rule);
		}

		return model;
	}

	/**
	 * Computes the heuristic information based on the specified dataset.
	 * 
	 * @param graph the contruction graph.
	 * @param dataset the dataset.
	 */
	private void heuristic(Graph<Term> graph, Dataset dataset)
	{
		DistanceContent content = new DistanceContent(dataset);
		double maximum = content.maximum();

		for (Vertex<Term> vertex : graph.vertices())
		{
			if (!VIRTUAL_VERTEX.equals(vertex.getLabel()))
			{
				double v = content.variance(vertex.getLabel());
				v = Double.isInfinite(v) ? 0 : (maximum - v) / maximum;
				vertex.setHeuristicValue(new HeuristicValue(v));
			}
		}
	}

	/**
	 * Updates the pheromone levels, using the specified model quality.
	 * 
	 * @param graph the contruction graph.
	 * @param model the probabilistic rule model.
	 */
	@SuppressWarnings("unchecked")
	private void pheromone(Graph<Term> graph, ProbabilisticRuleList model)
	{
		HashMap<Vertex<Term>, HashSet<Vertex<Term>>> mapping =
			new HashMap<Vertex<Term>, HashSet<Vertex<Term>>>();

		for (Rule rule : model.getRules())
		{
			Vertex<Term> from = graph.get(VIRTUAL_VERTEX);

			for (Vertex<?> to : rule.getVertices())
			{
				HashSet<Vertex<Term>> neighbors = mapping.get(from);

				if (neighbors == null)
				{
					neighbors = new HashSet<Vertex<Term>>();
					mapping.put(from, neighbors);
				}

				from = (Vertex<Term>) to;
				neighbors.add(from);
			}
		}

		double quality = Double.parseDouble(model.getConfidence().toString());

		for (Vertex<Term> from : mapping.keySet())
		{
			for (Vertex<Term> to : mapping.get(from))
			{
				Edge<?> edge = graph.getEdge(from, to);
				double value = edge.getPheromone().getValue();
				edge.setPheromone(new Pheromone(value + (value * quality)));
			}
		}

		graph.apply(EDGE_EVAPORATION);
	}

	/**
	 * This class represents the ant walk's runnable task.
	 */
	private static class Task
		implements Runnable
	{
		/**
		 * The ant reference of this task.
		 */
		private Ant ant;

		/**
		 * The pruner reference of this task.
		 */
		private LocalSearch pruner;

		/**
		 * The objective function reference of this task.
		 */
		private ObjectiveFunction function;

		/**
		 * The synchronization count down aid.
		 */
		private CountDownLatch latch;

		/**
		 * Default constructor.
		 * 
		 * @param ant the ant reference of this task.
		 * @param pruner the pruner reference of this task.
		 * @param function the objective function reference of this task.
		 * @param latch the synchronization count down aid.
		 */
		Task(Ant ant, LocalSearch pruner, ObjectiveFunction function,
				CountDownLatch latch)
		{
			this.ant = ant;
			this.pruner = pruner;
			this.function = function;
			this.latch = latch;
		}

		public void run()
		{
			Trail trail = pruner.explore(ant.walk());
			ant.setTrail(trail);
			trail.setCost(function.evaluate(trail));
			latch.countDown();
		}
	}
}