/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2009  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.hantminer.probabilistic;

import net.sf.myra.cantminer.EdgePheromoneUpdater;
import net.sf.myra.cantminer.EdgeTrailFactory;
import net.sf.myra.cantminer.cAntMiner2;
import net.sf.myra.datamining.AbstractObjectiveFunction;
import net.sf.myra.datamining.BacktrackPruner;
import net.sf.myra.datamining.ClassAssignator;
import net.sf.myra.datamining.ClassFrequencyAssignator;
import net.sf.myra.datamining.Classifier;
import net.sf.myra.datamining.ConvergenceTest;
import net.sf.myra.datamining.Evaluator;
import net.sf.myra.datamining.IterationTest;
import net.sf.myra.datamining.ProbabilisticRule;
import net.sf.myra.datamining.Rule;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.DistanceHeuristicInformation;
import net.sf.myra.datamining.data.Metadata;
import net.sf.myra.datamining.data.Term;
import net.sf.myra.datamining.function.hierarchical.VarianceGainFunction;
import net.sf.myra.datamining.model.AbstractRuleModel;
import net.sf.myra.datamining.model.ProbabilisticRuleList;
import net.sf.myra.framework.Colony;
import net.sf.myra.framework.CompoundStopCondition;
import net.sf.myra.framework.Configuration;
import net.sf.myra.framework.Graph;
import net.sf.myra.framework.ObjectFactory;
import net.sf.myra.framework.ParallelScheduler;
import net.sf.myra.framework.Scheduler;
import net.sf.myra.framework.TrailFactory;
import net.sf.myra.framework.Vertex;

/**
 * This class represents the hmAnt-Miner classification algorithm.
 * 
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class hmAntMiner
	extends Classifier
{
	/**
	 * Default options filename.
	 */
	public static final String OPTIONS_FILE = "myra-hmantminer.messages";

	/**
	 * Default constructor.
	 */
	public hmAntMiner()
	{
		super(OPTIONS_FILE);
	}

	public AbstractRuleModel run(Dataset dataset)
	{
		if (!dataset.getMetadata().isHierarchical())
		{
			throw new IllegalArgumentException("Not a hierarchical dataset.");
		}

		AbstractRuleModel discovered = new ProbabilisticRuleList();
		Metadata metadata = dataset.getMetadata();

		int maxUncoveredCases =
			Configuration.getIntProperty(MAXIMUM_UNCOVERED_CASES,
				DEFAULT_MAXIMUM_UNCOVERED_CASES);

		ClassAssignator assignator = new ClassFrequencyAssignator();
		ProbabilisticRule defaultRule = new ProbabilisticRule(metadata);
		assignator.assign(defaultRule, dataset);

		while (dataset.getSize() > maxUncoveredCases)
		{
			Rule rule = evolve(dataset);

			// checks if the quality of the rule if greater than 0 (if it less,
			// it is not worth to use the rule)
			if (rule.getCost().getValue() > 0)
			{
				dataset.remove(Evaluator.findCoveredCases(rule, dataset));
				discovered.add(rule.clone());
			}
			// we shouldn't use the rule. The remaining uncovered cases will
			// be classified by the default rule, since we couldn't find a
			// good rule
			else
			{
				break;
			}
		}

		if (!discovered.hasDefault())
		{
			if (!dataset.isEmpty())
			{
				assignator.assign(defaultRule, dataset);
			}

			discovered.add(defaultRule);
		}

		return discovered;
	}

	/**
	 * Ant Colony Optimization (ACO) procedure to evolve a rule.
	 * 
	 * @param dataset
	 * @return
	 */
	private Rule evolve(Dataset dataset)
	{
		Graph<Term> graph = dataset.getMetadata().toGraph();
		Vertex<Term> start = new Vertex<Term>(cAntMiner2.VIRTUAL_VERTEX);
		graph.add(start);

		// creates an edge between the start vertex to all vertices (excluding
		// the start vertex) of the graph
		Vertex<Term>[] vertices = graph.vertices();

		for (int i = 0; i < vertices.length; i++)
		{
			if (!vertices[i].equals(start))
			{
				graph.connect(start, vertices[i]);
			}
		}

		TrailFactory factory = new EdgeTrailFactory(graph, dataset);

		Colony colony = new Colony();
		colony.setTrailFactory(factory);

		CompoundStopCondition condition = new CompoundStopCondition();
		condition.add(new ConvergenceTest());
		condition.add(new IterationTest());

		ClassAssignator assignator = new ClassFrequencyAssignator();
		AbstractObjectiveFunction function =
			(AbstractObjectiveFunction) ObjectFactory.create(Configuration
				.getProperty(RULE_QUALITY_FUNCTION, VarianceGainFunction.class
					.getName()), new Class<?>[] { Dataset.class },
				new Object[] { dataset });

		Scheduler scheduler =
			new ParallelScheduler(colony, condition, function,
				new EdgePheromoneUpdater(), new DistanceHeuristicInformation(
					dataset));
		scheduler.setLocalSearch(new BacktrackPruner(function, assignator));

		scheduler.run();

		// get the best ant trail constructed
		return (Rule) colony.getBest().getBestTrail();
	}
}