/*
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008-2012 Fernando Esteban Barril Otero
 */

package net.sf.myra.hantminer.util;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Map;

import net.sf.myra.cantminer.cAntMiner;
import net.sf.myra.datamining.Classifier;
import net.sf.myra.datamining.Gatherer;
import net.sf.myra.datamining.Model;
import net.sf.myra.datamining.data.ClassHierarchy;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Label;
import net.sf.myra.datamining.data.Node;
import net.sf.myra.datamining.data.Strategy;
import net.sf.myra.datamining.model.AbstractRuleModel;

/**
 * @author Fernando Esteban Barril Otero
 */
public class InternalNodeTopDown
	extends Classifier
{
	public InternalNodeTopDown()
	{
		super(null);
	}

	@Override
	public TopDownRuleModel run(Dataset dataset)
	{
		ClassHierarchy hierarchy = dataset.getMetadata().getClassHierarchy();
		TopDownRuleModel model = new TopDownRuleModel(hierarchy);
		Node root = hierarchy.getRoot();

		Map<Node, Dataset> mapping = new HashMap<Node, Dataset>();
		mapping.put(root, dataset);

		while (!mapping.isEmpty())
		{
			Node node = mapping.keySet().iterator().next();
			Dataset local = mapping.get(node);
			mapping.remove(node);

			// produces a classifier for the current node

			AbstractRuleModel rules = getClassifier(node, local);
			model.set(node.getLabel(), rules);

			// splits the dataset

			Map<String, Dataset> children = split(node, local, rules);

			for (String label : children.keySet())
			{
				Node child = hierarchy.find(label);

				if (!child.isLeaf())
				{
					mapping.put(child, children.get(label));
				}
			}
		}

		return model;
	}

	public AbstractRuleModel getClassifier(Node node, Dataset dataset)
	{
		Dataset local = Strategy.Type.MULTI_CLASS.transform(node, dataset);
		
		// TODO decides which classifier to run
		cAntMiner classifier = new cAntMiner();
		return classifier.run(local);
	}

	public Map<String, Dataset> split(Node node, Dataset dataset,
		AbstractRuleModel model)
	{
		Map<String, Dataset> children = new HashMap<String, Dataset>();

		for (Node child : node.getChildren())
		{
			Dataset d = new Dataset(dataset.getMetadata());
			children.put(child.getLabel(), d);
		}

		for (Instance instance : dataset)
		{
			// this would divide the dataset based on the prediction
			// of the model
			//String label = model.classify(instance).toString();
			
			String label = null;
			
			for (String l : children.keySet())
			{
				if (instance.getLabel().active(l))
				{
					label = l;
					break;
				}
			}

			children.get(label).add(instance.getValues());
		}

		return children;
	}

	/**
	 * This class represents the hierarchical rule model.
	 */
	private class TopDownRuleModel
		implements Model
	{
		/**
		 * The mapping of (class label, decision tree) instances.
		 */
		private Map<String, AbstractRuleModel> mapping;

		/**
		 * The class hierarchy.
		 */
		private ClassHierarchy hierarchy;

		/**
		 * Default constructor.
		 * 
		 * @param hierarchy the class hierarchy.
		 */
		public TopDownRuleModel(ClassHierarchy hierarchy)
		{
			this.hierarchy = hierarchy;
			mapping =
				new HashMap<String, AbstractRuleModel>(hierarchy.getSize() - 1,
					1.0f);
		}

		/**
		 * Sets the model for the specified class label.
		 * 
		 * @param label the class label.
		 * @param model the model to set.
		 */
		public void set(String label, AbstractRuleModel model)
		{
			mapping.put(label, model);
		}

		public Gatherer getGatherer()
		{
			throw new UnsupportedOperationException();
		}

		/**
		 * Returns the predicted label for the specified instance. The instance
		 * is classified in a top-down fashion.
		 * 
		 * @param instance the instance to predict the class label.
		 * 
		 * @return the predicted label for the specified instance.
		 */
		public Label classify(Instance instance)
		{
			HashSet<String> predicted = new HashSet<String>();
			LinkedList<Node> candidates = new LinkedList<Node>();
			candidates.add(hierarchy.getRoot());

			while (!candidates.isEmpty())
			{
				Node node = candidates.removeFirst();
				predicted.add(node.getLabel());

				AbstractRuleModel model = mapping.get(node);

				if (model == null)
				{
					// we probably reached a leaf node
					break;
				}

				candidates.add(hierarchy.find(model.classify(instance)
					.toString()));
			}

			return instance.getLabel().toLabel(predicted);
		}

		@Override
		public String toString()
		{
			return "(TopDown Rule Model)\n";
		}
	}
}