/*
 * $Id$
 * 
 * ============================================================================
 *
 * This file is part of Myra.
 *
 * Myra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Myra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Myra. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2008  Fernando Esteban Barril Otero
 * 
 * ============================================================================
 */

package net.sf.myra.hantminer.util;

import java.io.StringReader;
import java.io.StringWriter;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Map;
import net.sf.myra.datamining.Classifier;
import net.sf.myra.datamining.Gatherer;
import net.sf.myra.datamining.Model;
import net.sf.myra.datamining.data.Attribute;
import net.sf.myra.datamining.data.ClassHierarchy;
import net.sf.myra.datamining.data.ContinuousAttribute;
import net.sf.myra.datamining.data.Dataset;
import net.sf.myra.datamining.data.Instance;
import net.sf.myra.datamining.data.Label;
import net.sf.myra.datamining.data.Node;
import net.sf.myra.datamining.data.Strategy;
import net.sf.myra.datamining.io.ArffHelper;
import net.sf.myra.framework.Configuration;
import weka.classifiers.trees.J48;
import weka.core.Instances;

/**
 * @author Fernando Esteban Barril Otero
 * @version $Revision$ $Date::                     $
 */
public class J48TD
	extends Classifier
{
	/**
	 * Default constructor.
	 */
	public J48TD()
	{
		super(null);
	}

	public TopDownTreeModel run(Dataset dataset)
	{
		ClassHierarchy hierarchy = dataset.getMetadata().getClassHierarchy();
		TopDownTreeModel model = new TopDownTreeModel(hierarchy);

		Strategy.Type strategy =
			Strategy.get(Configuration.getProperty(Strategy.TYPE_KEY,
				Strategy.DEFAULT_TYPE));
		ArffHelper helper = new ArffHelper();

		for (Node node : dataset.getMetadata().getClassHierarchy().flat())
		{
			if (!node.isRoot())
			{
				try
				{
					Dataset local = strategy.transform(node, dataset);
					StringWriter output = new StringWriter();
					helper.write(output, local);

					Instances training =
						new Instances(new StringReader(output.toString()));
					training.setClassIndex(training.numAttributes() - 1);

					J48 j48 = new J48();
					j48.buildClassifier(training);

					DecisionTree t = (DecisionTree) model.get(node.getLabel());
					t.setClassifier(j48);
					t.setDataset(training);
				}
				catch (Exception e)
				{
					throw new RuntimeException(e);
				}
			}
		}

		return model;
	}

	/**
	 * This class represents the hierarchical decision tree model.
	 */
	private class TopDownTreeModel
		implements Model
	{
		/**
		 * The mapping of (class label, decision tree) instances.
		 */
		private Map<String, DecisionTree> mapping;

		/**
		 * The class hierarchy.
		 */
		private ClassHierarchy hierarchy;

		/**
		 * Default constructor.
		 * 
		 * @param hierarchy the class hierarchy.
		 */
		public TopDownTreeModel(ClassHierarchy hierarchy)
		{
			this.hierarchy = hierarchy;
			mapping =
				new HashMap<String, DecisionTree>(hierarchy.getSize() - 1, 1.0f);

			for (Node node : hierarchy.flat())
			{
				if (!node.isRoot())
				{
					mapping.put(node.getLabel(), new DecisionTree());
				}
			}
		}

		/**
		 * Returns the decision tree for the specified class label.
		 * 
		 * @param label the class label.
		 * 
		 * @return the decision tree for the specified class label.
		 */
		public DecisionTree get(String label)
		{
			return mapping.get(label);
		}
		
		public Gatherer getGatherer()
		{
			throw new UnsupportedOperationException();
		}

		/**
		 * Returns the predicted label for the specified instance. The instance
		 * is classified in a top-down fashion.
		 * 
		 * @param instance the instance to predict the class label.
		 * 
		 * @return the predicted label for the specified instance.
		 */
		public Label classify(Instance instance)
		{
			HashSet<String> predicted = new HashSet<String>();
			LinkedList<String> candidates = new LinkedList<String>();

			for (Node node : hierarchy.getRoot().getChildren())
			{
				candidates.add(node.getLabel());
			}

			predicted.add(hierarchy.getRoot().getLabel());

			while (!candidates.isEmpty())
			{
				String current = candidates.removeFirst();
				DecisionTree tree = mapping.get(current);

				if (Strategy.POSITIVE.equals(tree.classify(instance)))
				{
					predicted.add(current);

					for (Node node : hierarchy.find(current).getChildren())
					{
						candidates.add(node.getLabel());
					}
				}
			}

			return instance.getLabel().toLabel(predicted);
		}

		@Override
		public String toString()
		{
			return "(TopDown Decision Tree)\n";
		}
	}

	/**
	 * This class represents a J48 decision tree.
	 */
	private class DecisionTree
	{
		/**
		 * The (trained) J48 classifier.
		 */
		private J48 classifier;

		/**
		 * The (WEKA) training dataset.
		 */
		private Instances dataset;

		/**
		 * Sets the classifier instance.
		 * 
		 * @param classifier the classifier to set
		 */
		public void setClassifier(J48 classifier)
		{
			this.classifier = classifier;
		}

		/**
		 * Sets the dataset instance.
		 * 
		 * @param dataset the dataset to set
		 */
		public void setDataset(Instances dataset)
		{
			this.dataset = dataset;
		}

		/**
		 * Returns the predicted class label for the specified instance.
		 * 
		 * @param instance the instance to predict the class label.
		 *  
		 * @return the predicted class label for the specified instance.
		 */
		public String classify(Instance instance)
		{
			weka.core.Instance wInstance =
				new weka.core.Instance(dataset.numAttributes());
			wInstance.setDataset(dataset);

			for (Attribute attribute : instance.getMetadata().getPredictor())
			{
				if (attribute instanceof ContinuousAttribute)
				{
					int index =
						instance.getMetadata().getIndex(attribute.getName());

					wInstance.setValue(dataset.attribute(attribute.getName()),
						instance.values()[index]);
				}
				else
				{
					wInstance.setValue(dataset.attribute(attribute.getName()),
						instance.getValue(attribute));
				}
			}

			try
			{
				return wInstance.dataset().classAttribute().value(
					(int) classifier.classifyInstance(wInstance));
			}
			catch (Exception e)
			{
				throw new RuntimeException(e);
			}
		}
	}
}